/*
 * BurgnerIsofsDirectoryDotRecord - BurgnerIsofsDirectoryDotRecord object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-rrip-nm.h"
#include "burgner-isofs-susp-ce.h"
#include "burgner-isofs-directory.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-extent.h"
#include "burgner-isofs-directory-dot-record.h"


static BurgnerIsofsDirectoryRecordClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_DIR,
	PROP_WITH_RRIP,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsDirectoryDotRecord *self = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (object);

	if (self->dir) {
		g_object_unref (self->dir);
		self->dir = NULL;
	}

	self->with_rrip = 1;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryDotRecord *self = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (object);

	switch (prop_id) {
		case PROP_DIR:
			if (self->dir)
				g_object_unref (self->dir);

			self->dir = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->dir)
				self->dir = BURGNER_ISOFS_ITEM (g_object_ref (self->dir));
			g_object_notify (object, "dir");
			break;
		case PROP_WITH_RRIP:
			self->with_rrip = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_rrip");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryDotRecord *self = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (object);

	switch (prop_id) {
		case PROP_DIR:
			g_value_set_object (value, self->dir);
			break;
		case PROP_WITH_RRIP:
			g_value_set_boolean (value, self->with_rrip);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static GSList *
_biddr_get_susp_suf_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectoryDotRecord *rec = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (r);

	if (!rec->with_rrip || !rec->suf_ent)
		return NULL;

	return burgner_isofs_item_g_slist_ref (rec->suf_ent);
}

static GSList *
_biddr_get_susp_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectoryDotRecord *rec = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (r);
	BurgnerIsofsItem *ent;
	GSList *l;

	if (!rec->with_rrip || !rec->ent)
		return NULL;

	l = burgner_isofs_item_g_slist_ref (rec->ent);
	{
		GSList *it;
		for (it = l; it != NULL; it = it->next) {
			it->data = g_slist_prepend (NULL, it->data);
		}
	}

	ent = burgner_isofs_rrip_nm_new ();
	g_object_set (ent, "current", TRUE, NULL);
	l = g_slist_prepend (l, g_slist_prepend (NULL, ent));

	return l;
}

static BurgnerIsofsItem *
_biddr_get_ce_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectoryDotRecord *rec = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (r);
	BurgnerIsofsItem *ce = NULL, *ca;

	g_return_val_if_fail (rec->dir != NULL, NULL);

	if (!rec->with_rrip)
		return NULL;

	ca = burgner_isofs_directory_get_continuation_area (BURGNER_ISOFS_DIRECTORY (rec->dir));
	if (ca) {
		ce = burgner_isofs_susp_ce_new ();
		g_object_set (ce, "ca", ca, NULL);
	}

	return ce;
}

static guint
_biddr_get_base_size (BurgnerIsofsDirectoryRecord *r)
{
	return 34;
}

static void
_biddr_write_base (BurgnerIsofsDirectoryRecord *rec,
		   BurgnerIsofsMapTable *table,
		   gpointer buf,
		   guint len_dr,
		   guint n_bytes)
{
	__BurgnerIsofsImplDirectoryRecord *impl;
	BurgnerIsofsDirectoryDotRecord *dot = BURGNER_ISOFS_DIRECTORY_DOT_RECORD (rec);
	guint32 addr;
	gboolean ret;

	g_return_if_fail (dot->dir != NULL);

	impl = (__BurgnerIsofsImplDirectoryRecord *) buf;
	memset (impl, 0, sizeof (*impl));
	buf += sizeof (*impl);

	burgner_isofs_set_711 (&impl->len_dr, len_dr);
	burgner_isofs_set_711 (&impl->ear_len, 0);

	addr = 0;
	ret = burgner_isofs_map_table_get_addr (table, dot->dir, &addr, NULL);
	g_return_if_fail (ret != FALSE);
	burgner_isofs_set_733 (impl->location, addr);

	if (BURGNER_IS_ISOFS_EXTENT (dot->dir)) {
		addr = burgner_isofs_extent_get_size (BURGNER_ISOFS_EXTENT (dot->dir));
	} else {
		guint32 n_s = 0, n_b = 0;
		burgner_isofs_item_get_size (dot->dir, &n_s, &n_b);
		addr = (n_s << 11) + n_b;
	}
	g_return_if_fail (addr != 0);
	burgner_isofs_set_733 (impl->data_len, addr);

	impl->file_flags = __BURGNER_ISOFS_IMPL_DIRECTORY_RECORD_FLAGS_DIRECTORY;

	burgner_isofs_set_711 (&impl->file_unit_size, 0);
	burgner_isofs_set_711 (&impl->interleave_gap_size, 0);
	burgner_isofs_set_733 (impl->volume_sequence_number, 1);
	burgner_isofs_set_711 (&impl->len_fi, 1);

	*((guchar *) buf) = 0;
}


static void
class_init (BurgnerIsofsDirectoryDotRecordClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsDirectoryRecordClass *dir_rec_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	dir_rec_class = (BurgnerIsofsDirectoryRecordClass *) klass;
	dir_rec_class->get_susp_suf_ent = _biddr_get_susp_suf_ent;
	dir_rec_class->get_susp_ent = _biddr_get_susp_ent;
	dir_rec_class->get_ce_ent = _biddr_get_ce_ent;
	dir_rec_class->get_base_size = _biddr_get_base_size;
	dir_rec_class->write_base = _biddr_write_base;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_DIR,
					 g_param_spec_object ("dir",
					            	      "dir",
							      "Directory",
							      BURGNER_TYPE_ISOFS_DIRECTORY,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WITH_RRIP,
					 g_param_spec_boolean ("with_rrip",
					            	       "with_rrip",
							       "With RockRidge",
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsDirectoryDotRecord *self)
{
	self->dir = NULL;
	self->with_rrip = 1;
}

GType
burgner_isofs_directory_dot_record_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsDirectoryDotRecordClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsDirectoryDotRecord),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_DIRECTORY_RECORD, "BurgnerIsofsDirectoryDotRecord", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_directory_dot_record_new (void)
{
	BurgnerIsofsDirectoryDotRecord *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_DIRECTORY_DOT_RECORD, NULL);


	return BURGNER_ISOFS_ITEM (self);
}

GQuark
burgner_isofs_directory_dot_record_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsDirectoryDotRecord");
}

void
burgner_isofs_directory_dot_record_append_suf_ents_valist_real (BurgnerIsofsDirectoryDotRecord *rec,
								BurgnerIsofsItem *first,
								va_list args)
{
	BurgnerIsofsItem *i = first;

	do {
		g_object_ref (i);
		rec->suf_ent = g_slist_append (rec->suf_ent, i);
	} while ((i = va_arg (args, BurgnerIsofsItem*)));
}


void
burgner_isofs_directory_dot_record_append_suf_ents (BurgnerIsofsDirectoryDotRecord *rec,
						    BurgnerIsofsItem *first,
						    ...)
{
	va_list args;

	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_DOT_RECORD (rec));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (first));

	va_start (args, first);
	burgner_isofs_directory_dot_record_append_suf_ents_valist_real (rec, first, args);
	va_end (args);
}

void
burgner_isofs_directory_dot_record_append_suf_ents_valist (BurgnerIsofsDirectoryDotRecord *rec,
							   BurgnerIsofsItem *first,
						    	   va_list args)
{
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_DOT_RECORD (rec));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (first));

	burgner_isofs_directory_dot_record_append_suf_ents_valist_real (rec, first, args);
}
