/*
 * BurgnerIsofsDirectoryFileRecord - BurgnerIsofsDirectoryFileRecord object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-tree.h"
#include "burgner-isofs-rrip-px.h"
#include "burgner-isofs-rrip-pn.h"
#include "burgner-isofs-rrip-nm.h"
#include "burgner-isofs-rrip-tf.h"
#include "burgner-isofs-rrip-sl.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-extent.h"
#include "burgner-isofs-directory-file-record.h"


static BurgnerIsofsDirectoryRecordClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_FILE,
	PROP_INFO,
	PROP_WITH_RRIP,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsDirectoryFileRecord *self = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (object);

	if (self->file) {
		g_object_unref (self->file);
		self->file = NULL;
	}

	if (self->info) {
		burgner_isofs_file_info_unref (self->info);
		self->info = NULL;
	}

	if (!self->with_rrip)
		self->with_rrip = 1;


	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryFileRecord *self = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (object);

	switch (prop_id) {
		case PROP_FILE:
			if (self->file)
				g_object_unref (self->file);

			self->file = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->file)
				self->file = BURGNER_ISOFS_ITEM (g_object_ref (self->file));
			g_object_notify (object, "file");
			break;
		case PROP_INFO:
			if (self->info)
				burgner_isofs_file_info_unref (self->info);
			self->info = (BurgnerIsofsFileInfo *) g_value_get_pointer (value);
			if (self->info)
				self->info = burgner_isofs_file_info_ref (self->info);
			g_object_notify (object, "info");
			break;
		case PROP_WITH_RRIP:
			self->with_rrip = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_rrip");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryFileRecord *self = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (object);

	switch (prop_id) {
		case PROP_FILE:
			g_value_set_object (value, self->file);
			break;
		case PROP_INFO:
			g_value_set_pointer (value, self->info);
			break;
		case PROP_WITH_RRIP:
			g_value_set_boolean (value, self->with_rrip?TRUE:FALSE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static GSList *
_bidfr_get_susp_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectoryFileRecord *rec = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (r);
	BurgnerIsofsItem *i;
	GSList *l = NULL;

	g_return_val_if_fail (rec->info != NULL, 0);

	if (rec->with_rrip) {
		i = burgner_isofs_rrip_px_new ();
		g_object_set (i,
			      "info", rec->info,
			      NULL);
		l = g_slist_prepend (l, g_slist_prepend (NULL, i));


		if (rec->info->type == BURGNER_ISOFS_FILE_TYPE_CHARACTER_DEVICE ||
		    rec->info->type == BURGNER_ISOFS_FILE_TYPE_BLOCK_DEVICE) {
			i = burgner_isofs_rrip_pn_new ();
			g_object_set (i, "info", rec->info, NULL);
			l = g_slist_prepend (l, g_slist_prepend (NULL, i));
		}


		if (rec->info->name) {
			GSList *nm = burgner_isofs_rrip_nm_new_slist_from_id (rec->info->name);
			l = g_slist_prepend (l, nm);
		}

		if (rec->info->creation_time || rec->info->modify_time ||
		    rec->info->access_time || rec->info->attributes_time ||
		    rec->info->backup_time || rec->info->expiration_time ||
		    rec->info->effective_time) {
			i = burgner_isofs_rrip_tf_new ();
			g_object_set (i, "info", rec->info, NULL);
			l = g_slist_prepend (l, g_slist_prepend (NULL, i));
		}

		if (rec->info->symlink) {
			GSList *sl = burgner_isofs_rrip_sl_new_slist_from_symlink (rec->info->symlink);
			l = g_slist_prepend (l, sl);
		}
	}

	return l;
}

static guint
_bidfr_get_base_size (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectoryFileRecord *rec = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (r);
	guint n_b;

	g_return_val_if_fail (rec->info != NULL, 0);

	n_b = 33 + strlen (rec->info->identifier);
	if (n_b & 1)
		n_b += 1;

	return n_b;
}

static void
_bidfr_write_base (BurgnerIsofsDirectoryRecord *r,
		   BurgnerIsofsMapTable *table,
		   gpointer buf,
		   guint len_dr,
		   guint n_bytes)
{
	
	__BurgnerIsofsImplDirectoryRecord *impl;
	BurgnerIsofsDirectoryFileRecord *rec = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (r);
	guint32 loc, len;

	g_return_if_fail (n_bytes != 0);
	g_return_if_fail (rec->info != NULL);
	g_return_if_fail (rec->info->identifier != NULL);

	impl = (__BurgnerIsofsImplDirectoryRecord *) buf;
	memset (impl, 0, sizeof (*impl));
	buf += sizeof (*impl);

	burgner_isofs_set_711 (&impl->len_dr, len_dr);
	burgner_isofs_set_711 (&impl->ear_len, 0);

	if (rec->file) {
		burgner_isofs_map_table_get_addr (table, BURGNER_ISOFS_ITEM (rec->file),
						  &loc, NULL);

		if (BURGNER_IS_ISOFS_EXTENT (rec->file)) {
			len = burgner_isofs_extent_get_size (BURGNER_ISOFS_EXTENT (rec->file));
		} else {
			guint32 n_s = 0, n_b = 0;
			burgner_isofs_item_get_size (rec->file, &n_s, &n_b);
			len = (n_s << 11) + n_b;
		}
	} else if (rec->info->is_in_prev_session) {
		loc = rec->info->addr;
		len = rec->info->size;
	} else {
		loc = len = 0;
	}

	/*
	 * TODO: Add support for LARGEFILEs (64 bits coded length)
	 * using RockRidge's SF entry.
	 */
	burgner_isofs_set_733 (impl->location, loc);
	burgner_isofs_set_733 (impl->data_len, len);

	burgner_isofs_rrip_tf_fill_951_from_time (&impl->recording_date, rec->info->modify_time);

	/* TODO: Flags: Support Associated files */

	burgner_isofs_set_723 (impl->volume_sequence_number, 1);

	{
		guint l;

		l = strlen (rec->info->identifier);
		g_return_if_fail (l != 0);

		burgner_isofs_set_711 (&impl->len_fi, l);
		memcpy (buf, rec->info->identifier, l);
		buf += l;

		if (!(l & 1)) {
			*((guchar *) buf) = 0;
		}
	}
}

static const gchar *
_burgner_isofs_directory_file_record_get_identifier (BurgnerIsofsDirectoryRecord *rec)
{
	BurgnerIsofsDirectoryFileRecord *fr = BURGNER_ISOFS_DIRECTORY_FILE_RECORD (rec);

	g_return_val_if_fail (fr->info != NULL, NULL);
	g_return_val_if_fail (fr->info->identifier != NULL, NULL);

	return fr->info->identifier;
}

static void
class_init (BurgnerIsofsDirectoryFileRecordClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsDirectoryRecordClass *dir_rec_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	dir_rec_class = (BurgnerIsofsDirectoryRecordClass *) klass;
	dir_rec_class->get_susp_ent = _bidfr_get_susp_ent;
	dir_rec_class->get_base_size = _bidfr_get_base_size;
	dir_rec_class->write_base = _bidfr_write_base;
	dir_rec_class->get_identifier = _burgner_isofs_directory_file_record_get_identifier;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_FILE,
					 g_param_spec_object ("file",
					            	      _("file"),
							      _("File"),
							      BURGNER_TYPE_ISOFS_ITEM,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_INFO,
					 g_param_spec_pointer ("info",
					            	       _("info"),
							       _("Info"),
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WITH_RRIP,
					 g_param_spec_boolean ("with_rrip",
					            	       _("with_rrip"),
							       _("With RockRidge"),
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsDirectoryFileRecord *self)
{
	self->file = NULL;
	self->info = NULL;
	self->with_rrip = 1;
}

GType
burgner_isofs_directory_file_record_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsDirectoryFileRecordClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsDirectoryFileRecord),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_DIRECTORY_RECORD, "BurgnerIsofsDirectoryFileRecord", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_directory_file_record_new (void)
{
	BurgnerIsofsDirectoryFileRecord *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_DIRECTORY_FILE_RECORD, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_directory_file_record_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsDirectoryFileRecord");
}

