/*
 * BurgnerIsofsDirectoryRecord - BurgnerIsofsDirectoryRecord object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-susp-ce.h"
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-susp-continuation-area.h"
#include "burgner-isofs-directory-record.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_WITH_SUSP,
	LAST_ARG
};

static void
_bidr_rec_unref (GSList *l)
{
	GSList *it;
	for (it = l; it != NULL; it = it->next)
		burgner_isofs_item_g_slist_unref ((GSList *) it->data);
	g_slist_free (l);
}

static void
finalize (GObject *object)
{
	BurgnerIsofsDirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_RECORD (object);

	if (self->suf_ent) {
		burgner_isofs_item_g_slist_unref (self->suf_ent);
		self->suf_ent = NULL;
		self->suf_ent_size = 0;
	}

	if (self->bonus_ent) {
		burgner_isofs_item_g_slist_unref (self->bonus_ent);
		self->bonus_ent = NULL;
		self->bonus_ent_size = 0;
	}

	if (self->ca_ent) {
		burgner_isofs_item_g_slist_unref (self->ca_ent);
		self->ca_ent = NULL;
	}

	if (self->ent) {
		_bidr_rec_unref (self->ent);
		self->ent = NULL;
		self->ent_size = 0;
	}

	if (self->ce) {
		g_object_unref (self->ce);
		self->ce = NULL;
	}

	if (self->inter_ent_size) {
		if (self->inter_is_ce) {
			g_object_unref (self->inter.ce);
			self->inter.ce = NULL;
			self->inter_is_ce = 0;
		} else {
			burgner_isofs_item_g_slist_unref (self->inter.inter_ent);
			self->inter.inter_ent = NULL;
		}

		self->inter_ent_size = 0;
		self->inter_is_our_ce = 0;
	}

	self->frigo_fetched = 0;
	self->with_susp = 1;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_RECORD (object);

	switch (prop_id) {
		case PROP_WITH_SUSP:
			self->with_susp = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_susp");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_RECORD (object);

	switch (prop_id) {
		case PROP_WITH_SUSP:
			g_value_set_boolean (value, self->with_susp?TRUE:FALSE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static guint
_bidr_entries_size (GSList *l)
{
	GSList *it;
	guint32 n_b, c = 0;

	for (it = l; it != NULL; it = it->next) {
		n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data),
					     NULL, &n_b);
		c += n_b;
	}

	return c;
}

static guint
_bidr_entries_rec_size (GSList *l)
{
	GSList *it;
	guint c = 0;

	for (it = l; it != NULL; it = it->next)
		c += _bidr_entries_size ((GSList *) it->data);

	return c;
}

static GSList *
_bidr_rec_to_lin (GSList *l)
{
	GSList *lin = NULL, *it;

	l = g_slist_reverse (l);
	for (it = l; it != NULL; it = it->next)
		lin = g_slist_concat ((GSList *) it->data, lin);
	g_slist_free (l);

	return lin;
}

static BurgnerIsofsItem *
_bidr_rec_get_remove_nearest (GSList **l, guint max, guint *n_b)
{
	BurgnerIsofsItem *i, *nearest = NULL;
	GSList *it, *c, *it_n = NULL, *c_n = NULL;
	guint d = 0, s = 0, b = 0;

	for (it = *l; it != NULL; it = it->next) {
		if ((c = (GSList *) it->data) &&
		    (i = BURGNER_ISOFS_ITEM (c->data))) {
			s = 0;
			burgner_isofs_item_get_size (i, NULL, &s);
			if (s <= max &&
			    (!nearest || (max-s) < d)) {
				b = s;
				d = max - s;
				nearest = i;
				it_n = it;
				c_n = c;
			}
		}

		if (nearest && d == 0)
			break;
	}

	if (nearest) {
		if (n_b)
			*n_b = b;
		it_n->data = g_slist_delete_link (it_n->data, c_n);
		if (!it_n->data)
			*l = g_slist_delete_link (*l, it_n);
	}

	return nearest;
}

static void
_bidr_get_size (BurgnerIsofsItem *item,
		guint32 *n_sectors,
		guint32 *n_bytes)
{
	BurgnerIsofsDirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_RECORD (item);
	guint32 n_b;

	if (!rec->with_susp) {
		n_b = burgner_isofs_directory_record_get_base_size (rec);
		if (n_sectors)
			*n_sectors = 0;
		if (n_bytes)
			*n_bytes = n_b;
		return;
	}


	if (!rec->frigo_fetched) {
		g_assert (rec->suf_ent == NULL);
		g_assert (rec->suf_ent_size == 0);
		g_assert (rec->ent == NULL);
		g_assert (rec->ent_size == 0);
		g_assert (rec->ce == NULL);

		rec->suf_ent = burgner_isofs_directory_record_get_susp_suf_ent (rec);
		rec->suf_ent_size = _bidr_entries_size (rec->suf_ent);
		rec->ent = burgner_isofs_directory_record_get_susp_ent (rec);
		rec->ent_size = _bidr_entries_rec_size (rec->ent);
		rec->ce = burgner_isofs_directory_record_get_ce_ent (rec);
		g_return_if_fail ((rec->ce == NULL) || (BURGNER_IS_ISOFS_SUSP_CE (rec->ce)));
		rec->frigo_fetched = 1;

		if (rec->ce && !rec->ent_size) {
			rec->inter_is_ce = 1;
			rec->inter_is_our_ce = 0;
			rec->inter_ent_size = 28;
			rec->inter.ce = rec->ce;
			rec->ce = NULL;
		} else if (!rec->ce && rec->ent_size <= 28) {
			rec->inter_is_ce = rec->inter_is_our_ce = 0;
			rec->inter_ent_size = rec->ent_size;
			rec->inter.inter_ent = _bidr_rec_to_lin (rec->ent);
			rec->ent = NULL;
			rec->ent_size = 0;
		}
	}

	/* If the Inter is filled, the frigo must be empty */
	g_assert (!(rec->inter_ent_size && (rec->ent_size || rec->ce)));

	n_b = burgner_isofs_directory_record_get_base_size (rec)
		+ rec->suf_ent_size + rec->inter_ent_size;

	if (rec->ent_size || rec->ce) {
		g_return_if_fail (rec->bonus_ent_size == 0);
		n_b += 28;
	} else {
		n_b += rec->bonus_ent_size;
	}

	if (n_b & 1)
		n_b += 1;


	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = n_b;
}

static void
check_sizes (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec));

	g_return_if_fail (_bidr_entries_size (rec->suf_ent) == rec->suf_ent_size);
	g_return_if_fail (_bidr_entries_size (rec->bonus_ent) == rec->bonus_ent_size);
	g_return_if_fail (_bidr_entries_rec_size (rec->ent) == rec->ent_size);

	if (rec->inter_is_ce) {
		g_return_if_fail (rec->inter_ent_size == 28);
	} else {
		g_assert (rec->inter_is_our_ce == 0);
		g_return_if_fail (rec->inter_ent_size == _bidr_entries_size
(rec->inter.inter_ent));
	}
}

static void
_bidr_prepare (BurgnerIsofsItem *item,
	       BurgnerIsofsMapTable *table,
	       guint32 sector_n,
	       guint32 byte_o,
	       guint32 last_sector_n,
	       guint32 last_byte_o)
{
	BurgnerIsofsDirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_RECORD (item);
	BurgnerIsofsItem *mi;
	GSList *it, *l[6];
	guint32 n_s, n_b, s_c, b_c, o_s, o_b;
	gint i;

	g_return_if_fail (rec->prepared == 0);
	g_return_if_fail (!rec->with_susp || rec->frigo_fetched == 1);
	g_return_if_fail ((sector_n == last_sector_n) ||
			  (sector_n+1 == last_sector_n && last_byte_o == 0));

	check_sizes (rec);

	s_c = sector_n;
	b_c = byte_o;

	rec->base_b = burgner_isofs_directory_record_get_base_size (rec);
	b_c += rec->base_b;

	g_assert (b_c <= 2048);
	if (b_c == 2048) {
		g_assert (rec->suf_ent == NULL);
		g_assert (rec->inter.inter_ent == NULL);
		g_assert (rec->bonus_ent == NULL);
		g_assert (rec->ca_ent == NULL);
		s_c += 1;
		b_c = 0;
	}

	if (!rec->with_susp) {
		g_return_if_fail (s_c == last_sector_n);
		g_return_if_fail (b_c == last_byte_o);
		return;
	}

	g_assert (rec->suf_ent_b == NULL);
	g_assert (rec->inter_ent_b == NULL);
	g_assert (rec->bonus_ent_b == NULL);

	l[0] = rec->suf_ent;
	l[1] = rec->bonus_ent;
	l[2] = rec->inter_is_ce ? NULL : rec->inter.inter_ent;
	l[3] = l[4] = l[5] = NULL;
	for (i = 0; i < 3; i++) {
		for (it = l[i]; it != NULL; it = it->next) {
			mi = BURGNER_ISOFS_ITEM (it->data);

			burgner_isofs_map_table_set (table, mi, s_c, b_c);

			n_s = n_b = 0;
			burgner_isofs_item_get_size (mi, &n_s, &n_b);
			g_assert (n_s == 0);
			l[i+3] = g_slist_prepend (l[i+3], GUINT_TO_POINTER (n_b));

			o_s = s_c;
			o_b = b_c;
			b_c += n_b;
			g_assert (b_c <= 2048);
			if (b_c == 2048) {
				g_assert (it->next == NULL);
				s_c += 1;
				b_c = 0;
			}

			burgner_isofs_item_prepare (mi, table, o_s, o_b, s_c, b_c);
		}
		l[i+3] = g_slist_reverse (l[i+3]);
	}

	rec->suf_ent_b = l[3];
	rec->bonus_ent_b = l[4];
	rec->inter_ent_b = rec->inter_is_ce ? NULL : l[5];

	if (rec->inter_is_ce && rec->inter_ent_size) {
		g_assert (rec->inter_ent_size == 28);

		burgner_isofs_map_table_set (table, rec->inter.ce, s_c, b_c);

		n_s = n_b = 0;
		burgner_isofs_item_get_size (rec->inter.ce, &n_s, &n_b);
		g_assert (n_s == 0);
		g_assert (n_b == 28);

		o_s = s_c;
		o_b = b_c;
		b_c += n_b;
		g_assert (b_c <= 2048);
		if (b_c == 2048) {
			s_c += 1;
			b_c = 0;
		}

		burgner_isofs_item_prepare (rec->inter.ce, table, o_s, o_b, s_c, b_c);
	}

	rec->prepared = 1;

/*	g_return_if_fail (s_c == last_sector_n);
	g_return_if_fail (b_c == last_byte_o);*/
}

static void
_bidr_write_sync (BurgnerIsofsItem *item,
		  BurgnerIsofsMapTable *table,
		  gpointer buf, guint32 size)
{
	BurgnerIsofsDirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_RECORD (item);
	GSList *it, *it_b, *l[6];
	guint n_b;
	gint i;

	g_return_if_fail (!rec->with_susp || rec->frigo_fetched == 1);
	check_sizes (rec);

	burgner_isofs_directory_record_write_base (rec,
						   table,
						   buf,
						   size,
						   rec->base_b);
	buf += rec->base_b;
	size -= rec->base_b;

	if (!rec->with_susp) {
		g_assert (size == 0);
		return;
	}

	l[0] = rec->suf_ent;
	l[1] = rec->bonus_ent;
	l[2] = rec->inter_is_ce ? NULL : rec->inter.inter_ent;
	l[3] = rec->suf_ent_b;
	l[4] = rec->bonus_ent_b;
	l[5] = rec->inter_is_ce ? NULL : rec->inter_ent_b;

	for (i = 0; i < 3; i++) {
		guint a = g_slist_length (l[i]), b = g_slist_length (l[i+3]);
		g_return_if_fail (a == b);

		for (it = l[i], it_b = l[i+3];
		     it != NULL;
		     it = it->next, it_b = it_b->next) {
#if 0
			burgner_isofs_writer_proxy_begin_item (proxy, BURGNER_ISOFS_ITEM (it->data));
			burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it->data),
						       proxy, table, 0,
						       GPOINTER_TO_UINT (it_b->data));
			burgner_isofs_writer_proxy_end_item (proxy, BURGNER_ISOFS_ITEM (it->data));
#else
			burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it->data),
							table,
							buf, GPOINTER_TO_UINT (it_b->data));
			buf += GPOINTER_TO_UINT (it_b->data);
			size -= GPOINTER_TO_UINT (it_b->data);
#endif
		}
	}

	n_b = rec->base_b + rec->suf_ent_size + rec->bonus_ent_size + rec->inter_ent_size;

	if (rec->inter_is_ce) {
		g_assert (rec->inter_ent_size == 28);
#if 0
		burgner_isofs_writer_proxy_begin_item (proxy, rec->inter.ce);
		burgner_isofs_item_write_sync (rec->inter.ce, proxy, table, 0, 28);
		burgner_isofs_writer_proxy_end_item (proxy, rec->inter.ce);
#else
		burgner_isofs_item_write_sync (rec->inter.ce, table, buf, 28);
		buf += 28;
		size -= 28;
#endif
	}

	if (n_b & 1) {
#if 0
		gchar c = 0;
		burgner_isofs_writer_proxy_write (proxy, &c, 1);
#else
		*((guchar*) buf++) = 0;
		size -= 1;
#endif
	}

	g_assert (size == 0);
}

static guint
_bidr_bonus_size (BurgnerIsofsDirectoryRecord *rec,
		  guint bonus_size)
{
	BurgnerIsofsItem *i;
	guint obonus, mbonus, obase, base, n_b;

	g_return_val_if_fail (!rec->with_susp || rec->frigo_fetched == 1, bonus_size);

	if (!rec->with_susp ||
	    (!rec->ent_size && !rec->ce))
		return bonus_size;

	g_assert (rec->inter_ent_size == 0);

	obase = burgner_isofs_directory_record_get_base_size (rec)
		+ rec->suf_ent_size;
	base = 254 - obase;
	mbonus = obonus = MIN (base, bonus_size);

	n_b = 0;
	while ((i = _bidr_rec_get_remove_nearest (&rec->ent, mbonus, &n_b))) {
		g_return_val_if_fail (n_b <= mbonus, bonus_size);
		rec->ent_size -= n_b;

		rec->bonus_ent = g_slist_prepend (rec->bonus_ent, i);
		rec->bonus_ent_size += n_b;

		mbonus -= n_b;
		if (mbonus < 4 || !rec->ent)
			break;

		n_b = 0;
	}
	rec->bonus_ent = g_slist_reverse (rec->bonus_ent);

	if (rec->ent) {
		if (rec->ent_size <= (28+mbonus) && !rec->ce) {
			rec->inter.inter_ent = _bidr_rec_to_lin (rec->ent);
			rec->inter_ent_size = rec->ent_size;
			rec->ent = NULL;
			rec->ent_size = 0;
			rec->inter_is_ce = rec->inter_is_our_ce = 0;

			if (rec->inter_ent_size > 28)
				mbonus -= rec->inter_ent_size - 28;
		} else {
			BurgnerIsofsItem *ce = burgner_isofs_susp_ce_new ();
			rec->inter.ce = ce;
			rec->inter_ent_size = 28;
			rec->inter_is_ce = rec->inter_is_our_ce = 1;

			rec->ca_ent = _bidr_rec_to_lin (rec->ent);
			rec->ent = NULL;
			rec->ent_size = 0;

			if (rec->ce) {
				rec->ca_ent = g_slist_append (rec->ca_ent, rec->ce);
				rec->ce = NULL;
			}
		}
	} else if (rec->ce) {
		rec->inter.ce = rec->ce;
		rec->ce = NULL;
		rec->inter_ent_size = 28;
		rec->inter_is_ce = 1;
		rec->inter_is_our_ce = 0;
	} else {
		mbonus += 28;
	}

	if (mbonus & 1)
		mbonus -= 1;

/*	check_sizes (rec);*/

	return bonus_size - obonus + mbonus;
}

static gboolean
_bidr_need_continuation_area (BurgnerIsofsDirectoryRecord *rec)
{
	return (rec->with_susp && rec->inter_is_ce && rec->inter_is_our_ce);
}

static void
_bidr_set_continuation_area (BurgnerIsofsDirectoryRecord *rec,
			     BurgnerIsofsItem *ca)
{
	GSList *it;

	g_assert (rec->with_susp);
	g_assert (rec->inter_is_ce && rec->inter_is_our_ce);

	g_object_set (rec->inter.ce,
		      "ca", ca,
		      NULL);

	rec->ca_ent = g_slist_reverse (rec->ca_ent);
	for (it = rec->ca_ent; it != NULL; it = it->next) {
		burgner_isofs_susp_continuation_area_prepend (BURGNER_ISOFS_SUSP_CONTINUATION_AREA (ca),
							      BURGNER_ISOFS_ITEM (it->data));
	}
	rec->ca_ent = g_slist_reverse (rec->ca_ent);
}


static void
class_init (BurgnerIsofsDirectoryRecordClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bidr_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_none;
	item_class->prepare = _bidr_prepare;
	item_class->write_sync = _bidr_write_sync;

	/* Virtual functions */
	klass->bonus_size = _bidr_bonus_size;
	klass->need_continuation_area = _bidr_need_continuation_area;
	klass->set_continuation_area = _bidr_set_continuation_area;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_WITH_SUSP,
					 g_param_spec_boolean ("with_susp",
					            	       "with_susp",
							       "With SUSP",
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsDirectoryRecord *self)
{
	self->base_b = self->suf_ent_size = self->inter_ent_size =
		self->bonus_ent_size = self->ent_size = 0;

	self->suf_ent = self->suf_ent_b =
		self->inter.inter_ent = self->inter_ent_b =
		self->bonus_ent = self->bonus_ent_b =
		self->ca_ent = self->ent = NULL;
	self->ce = NULL;

	self->frigo_fetched = self->inter_is_ce = self->inter_is_our_ce = 0;

	self->with_susp = 1;
}

GType
burgner_isofs_directory_record_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsDirectoryRecordClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsDirectoryRecord),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsDirectoryRecord", &info, 0);
	}

	return type;
}

GQuark
burgner_isofs_directory_record_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsDirectoryRecord");
}

GSList *
burgner_isofs_directory_record_get_susp_suf_ent (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), NULL);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_susp_suf_ent)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_susp_suf_ent (rec);

	return NULL;
}

GSList *
burgner_isofs_directory_record_get_susp_ent (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), NULL);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_susp_ent)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_susp_ent (rec);

	return NULL;
}

BurgnerIsofsItem *
burgner_isofs_directory_record_get_ce_ent (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), NULL);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_ce_ent)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_ce_ent (rec);

	return NULL;
}

guint
burgner_isofs_directory_record_get_base_size (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, 0);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), 0);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_base_size)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_base_size (rec);

	return 0;
}

void
burgner_isofs_directory_record_write_base (BurgnerIsofsDirectoryRecord *rec, 
		    			   BurgnerIsofsMapTable *table,
					   gpointer buf,
		                           guint len_dr,
		    			   guint n_bytes)
{
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec));

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->write_base)
		BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->write_base (rec, table, buf, len_dr, n_bytes);
}


guint
burgner_isofs_directory_record_bonus_size (BurgnerIsofsDirectoryRecord *rec,
		                           guint bonus_size)
{
	g_return_val_if_fail (rec != NULL, bonus_size);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), bonus_size);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->bonus_size)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->bonus_size (rec, bonus_size);

	return bonus_size;
}

gboolean
burgner_isofs_directory_record_need_continuation_area (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), FALSE);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->need_continuation_area)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->need_continuation_area (rec);

	return FALSE;
}

void
burgner_isofs_directory_record_set_continuation_area (BurgnerIsofsDirectoryRecord *rec,
			                              BurgnerIsofsItem *ca)
{
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec));
	g_return_if_fail (ca != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (ca));

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->set_continuation_area)
		BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->set_continuation_area (rec, ca);
}

const gchar *
burgner_isofs_directory_record_get_identifier (BurgnerIsofsDirectoryRecord *rec)
{
	g_return_val_if_fail (rec != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec), NULL);

	if (BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_identifier)
		return BURGNER_ISOFS_DIRECTORY_RECORD_GET_CLASS (rec)->get_identifier (rec);

	return NULL;
}

