/*
 * BurgnerIsofsDirectorySubdirectoryRecord - BurgnerIsofsDirectorySubdirectoryRecord object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-tree.h"
#include "burgner-isofs-susp-entry.h"
#include "burgner-isofs-rrip-loc.h"
#include "burgner-isofs-rrip-tf.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-extent.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-directory-subdirectory-record.h"


static BurgnerIsofsDirectoryRecordClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_DIRECTORY,
	PROP_INFO,
	PROP_ID,
	PROP_WITH_RRIP,
	PROP_IS_RELOCATED,
	PROP_IS_CHILD_LINK,
	PROP_FILE,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsDirectorySubdirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (object);

	if (self->directory) {
		g_object_unref (self->directory);
		self->directory = NULL;
	}

	if (self->file) {
		g_object_unref (self->file);
		self->file = NULL;
	}

	if (self->info) {
		burgner_isofs_file_info_unref (self->info);
		self->info = NULL;
	}

	self->with_rrip = 1;
	self->is_relocated = 0;
	self->is_child_link = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectorySubdirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (object);

	switch (prop_id) {
		case PROP_DIRECTORY:
			if (self->directory)
				g_object_unref (self->directory);
			self->directory = (BurgnerIsofsDirectory *) g_value_get_object (value);
			if (self->directory)
				self->directory = BURGNER_ISOFS_DIRECTORY (g_object_ref (self->directory));
			g_object_notify (object, "dir");
			break;
		case PROP_INFO:
			if (self->info)
				burgner_isofs_file_info_unref (self->info);
			self->info = g_value_get_pointer (value);
			if (self->info)
				self->info = burgner_isofs_file_info_ref (self->info);
			g_object_notify (object, "info");
			break;
		case PROP_ID:
			if (self->id)
				g_free (self->id);
			self->id = g_value_dup_string (value);
			g_object_notify (object, "id");
			break;
		case PROP_WITH_RRIP:
			self->with_rrip = g_value_get_boolean (value) ? 1 : 0;
			g_object_notify (object, "with_rrip");
			break;
		case PROP_IS_RELOCATED:
			self->is_relocated = g_value_get_boolean (value) ? 1 : 0;
			g_object_notify (object, "is_relocated");
			break;
		case PROP_IS_CHILD_LINK:
			self->is_child_link = g_value_get_boolean (value) ? 1 : 0;
			g_object_notify (object, "is_child_link");
			break;
		case PROP_FILE:
			if (self->file)
				g_object_unref (self->file);
			self->file = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->file)
				self->file = BURGNER_ISOFS_ITEM (g_object_ref (self->file));
			g_object_notify (object, "file");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectorySubdirectoryRecord *self = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (object);

	switch (prop_id) {
		case PROP_DIRECTORY:
			g_value_set_object (value, self->directory);
			break;
		case PROP_INFO:
			g_value_set_pointer (value, self->info);
			break;
		case PROP_ID:
			g_value_set_string (value, self->id);
			break;
		case PROP_WITH_RRIP:
			g_value_set_boolean (value, self->with_rrip);
			break;
		case PROP_IS_RELOCATED:
			g_value_set_boolean (value, self->is_relocated);
			break;
		case PROP_IS_CHILD_LINK:
			g_value_set_boolean (value, self->is_child_link);
			break;
		case PROP_FILE:
			g_value_set_object (value, self->file);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static GSList *
_bidsr_get_susp_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectorySubdirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (r);
	GSList *l = NULL;
	BurgnerIsofsItem *i;

	if (!rec->with_rrip)
		return NULL;

	if (rec->is_relocated) {
		i = burgner_isofs_rrip_re_new ();
		l = g_slist_prepend (l, g_slist_prepend (NULL, i));
	} else if (rec->is_child_link) {
		i = burgner_isofs_rrip_cl_new ();
		g_object_set (i, "item", rec->directory, NULL);
	}

	return l;
}

static BurgnerIsofsItem *
_bidsr_get_ce_ent (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectorySubdirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (r);
	BurgnerIsofsItem *ca, *ce;

	g_return_val_if_fail (rec->directory != NULL, NULL);

	if (!rec->with_rrip || rec->is_relocated || rec->is_child_link)
		return NULL;

	ca = burgner_isofs_directory_get_continuation_area (BURGNER_ISOFS_DIRECTORY (rec->directory));
	g_return_val_if_fail (ca != NULL, NULL);

	ce = burgner_isofs_susp_ce_new ();
	g_object_set (ce, "ca", ca, NULL);

	return ce;
}

static guint
_bidsr_get_base_size (BurgnerIsofsDirectoryRecord *r)
{
	BurgnerIsofsDirectorySubdirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (r);
	guint n_b;

	g_return_val_if_fail (rec->id || rec->info, 34);

	n_b = 33;

	if (rec->id)
		n_b += strlen (rec->id);
	else if (rec->info->identifier)
		n_b += strlen (rec->info->identifier);

	if (n_b & 1)
		n_b += 1;

	return n_b;
}

static void
_bidsr_write_base (BurgnerIsofsDirectoryRecord *r,
		   BurgnerIsofsMapTable *table,
		   gpointer buf,
		   guint len_dr,
		   guint n_bytes)
{
	__BurgnerIsofsImplDirectoryRecord *impl;
	BurgnerIsofsDirectorySubdirectoryRecord *rec = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (r);
	BurgnerIsofsItem *extent;
	guint l;
	guint32 loc, len;

	g_return_if_fail (rec->id || rec->info);
	g_return_if_fail (rec->id || rec->info->identifier);

	impl = (__BurgnerIsofsImplDirectoryRecord *) buf;
	memset (impl, 0, sizeof (*impl));
	buf += sizeof (*impl);

	burgner_isofs_set_711 (&impl->len_dr, len_dr);
	burgner_isofs_set_711 (&impl->ear_len, 0);

	if (rec->is_child_link && rec->file) {
		extent = BURGNER_ISOFS_ITEM (rec->file);
	} else if (!rec->is_child_link && rec->directory) {
		extent = BURGNER_ISOFS_ITEM (rec->directory);
	} else {
		extent = NULL;
	}

	if (extent) {
		burgner_isofs_map_table_get_addr (table, extent,
						  &loc, NULL);

		if (BURGNER_IS_ISOFS_EXTENT (extent)) {
			len = burgner_isofs_extent_get_size (BURGNER_ISOFS_EXTENT (extent));
		} else {
			guint32 n_s = 0, n_b = 0;
			burgner_isofs_item_get_size (extent, &n_s, &n_b);
			len = (n_s << 11) + n_b;
		}
	} else if (rec->info->is_in_prev_session) {
		loc = rec->info->addr;
		len = rec->info->size;
	} else {
		loc = len = 0;
	}

	burgner_isofs_set_733 (impl->location, loc);
	burgner_isofs_set_733 (impl->data_len, len);

	if (rec->info)
		burgner_isofs_rrip_tf_fill_951_from_time (&impl->recording_date, rec->info->modify_time);

	impl->file_flags = __BURGNER_ISOFS_IMPL_DIRECTORY_RECORD_FLAGS_DIRECTORY;

	burgner_isofs_set_723 (impl->volume_sequence_number, 1);

	if (rec->id) {
		l = strlen (rec->id);
		memcpy (buf, rec->id, l);
		buf += l;
	} else if (rec->info->identifier) {
		l = strlen (rec->info->identifier);
		memcpy (buf, rec->info->identifier, l);
		buf += l;
	} else {
		l = 0;
	}

	burgner_isofs_set_711 (&impl->len_fi, l);

	if (!(l & 1)) {
		*((guchar *) buf) = 0;
	}
}

static const gchar *
_burgner_isofs_directory_subdirectory_record_get_identifier (BurgnerIsofsDirectoryRecord *rec)
{
	BurgnerIsofsDirectorySubdirectoryRecord *sr = BURGNER_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD (rec);
	g_return_val_if_fail (sr->id || sr->info, NULL);
	return sr->id ? sr->id : sr->info->identifier;
}

static void
class_init (BurgnerIsofsDirectorySubdirectoryRecordClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsDirectoryRecordClass *dir_rec_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	dir_rec_class = (BurgnerIsofsDirectoryRecordClass *) klass;
	dir_rec_class->get_susp_ent = _bidsr_get_susp_ent;
	dir_rec_class->get_ce_ent = _bidsr_get_ce_ent;
	dir_rec_class->get_base_size = _bidsr_get_base_size;
	dir_rec_class->write_base = _bidsr_write_base;
	dir_rec_class->get_identifier = _burgner_isofs_directory_subdirectory_record_get_identifier;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_DIRECTORY,
					 g_param_spec_object ("dir",
					            	      _("dir"),
							      _("Directory"),
							      BURGNER_TYPE_ISOFS_DIRECTORY,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_INFO,
					 g_param_spec_pointer ("info",
					            	       _("info"),
							       _("A BurgnerIsofsFileInfo"),
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_ID,
					 g_param_spec_string ("id",
					                       _("id"),
							       _("A string to use as identifier"),
							       NULL,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WITH_RRIP,
					 g_param_spec_boolean ("with_rrip",
					            	       _("with_rrip"),
							       _("With RockRidge"),
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_IS_RELOCATED,
					 g_param_spec_boolean ("is_relocated",
					                       _("is_relocated"),
							       _("Is Relocated"),
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_IS_CHILD_LINK,
					 g_param_spec_boolean ("is_child_link",
					                       _("is_child_link"),
							       _("Is Child Link"),
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_FILE,
					 g_param_spec_object ("file",
					                      _("file"),
							      _("File"),
							      BURGNER_TYPE_ISOFS_ITEM,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsDirectorySubdirectoryRecord *self)
{
	self->directory = NULL;
	self->file = NULL;
	self->info = NULL;
	self->with_rrip = 1;
	self->is_relocated = self->is_child_link = 0;
}

GType
burgner_isofs_directory_subdirectory_record_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsDirectorySubdirectoryRecordClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsDirectorySubdirectoryRecord),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_DIRECTORY_RECORD, "BurgnerIsofsDirectorySubdirectoryRecord", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_directory_subdirectory_record_new (void)
{
	BurgnerIsofsDirectorySubdirectoryRecord *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_DIRECTORY_SUBDIRECTORY_RECORD, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_directory_subdirectory_record_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsDirectorySubdirectoryRecord");
}

