/*
 * BurgnerIsofsDirectory - BurgnerIsofsDirectory object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <string.h>
#include <stdio.h>
#include <glib-object.h>
#include "burgner-isofs-susp-continuation-area.h"
#include "burgner-isofs-rrip-px.h"
#include "burgner-isofs-rrip-nm.h"
#include "burgner-isofs-rrip-tf.h"
#include "burgner-isofs-tree.h"
#include "burgner-isofs-extent.h"
#include "burgner-isofs-directory-dot-record.h"
#include "burgner-isofs-directory.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_INFO,
	PROP_WITH_RRIP,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsDirectory *self = BURGNER_ISOFS_DIRECTORY (object);

	if (self->info) {
		burgner_isofs_file_info_unref (self->info);
		self->info = NULL;
	}

	if (self->records) {
		g_slist_foreach (self->records, (GFunc) g_object_unref, NULL);
		g_slist_free (self->records);
		self->records = NULL;
	}

	if (self->ca) {
		g_slist_foreach (self->ca, (GFunc) g_object_unref, NULL);
		g_slist_free (self->ca);
		self->ca = NULL;
	}

	if (self->ce) {
		g_object_unref (self->ce);
		self->ce = NULL;
	}

	if (self->records_bytes) {
		g_slist_free (self->records_bytes);
		self->records_bytes = NULL;
	}

	self->with_rrip = 1;
	self->fixed = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectory *self = BURGNER_ISOFS_DIRECTORY (object);
	switch (prop_id) {
		case PROP_WITH_RRIP:
			g_return_if_fail (self->fixed == 0);
			self->with_rrip = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_rrip");
			break;
		case PROP_INFO:
			g_return_if_fail (self->fixed == 0);
			if (self->info)
				burgner_isofs_file_info_unref (self->info);
			self->info = (BurgnerIsofsFileInfo *) g_value_get_pointer (value);
			if (self->info)
				self->info = burgner_isofs_file_info_ref (self->info);
			g_object_notify (object, "info");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsDirectory *self = BURGNER_ISOFS_DIRECTORY (object);

	switch (prop_id) {
		case PROP_WITH_RRIP:
			g_value_set_boolean (value, self->with_rrip?TRUE:FALSE);
			break;
		case PROP_INFO:
			g_value_set_pointer (value, self->info);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

#if 0
static void
_bid_debug_print_rec_slist (GSList *l)
{
	GSList *it;
	guint32 n_b;
	for (it = l; it != NULL; it = it->next) {
		n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data), NULL, &n_b);
		fprintf (stderr, "-- %s: %u\n", G_OBJECT_TYPE_NAME (it->data), n_b);
	}
}
#endif

static gboolean _bid_perform_sec_debug = FALSE;

static void
_bid_perform_sec (BurgnerIsofsDirectory *dir,
		  GSList *sec,
		  guint bonus)
{
	GSList *it, *sec_bytes;
	guint nbonus, t = 0;
	guint32 n_b;

	for (it = sec; it != NULL; it = it->next) {
		nbonus = burgner_isofs_directory_record_bonus_size (BURGNER_ISOFS_DIRECTORY_RECORD (it->data),
								    bonus);
		g_assert (nbonus <= bonus);
		if (_bid_perform_sec_debug) {
			fprintf (stderr, "bb %s: %u\n", G_OBJECT_TYPE_NAME (it->data), bonus-nbonus);
		}
		bonus = nbonus;

		if (burgner_isofs_directory_record_need_continuation_area (BURGNER_ISOFS_DIRECTORY_RECORD (it->data))) {
			BurgnerIsofsItem *ca = burgner_isofs_susp_continuation_area_new ();
			burgner_isofs_directory_record_set_continuation_area (BURGNER_ISOFS_DIRECTORY_RECORD (it->data),
									      ca);
			dir->ca = g_slist_prepend (dir->ca, ca);
		}
	}

	sec_bytes = NULL;
	for (it = sec; it != NULL; it = it->next) {
		n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data),
					     NULL, &n_b);
		sec_bytes = g_slist_prepend (sec_bytes, GUINT_TO_POINTER (n_b));
		t += n_b;
	}
	g_assert (t <= 2048);
	dir->records_bytes = g_slist_concat (dir->records_bytes, sec_bytes);

	g_slist_free (sec);
}

static void
_bid_create_node_susp_ent (BurgnerIsofsDirectory *dir)
{
	BurgnerIsofsItem *ent;
	/*gchar *id;*/

	g_return_if_fail (dir->info != NULL);
	g_return_if_fail (dir->ce != NULL);

	ent = burgner_isofs_rrip_px_new ();
	g_object_set (ent,
		      "info", dir->info,
		      NULL);
	burgner_isofs_susp_continuation_area_prepend (BURGNER_ISOFS_SUSP_CONTINUATION_AREA (dir->ce), ent);
	g_object_unref (ent);

	if (dir->info->name) {
		GSList *l = burgner_isofs_rrip_nm_new_slist_from_id (dir->info->name);
		burgner_isofs_susp_continuation_area_prepend_slist_take_ownership (BURGNER_ISOFS_SUSP_CONTINUATION_AREA (dir->ce), l);
	}

	if (dir->info->creation_time || dir->info->modify_time ||
			dir->info->access_time || dir->info->attributes_time ||
			dir->info->backup_time || dir->info->expiration_time ||
			dir->info->effective_time) {
		ent = burgner_isofs_rrip_tf_new ();
		g_object_set (ent, "info", dir->info, NULL);
		burgner_isofs_susp_continuation_area_prepend (BURGNER_ISOFS_SUSP_CONTINUATION_AREA (dir->ce), ent);
		g_object_unref (ent);
	}
}

static gint
_bid_records_cmp (gconstpointer a,
		  gconstpointer b,
		  BurgnerIsofsDirectory *dir)
{
	if (BURGNER_ISOFS_DIRECTORY_GET_CLASS (dir)->records_cmp)
		return BURGNER_ISOFS_DIRECTORY_GET_CLASS (dir)->records_cmp (dir,
									     BURGNER_ISOFS_DIRECTORY_RECORD (a),
									     BURGNER_ISOFS_DIRECTORY_RECORD (b));
	g_warning ("BurgnerIsofsDirectory gives no records_cmp function !");
	return 0;
}

static void
_bid_realize (BurgnerIsofsDirectory *dir)
{
	GSList *it, *sec = NULL;
	guint32 sec_b = 0, sum, /*bonus,*/ n_b;

	g_return_if_fail (dir->fixed == 0);
	g_return_if_fail (dir->ca == NULL);
	g_return_if_fail (dir->ce != NULL);
	g_return_if_fail (dir->records_bytes == NULL);
	g_return_if_fail (dir->records != NULL);

	dir->records = g_slist_sort_with_data (dir->records,
				(GCompareDataFunc) _bid_records_cmp, dir);

//	dir->records = g_slist_reverse (dir->records);

	for (it = dir->records; it != NULL; it = it->next) {
		n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data),
					     NULL, &n_b);
		g_assert (n_b <= 2048);
		sum = sec_b + n_b;
		if (sum > 2048) {
			_bid_perform_sec (dir, sec, 2048 - sec_b);
			sec_b = n_b;
			sec = g_slist_prepend (NULL, it->data);
		} else if (sum == 2048) {
			sec = g_slist_prepend (sec, it->data);
			_bid_perform_sec (dir, sec, 0);
			sec = NULL;
			sec_b = 0;
		} else {
			sec_b = sum;
			sec = g_slist_prepend (sec, it->data);
		}

		if (sec && !it->next) {
			_bid_perform_sec (dir, sec, 2048 - sec_b);
		}
	}

//	dir->records = g_slist_reverse (dir->records);

	if (dir->with_rrip) {
/*		dir->ce = burgner_isofs_susp_continuation_area_new (void);*/
		_bid_create_node_susp_ent (dir);
	}

	dir->fixed = 1;
}

static void
_bid_get_size (BurgnerIsofsItem *item,
	       guint32 *n_sectors,
	       guint32 *n_bytes)
{
	BurgnerIsofsDirectory *dir = BURGNER_ISOFS_DIRECTORY (item);
	GSList *it;
	guint32 b = 0, s = 0, n_b, sum;

	if (!dir->fixed) {
		_bid_realize (dir);
	}

	{
		guint a = g_slist_length (dir->records),
			b = g_slist_length (dir->records_bytes);
		g_return_if_fail (a == b);
	}
	for (it = dir->records_bytes; it != NULL; it = it->next) {
		n_b = GPOINTER_TO_UINT (it->data);
		sum = b + n_b;
		if (sum > 2048) {
			s += 1;
			b = n_b;
		} else if (sum == 2048) {
			s += 1;
			b = 0;
		} else {
			b = sum;
		}
	}

	if (b) {
		s += 1;
		b = 0;
	}

	if (n_sectors)
		*n_sectors = s;
	if (n_bytes)
		*n_bytes = b;
}

static void
_bid_prepare (BurgnerIsofsItem *item,
	      BurgnerIsofsMapTable *table,
	      guint32 sector_n,
	      guint32 byte_o,
	      guint32 last_sector_n,
	      guint32 last_byte_o)
{
	BurgnerIsofsDirectory *dir = BURGNER_ISOFS_DIRECTORY (item);
	GSList *it, *it_i;
	guint32 s = sector_n, b = byte_o, sum, n_b, b_s, b_b;

	g_return_if_fail (dir->fixed == 1);
	g_return_if_fail (byte_o == 0);

	{
		guint a = g_slist_length (dir->records),
			b = g_slist_length (dir->records_bytes);
		g_return_if_fail (a == b);
	}
	for (it = dir->records_bytes, it_i = dir->records;
	     it != NULL;
	     it = it->next, it_i = it_i->next) {
		n_b = GPOINTER_TO_UINT (it->data);
		sum = b + n_b;
		if (sum > 2048) {
			s += 1;
			b = 0;
		}
		g_return_if_fail (burgner_isofs_map_table_get_addr (table,
BURGNER_ISOFS_ITEM (it_i->data), NULL, NULL) == FALSE);
		burgner_isofs_map_table_set (table,
					     BURGNER_ISOFS_ITEM (it_i->data),
					     s, b);
		b_s = s;
		b_b = b;
		b += n_b;
		g_return_if_fail (b <= 2048);
		if (b == 2048) {
			s += 1;
			b = 0;
		}
		burgner_isofs_item_prepare (BURGNER_ISOFS_ITEM (it_i->data),
					    table,
					    b_s, b_b,
					    s, b);
	}
	if (b) {
		s += 1;
		b = 0;
	}

	g_return_if_fail (s == last_sector_n);
	g_return_if_fail (b == last_byte_o);
}

static void
_bid_write_sync (BurgnerIsofsItem *item,
		 BurgnerIsofsMapTable *table,
		 gpointer buf, guint32 size)
{
	BurgnerIsofsDirectory *dir = BURGNER_ISOFS_DIRECTORY (item);
	GSList *it, *it_i;
	guint n_b, sum, b;

	{
		guint a = g_slist_length (dir->records),
			b = g_slist_length (dir->records_bytes);
		g_return_if_fail (a == b);
	}

	b = 0;
	for (it = dir->records_bytes, it_i = dir->records;
	     it != NULL;
	     it = it->next, it_i = it_i->next) {
		n_b = GPOINTER_TO_UINT (it->data);
		sum = b + n_b;
		if (sum > 2048) {
			memset (buf, 0, 2048 - b);
			buf += 2048 - b;
			size -= 2048 - b;
			b = 0;
		}
		burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it_i->data),
					       table, buf, n_b);
		buf += n_b;
		size -= n_b;

		b += n_b;
		g_return_if_fail (b <= 2048);
		if (b == 2048) {
			b = 0;
		}
	}
	if (b) {
		memset (buf, 0, 2048 - b);
		buf += 2048 - b;
		size -= 2048 - b;
		b = 0;
	}

	g_assert (size == 0);
}

static guint32
_bid_extent_get_size (BurgnerIsofsExtent *extent)
{
	BurgnerIsofsDirectory *dir = BURGNER_ISOFS_DIRECTORY (extent);
	GSList *it;
	guint32 n_b, sum, s = 0, b = 0;

	g_return_val_if_fail (dir->fixed == 1, 0);

	it = NULL;
	for (it = dir->records_bytes; it != NULL; it = it->next) {
		n_b = GPOINTER_TO_UINT (it->data);
		sum = b + n_b;
		if (sum > 2048) {
			s += 1;
			b = n_b;
		} else if (sum == 2048) {
			s += 1;
			b = 0;
		} else {
			b = sum;
		}
	}
	if (b) {
		s += 1;
		b = 0;
	}

	return s*2048;
}

static gint
_bid_impl_records_cmp (BurgnerIsofsDirectory *dir,
		       BurgnerIsofsDirectoryRecord *a,
		       BurgnerIsofsDirectoryRecord *b)
{
	const gchar *ai = burgner_isofs_directory_record_get_identifier (a),
		    *bi = burgner_isofs_directory_record_get_identifier (b);

#if 0
	g_return_val_if_fail (ai != NULL, -1);
	g_return_val_if_fail (bi != NULL, 1);
#endif
	if (ai && bi)
		return strcmp (ai, bi);

	if (ai)
		return 1;
	if (bi)
		return -1;

	if (BURGNER_IS_ISOFS_DIRECTORY_DOT_RECORD (a))
		return -1;

	return (BURGNER_IS_ISOFS_DIRECTORY_DOT_RECORD (b));
}

static GList *
_burgner_isofs_directory_get_free_children (BurgnerIsofsItem *item)
{
	BurgnerIsofsDirectory *self = BURGNER_ISOFS_DIRECTORY (item);
	GList *l = NULL;
	GSList *it;

	if (!self->fixed) {
		_bid_realize (self);
	}

	if (!self->with_rrip)
		return NULL;

	for (it = self->ca; it != NULL; it = it->next) {
		l = g_list_prepend (l, it->data);
	}

	if (self->ce)
		l = g_list_prepend (l, self->ce);

	return l;
}

static void
class_init (BurgnerIsofsDirectoryClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bid_get_size;
	item_class->prepare = _bid_prepare;
	item_class->write_sync = _bid_write_sync;
	item_class->get_free_children = _burgner_isofs_directory_get_free_children;

	/* Virtual Functions */
	klass->records_cmp = _bid_impl_records_cmp;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_WITH_RRIP,
					 g_param_spec_boolean ("with_rrip",
					            	       _("with_rrip"),
							       _("With RRIP"),
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_INFO,
					 g_param_spec_pointer ("info",
					            	       _("info"),
							       _("Info"),
							       G_PARAM_READABLE | G_PARAM_WRITABLE));

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsDirectory *self)
{
	self->records = self->ca =
	    self->records_bytes;
	self->info = NULL;
	self->fixed = 0;
	self->with_rrip = 1;

	self->ce = burgner_isofs_susp_continuation_area_new ();
}

static void
_bid_extent_iface_init (BurgnerIsofsExtentIface *iface)
{
	iface->get_size = _bid_extent_get_size;
}


GType
burgner_isofs_directory_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (BurgnerIsofsDirectoryClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsDirectory),
			0,
			(GInstanceInitFunc) init,
		};

		static const GInterfaceInfo extent_info = {
			(GInterfaceInitFunc) _bid_extent_iface_init,
			NULL,
			NULL
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsDirectory", &info, 0);
		g_type_add_interface_static (type, BURGNER_TYPE_ISOFS_EXTENT, &extent_info);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_directory_new (void)
{
	BurgnerIsofsDirectory *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_DIRECTORY, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_directory_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsDirectory");
}

static void
burgner_isofs_directory_append_record_real (BurgnerIsofsDirectory *dir,
				            BurgnerIsofsItem *rec)
{
	g_object_ref (rec);
	dir->records = g_slist_prepend (dir->records, rec);
}


void
burgner_isofs_directory_append_record (BurgnerIsofsDirectory *dir,
				       BurgnerIsofsItem *rec)
{
	g_return_if_fail (dir != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY (dir));
	g_return_if_fail (dir->fixed == 0);
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY_RECORD (rec));

	burgner_isofs_directory_append_record_real (dir, rec);
}

BurgnerIsofsItem *
burgner_isofs_directory_get_continuation_area (BurgnerIsofsDirectory *dir)
{
	g_return_val_if_fail (dir != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_DIRECTORY (dir), NULL);
	g_return_val_if_fail (dir->ce != NULL, NULL);

	return dir->ce;
}

static void
burgner_isofs_directory_append_records_valist_real (BurgnerIsofsDirectory *dir,
						    BurgnerIsofsItem *first,
						    va_list args)
{
	BurgnerIsofsItem *i = first;

	do {
		g_return_if_fail (BURGNER_IS_ISOFS_ITEM (i));
		burgner_isofs_directory_append_record_real (dir, i);
	} while ((i = va_arg (args, BurgnerIsofsItem*)));
}

void
burgner_isofs_directory_append_records_valist (BurgnerIsofsDirectory *dir,
					       BurgnerIsofsItem *first,
					       va_list args)
{
	g_return_if_fail (dir != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY (dir));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (first));

	burgner_isofs_directory_append_records_valist_real (dir, first, args);
}

void
burgner_isofs_directory_append_records (BurgnerIsofsDirectory *dir,
					BurgnerIsofsItem *first,
					...)
{
	va_list args;

	g_return_if_fail (dir != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_DIRECTORY (dir));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (first));

	va_start (args, first);
	burgner_isofs_directory_append_records_valist_real (dir, first, args);
	va_end (args);
}

