/*
 * BurgnerIsofsEltoritoBootCatalog - BurgnerIsofsEltoritoBootCatalog object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-eltorito-boot-catalog-entry.h"
#include "burgner-isofs-eltorito-boot-catalog.h"


static BurgnerIsofsItemClass *parent_class = NULL;

static void
burgner_isofs_eltorito_boot_catalog_finalize (GObject *object)
{
	BurgnerIsofsEltoritoBootCatalog *self = BURGNER_ISOFS_ELTORITO_BOOT_CATALOG (object);

	if (self->entries) {
		burgner_isofs_item_g_slist_unref (self->entries);
		self->entries = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_biebc_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	BurgnerIsofsEltoritoBootCatalog *bc = BURGNER_ISOFS_ELTORITO_BOOT_CATALOG (item);
	guint32 n, s;

	n = g_slist_length (bc->entries) * 32;
	s = n >> 11;

	if (n & 0x7ff)
		s += 1;

	if (n_sectors)
		*n_sectors = s;
	if (n_bytes)
		*n_bytes = 0;
}

static void
_biebc_prepare (BurgnerIsofsItem *item,
		BurgnerIsofsMapTable *table,
		guint32 sector_n,
		guint32 byte_o,
		guint32 last_sector_n,
		guint32 last_byte_o)
{
	BurgnerIsofsEltoritoBootCatalog *bc = BURGNER_ISOFS_ELTORITO_BOOT_CATALOG (item);
	GSList *it;
	guint32 s, b;

	g_return_if_fail (last_byte_o == 0);
	g_return_if_fail (byte_o == 0);

	s = sector_n;
	b = 0;
	for (it = bc->entries; it != NULL; it = it->next) {
		BurgnerIsofsItem *c = BURGNER_ISOFS_ITEM (it->data);
		guint32 os, ob;

		burgner_isofs_map_table_set (table, c, s, b);

		os = s;
		ob = b;
		b += 32;
		if (b == 2048) {
			s += 1;
			b = 0;
		}

		burgner_isofs_item_prepare (c, table, os, ob, s, b);
	}

#if 0
	if (last_byte_o == 2048) {
		last_sector_n += 1;
		last_byte_o = 0;
	}
#endif

	if (b)
		s++;

	g_return_if_fail (s == last_sector_n);
}

static void
_biebc_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	BurgnerIsofsEltoritoBootCatalog *bc = BURGNER_ISOFS_ELTORITO_BOOT_CATALOG (item);
	GSList *it;
	guint n, s;

	g_return_if_fail ((size & 0x7ff) == 0);

	if (n & 0x7ff)
		s += 1;

	n = g_slist_length (bc->entries) * 32;
	s = n >> 11;

	if (n & 0x7ff)
		s += 1;

	g_return_if_fail (size == (s << 11));

	for (it = bc->entries; it != NULL; it = it->next) {
		burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it->data), table, buf, 32);
		buf += 32;
	}
}

static void
burgner_isofs_eltorito_boot_catalog_class_init (BurgnerIsofsEltoritoBootCatalogClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = burgner_isofs_eltorito_boot_catalog_finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _biebc_get_size;
	item_class->prepare = _biebc_prepare;
	item_class->write_sync = _biebc_write_sync;

	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_isofs_eltorito_boot_catalog_init (BurgnerIsofsEltoritoBootCatalog *self)
{
	self->entries = NULL;
}

GType
burgner_isofs_eltorito_boot_catalog_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsEltoritoBootCatalogClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_isofs_eltorito_boot_catalog_class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsEltoritoBootCatalog),
			0,
			(GInstanceInitFunc) burgner_isofs_eltorito_boot_catalog_init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsEltoritoBootCatalog", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_eltorito_boot_catalog_new (void)
{
	BurgnerIsofsEltoritoBootCatalog *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_ELTORITO_BOOT_CATALOG, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_eltorito_boot_catalog_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsEltoritoBootCatalog");
}

static void
_biebc_append_entry_real (BurgnerIsofsEltoritoBootCatalog *bc,
			  BurgnerIsofsItem *ent)
{
	bc->entries = g_slist_append (bc->entries, g_object_ref (ent));
}

void
burgner_isofs_eltorito_boot_catalog_append_entry (BurgnerIsofsEltoritoBootCatalog *bc,
						  BurgnerIsofsItem *entry)
{
	g_return_if_fail (bc != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG (bc));
	g_return_if_fail (entry != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG_ENTRY (entry));

	_biebc_append_entry_real (bc, entry);
}

void
burgner_isofs_eltorito_boot_catalog_append_entries (BurgnerIsofsEltoritoBootCatalog *bc,
						    BurgnerIsofsItem *first,
						    ...)
{
	BurgnerIsofsItem *item = first;
	va_list args;

	g_return_if_fail (bc != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG (bc));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG_ENTRY (first));

	va_start (args, first);
	do {
		_biebc_append_entry_real (bc, item);
	} while ((item = va_arg (args, BurgnerIsofsItem*)));
	va_end (args);
}

void
burgner_isofs_eltorito_boot_catalog_append_entries_valist (BurgnerIsofsEltoritoBootCatalog *bc,
						           BurgnerIsofsItem *first,
							   va_list args)
{
	BurgnerIsofsItem *item;

	g_return_if_fail (bc != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG (bc));
	g_return_if_fail (first != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ELTORITO_BOOT_CATALOG_ENTRY (first));

	do {
		_biebc_append_entry_real (bc, item);
	} while ((item = va_arg (args, BurgnerIsofsItem*)));
}

