/*
 * BurgnerIsofsFile - BurgnerIsofsFile object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <glib-object.h>
#include <libburgnerutils/burgner-thread.h>
#include <libburgnerutils/g-uri-string.c>
#include "burgner-isofs-file.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_INFO,
	BURGNER_ISOFS_FILE_LAST_ARG
};


G_LOCK_DEFINE_STATIC (_burgner_isofs_file_thread_list);
static GList *_burgner_isofs_file_thread_list = NULL;

static BurgnerThread *
_burgner_isofs_file_get_thread (GError **error)
{
	BurgnerThread *t = NULL;

	G_LOCK (_burgner_isofs_file_thread_list);
	if (_burgner_isofs_file_thread_list) {
		t = _burgner_isofs_file_thread_list->data;
		_burgner_isofs_file_thread_list = g_list_delete_link (_burgner_isofs_file_thread_list, _burgner_isofs_file_thread_list);
	}
	G_UNLOCK (_burgner_isofs_file_thread_list);

	if (!t)
		t = burgner_thread_new (error);

	return t;
}

static void
_burgner_isofs_file_release_thread (BurgnerThread *thread)
{
	G_LOCK (_burgner_isofs_file_thread_list);
	_burgner_isofs_file_thread_list = g_list_prepend (_burgner_isofs_file_thread_list, thread);
	G_UNLOCK (_burgner_isofs_file_thread_list);
}

static void
_burgner_isofs_file_finalize (GObject *object)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (object);

	g_return_if_fail (self->write_data_list == NULL);

	if (self->info) {
		burgner_isofs_file_info_unref (self->info);
		self->info = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_isofs_file_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (object);

	switch (prop_id) {
		case PROP_INFO:
			if (self->info)
				burgner_isofs_file_info_unref (self->info);
			self->info = (BurgnerIsofsFileInfo *) g_value_get_pointer (value);
			if (self->info)
				self->info = burgner_isofs_file_info_ref (self->info);
			g_object_notify (object, "info");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_isofs_file_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (object);

	switch (prop_id) {
		case PROP_INFO:
			g_value_set_pointer (value, self->info);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static BurgnerIsofsItemFlags
_burgner_isofs_file_get_flags (BurgnerIsofsItem *item)
{
	return BURGNER_ISOFS_ITEM_WRITE_ASYNC | BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED;
}

static void
_burgner_isofs_file_get_size (BurgnerIsofsItem *item,
			      guint32 *n_s,
			      guint32 *n_b)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (item);

	g_return_if_fail (self->info != NULL);

	if (n_s)
		*n_s = self->info->size >> 11;
	if (n_b)
		*n_b = self->info->size & 0x7ff;
}

static void
_burgner_isofs_file_prepare (BurgnerIsofsItem *item,
			     BurgnerIsofsMapTable *table,
			     guint32 sector_n,
			     guint32 byte_o,
			     guint32 last_sector_n,
			     guint32 last_byte_o)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (item);
	guint32 s;

	g_return_if_fail (self->info != NULL);

	s = ((last_sector_n - sector_n) << 11) + last_byte_o - byte_o;
	g_return_if_fail (s == self->info->size);
}


typedef struct {
	BurgnerIsofsFile *self;

	guint idle_id;
	BurgnerThread *thread;

	GIOChannel *channel;
	GError *error;

	guint32 remaining;

	/* Opening data */
	gchar *file_name;

	BurgnerIsofsItemWriteAsyncInitCallback init_cb;
	gpointer init_cb_data;

	/* Reading data */
	gpointer buf;
	guint32 buf_size;
	guint32 bytes_read;

	BurgnerIsofsItemWriteAsyncCallback cb;
	gpointer cb_data;

	guint opening : 1;
	guint reading : 1;
} _BurgnerIsofsFile_WriteData;

static gboolean
_burgner_isofs_file_open_thread_func (gpointer thread_data)
{
	_BurgnerIsofsFile_WriteData *d = (_BurgnerIsofsFile_WriteData *) thread_data;
	GIOFlags f;

	g_return_val_if_fail (d->file_name != NULL, FALSE);
	g_return_val_if_fail (d->opening == 1, FALSE);
	g_return_val_if_fail (d->error == NULL, FALSE);

	if (!(d->channel = g_io_channel_new_file (d->file_name, "r", &d->error))) {
		return FALSE;
	}

	if ((f = g_io_channel_get_flags (d->channel)) & G_IO_FLAG_NONBLOCK) {
		if (g_io_channel_set_flags (d->channel, f & ~G_IO_FLAG_NONBLOCK, &d->error) != G_IO_STATUS_NORMAL) {
			g_io_channel_unref (d->channel);
			d->channel = NULL;
			return FALSE;
		}
	}

	if (g_io_channel_set_encoding (d->channel, NULL, &d->error) != G_IO_STATUS_NORMAL) {
		g_io_channel_unref (d->channel);
		d->channel = NULL;
		return FALSE;
	}

	g_io_channel_set_buffered (d->channel, FALSE);

	return FALSE;
}

static gboolean
_burgner_isofs_file_open_thread_cb_func (BurgnerThread *thread,
					 BurgnerThreadCallId call_id,
					 BurgnerThreadFunction function,
					 gpointer function_data,
					 gpointer user_data)
{
	_BurgnerIsofsFile_WriteData *d = (_BurgnerIsofsFile_WriteData *) user_data;

	g_assert (d->opening == 1);

	g_free (d->file_name);
	d->file_name = NULL;
	d->opening = 0;

	if (d->error) {
		g_return_val_if_fail (d->channel == NULL, FALSE);
		d->self->write_data_list = g_list_remove (d->self->write_data_list, d);

		if (d->init_cb)
			d->init_cb (BURGNER_ISOFS_ITEM (d->self), NULL, d->error, d->init_cb_data);

		_burgner_isofs_file_release_thread (d->thread);
		g_error_free (d->error);
		g_free (d);
	} else {
		BurgnerIsofsItemWriteAsyncInitCallback cb = d->init_cb;
		gpointer cb_data = d->init_cb_data;

		g_return_val_if_fail (d->channel != NULL, FALSE);
		d->opening = 0;
		d->init_cb = NULL;
		d->init_cb_data = NULL;

		d->remaining = d->self->info->size;

		if (cb)
			cb (BURGNER_ISOFS_ITEM (d->self), d, NULL, cb_data);
	}

	return FALSE;
}

static gboolean
_burgner_isofs_file_write_async_init_idle_func (gpointer user_data)
{
	_BurgnerIsofsFile_WriteData *d = (_BurgnerIsofsFile_WriteData *) user_data;
	GError *error = NULL;

	d->idle_id = 0;

	g_return_val_if_fail (d->self->info != NULL, FALSE);
	g_return_val_if_fail (d->self->info->uri != NULL, FALSE);
	g_return_val_if_fail (g_uri_get_path (d->self->info->uri) != NULL, FALSE);

	d->file_name = g_uri_string_compress (g_uri_get_path (d->self->info->uri));
	g_return_val_if_fail (d->file_name != NULL, FALSE);

	if (!(d->thread = _burgner_isofs_file_get_thread (&error))) {
		d->self->write_data_list = g_list_remove (d->self->write_data_list, d);
		g_free (d->file_name);
		if (d->init_cb)
			d->init_cb (BURGNER_ISOFS_ITEM (d->self), NULL, error, d->init_cb_data);
		g_free (d);
	} else {
		d->opening = 1;
		burgner_thread_call (d->thread,
				     _burgner_isofs_file_open_thread_func, d,
				     _burgner_isofs_file_open_thread_cb_func, d);
	}

	return FALSE;
}

static void
_burgner_isofs_file_write_async_init (BurgnerIsofsItem *item,
				      BurgnerIsofsMapTable *table,
				      BurgnerIsofsItemWriteAsyncInitCallback cb,
				      gpointer cb_data)
{
	BurgnerIsofsFile *self = BURGNER_ISOFS_FILE (item);
	_BurgnerIsofsFile_WriteData *d;

	g_return_if_fail (self->info != NULL);
	g_return_if_fail (self->info->uri != NULL);
	g_return_if_fail (g_uri_get_path (self->info->uri) != NULL);

	d = (_BurgnerIsofsFile_WriteData *) g_new0 (_BurgnerIsofsFile_WriteData, 1);
	d->self = self;
	d->init_cb = cb;
	d->init_cb_data = cb_data;

	d->idle_id = g_idle_add (_burgner_isofs_file_write_async_init_idle_func, d);

	self->write_data_list = g_list_prepend (self->write_data_list, d);
}

static gboolean
_burgner_isofs_file_write_async_thread_func (gpointer thread_data)
{
	_BurgnerIsofsFile_WriteData *d = (_BurgnerIsofsFile_WriteData *) thread_data;

	g_return_val_if_fail (d->reading == 1, FALSE);
	g_return_val_if_fail (d->channel != NULL, FALSE);

	g_io_channel_read_chars (d->channel,
				 d->buf, MIN (d->remaining, d->buf_size), &d->bytes_read,
				 &d->error);
	return FALSE;
}

static gboolean
_burgner_isofs_file_write_async_thread_cb_func (BurgnerThread *thread,
						BurgnerThreadCallId call_id,
						BurgnerThreadFunction function,
						gpointer function_data,
						gpointer user_data)
{
	_BurgnerIsofsFile_WriteData *d = (_BurgnerIsofsFile_WriteData *) user_data;
	BurgnerIsofsItemWriteAsyncCallback cb = d->cb;
	gpointer cb_data = d->cb_data, buf = d->buf;
	guint32 buf_size = d->buf_size, bytes_read = d->bytes_read, remaining = d->remaining;
	GError *error = d->error;

	d->cb = NULL;
	d->cb_data = NULL;
	d->buf = NULL;
	d->buf_size = d->bytes_read = 0;
	d->error = NULL;
	d->reading = 0;

	if (d->error) {
		if (cb)
			cb (BURGNER_ISOFS_ITEM (d->self), d,
			    buf, buf_size, 0,
			    error, (remaining == 0),
			    cb_data);
		g_error_free (error);
	} else {
		BurgnerIsofsFile *self = d->self;
		d->remaining -= bytes_read;
		remaining = d->remaining;

		if (remaining == 0) {
			g_io_channel_unref (d->channel);
			_burgner_isofs_file_release_thread (d->thread);
			g_free (d);
			self->write_data_list = g_list_remove (self->write_data_list, d);
		}

		if (cb)
			cb (BURGNER_ISOFS_ITEM (self), d,
			    buf, buf_size, bytes_read,
			    NULL, (remaining == 0),
			    cb_data);
	}

	return FALSE;
}

static void
_burgner_isofs_file_write_async (BurgnerIsofsItem *item,
				 BurgnerIsofsMapTable *table,
				 gpointer write_async_data,
				 gpointer buf,
				 guint32 bytes_requested,
				 BurgnerIsofsItemWriteAsyncCallback cb,
				 gpointer cb_data)
{
	BurgnerIsofsFile *self;
	_BurgnerIsofsFile_WriteData *d;

	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_FILE (item));
	g_return_if_fail (write_async_data != NULL);

	self = BURGNER_ISOFS_FILE (item);
	d = (_BurgnerIsofsFile_WriteData *) write_async_data;

	g_return_if_fail (g_list_index (self->write_data_list, write_async_data) >= 0);
	g_return_if_fail (d->opening == 0);
	g_return_if_fail (d->reading == 0);
	g_return_if_fail (d->channel != NULL);

	d->buf = buf; d->buf_size = bytes_requested;
	d->cb = cb; d->cb_data = cb_data;
	d->reading = 1;

	burgner_thread_call (d->thread,
			     _burgner_isofs_file_write_async_thread_func, d,
			     _burgner_isofs_file_write_async_thread_cb_func, d);
}

static void
_burgner_isofs_file_write_async_cancel (BurgnerIsofsItem *item,
					gpointer write_async_data)
{
	BurgnerIsofsFile *self;
	_BurgnerIsofsFile_WriteData *d;

	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_FILE (item));
	g_return_if_fail (write_async_data != NULL);

	self = BURGNER_ISOFS_FILE (item);
	d = (_BurgnerIsofsFile_WriteData *) write_async_data;

	g_return_if_fail (g_list_index (self->write_data_list, write_async_data) >= 0);
	g_return_if_fail (d->opening == 0);
	g_return_if_fail (d->reading == 0);

	if (d->idle_id) {
		g_source_remove (d->idle_id);
		d->idle_id = 0;
	}

	if (d->channel) {
		g_io_channel_unref (d->channel);
		d->channel = NULL;
	}

	if (d->thread) {
		_burgner_isofs_file_release_thread (d->thread);
		d->thread = NULL;
	}

	if (d->file_name) {
		g_free (d->file_name);
		d->file_name = NULL;
	}

	self->write_data_list = g_list_remove (self->write_data_list, d);
	g_free (d);
}

static void
_burgner_isofs_file_class_init (BurgnerIsofsFileClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_isofs_file_set_property;
	object_class->get_property = _burgner_isofs_file_get_property;
	object_class->finalize = _burgner_isofs_file_finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_flags = _burgner_isofs_file_get_flags;
	item_class->get_size = _burgner_isofs_file_get_size;
	item_class->prepare = _burgner_isofs_file_prepare;
	item_class->write_async_init = _burgner_isofs_file_write_async_init;
	item_class->write_async = _burgner_isofs_file_write_async;
	item_class->write_async_cancel = _burgner_isofs_file_write_async_cancel;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_INFO,
					 g_param_spec_pointer ("info",
					                _("info"),
							_("A BurgnerIsofsFileInfo"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_isofs_file_init (BurgnerIsofsFile *self)
{
	self->info = NULL;
	self->write_data_list = NULL;
}

GType
burgner_isofs_file_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsFileClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_isofs_file_class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsFile),
			0,
			(GInstanceInitFunc) _burgner_isofs_file_init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsFile", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_file_new (void)
{
	BurgnerIsofsFile *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_FILE, NULL);

	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_file_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsFile");
}

