/*
 * 
 *
 * This file is part of the libburgnerisofs library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include "burgner-isofs-identifier.h"

static struct {
	const gchar *long_ext;
	const gchar *short_ext;
	guint long_ext_len;
} long2short_ext[] = {
	{ ".tar.gz", "TGZ", 7 },
	{ ".tar.bz2", "TB2", 8 },
	{ ".ps.gz", "PSZ", 6 },
	{ NULL, NULL, 0 }
};

gchar *
burgner_isofs_file_identifier_from_name (const gchar *name, BurgnerIsofsIdentifierOptions options)
{
	gchar *name_part = NULL, *ext = NULL, *ret;
	guint name_len, c, name_part_len = 0, ext_len;

	g_return_val_if_fail (name != NULL, NULL);
	g_return_val_if_fail (*name != 0, NULL);

	name_len = strlen (name);

	if (!(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LONG_EXTENSIONS)) {
		for (c = 0; long2short_ext[c].long_ext_len != 0; c++) {
			if (name_len > long2short_ext[c].long_ext_len &&
			    strcmp (long2short_ext[c].long_ext, &name[name_len - long2short_ext[c].long_ext_len]) == 0) {
				name_part_len = name_len - long2short_ext[c].long_ext_len;
				name_part = g_strndup (name, name_part_len);

				ext = g_strdup (long2short_ext[c].short_ext);
				ext_len = strlen (ext);
			} else {
				continue;
			}

			break;
		}
	}

	if (!ext) {
		const gchar *rdot = strrchr (name, '.');

		if (rdot) {
			guint rdot_l = name_len - (1 + rdot - name);

			ext_len = rdot_l;

			if (rdot_l > 3) {
				if (options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LONG_EXTENSIONS) {
					ext = g_strndup (rdot + 1, MIN (29, rdot_l));
				} else {
					ext = g_strndup (rdot + 1, 3);
					ext_len = 3;
				}
			} else if (rdot_l > 0) {
				ext = g_strndup (rdot + 1, rdot_l);
			}

			name_part_len = rdot - name;
			name_part = g_strndup (name, name_part_len);
		}

		if (!ext) {
			if (!(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_NO_EXTENSION)) {
				ext = g_strdup ("EXT");
				ext_len = 3;
			}

			name_part = g_strdup (name);
			name_part_len = name_len;
		}
	}

	g_assert (name_part);

	if (name_part_len > 8 &&
	    !(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LONG_NAMES)) {
		name_part[8] = 0;
		name_part_len = 8;
	}

	if ((name_part_len + (ext ? ext_len + 1 : 0)) > 31) {
		name_part_len = 31 - (ext ? ext_len + 1 : 0);
		name_part[name_part_len] = 0;
	}

	for (c = 0; name_part[c] != 0; c++) {
		if ((name_part[c] >= '0' && name_part[c] <= '9') ||
		    (name_part[c] >= 'A' && name_part[c] <= 'Z') ||
		    name_part[c] == '_' ||
		    (name_part[c] == '.' && (options & BURGNER_ISOFS_IDENTIFIER_ALLOW_MULTIDOT)))
			continue;

		if (name_part[c] >= 'a' && name_part[c] <= 'z') {
			if (!(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LOWER_CASE))
				name_part[c] += 'A' - 'a';
			continue;
		}

		if (options & BURGNER_ISOFS_IDENTIFIER_RELAX_CHARACTER_SET)
			continue;

		name_part[c] = '_';
	}

	ret = g_strdup_printf ("%s%s%s", name_part, ext ? "." : "", ext ? ext : "");
	g_free (name_part);
	g_free (ext);

	return ret;
}

gchar *
burgner_isofs_directory_identifier_from_name (const gchar *name, BurgnerIsofsIdentifierOptions options)
{
	gchar *name_part = NULL;
	guint name_len, c;

	g_return_val_if_fail (name != NULL, NULL);
	g_return_val_if_fail (*name != 0, NULL);

	name_len = strlen (name);

	name_part = g_strndup (name, (name_len > 8 && !(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LONG_NAMES)) ? 8 : (MIN (31, name_len)));

	for (c = 0; name_part[c] != 0; c++) {
		if ((name_part[c] >= '0' && name_part[c] <= '9') ||
		    (name_part[c] >= 'A' && name_part[c] <= 'Z') ||
		    name_part[c] == '_' ||
		    (name_part[c] == '.' && (options & BURGNER_ISOFS_IDENTIFIER_ALLOW_MULTIDOT)))
			continue;

		if (name_part[c] >= 'a' && name_part[c] <= 'z') {
			if (!(options & BURGNER_ISOFS_IDENTIFIER_ALLOW_LOWER_CASE))
				name_part[c] += 'A' - 'a';
			continue;
		}

		if (options & BURGNER_ISOFS_IDENTIFIER_RELAX_CHARACTER_SET)
			continue;

		name_part[c] = '_';
	}

	return name_part;
}

