/*
 * 
 *
 * This file is part of the libburgnerisofs library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../libburgnerutils/burgner-g-list-utils.h"
#include "../libburgnerutils/burgner-math.h"
#include "burgner-isofs-item.h"
#include "burgner-isofs-void-item.h"
#include "burgner-isofs-item-list.h"


static guint
_burgner_isofs_item_list_extract_best_interval_sector_num_func (gpointer data, gpointer user_data)
{
	guint32 s = 0, b = 0;
	burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (data), &s, &b);
	return s + (b >> 11) + (b & 0x7ff ? 1 : 0);
}

static guint
_burgner_isofs_item_list_extract_best_interval_bytes_num_func (gpointer data, gpointer user_data)
{
	guint32 s = 0, b = 0;
	burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (data), &s, &b);
	return (s << 11) + b;
}

static gint
_burgner_isofs_list_item_extract_best_interval_list_sort_func (gconstpointer a, gconstpointer b,
							       gpointer user_data)
{
	gint ai = g_list_index ((GList *) user_data, a), bi = g_list_index ((GList *) user_data, b);
	if (ai < bi)
		return -1;
	return (ai != bi);
}

GList *
burgner_isofs_item_list_extract_best_interval (GList **list,
						GList **sorted_list,
						guint32 interval_size,
						guint32 *extracted_interval_size,
						gboolean sector_round)
{
	GList *nl = NULL, *it;
	guint f;

	g_return_val_if_fail (list != NULL, NULL);
	g_return_val_if_fail (*list != NULL, NULL);
	g_return_val_if_fail (sorted_list != NULL, NULL);
	g_return_val_if_fail (*sorted_list != NULL, NULL);
	g_return_val_if_fail (interval_size != 0, NULL);

	f = burgner_math_get_nearest_sum (*sorted_list, interval_size,
					  sector_round
					  	? _burgner_isofs_item_list_extract_best_interval_sector_num_func
					  	: _burgner_isofs_item_list_extract_best_interval_bytes_num_func,
					  NULL, &nl);
	nl = g_list_sort_with_data (nl, _burgner_isofs_list_item_extract_best_interval_list_sort_func, *list);

	for (it = nl; it != NULL; it = it->next) {
		*list = g_list_remove (*list, it->data);
		*sorted_list = g_list_remove (*sorted_list, it->data);
	}

	if (extracted_interval_size)
		*extracted_interval_size = f;

	return nl;
}


static gint
_burgner_isofs_item_list_arrange_sector_aligned_first_sort_func (gconstpointer a, gconstpointer b,
								 gpointer user_data)
{
	guint32 a_ma = 0, b_ma = 0;
	gboolean a_has_ma = burgner_isofs_item_get_mandatory_addr (BURGNER_ISOFS_ITEM (a), &a_ma),
		 b_has_ma = burgner_isofs_item_get_mandatory_addr (BURGNER_ISOFS_ITEM (b), &b_ma);

	if (a_has_ma != b_has_ma)
		return a_has_ma ? -1 : 1;

	if (!a_has_ma) {
		guint ai = g_list_index ((GList *) user_data, a),
		      bi = g_list_index ((GList *) user_data, b);

		if (ai < bi)
			return -1;
		return (ai != bi);
	}

	if (a_ma < b_ma)
		return -1;

	return (a_ma != b_ma);
}

GList *
burgner_isofs_item_list_arrange_sector_aligned (GList *list,
						guint32 start_addr)
{
	GList *nma, *n, *it, *nma_sorted;
	guint32 addr = start_addr;

	g_return_val_if_fail (list != NULL, NULL);

	n = g_list_sort_with_data (g_list_copy (list), _burgner_isofs_item_list_arrange_sector_aligned_first_sort_func, list);

	for (nma = n; nma && burgner_isofs_item_get_mandatory_addr (BURGNER_ISOFS_ITEM (nma->data), NULL); nma = nma->next);

	if (nma == n) {
		n = NULL;
	} else if (nma && nma->prev) {
		nma->prev->next = NULL;
		nma->prev = NULL;
	}

	nma_sorted = burgner_math_get_nearest_sum_sort_list (g_list_copy (nma), _burgner_isofs_item_list_extract_best_interval_sector_num_func, NULL);

	for (it = n; it != NULL; it = it->next) {
		BurgnerIsofsItem *i = BURGNER_ISOFS_ITEM (it->data);
		GList *p;
		guint32 a = 0, n_s = 0, n_b = 0, s, sp;

		g_assert (burgner_isofs_item_get_mandatory_addr (i, NULL));

		burgner_isofs_item_get_mandatory_addr (i, &a);
		g_return_val_if_fail (a >= addr, NULL);

		if (a != addr) {
			if (nma) {
				p = burgner_isofs_item_list_extract_best_interval (&nma, &nma_sorted,
						a - addr, &s, TRUE);
				n = g_list_insert_list_before (n, it, p);
				sp = (a - addr) - s;
			} else {
				sp = a - addr;
			}

			if (sp) {
				BurgnerIsofsItem *vi = burgner_isofs_void_item_new ();
				g_object_set (vi, "size", sp << 11, NULL);
				g_object_set_data (G_OBJECT (vi), "removable", GUINT_TO_POINTER (TRUE));
				n = g_list_insert_before (n, it, vi);
			}
		}

		burgner_isofs_item_get_size (i, &n_s, &n_b);
		addr = a + n_s + (n_b >> 11) + (n_b & 0x7ff ? 1 : 0);
	}

	if (nma) {
		n = g_list_concat (n, nma);
		g_assert (nma_sorted);
		g_list_free (nma_sorted);
	} else {
		g_assert (!nma_sorted);
	}

	return n;
}

static gint
_burgner_isofs_item_list_arrange_first_sort_func (gconstpointer a, gconstpointer b,
						  gpointer user_data)
{
	gboolean a_sa = (burgner_isofs_item_get_flags (BURGNER_ISOFS_ITEM (a)) & BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED),
		 b_sa = (burgner_isofs_item_get_flags (BURGNER_ISOFS_ITEM (b)) & BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED);
	guint ai, bi;

	if (a_sa != b_sa) {
		return a_sa ? -1 : 1;
	}

	ai = g_list_index ((GList *) user_data, a);
	bi = g_list_index ((GList *) user_data, b);

	if (ai < bi)
		return -1;

	return (ai != bi);
}

GList *
burgner_isofs_item_list_arrange (GList *list,
				 guint32 start_addr,
				 gboolean optimize)
{
	GList *sa, *nsa, *n, *it, *nsa_sorted;
	guint32 a_s, a_b, l_s, l_b;

	if (!list)
		return NULL;

	sa = g_list_sort_with_data (g_list_copy (list), _burgner_isofs_item_list_arrange_first_sort_func, list);

	for (nsa = sa;
	     nsa && burgner_isofs_item_get_flags (BURGNER_ISOFS_ITEM (nsa->data)) & BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED;
	     nsa = nsa->next);

	if (sa == nsa) {
		sa = NULL;
	} else if (nsa && nsa->prev) {
		nsa->prev->next = NULL;
		nsa->prev = NULL;
	}

	n = burgner_isofs_item_list_arrange_sector_aligned (sa, start_addr);
	g_list_free (sa);

	nsa_sorted = burgner_math_get_nearest_sum_sort_list (g_list_copy (nsa), _burgner_isofs_item_list_extract_best_interval_bytes_num_func, NULL);

	l_s = a_s = start_addr;
	l_b = a_b = 0;
	it = n;

	if (optimize) {
		while (it) {
			GList *f_list = NULL;
			guint32 n_s = 0, n_b = 0;
			BurgnerIsofsItem *i = BURGNER_ISOFS_ITEM (it->data);

			burgner_isofs_item_get_size (i, &n_s, &n_b);

			if (G_OBJECT_TYPE (i) == BURGNER_TYPE_ISOFS_VOID_ITEM &&
					GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (i), "removable"))) {
				GList *self_it = it;
				it = it->next;
				n = g_list_delete_link (n, self_it);

				a_s += n_s; a_b += n_b;
				a_s += a_b >> 11;
				if (a_b & 0x7ff) {
					a_s++; a_b = 0;
				}

				continue;
			}

			while (l_s != a_s && l_b != a_b) {
				GList *sp_list;
				guint sp, sp_used = 0;

				g_assert ((l_s < a_s) || (l_s == a_s && l_b < a_b));

				if (l_s < a_s) {
					sp = 2048 - l_b;
					l_s++;
					l_b = 0;
				} else {
					sp = a_b - l_b;
					l_b = a_b;
				}

				sp_list = nsa ? burgner_isofs_item_list_extract_best_interval (&nsa, &nsa_sorted,
						sp, &sp_used, FALSE) : NULL;
				if (sp_used != sp) {
					BurgnerIsofsItem *vi = burgner_isofs_void_item_new ();
					g_object_set (vi,
							"size", sp - sp_used,
							NULL);
					g_object_set_data (G_OBJECT (vi), "removable", GUINT_TO_POINTER (TRUE));
					burgner_isofs_void_item_remove_flags (BURGNER_ISOFS_VOID_ITEM (vi),
							BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED);
					sp_list = g_list_prepend (sp_list, vi);
				}

				f_list = g_list_concat (f_list, sp_list);
			}

			n = g_list_insert_list_before (n, it, f_list);

			a_s += n_s; a_b += n_b;
			a_s += a_b >> 11; a_b &= 0x7ff;
			l_s = a_s; l_b = a_b;
			if (a_b) {
				a_s++; a_b = 0;
			}

			it = it->next;
		}
	} else {
		BurgnerIsofsItem *last_item = BURGNER_ISOFS_ITEM (g_list_last (n)->data);
		burgner_isofs_item_get_size (last_item, NULL, &l_b);
	}

	if (nsa) {
		GList *f_list = NULL;

		g_assert (nsa_sorted);

		while (nsa) {
			GList *sp_list = NULL;
			guint sp, sp_used = 0;

			sp = l_b ? 2048 - l_b : 2048;

			sp_list = burgner_isofs_item_list_extract_best_interval (&nsa, &nsa_sorted,
									sp, &sp_used, FALSE);
			if (!sp_list) {
				sp_list = nsa;
				g_list_free (nsa_sorted);
				nsa = nsa_sorted = NULL;
			} else if (sp_used != sp) {
				BurgnerIsofsItem *vi = burgner_isofs_void_item_new ();
				g_object_set (vi,
						"size", sp - sp_used,
						NULL);
				g_object_set_data (G_OBJECT (vi), "removable", GUINT_TO_POINTER (TRUE));
				burgner_isofs_void_item_remove_flags (BURGNER_ISOFS_VOID_ITEM (vi),
								      BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED);
				sp_list = g_list_prepend (sp_list, vi);
			}

			f_list = g_list_concat (f_list, sp_list);
		}

		n = g_list_concat (n, f_list);
	} else {
		g_assert (!nsa_sorted);
	}

	return n;
}

