/*
 * BurgnerIsofsItem
 *
 * This file is part of the libburgner library.
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-marshal.h"
#include "burgner-isofs-item.h"

static GObjectClass *parent_class = NULL;

static void
finalize (GObject *object)
{
	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
item_get_size (BurgnerIsofsItem *item,
	       guint32 *n_sectors,
	       guint32 *n_bytes)
{
	if (n_sectors)
		*n_sectors = 0;

	if (n_bytes)
		*n_bytes = 0;
}

static BurgnerIsofsItemFlags
_burgner_isofs_item_get_flags_default_impl (BurgnerIsofsItem *item)
{
	return BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED;
}

static gboolean
_burgner_isofs_item_get_mandatory_addr_default_impl (BurgnerIsofsItem *item,
						     guint32 *addr)
{
	if (item->has_mandatory_addr && addr)
		*addr = item->mandatory_addr;

	return item->has_mandatory_addr;
}

static void
_burgner_isofs_item_set_mandatory_addr_default_impl (BurgnerIsofsItem *item,
						     gboolean has_mandatory_addr,
						     guint32 addr)
{
	item->has_mandatory_addr = has_mandatory_addr;
	item->mandatory_addr = item->has_mandatory_addr ? addr : 0;
}

static void
class_init (BurgnerIsofsItemClass *klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (klass);
	object_class->finalize = finalize;

	/* Virtual functions */
	klass->get_size = item_get_size;
	klass->get_flags = _burgner_isofs_item_get_flags_default_impl;
	klass->get_mandatory_addr = _burgner_isofs_item_get_mandatory_addr_default_impl;
	klass->set_mandatory_addr = _burgner_isofs_item_set_mandatory_addr_default_impl;

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsItem *item)
{
	item->mandatory_addr = 0;
	item->has_mandatory_addr = 0;
	item->prepared = 0;
}

GType
burgner_isofs_item_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (BurgnerIsofsItemClass),
			NULL, NULL, (GClassInitFunc) class_init, NULL, NULL,
			sizeof (BurgnerIsofsItem), 0, (GInstanceInitFunc) init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerIsofsItem", &info, G_TYPE_FLAG_ABSTRACT);
	}

	return type;
}


void
burgner_isofs_item_get_size (BurgnerIsofsItem *item,
			     guint32 *n_sectors,
			     guint32 *n_bytes)
{
	BurgnerIsofsItemClass *klass;

	g_return_if_fail (item != NULL);

	klass = BURGNER_ISOFS_ITEM_GET_CLASS (item);
	if (klass->get_size)
		klass->get_size (item, n_sectors, n_bytes);
}

void
burgner_isofs_item_prepare (BurgnerIsofsItem *item,
			    BurgnerIsofsMapTable *table,
			    guint32 sector_n,
			    guint32 byte_o,
			    guint32 last_sector_n,
			    guint32 last_byte_o)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));
	g_return_if_fail ((last_sector_n > sector_n) ||
			  (last_sector_n == sector_n && last_byte_o >= byte_o));
	g_return_if_fail (item->prepared == 0);

	item->prepared = 1;

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->prepare)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->prepare (item, table, sector_n, byte_o, last_sector_n, last_byte_o);
}

BurgnerIsofsItemFlags
burgner_isofs_item_get_flags (BurgnerIsofsItem *item)
{
	BurgnerIsofsItemFlags flags;

	g_return_val_if_fail (item != NULL, BURGNER_ISOFS_ITEM_FLAGS_NONE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_ITEM (item), BURGNER_ISOFS_ITEM_FLAGS_NONE);

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_flags)
		flags = BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_flags (item);
	else
		flags = BURGNER_ISOFS_ITEM_FLAGS_NONE;

	if (item->has_mandatory_addr)
		flags |= BURGNER_ISOFS_ITEM_HAS_MANDATORY_ADDR;

	return flags;
}

void
burgner_isofs_item_set_mandatory_addr (BurgnerIsofsItem *item,
				    gboolean has_mandatory_addr,
				    guint32 mandatory_addr)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->set_mandatory_addr)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->set_mandatory_addr (item, has_mandatory_addr, mandatory_addr);
}

gboolean
burgner_isofs_item_get_mandatory_addr (BurgnerIsofsItem *item,
					guint32 *mandatory_addr)
{
	g_return_val_if_fail (item != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_ITEM (item), FALSE);

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_mandatory_addr)
		return BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_mandatory_addr (item, mandatory_addr);

	return FALSE;
}

void
burgner_isofs_item_write_sync (BurgnerIsofsItem *item,
			    BurgnerIsofsMapTable *table,
		            gpointer buf, guint32 size)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_sync)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_sync (item, table, buf, size);
}

void
burgner_isofs_item_write_async_init (BurgnerIsofsItem *item,
				  BurgnerIsofsMapTable *table,
				  BurgnerIsofsItemWriteAsyncInitCallback cb,
				  gpointer cb_data)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async_init)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async_init (item, table, cb, cb_data);
}

void
burgner_isofs_item_write_async (BurgnerIsofsItem *item,
			     BurgnerIsofsMapTable *table,
		             gpointer write_async_data,
			     gpointer buf,
			     guint32 bytes_requested,
			     BurgnerIsofsItemWriteAsyncCallback cb,
			     gpointer cb_data)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));
	g_return_if_fail (!bytes_requested || buf);

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async (item, table, write_async_data, buf, bytes_requested, cb, cb_data);
}

void
burgner_isofs_item_write_async_cancel (BurgnerIsofsItem *item,
				       gpointer write_async_data)
{
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async_cancel)
		BURGNER_ISOFS_ITEM_GET_CLASS (item)->write_async_cancel (item, write_async_data);
}

GList *
burgner_isofs_item_get_free_children (BurgnerIsofsItem *item)
{
	g_return_val_if_fail (item != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_ITEM (item), NULL);

	if (BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_free_children)
		return BURGNER_ISOFS_ITEM_GET_CLASS (item)->get_free_children (item);

	return NULL;
}

