/*
 * BurgnerIsofsListMaker - BurgnerIsofsListMaker object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include "burgner-isofs-void-item.h"
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-item-list.h"
#include "burgner-isofs-volume-descriptor.h"
#include "burgner-isofs-list-maker.h"

#if 0
#define _BURGNER_ISOFS_LIST_MAKER_USE_MEM_CHUNK
#define _BURGNER_ISOFS_LIST_MAKER_N_PREALLOC 1
#endif

#ifdef _BURGNER_ISOFS_LIST_MAKER_USE_MEM_CHUNK
static GMemChunk *_burgner_isofs_list_maker_mem_chunk = NULL;
G_LOCK_DEFINE_STATIC (_burgner_isofs_list_maker_mem_chunk);
#endif

BurgnerIsofsListMaker *
burgner_isofs_list_maker_new (void)
{
	BurgnerIsofsListMaker *self;

#ifdef _BURGNER_ISOFS_LIST_MAKER_USE_MEM_CHUNK
	G_LOCK (_burgner_isofs_list_maker_mem_chunk);
	if (!_burgner_isofs_list_maker_mem_chunk) {
		_burgner_isofs_list_maker_mem_chunk = g_mem_chunk_new ("BurgnerIsofsListMaker mem chunk",
		 			sizeof (BurgnerIsofsListMaker),
					_BURGNER_ISOFS_LIST_MAKER_N_PREALLOC * sizeof (BurgnerIsofsListMaker),
					G_ALLOC_AND_FREE);
	}
	self = (BurgnerIsofsListMaker *) g_mem_chunk_alloc0 (_burgner_isofs_list_maker_mem_chunk);
	G_UNLOCK (_burgner_isofs_list_maker_mem_chunk);
#else
	self = g_new0 (BurgnerIsofsListMaker, 1);
#endif

	self->name_to_builder = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	self->refcount = 1;

	return self;
}


BurgnerIsofsListMaker *
burgner_isofs_list_maker_ref (BurgnerIsofsListMaker *self)
{
	g_return_val_if_fail (self != NULL, NULL);
	self->refcount++;
	return self;
}

void
burgner_isofs_list_maker_unref (BurgnerIsofsListMaker *self)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (self->refcount != 0);

	self->refcount--;
	if (self->refcount == 0) {
		burgner_isofs_list_maker_free (self);
	}
}

void
burgner_isofs_list_maker_free (BurgnerIsofsListMaker *self)
{
	g_return_if_fail (self != NULL);

	

#ifdef _BURGNER_ISOFS_LIST_MAKER_USE_MEM_CHUNK
	G_LOCK (_burgner_isofs_list_maker_mem_chunk);
	g_mem_chunk_free (_burgner_isofs_list_maker_mem_chunk, self);
	G_UNLOCK (_burgner_isofs_list_maker_mem_chunk);
#else
	g_free (self);
#endif
}


void
burgner_isofs_list_maker_add_builder (BurgnerIsofsListMaker *self,
					const gchar *name,
					BurgnerIsofsTreeBuilder *builder)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (builder != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_TREE_BUILDER (builder));

	self->builders = g_list_append (self->builders, g_object_ref (builder));

	if (name)
		g_hash_table_insert (self->name_to_builder, g_strdup (name), builder);
}

typedef struct {
	BurgnerIsofsListMaker *self;
	GList *files;
} _BurgnerIsofsListMakerCreateListData;

static gboolean
_burgner_isofs_list_maker_create_list_tree_traverse_func (GNode *node,
							  gpointer user_data)
{
	_BurgnerIsofsListMakerCreateListData *d = (_BurgnerIsofsListMakerCreateListData *) user_data;
	BurgnerIsofsListMaker *self = d->self;
	BurgnerIsofsFileInfo *info = (BurgnerIsofsFileInfo *) node->data,
			     *parent_info = (BurgnerIsofsFileInfo *) (node->parent ? node->parent->data : NULL);
	GList *it;

	if (info->is_directory) {
		for (it = self->builders; it != NULL; it = it->next) {
			burgner_isofs_tree_builder_directory (BURGNER_ISOFS_TREE_BUILDER (it->data),
								parent_info, info,
								g_node_depth (node));
		}
	} else {
		BurgnerIsofsItem *file;
		BurgnerIsofsTreeBuilder *tb;

		g_return_val_if_fail (node->children == NULL, TRUE);
		g_return_val_if_fail (info->uri != NULL, TRUE);

		tb = BURGNER_ISOFS_TREE_BUILDER (g_hash_table_lookup (self->name_to_builder,
								      g_uri_get_authority (info->uri)));
		g_return_val_if_fail (tb != NULL, TRUE);

		file = burgner_isofs_tree_builder_create_file (tb, info);
		if (file && (g_list_index (d->files, file) < 0)) {
			d->files = g_list_prepend (d->files, file);
		}

		for (it = self->builders; it != NULL; it = it->next) {
			burgner_isofs_tree_builder_file (BURGNER_ISOFS_TREE_BUILDER (it->data),
							 parent_info, info, file);
		}
	}

	return FALSE;
}

#if 0
static gpointer
_burgner_isofs_find_double (GList *l)
{
	GList *it;

	if (!l)
		return NULL;

	for (it = l; it && it->next; it = it->next) {
		if (g_list_find (it->next, it->data)) {
			return it->data;
		}
	}

	return NULL;
}
#endif

GList *
burgner_isofs_list_maker_create_list (BurgnerIsofsListMaker *self,
				      GNode *tree,
				      gboolean optimize,
				      gboolean void_first_16_sectors)
{
	GList *items = NULL, *vd = NULL, *it, *list, *nl;
	BurgnerIsofsItem *last_item, *vdt, *first_item;
	_BurgnerIsofsListMakerCreateListData d;
	guint32 a;

	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (tree != NULL, NULL);

	last_item = burgner_isofs_void_item_new ();
	g_object_set (last_item, "size", 0, NULL);

	for (it = self->builders; it != NULL; it = it->next) {
		burgner_isofs_tree_builder_begin (BURGNER_ISOFS_TREE_BUILDER (it->data),
						  tree, last_item);
	}

	d.self = self; d.files = NULL;
	g_node_traverse (tree,
			 G_PRE_ORDER, G_TRAVERSE_ALL, -1,
			 _burgner_isofs_list_maker_create_list_tree_traverse_func, &d);

	for (it = self->builders; it != NULL; it = it->next) {
		burgner_isofs_tree_builder_end (BURGNER_ISOFS_TREE_BUILDER (it->data),
						tree, last_item);

		items = g_list_concat (burgner_isofs_item_g_list_ref (burgner_isofs_tree_builder_get_items (BURGNER_ISOFS_TREE_BUILDER (it->data))), items);
		vd = g_list_concat (burgner_isofs_item_g_list_ref (burgner_isofs_tree_builder_get_volume_descriptors (BURGNER_ISOFS_TREE_BUILDER (it->data))), vd);
	}

	list = burgner_isofs_item_list_arrange_sector_aligned (vd, 0x10);
	g_list_free (vd);

	a = 0x10;
	for (it = list; it != NULL; it = it->next) {
		guint n_s = 0, n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data), &n_s, &n_b);
		burgner_isofs_item_set_mandatory_addr (BURGNER_ISOFS_ITEM (it->data), TRUE, a);
		n_s += n_b >> 11;
		g_assert ((n_b & 0x7ff) == 0);
		a += n_s;
	}

	vdt = burgner_isofs_volume_descriptor_new ();
	burgner_isofs_item_set_mandatory_addr (vdt, TRUE, a);
	list = g_list_append (list, vdt);

	nl = items;
	items = NULL;
	while (nl) {
		GList *t_l = nl;
		nl = NULL;

		for (it = t_l; it; it = it->next) {
			GList *fc = burgner_isofs_item_get_free_children (BURGNER_ISOFS_ITEM (it->data));
			if (fc) {
				nl = g_list_concat (fc, nl);
			}
		}

		items = g_list_concat (items, t_l);
	}

	items = g_list_concat (g_list_concat (list, items), d.files);

	first_item = burgner_isofs_void_item_new ();
	g_object_set (first_item, "size", 2048 * (void_first_16_sectors ? 16 : 1), NULL);
	burgner_isofs_item_set_mandatory_addr (first_item, TRUE, 0);
	items = g_list_prepend (items, first_item);

	nl = burgner_isofs_item_list_arrange (items, 0, optimize);
	g_list_free (items);

	return g_list_append (nl, last_item);
}

