/*
 * BurgnerIsofsMapTable - BurgnerIsofsMapTable object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-map-table.h"


typedef struct {
	guint32 s;
	guint32 b;
} _BimpDictData;

#ifdef USE_MUTEX
static GStaticMutex _bimp_dict_data_mem_chunk_mutex = G_STATIC_MUTEX_INIT;
#endif

static GMemChunk *_bimp_dict_data_mem_chunk = NULL;

static _BimpDictData *
_bimp_dict_data_new (void)
{
	_BimpDictData *d;

#ifdef USE_MUTEX
	g_static_mutex_lock (_bimp_dict_data_mem_chunk_mutex);
#endif
	if (!_bimp_dict_data_mem_chunk) {
		_bimp_dict_data_mem_chunk = g_mem_chunk_new ("_BimpDictData mem chunk",
							     sizeof (_BimpDictData),
							     1024,
							     G_ALLOC_AND_FREE);
	}

	d = g_chunk_new0 (_BimpDictData, _bimp_dict_data_mem_chunk);
#ifdef USE_MUTEX
	g_static_mutex_unlock (_bimp_dict_data_mem_chunk_mutex);
#endif

	return d;
}

static void
_bimp_dict_data_free (_BimpDictData *d)
{
#ifdef USE_MUTEX
	g_static_mutex_lock (_bimp_dict_data_mem_chunk_mutex);
#endif
	g_mem_chunk_free (_bimp_dict_data_mem_chunk, d);
#ifdef USE_MUTEX
	g_static_mutex_unlock (_bimp_dict_data_mem_chunk_mutex);
#endif
}



static GObjectClass *parent_class = NULL;


static void
finalize (GObject *object)
{
	BurgnerIsofsMapTable *self = BURGNER_ISOFS_MAP_TABLE (object);

	if (self->dict) {
		g_hash_table_destroy (self->dict);
		self->dict = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
class_init (BurgnerIsofsMapTableClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = finalize;

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsMapTable *self)
{
	self->dict = g_hash_table_new_full (NULL, NULL, NULL,
					    (GDestroyNotify) _bimp_dict_data_free);
}

GType
burgner_isofs_map_table_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (BurgnerIsofsMapTableClass),
			NULL, NULL, (GClassInitFunc) class_init, NULL, NULL,
			sizeof (BurgnerIsofsMapTable), 0, (GInstanceInitFunc) init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerIsofsMapTable", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsMapTable *
burgner_isofs_map_table_new (void)
{
	BurgnerIsofsMapTable *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_MAP_TABLE, NULL);


	return self;

}


GQuark
burgner_isofs_map_table_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsMapTable");
}

void
burgner_isofs_map_table_set (BurgnerIsofsMapTable *table,
			     BurgnerIsofsItem *item,
			     guint32 sector_n,
			     guint32 byte_o)
{
	_BimpDictData *d;

	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));

	d = g_hash_table_lookup (table->dict, item);
	if (!d) {
		d = _bimp_dict_data_new ();
		g_hash_table_insert (table->dict, item, d);
	}

	d->s = sector_n;
	d->b = byte_o;
}

void
burgner_isofs_map_table_unset (BurgnerIsofsMapTable *table,
			       BurgnerIsofsItem *item)
{
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));
	g_return_if_fail (item != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_ITEM (item));

	if (g_hash_table_lookup (table->dict, item))
		g_hash_table_remove (table->dict, item);
}

gboolean
burgner_isofs_map_table_get_addr (BurgnerIsofsMapTable *table,
				  BurgnerIsofsItem *item,
				  guint32 *sector_n,
				  guint32 *byte_o)
{
	_BimpDictData *d;

	g_return_val_if_fail (table != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table), FALSE);
	g_return_val_if_fail (item != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_ITEM (item), FALSE);

	d = g_hash_table_lookup (table->dict, item);
	if (!d) {
		if (sector_n)
			*sector_n = 0;
		if (byte_o)
			*byte_o = 0;
		return FALSE;
	}

	if (sector_n)
		*sector_n = d->s;
	if (byte_o)
		*byte_o = d->b;

	return TRUE;
}
