/*
 * BurgnerIsofsPathTableRecord - BurgnerIsofsPathTableRecord object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-directory.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-path-table-record.h"


static BurgnerIsofsItemClass *parent_class = NULL;


enum {
	PROP_0,
	PROP_DIR,
	PROP_PAR,
	PROP_PAR_IDX,
	PROP_IS_ROOT_DIR,
	PROP_ID,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsPathTableRecord *self = BURGNER_ISOFS_PATH_TABLE_RECORD (object);

	if (self->id) {
		g_free (self->id);
		self->id = NULL;
	}

	if (self->dir) {
		g_object_unref (self->dir);
		self->dir = NULL;
	}

	if (self->par) {
		g_object_unref (self->par);
		self->par = NULL;
	}

	self->par_idx = 1;
	self->is_root_dir = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize)(object);
}

static void
_biptr_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	BurgnerIsofsPathTableRecord *rec = BURGNER_ISOFS_PATH_TABLE_RECORD (item);
	guint len = 8;

	g_return_if_fail ((rec->is_root_dir) || (rec->id != NULL));

	len += rec->is_root_dir ? 1 : strlen (rec->id);

	if (len & 1)
		len += 1;

	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = len;
}

static void
_biptr_write (BurgnerIsofsItem *item,
	      BurgnerIsofsMapTable *table,
	      gpointer buf, guint32 size,
	      BurgnerIsofsPathTableRecordType type)
{
	__BurgnerIsofsImplPathTableRecord *impl;
	BurgnerIsofsPathTableRecord *rec = BURGNER_ISOFS_PATH_TABLE_RECORD (item);
	guint len;
	guint32 dir_addr = 0;

	g_return_if_fail ((rec->id != NULL) || (rec->is_root_dir));
	g_return_if_fail (rec->dir != NULL);
	g_return_if_fail (rec->par != NULL);
	g_return_if_fail (rec->par_idx != 0);

	impl = (__BurgnerIsofsImplPathTableRecord *) buf;
	memset (impl, 0, sizeof (*impl));

	len = rec->is_root_dir ? 1 : strlen (rec->id);

	{
		gboolean ret = burgner_isofs_map_table_get_addr (table, rec->dir,
					                         &dir_addr, NULL);
		g_return_if_fail (ret != FALSE);
	}

	burgner_isofs_set_711 (&impl->len_di, len & 0xff);
	burgner_isofs_set_711 (&impl->ear_len, 0);

	switch (type) {
		case BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L:
			burgner_isofs_set_731 (impl->location_of_extent, dir_addr);
			burgner_isofs_set_721 (impl->parent_dir_num, rec->par_idx);
			break;
		case BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_M:
			burgner_isofs_set_732 (impl->location_of_extent, dir_addr);
			burgner_isofs_set_722 (impl->parent_dir_num, rec->par_idx);
			break;
		default:
			g_return_if_reached ();
			break;
	}

	buf += sizeof (*impl);
	size -= sizeof (*impl);

	if (rec->is_root_dir) {
		*((guchar *) buf++) = 0;
		size -= 1;
	} else {
		memcpy (buf, rec->id, len);
		buf += len;
		size -= len;
	}

	if (len & 1) {
		*((guchar *) buf++) = 0;
		size -= 1;
	}

	g_assert (size == 0);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsPathTableRecord *rec = BURGNER_ISOFS_PATH_TABLE_RECORD (object);

	switch (prop_id) {
		case PROP_DIR:
			if (rec->dir)
				g_object_unref (rec->dir);

			rec->dir = (BurgnerIsofsItem *) g_value_get_object (value);
			if (rec->dir)
				rec->dir = BURGNER_ISOFS_ITEM (g_object_ref (rec->dir));
			g_object_notify (object, "dir");
			break;
		case PROP_PAR:
			if (rec->par)
				g_object_unref (rec->par);

			rec->par = (BurgnerIsofsItem *) g_value_get_object (value);
			if (rec->par)
				rec->par = BURGNER_ISOFS_ITEM (g_object_ref (rec->par));
			g_object_notify (object, "par");
			break;
		case PROP_PAR_IDX:
			rec->par_idx = g_value_get_uint (value);
			g_object_notify (object, "par_index");
			break;
		case PROP_ID:
			if (rec->id)
				g_free (rec->id);
			rec->id = g_value_dup_string (value);
			g_object_notify (object, "id");
			break;
		case PROP_IS_ROOT_DIR:
			rec->is_root_dir = g_value_get_boolean (value) ? 1 : 0;
			g_object_notify (object, "is_root_dir");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsPathTableRecord *rec = BURGNER_ISOFS_PATH_TABLE_RECORD (object);

	switch (prop_id) {
		case PROP_DIR:
			g_value_set_object (value, rec->dir);
			break;
		case PROP_PAR:
			g_value_set_object (value, rec->par);
			break;
		case PROP_PAR_IDX:
			g_value_set_uint (value, rec->par_idx);
			break;
		case PROP_ID:
			g_value_set_string (value, rec->id);
			break;
		case PROP_IS_ROOT_DIR:
			g_value_set_boolean (value, rec->is_root_dir?TRUE:FALSE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_biptr_prepare (BurgnerIsofsItem *item,
		BurgnerIsofsMapTable *table,
		guint32 s_c, guint32 b_o,
		guint32 n_s, guint32 n_b)
{
	item->prepared = 0;
}

static void
class_init (BurgnerIsofsPathTableRecordClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = finalize;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
 
	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _biptr_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_none;
	item_class->prepare = _biptr_prepare;

	klass->write = _biptr_write;


	g_object_class_install_property (object_class,
					 PROP_DIR,
					 g_param_spec_object ("dir",
					                      "dir",
							      "Directory",
							      BURGNER_TYPE_ISOFS_DIRECTORY,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_PAR,
					 g_param_spec_object ("par",
					                      "par",
							      "Parent Directory",
							      BURGNER_TYPE_ISOFS_DIRECTORY,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_PAR_IDX,
					 g_param_spec_uint ("par_index",
					                    "par_index",
							    "Parent index",
							    1,
							    0xffff,
							    1,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_ID,
					 g_param_spec_string ("id",
					            	       "id",
							       "Identifier",
							       NULL,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_IS_ROOT_DIR,
					 g_param_spec_boolean ("is_root_dir",
					            	       "is_root_dir",
							       "Is Root Directory",
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsPathTableRecord *self)
{
	self->dir = self->par = NULL;
	self->id = NULL;
	self->par_idx = 1;
	self->is_root_dir = 0;
}

GType
burgner_isofs_path_table_record_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsPathTableRecordClass),
			NULL, NULL, (GClassInitFunc) class_init, NULL, NULL,
			sizeof (BurgnerIsofsPathTableRecord), 0, (GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsPathTableRecord", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_path_table_record_new (void)
{
	BurgnerIsofsPathTableRecord *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_PATH_TABLE_RECORD, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_path_table_record_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsPathTableRecord");
}

void
burgner_isofs_path_table_record_write (BurgnerIsofsPathTableRecord *rec,
				       BurgnerIsofsMapTable *table,
		                       gpointer buf, guint32 size,
		                       BurgnerIsofsPathTableRecordType type)
{
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_PATH_TABLE_RECORD (rec));
	g_return_if_fail (table != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_MAP_TABLE (table));

	if (BURGNER_ISOFS_PATH_TABLE_RECORD_GET_CLASS (rec)->write)
		BURGNER_ISOFS_PATH_TABLE_RECORD_GET_CLASS (rec)->write (BURGNER_ISOFS_ITEM (rec), table, buf, size, type);
}

GType
burgner_isofs_path_table_record_type_get_type (void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			{ BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L, "BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L", "L" },
			{ BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_M, "BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_M", "M" },
			{ 0, NULL, NULL }
		};
		etype = g_enum_register_static ("BurgnerIsofsPathTableRecordType", values);
	}

	return etype;
}
