/*
 * BurgnerIsofsPathTable - BurgnerIsofsPathTable object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-path-table.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_TYPE,
	PROP_LAST
};



static void
finalize (GObject *object)
{
	BurgnerIsofsPathTable *self = BURGNER_ISOFS_PATH_TABLE (object);

	if (self->dir_to_idx) {
		g_hash_table_destroy (self->dir_to_idx);
		self->dir_to_idx = NULL;
	}

	if (self->n_bytes) {
		g_slist_free (self->n_bytes);
		self->n_bytes = NULL;
	}

	if (self->records) {
		burgner_isofs_item_g_slist_unref (self->records);
		self->records = NULL;
	}

	self->type = BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize)(object);
}

static void
_bipt_get_size (BurgnerIsofsItem *item,
		guint32 *n_sectors,
		guint32 *n_bytes)
{
	BurgnerIsofsPathTable *pt = BURGNER_ISOFS_PATH_TABLE (item);
	GSList *it;
	guint32 n = 0, c;

	for (it = pt->records; it != NULL; it = it->next) {
		c = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data), NULL, &c);
		n += c;

	}

	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = n;
}

static void
_bipt_write_sync (BurgnerIsofsItem *item,
		  BurgnerIsofsMapTable *table,
		  gpointer buf, guint32 size)
{
	BurgnerIsofsPathTable *pt = BURGNER_ISOFS_PATH_TABLE (item);
	GSList *it, *b_it;

	{
		guint a = g_slist_length (pt->records),
			b = g_slist_length (pt->n_bytes);
		g_return_if_fail (a == b);
	}

	for (it = pt->records, b_it = pt->n_bytes;
	     it != NULL;
	     it = it->next, b_it = b_it->next) {
#if 0
		burgner_isofs_writer_proxy_begin_item (proxy, BURGNER_ISOFS_ITEM (it->data));
		burgner_isofs_path_table_record_write (BURGNER_ISOFS_PATH_TABLE_RECORD (it->data),
						       proxy,
						       table,
						       0,
						       GPOINTER_TO_UINT (b_it->data),
						       pt->type);
		burgner_isofs_writer_proxy_end_item (proxy, BURGNER_ISOFS_ITEM (it->data));
#else
		burgner_isofs_path_table_record_write (BURGNER_ISOFS_PATH_TABLE_RECORD (it->data),
							table, buf, GPOINTER_TO_UINT (b_it->data), pt->type);
		buf += GPOINTER_TO_UINT (b_it->data);
		size -= GPOINTER_TO_UINT (b_it->data);
#endif
	}

	g_assert (size == 0);
}

static void
_bipt_add_record (BurgnerIsofsPathTable *pt,
		  BurgnerIsofsItem *rec)
{
	BurgnerIsofsItem *d, *p;
	guint16 idx;

	g_return_if_fail (pt->n_bytes == NULL);

	d = p = NULL;
	g_object_get (rec,
		      "dir", &d,
		      "par", &p,
		      NULL);
	g_return_if_fail (d != NULL);
	g_return_if_fail (p != NULL);

	idx = g_slist_length (pt->records) + 1;
	g_hash_table_insert (pt->dir_to_idx, d, GUINT_TO_POINTER ((guint) idx));

	idx = GPOINTER_TO_UINT (g_hash_table_lookup (pt->dir_to_idx, p));
	g_return_if_fail (idx != 0);
	g_object_set (rec, "par_index", idx, NULL);

	pt->records = g_slist_append (pt->records, g_object_ref (rec));
}

static void
_bipt_set_property (GObject *object,
	            guint prop_id,
	            const GValue *value,
	            GParamSpec *pspec)
{
	BurgnerIsofsPathTable *pt = BURGNER_ISOFS_PATH_TABLE (object);

	switch (prop_id) {
		case PROP_TYPE:
			pt->type = g_value_get_enum (value);
			g_object_notify (object, "type");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bipt_get_property (GObject *object,
	            guint prop_id,
	            GValue *value,
	            GParamSpec *pspec)
{
	BurgnerIsofsPathTable *pt = BURGNER_ISOFS_PATH_TABLE (object);

	switch (prop_id) {
		case PROP_TYPE:
			g_value_set_enum (value, pt->type);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bipt_prepare (BurgnerIsofsItem *item,
	       BurgnerIsofsMapTable *table,
	       guint32 sector_n,
	       guint32 byte_o,
	       guint32 last_sector_n,
	       guint32 last_byte_o)
{
	BurgnerIsofsPathTable *pt = BURGNER_ISOFS_PATH_TABLE (item);
	BurgnerIsofsItem *rec;
	GSList *it;
	guint32 s = sector_n, b = byte_o, n_b, l_s, l_b;

	item->prepared = 0;

	if (pt->n_bytes) {
		g_slist_free (pt->n_bytes);
		pt->n_bytes = NULL;
	}

	for (it = pt->records; it != NULL; it = it->next) {
		if ((s > last_sector_n) ||
		    (s == last_sector_n && b >= last_byte_o)) {
			g_warning ("Not enough space!");
			break;
		}

		rec = BURGNER_ISOFS_ITEM (it->data);
		burgner_isofs_map_table_set (table, rec, s, b);
		/* XXX fr: Deux adresses correspondent a cet item:
			celle de la version BigEndian, et celle de
			la version LittleEndian, mais seule la derniere
			sera prise en compte :/ */

		n_b = 0;
		burgner_isofs_item_get_size (rec, NULL, &n_b);
		pt->n_bytes = g_slist_prepend (pt->n_bytes, GUINT_TO_POINTER (n_b));
		l_s = s;
		l_b = b;
		b += n_b;
		if (b >= 2048) {
			s += b / 2048;
			b %= 2048;
		}
		if ((s > last_sector_n) ||
		    (s == last_sector_n && b > last_byte_o)) {
			g_warning ("Too big item!");
			break;
		}
		burgner_isofs_item_prepare (rec, table, l_s, l_b, s, b);
	}
	pt->n_bytes = g_slist_reverse (pt->n_bytes);
}

static void
class_init (BurgnerIsofsPathTableClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = finalize;
	object_class->set_property = _bipt_set_property;
	object_class->get_property = _bipt_get_property;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bipt_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_sector_start_aligned;
	item_class->prepare = _bipt_prepare;
	item_class->write_sync = _bipt_write_sync;

	/* Virtual functions */
	klass->add_record = _bipt_add_record;

	g_object_class_install_property (object_class,
					 PROP_TYPE,
					 g_param_spec_enum ("type",
					                    "type",
							    "Type",
							    BURGNER_TYPE_ISOFS_PATH_TABLE_RECORD_TYPE,
							    BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L,
							    G_PARAM_READABLE|G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsPathTable *self)
{
	self->dir_to_idx = g_hash_table_new (NULL, NULL);
	self->records = self->n_bytes = NULL;
	self->type = BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L;
}

GType
burgner_isofs_path_table_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsPathTableClass),
			NULL, NULL, (GClassInitFunc) class_init, NULL, NULL,
			sizeof (BurgnerIsofsPathTable), 0, (GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsPathTable", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_path_table_new (void)
{
	BurgnerIsofsPathTable *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_PATH_TABLE, NULL);


	return BURGNER_ISOFS_ITEM (self);
}

GQuark
burgner_isofs_path_table_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsPathTable");
}

void
burgner_isofs_path_table_add_record (BurgnerIsofsPathTable *pt,
				     BurgnerIsofsItem *rec)
{
	g_return_if_fail (pt != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_PATH_TABLE (pt));
	g_return_if_fail (rec != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_PATH_TABLE_RECORD (rec));

	if (BURGNER_ISOFS_PATH_TABLE_GET_CLASS (pt)->add_record)
		BURGNER_ISOFS_PATH_TABLE_GET_CLASS (pt)->add_record (pt, rec);
}
