/*
 * BurgnerIsofsPrimaryBuilder - BurgnerIsofsPrimaryBuilder object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "../libburgnerutils/burgner-hash.h"
#include "burgner-isofs-item-g-list.h"
#include "burgner-isofs-directory.h"
#include "burgner-isofs-directory-dot-record.h"
#include "burgner-isofs-susp-sp.h"
#include "burgner-isofs-susp-er.h"
#include "burgner-isofs-directory-dotdot-record.h"
#include "burgner-isofs-directory-subdirectory-record.h"
#include "burgner-isofs-directory-file-record.h"
#include "burgner-isofs-tvd-info.h"
#include "burgner-isofs-path-table-record.h"
#include "burgner-isofs-path-table.h"
#include "burgner-isofs-typed-path-table.h"
#include "burgner-isofs-tree-volume-descriptor.h"
#include "burgner-isofs-primary-builder.h"


typedef struct Bipb_PtTreeData __Bipb_PtTreeData;
struct Bipb_PtTreeData {
	BurgnerIsofsItem *par;
	BurgnerIsofsItem *dir;
	gchar *id;

	guint level;
	__Bipb_PtTreeData *par_d;
};

static GMemChunk *pt_tree_data_mem_chunk = NULL;

#ifdef USE_MUTEX
static GStaticMutex pt_tree_data_mutex = G_STATIC_MUTEX_INIT;
#else
G_LOCK_DEFINE_STATIC (pt_tree_data_mem_chunk);
#endif

static __Bipb_PtTreeData *
_bipb_pt_tree_data_new (void)
{
	__Bipb_PtTreeData *d;

#ifdef USE_MUTEX
	g_static_mutex_lock (&pt_tree_data_mutex);
#else
	G_LOCK (pt_tree_data_mem_chunk);
#endif
	if (!pt_tree_data_mem_chunk) {
		pt_tree_data_mem_chunk = g_mem_chunk_new ("__Bipb_PtTreeData mem chunk",
							  sizeof (__Bipb_PtTreeData),
							  sizeof (__Bipb_PtTreeData) * 256,
							  G_ALLOC_AND_FREE);
	}
	d = (__Bipb_PtTreeData *) g_mem_chunk_alloc0 (pt_tree_data_mem_chunk);
#ifdef USE_MUTEX
	g_static_mutex_unlock (&pt_tree_data_mutex);
#else
	G_UNLOCK (pt_tree_data_mem_chunk);
#endif

	return d;
}

static void
_bipb_pt_tree_data_free (__Bipb_PtTreeData *d)
{
	g_return_if_fail (d != NULL);
#ifdef USE_MUTEX
	g_static_mutex_lock (&pt_tree_data_mutex);
#else
	G_LOCK (pt_tree_data_mem_chunk);
#endif
	g_mem_chunk_free (pt_tree_data_mem_chunk, (gpointer) d);
#ifdef USE_MUTEX
	g_static_mutex_unlock (&pt_tree_data_mutex);
#else
	G_UNLOCK (pt_tree_data_mem_chunk);
#endif
}


static BurgnerIsofsTreeBuilderClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_RELOC_LEVEL,
	PROP_RELOC_DIR_ID,
	PROP_WITH_RELOC,
	PROP_WITH_RRIP,
	PROP_TVD_INFO,
	PROP_CACHE_INODES,
	PROP_NO_PATH_TABLES,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (object);

	if (self->vd_list) {
		burgner_isofs_item_g_list_unref (self->vd_list);
		self->vd_list = NULL;
	}

	if (self->items_list) {
		burgner_isofs_item_g_list_unref (self->items_list);
		self->items_list = NULL;
	}

	if (self->fi_to_pt_d) {
		g_hash_table_destroy (self->fi_to_pt_d);
		self->fi_to_pt_d = NULL;
	}

	self->root_pt_d = self->reloc_pt_d = NULL;

	if (self->pt_list) {
		g_list_foreach (self->pt_list, (GFunc) _bipb_pt_tree_data_free, NULL);
		g_list_free (self->pt_list);
		self->pt_list = NULL;
	}

	if (self->reloc_dir_id) {
		g_free (self->reloc_dir_id);
		self->reloc_dir_id = NULL;
	}

	if (self->reloc_dir) {
		g_object_unref (self->reloc_dir);
		self->reloc_dir = NULL;
	}

	self->reloc_id_count = 0;

	if (self->fi_to_dir) {
		g_hash_table_destroy (self->fi_to_dir);
		self->fi_to_dir = NULL;
	}

	if (self->tvd_info) {
		burgner_isofs_tvd_info_static_free (self->tvd_info);
		g_free (self->tvd_info);
		self->tvd_info = NULL;
	}

	if (self->root_dir) {
		g_object_unref (self->root_dir);
		self->root_dir = NULL;
	}

	if (self->root_dir_rec) {
		g_object_unref (self->root_dir_rec);
		self->root_dir_rec = NULL;
	}

	if (self->inode_to_file) {
		g_hash_table_destroy (self->inode_to_file);
		self->inode_to_file = NULL;
	}

	self->with_reloc = 1;
	self->with_rrip = 1;
	self->no_path_tables = 0;

	self->reloc_level = 9;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (object);

	switch (prop_id) {
		case PROP_RELOC_LEVEL:
			self->reloc_level = g_value_get_uint (value);
			g_object_notify (object, "reloc_level");
			break;
		case PROP_RELOC_DIR_ID:
			if (self->reloc_dir_id)
				g_free (self->reloc_dir_id);
			self->reloc_dir_id = g_value_dup_string (value);
			g_object_notify (object, "reloc_dir_id");
			break;
		case PROP_WITH_RELOC:
			self->with_reloc = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_reloc");
			break;
		case PROP_WITH_RRIP:
			self->with_rrip = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "with_rrip");
			break;
		case PROP_TVD_INFO: {
			BurgnerIsofsTvdInfo *t;
				burgner_isofs_tvd_info_static_free (self->tvd_info);

			if ((t = g_value_get_pointer (value)))
				burgner_isofs_tvd_info_static_copy (t, self->tvd_info);

			g_object_notify (object, "tvd_info");
			} break;
		case PROP_CACHE_INODES:
			self->cache_inodes = g_value_get_boolean (value) ? 1 : 0;
			break;
		case PROP_NO_PATH_TABLES:
			self->no_path_tables = g_value_get_boolean (value) ? 1 : 0;
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (object);

	switch (prop_id) {
		case PROP_RELOC_LEVEL:
			g_value_set_uint (value, self->reloc_level);
			break;
		case PROP_RELOC_DIR_ID:
			g_value_set_pointer (value, self->reloc_dir_id);
			break;
		case PROP_WITH_RELOC:
			g_value_set_boolean (value, self->with_reloc);
			break;
		case PROP_WITH_RRIP:
			g_value_set_boolean (value, self->with_rrip);
			break;
		case PROP_TVD_INFO:
			g_value_set_pointer (value, self->tvd_info);
			break;
		case PROP_CACHE_INODES:
			g_value_set_boolean (value, self->cache_inodes);
			break;
		case PROP_NO_PATH_TABLES:
			g_value_set_boolean (value, self->no_path_tables);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bipb_begin (BurgnerIsofsTreeBuilder *tb,
	     GNode *root,
	     BurgnerIsofsItem *last_item)
{
	BurgnerIsofsPrimaryBuilder *pb = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	BurgnerIsofsItem *dir, *cur, *par;
	__Bipb_PtTreeData *d;

	g_return_if_fail (pb->pvd == NULL);
	g_return_if_fail (pb->fi_to_pt_d == NULL);
	g_return_if_fail (pb->pt_list == NULL);
	g_return_if_fail (pb->fi_to_dir == NULL);
	g_return_if_fail (pb->root_dir == NULL);

	/* TODO: ref the BurgnerIsofsFileInfo in the two first hash tables */
	pb->fi_to_dir = g_hash_table_new (NULL, NULL);
	pb->fi_to_pt_d = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify) _bipb_pt_tree_data_free);

	pb->fi_to_file = g_hash_table_new_full (NULL, NULL,
						(GDestroyNotify) burgner_isofs_file_info_unref,
						(GDestroyNotify) g_object_unref);

	if (pb->cache_inodes) {
		pb->inode_to_file = g_hash_table_new_full (burgner_hash_guint64_pointer_hash,
							   burgner_hash_guint64_pointer_equal,
							   g_free, NULL);
	}

	dir = burgner_isofs_directory_new ();
	g_object_set (dir,
		      "info", root->data,
		      "with_rrip", pb->with_rrip,
		      NULL);
	pb->root_dir = BURGNER_ISOFS_ITEM (g_object_ref (dir));
	cur = burgner_isofs_directory_dot_record_new ();
	g_object_set (cur,
		      "dir", dir,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);
	if (pb->with_rrip) {
		BurgnerIsofsItem *sp, *er;

		sp = burgner_isofs_susp_sp_new ();
		g_object_set (sp, "bs", 0, NULL);

		er = burgner_isofs_susp_er_new ();
		g_object_set (er,
			      "id", "IEEE_P1282",
			      "desc", "THE IEEE P1282 PROTOCOL PROVIDES SUPPORT FOR POSIX FILE SYSTEM SEMANTICS.",
			      "src", "PLEASE CONTACT THE IEEE STANDARDS DEPARTMENT, PISCATAWAY, NJ, USA FOR THE P1282 SPECIFICATION.",
			      "ext_ver", 1,
			      NULL);

		burgner_isofs_directory_dot_record_append_suf_ents (BURGNER_ISOFS_DIRECTORY_DOT_RECORD (cur),
								    sp, er, NULL);
		g_object_unref (sp);
		g_object_unref (er);
	}

	par = burgner_isofs_directory_dotdot_record_new ();
	g_object_set (par,
		      "dir", dir,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);

	burgner_isofs_directory_append_records (BURGNER_ISOFS_DIRECTORY (dir), cur, par, NULL);

	pb->root_dir_rec = burgner_isofs_directory_dot_record_new ();
	g_object_set (pb->root_dir_rec,
		      "dir", dir,
		      "with_susp", FALSE,
		      "with_rrip", FALSE,
		      NULL);

	g_object_unref (cur);
	g_object_unref (par);

	g_hash_table_insert (pb->fi_to_dir, root->data, dir);
	pb->items_list = g_list_prepend (pb->items_list, dir);

	d = _bipb_pt_tree_data_new ();
	d->par = d->dir = dir;
	d->level = 1;
	g_hash_table_insert (pb->fi_to_pt_d, root->data, d);
	pb->root_pt_d = (gpointer) d;
	pb->pt_list = g_list_prepend (pb->pt_list, d);
}

static gint
_bipb_pt_list_sort_func (__Bipb_PtTreeData *a,
			 __Bipb_PtTreeData *b)
{
	if (a->level != b->level)
		return a->level - b->level;

	if (a->par == b->par) {
		g_assert (a->par_d == b->par_d);
		return strcmp (a->id, b->id);
	}

	return _bipb_pt_list_sort_func (a->par_d, b->par_d);
}

static void
_bipb_end (BurgnerIsofsTreeBuilder *tb,
	   GNode *root,
	   BurgnerIsofsItem *last_item)
{
	BurgnerIsofsPrimaryBuilder *pb = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	BurgnerIsofsItem *pvd, *pt, *l_pt = NULL, *m_pt = NULL, *rec;
	__Bipb_PtTreeData *d;
	GList *it;

	g_return_if_fail (pb->vd_list == NULL);

	if (!pb->no_path_tables) {
		pb->pt_list = g_list_sort (g_list_reverse (pb->pt_list), (GCompareFunc) _bipb_pt_list_sort_func);

		pt = burgner_isofs_path_table_new ();
		for (it = pb->pt_list; it != NULL; it = it->next) {
			d = (__Bipb_PtTreeData *) it->data;
			rec = burgner_isofs_path_table_record_new ();
			g_object_set (rec,
					"dir", d->dir,
					"par", d->par,
					"is_root_dir", (d->level == 1)?TRUE:FALSE,
					"id", d->id,
					NULL);
			burgner_isofs_path_table_add_record (BURGNER_ISOFS_PATH_TABLE (pt), rec);
			g_object_unref (rec);
		}

		l_pt = burgner_isofs_typed_path_table_new ();
		g_object_set (l_pt,
				"path_table", pt,
				"type", BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L,
				NULL);
		m_pt = burgner_isofs_typed_path_table_new ();
		g_object_set (m_pt,
				"path_table", pt,
				"type", BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_M,
				NULL);
		g_object_unref (pt);

		pb->items_list = g_list_prepend (pb->items_list, m_pt);
		pb->items_list = g_list_prepend (pb->items_list, l_pt);
	}

	pvd = burgner_isofs_tree_volume_descriptor_new ();
	g_object_set (pvd,
		      "sector", 16,
		      "is_svd", FALSE,
		      "file_structure_version", 1,
		      "lb_size", 2048,
		      "l_pt", l_pt,
		      "m_pt", m_pt,
		      "root_dir_rec", pb->root_dir_rec,
		      "last_item", last_item,
		      NULL);
	burgner_isofs_tvd_info_setup_tvd (pb->tvd_info, pvd);
	burgner_isofs_item_set_mandatory_addr (pvd, TRUE, 0x10);
	pb->vd_list = g_list_prepend (NULL, pvd);

	g_list_foreach (pb->pt_list, (GFunc) _bipb_pt_tree_data_free, NULL);
	g_list_free (pb->pt_list);
	pb->pt_list = NULL;
	pb->root_pt_d = NULL;
	pb->reloc_pt_d = NULL;
	g_hash_table_destroy (pb->fi_to_pt_d);
	pb->fi_to_pt_d = NULL;

	g_hash_table_destroy (pb->fi_to_dir);
	pb->fi_to_dir = NULL;

	g_hash_table_destroy (pb->fi_to_file);
	pb->fi_to_file = NULL;

	if (pb->cache_inodes) {
		g_return_if_fail (pb->inode_to_file != NULL);
		g_hash_table_destroy (pb->inode_to_file);
		pb->inode_to_file = NULL;
	} else {
		g_return_if_fail (pb->inode_to_file == NULL);
	}
}

static void
_bipb_directory (BurgnerIsofsTreeBuilder *tb,
		 BurgnerIsofsFileInfo *parent_info,
		 BurgnerIsofsFileInfo *info,
		 guint level)
{
	BurgnerIsofsPrimaryBuilder *pb = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	BurgnerIsofsItem *rec, *dir, *cur, *par, *par_dir;
	gboolean must_reloc;
	__Bipb_PtTreeData *d;

	if (level == 1)
		return;


	d = _bipb_pt_tree_data_new ();

	must_reloc = (pb->with_reloc && pb->reloc_level > 3 &&
		      level > 3 && ((level - 3) % (pb->reloc_level - 3) == 0));

	par_dir = BURGNER_ISOFS_ITEM (g_hash_table_lookup (pb->fi_to_dir, parent_info));
	g_return_if_fail (par_dir != NULL);

	dir = burgner_isofs_directory_new ();
	g_object_set (dir,
		      "info", info,
		      "with_rrip", pb->with_rrip,
		      NULL);
	g_hash_table_insert (pb->fi_to_dir, info, dir);

	cur = burgner_isofs_directory_dot_record_new ();
	g_object_set (cur,
		      "dir", dir,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);
	par = burgner_isofs_directory_dotdot_record_new ();
	g_object_set (par,
		      "dir", par_dir,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);
	burgner_isofs_directory_append_records (BURGNER_ISOFS_DIRECTORY (dir), cur, par, NULL);
	g_object_unref (cur);
	g_object_unref (par);

	rec = burgner_isofs_directory_subdirectory_record_new ();
	g_object_set (rec,
		      "dir", dir,
		      "info", info,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);

	burgner_isofs_directory_append_record (BURGNER_ISOFS_DIRECTORY (par_dir), rec);

	d->dir = dir;

	if (must_reloc) {
		BurgnerIsofsItem *m_rec;
		gchar *m_id;

		if (!pb->reloc_dir) {
			BurgnerIsofsItem *m_cur, *m_par;
			BurgnerIsofsFileInfo *info;
			GString *info_name;
			__Bipb_PtTreeData *m_d;

			g_return_if_fail (pb->root_dir != NULL);

			pb->reloc_dir = burgner_isofs_directory_new ();

			if (!pb->reloc_dir_id) {
				pb->reloc_dir_id = g_strdup ("RR_MOVED");
			}

			info = burgner_isofs_file_info_new ();

			info_name = g_string_new (pb->reloc_dir_id);
			g_string_prepend_c (info_name, '.');
			g_string_ascii_down (info_name);
			info->name = info_name->str;
			g_string_free (info_name, FALSE);

			info->type = BURGNER_ISOFS_FILE_TYPE_DIRECTORY;
			info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
						BURGNER_ISOFS_FILE_PERMISSION_OWNER_EXEC |
						BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
						BURGNER_ISOFS_FILE_PERMISSION_GROUP_EXEC |
						BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ |
						BURGNER_ISOFS_FILE_PERMISSION_OTHERS_EXEC;

			g_object_set (pb->reloc_dir, "info", info, NULL);
			burgner_isofs_file_info_unref (info);

			m_cur = burgner_isofs_directory_dot_record_new ();
			g_object_set (m_cur,
				      "dir", pb->reloc_dir,
				      "with_susp", pb->with_rrip,
				      "with_rrip", pb->with_rrip,
				      NULL);
			m_par = burgner_isofs_directory_dotdot_record_new ();
			g_object_set (m_par,
				      "dir", pb->root_dir,
				      "with_susp", pb->with_rrip,
				      "with_rrip", pb->with_rrip,
				      NULL);
			burgner_isofs_directory_append_records (BURGNER_ISOFS_DIRECTORY (pb->reloc_dir), m_cur, m_par, NULL);
			g_object_unref (m_cur);
			g_object_unref (m_par);

			m_d = _bipb_pt_tree_data_new ();
			m_d->dir = pb->reloc_dir;
			m_d->par = pb->root_dir;

			m_d->id = g_strdup (pb->reloc_dir_id);
			m_d->level = 2;
			g_return_if_fail (pb->root_pt_d != NULL);
			m_d->par_d = pb->root_pt_d;
			pb->reloc_pt_d = (gpointer) m_d;

			pb->pt_list = g_list_prepend (pb->pt_list, m_d);
			pb->reloc_id_count = 0;

			pb->items_list = g_list_prepend (pb->items_list, g_object_ref (pb->reloc_dir));
		}

		m_id = g_strdup_printf ("%8.8X", pb->reloc_id_count++);

		m_rec = burgner_isofs_directory_subdirectory_record_new ();
		g_object_set (m_rec,
			      "dir", dir,
			      "id", m_id,
			      "with_susp", pb->with_rrip,
			      "with_rrip", pb->with_rrip,
			      "is_relocated", TRUE,
			      NULL);
		d->id = m_id;
		burgner_isofs_directory_append_record (BURGNER_ISOFS_DIRECTORY (pb->reloc_dir), m_rec);
		g_object_unref (m_rec);

		d->par = pb->reloc_dir;
		d->par_d = (__Bipb_PtTreeData *) pb->reloc_pt_d;
	} else {
		__Bipb_PtTreeData *par_d;
		d->par = par_dir;
		d->id = g_strdup (info->identifier);
		par_d = (__Bipb_PtTreeData *) g_hash_table_lookup (pb->fi_to_pt_d, parent_info);
		g_return_if_fail (par_d != NULL);
		d->par_d = par_d;
	}

	d->level = pb->with_reloc && level > 3 ? ((level-3)%(pb->reloc_level-3))+3 : level;
	g_hash_table_insert (pb->fi_to_pt_d, info, d);
	pb->pt_list = g_list_prepend (pb->pt_list, d);
	pb->items_list = g_list_prepend (pb->items_list, dir);
}

static void
_bipb_file (BurgnerIsofsTreeBuilder *tb,
	    BurgnerIsofsFileInfo *parent_info,
	    BurgnerIsofsFileInfo *info,
	    BurgnerIsofsItem *file)
{
	BurgnerIsofsPrimaryBuilder *pb = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	BurgnerIsofsItem *rec, *par_dir;

	par_dir = BURGNER_ISOFS_ITEM (g_hash_table_lookup (pb->fi_to_dir, parent_info));
	g_return_if_fail (par_dir != NULL);

	rec = burgner_isofs_directory_file_record_new ();
	g_object_set (rec,
		      "file", file,
		      "info", info,
		      "with_susp", pb->with_rrip,
		      "with_rrip", pb->with_rrip,
		      NULL);
	burgner_isofs_directory_append_record (BURGNER_ISOFS_DIRECTORY (par_dir), rec);
	g_object_unref (rec);
}

static GList *
_burgner_isofs_primary_builder_get_volume_descriptors (BurgnerIsofsTreeBuilder *tb)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	return self->vd_list;
}

static GList *
_burgner_isofs_primary_builder_get_items (BurgnerIsofsTreeBuilder *tb)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	return self->items_list;
}

static BurgnerIsofsItem *
_burgner_isofs_primary_builder_create_file_real (BurgnerIsofsPrimaryBuilder *self,
						 BurgnerIsofsFileInfo *info,
						 GSList **list)
{
	BurgnerIsofsItem *file;

	if (!info) {
		return NULL;
	}

	if (info->type == BURGNER_ISOFS_FILE_TYPE_REGULAR) {
		g_return_val_if_fail (info->uri != NULL, NULL);
		g_return_val_if_fail (info->uri->path != NULL, NULL);

		if (self->cache_inodes &&
		    info->inode_number &&
		    (file = (BurgnerIsofsItem *) g_hash_table_lookup (self->inode_to_file, &info->inode_number))) {
			file = BURGNER_ISOFS_ITEM (g_object_ref (file));
		} else {
			file = burgner_isofs_file_new ();
			g_object_set (file, "info", info, NULL);

			if (self->cache_inodes && info->inode_number) {
				guint64 *inode = g_new (guint64, 1);

				*inode = info->inode_number;
				g_hash_table_insert (self->inode_to_file, inode, file);
			}
		}
	} else if (info->type == BURGNER_ISOFS_FILE_TYPE_SYMBOLIC_LINK) {
		g_return_val_if_fail (g_slist_index (*list, info->symlink_info) < 0, NULL);

		file = (BurgnerIsofsItem *) g_hash_table_lookup (self->fi_to_file, info->symlink_info);

		if (file) {
			file = BURGNER_ISOFS_ITEM (g_object_ref (file));
		} else {
			*list = g_slist_prepend (*list, info);

			file = _burgner_isofs_primary_builder_create_file_real (self, info->symlink_info, list);
		}
	} else {
		file = NULL;
	}

	return file;
}

static BurgnerIsofsItem *
_burgner_isofs_primary_builder_create_file (BurgnerIsofsTreeBuilder *tb,
					    BurgnerIsofsFileInfo *info)
{
	BurgnerIsofsPrimaryBuilder *self = BURGNER_ISOFS_PRIMARY_BUILDER (tb);
	GSList *l = NULL;

	g_return_val_if_fail (self->fi_to_file != NULL, NULL);

	return _burgner_isofs_primary_builder_create_file_real (self, info, &l);
}

static void
class_init (BurgnerIsofsPrimaryBuilderClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsTreeBuilderClass *builder_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	builder_class = (BurgnerIsofsTreeBuilderClass *) klass;
	builder_class->begin = _bipb_begin;
	builder_class->end = _bipb_end;
	builder_class->directory = _bipb_directory;
	builder_class->file = _bipb_file;
	builder_class->get_volume_descriptors = _burgner_isofs_primary_builder_get_volume_descriptors;
	builder_class->get_items = _burgner_isofs_primary_builder_get_items;
	builder_class->create_file = _burgner_isofs_primary_builder_create_file;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_RELOC_LEVEL,
					 g_param_spec_uint ("reloc_level",
					            	    _("reloc_level"),
							    _("Relocation Level"),
							    3,
							    G_MAXUINT,
							    8,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_RELOC_DIR_ID,
					 g_param_spec_string ("reloc_dir_id",
					            	      _("reloc_dir_id"),
							      _("Reloc Directory Identifier (RR_MOVED by default)"),
							      NULL,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WITH_RELOC,
					 g_param_spec_boolean ("with_reloc",
					            	       _("with_reloc"),
							       _("With Relocation"),
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WITH_RRIP,
					 g_param_spec_boolean ("with_rrip",
					                       _("with_rrip"),
							       _("With RockRidge"),
							       TRUE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_TVD_INFO,
					 g_param_spec_pointer ("tvd_info",
					                       _("tvd_info"),
							       _("a BurgnerIsofsTvdInfo"),
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CACHE_INODES,
					 g_param_spec_boolean ("cache_inodes",
					                       _("cache_inodes"),
							       _("Cache inodes"),
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_NO_PATH_TABLES,
					 g_param_spec_boolean ("no_path_tables",
					                       _("no_path_tables"),
							       _("Do not generate path tables"),
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsPrimaryBuilder *self)
{
	self->vd_list = self->items_list = NULL;

	self->fi_to_pt_d = NULL;
	self->root_pt_d = NULL;
	self->reloc_pt_d = NULL;
	self->pt_list = NULL;

	self->reloc_dir_id = NULL;
	self->reloc_dir = NULL;
	self->reloc_id_count = 0;

	self->fi_to_dir = NULL;
	self->inode_to_file = NULL;

	self->tvd_info = g_new0 (BurgnerIsofsTvdInfo, 1);
	self->tvd_info->volume_set_size = 1;
	self->tvd_info->volume_sequence_number = 1;
	self->tvd_info->application_id.str = g_strdup ("BURGNER ISOFS LIB");
	self->tvd_info->application_id.type = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR_IDENTIFIER_TYPE_STR;

	self->reloc_level = 9; /* -> max (9 - 1 =) 8 dir levels */

	self->with_reloc = 1;
	self->with_rrip = 1;
	self->cache_inodes = 0;
	self->no_path_tables = 0;
}

GType
burgner_isofs_primary_builder_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsPrimaryBuilderClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsPrimaryBuilder),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_TREE_BUILDER, "BurgnerIsofsPrimaryBuilder", &info, 0);
	}

	return type;
}

BurgnerIsofsTreeBuilder *
burgner_isofs_primary_builder_new (void)
{
	BurgnerIsofsPrimaryBuilder *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_PRIMARY_BUILDER, NULL);


	return BURGNER_ISOFS_TREE_BUILDER (self);
}


GQuark
burgner_isofs_primary_builder_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsPrimaryBuilder");
}

