/*
 * BurgnerIsofsRripLoc - BurgnerIsofsRripLoc object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-rrip-loc.h"


static BurgnerIsofsSuspEntryClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_ITEM,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsRripLoc *self = BURGNER_ISOFS_RRIP_LOC (object);

	if (self->item) {
		g_object_unref (self->item);
		self->item = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripLoc *self = BURGNER_ISOFS_RRIP_LOC (object);

	switch (prop_id) {
		case PROP_ITEM:
			if (self->item)
				g_object_unref (self->item);
			self->item = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->item)
				self->item = BURGNER_ISOFS_ITEM (g_object_ref (self->item));
			g_object_notify (object, "item");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripLoc *self = BURGNER_ISOFS_RRIP_LOC (object);

	switch (prop_id) {
		case PROP_ITEM:
			g_value_set_object (value, self->item);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_birl_get_size (BurgnerIsofsItem *item,
		guint32 *n_sectors,
		guint32 *n_bytes)
{
	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = 12;
}

static void
_birl_write_sync (BurgnerIsofsItem *item,
		  BurgnerIsofsMapTable *table,
		  gpointer buf, guint32 size)
{
	__BurgnerIsofsImplRripLoc *impl;
	BurgnerIsofsRripLoc *loc = BURGNER_ISOFS_RRIP_LOC (item);
	gchar sig[2] = { 0, 0 }, *s = NULL;
	gboolean ret;
	guint version = 1;
	guint32 addr;

	g_return_if_fail (size == 12);

	impl = (__BurgnerIsofsImplRripLoc *) buf;
	memset (impl, 0, sizeof (*impl));

	g_object_get (loc,
		      "sig", &s,
		      "version", &version,
		      NULL);
	if (s) {
		if (*s && *(s+1)) {
			sig[0] = *s;
			sig[1] = *(s+1);
		}
		g_free (s);
	}


	if (loc->item) {
		ret = burgner_isofs_map_table_get_addr (table, loc->item, &addr, NULL);
		g_return_if_fail (ret != FALSE);
	} else {
		addr = 0;
	}

	impl->header.sig[0] = sig[0];
	impl->header.sig[1] = sig[1];
	burgner_isofs_set_711 (&impl->header.len, 12);
	burgner_isofs_set_711 (&impl->header.version, version);

	burgner_isofs_set_733 (impl->loc, addr);

#if 0
	burgner_isofs_writer_proxy_write (proxy, (guchar *) &impl, 12);
#else
	buf += 12;
	size -= 12;
#endif

	g_assert (size == 0);
}

static void
class_init (BurgnerIsofsRripLocClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _birl_get_size;
	item_class->write_sync = _birl_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_ITEM,
					 g_param_spec_object ("item",
					            	      "item",
							      "Item",
							      BURGNER_TYPE_ISOFS_ITEM,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsRripLoc *self)
{
	self->item = NULL;
}

GType
burgner_isofs_rrip_loc_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsRripLocClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsRripLoc),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_SUSP_ENTRY, "BurgnerIsofsRripLoc", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_rrip_loc_new (void)
{
	BurgnerIsofsRripLoc *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_RRIP_LOC, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_rrip_loc_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsRripLoc");
}

BurgnerIsofsItem *
burgner_isofs_rrip_pl_new (void)
{
	BurgnerIsofsItem *it = burgner_isofs_rrip_loc_new ();
	g_object_set (it,
		      "sig", "PL",
		      "version", 1,
		      NULL);
	return it;
}

BurgnerIsofsItem *
burgner_isofs_rrip_cl_new (void)
{
	BurgnerIsofsItem *it = burgner_isofs_rrip_loc_new ();
	g_object_set (it,
		      "sig", "CL",
		      "version", 1,
		      NULL);
	return it;
}

