/*
 * BurgnerIsofsRripPx - BurgnerIsofsRripPx object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-impl.h"
#include "burgner-isofs-file-info.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-susp-entry.h"
#include "burgner-isofs-rrip-px.h"


static BurgnerIsofsSuspEntryClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_INFO,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsRripPx *self = BURGNER_ISOFS_RRIP_PX (object);

	if (self->info) {
		burgner_isofs_file_info_unref (self->info);
		self->info = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripPx *self = BURGNER_ISOFS_RRIP_PX (object);

	switch (prop_id) {
		case PROP_INFO:
			if (self->info)
				burgner_isofs_file_info_unref (self->info);
			self->info = (BurgnerIsofsFileInfo *) g_value_get_pointer (value);
			if (self->info)
				self->info = burgner_isofs_file_info_ref (self->info);
			g_object_notify (object, "info");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripPx *self = BURGNER_ISOFS_RRIP_PX (object);

	switch (prop_id) {
		case PROP_INFO:
			g_value_set_pointer (value, self->info);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_birpx_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = 44;
}

static void
_birpx_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	__BurgnerIsofsImplRripPX *impl;
	BurgnerIsofsRripPx *px = BURGNER_ISOFS_RRIP_PX (item);
	guint32 n = 0;

	g_return_if_fail (px->info != NULL);
	g_return_if_fail (size == 44);

	impl = (__BurgnerIsofsImplRripPX *) buf;
	memset (impl, 0, sizeof (*impl));

	impl->header.sig[0] = 'P';
	impl->header.sig[1] = 'X';
	burgner_isofs_set_711 (&impl->header.len, 44);
	burgner_isofs_set_711 (&impl->header.version, 1);

	n = (px->info->type != BURGNER_ISOFS_FILE_TYPE_UNKNOWN ? px->info->type : BURGNER_ISOFS_FILE_TYPE_REGULAR) |
	    px->info->permissions;
	burgner_isofs_set_733 (impl->mode, n);

	burgner_isofs_set_733 (impl->links, px->info->link_count);


	burgner_isofs_set_733 (impl->uid, px->info->uid);
	burgner_isofs_set_733 (impl->gid, px->info->gid);

	burgner_isofs_set_733 (impl->inode, px->info->inode_number);
}

static void
class_init (BurgnerIsofsRripPxClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _birpx_get_size;
	item_class->write_sync = _birpx_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_INFO,
					 g_param_spec_pointer ("info",
					            	       "info",
							       "Info",
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsRripPx *self)
{
	self->info = NULL;
}

GType
burgner_isofs_rrip_px_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsRripPxClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsRripPx),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_SUSP_ENTRY, "BurgnerIsofsRripPx", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_rrip_px_new (void)
{
	BurgnerIsofsRripPx *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_RRIP_PX, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_rrip_px_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsRripPx");
}

