/*
 * BurgnerIsofsRripSl - BurgnerIsofsRripSl object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "../libburgnerutils/g-uri-path-parser.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-rrip-sl.h"


static BurgnerIsofsSuspEntryClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_COMPONENT_AREA,
	PROP_COMPONENT_AREA_LEN,
	PROP_CONTINUE,
	BURGNER_ISOFS_RRIP_SL_LAST_ARG
};

static void
_burgner_isofs_rrip_sl_finalize (GObject *object)
{
	BurgnerIsofsRripSl *self = BURGNER_ISOFS_RRIP_SL (object);

	if (self->ca)
		self->ca = NULL;
	self->ca_len = 0;
	self->cont = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_isofs_rrip_sl_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripSl *self = BURGNER_ISOFS_RRIP_SL (object);

	switch (prop_id) {
		case PROP_COMPONENT_AREA:
			self->ca = g_value_get_pointer (value);
			g_object_notify (object, "ca");
			break;
		case PROP_COMPONENT_AREA_LEN:
			self->ca_len = g_value_get_uint (value);
			g_object_notify (object, "ca_len");
			break;
		case PROP_CONTINUE:
			self->cont = g_value_get_boolean (value);
			g_object_notify (object, "continue");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_isofs_rrip_sl_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsRripSl *self = BURGNER_ISOFS_RRIP_SL (object);

	switch (prop_id) {
		case PROP_COMPONENT_AREA:
			g_value_set_pointer (value, self->ca);
			break;
		case PROP_COMPONENT_AREA_LEN:
			g_value_set_uint (value, self->ca_len);
			break;
		case PROP_CONTINUE:
			g_value_set_boolean (value, self->cont);
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_isofs_rrip_sl_get_size (BurgnerIsofsItem *item,
				 guint32 *n_s, guint32 *n_b)
{
	BurgnerIsofsRripSl *self = BURGNER_ISOFS_RRIP_SL (item);

	g_return_if_fail ((!self->ca_len) || self->ca);

	if (n_s)
		*n_s = 0;
	if (n_b)
		*n_b = 5 + self->ca_len;
}

static void
_burgner_isofs_rrip_sl_write_sync (BurgnerIsofsItem *item,
				   BurgnerIsofsMapTable *table,
				   gpointer buf, guint32 size)
{
	__BurgnerIsofsImplRripSl *impl;
	BurgnerIsofsRripSl *self = BURGNER_ISOFS_RRIP_SL (item);

	g_return_if_fail ((!self->ca_len) || self->ca);
	g_return_if_fail (size == (5 + self->ca_len));

	impl = (__BurgnerIsofsImplRripSl *) buf;
	buf += 5;

	impl->header.sig[0] = 'S';
	impl->header.sig[1] = 'L';
	burgner_isofs_set_711 (&impl->header.len, size);
	burgner_isofs_set_711 (&impl->header.version, 1);

	impl->flags = self->cont ? __BURGNER_ISOFS_IMPL_RRIP_SL_FLAG_CONTINUE : 0;

	memcpy (buf, self->ca, self->ca_len);
}


static void
_burgner_isofs_rrip_sl_class_init (BurgnerIsofsRripSlClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_isofs_rrip_sl_set_property;
	object_class->get_property = _burgner_isofs_rrip_sl_get_property;
	object_class->finalize = _burgner_isofs_rrip_sl_finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _burgner_isofs_rrip_sl_get_size;
	item_class->write_sync = _burgner_isofs_rrip_sl_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_COMPONENT_AREA,
					 g_param_spec_pointer ("ca",
					                _("ca"),
							_("A component area"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_COMPONENT_AREA_LEN,
					 g_param_spec_uint ("ca_len",
					                _("ca_len"),
							_("The length of the component area"),
							0,
							0xff,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CONTINUE,
					 g_param_spec_boolean ("continue",
					                _("continue"),
							_("Whether the continue flag should be set"),
							FALSE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_isofs_rrip_sl_init (BurgnerIsofsRripSl *self)
{
	self->ca = NULL;
	self->ca_len = 0;
	self->cont = 0;
}

GType
burgner_isofs_rrip_sl_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsRripSlClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_isofs_rrip_sl_class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsRripSl),
			0,
			(GInstanceInitFunc) _burgner_isofs_rrip_sl_init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_SUSP_ENTRY, "BurgnerIsofsRripSl", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_rrip_sl_new (void)
{
	BurgnerIsofsRripSl *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_RRIP_SL, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_rrip_sl_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsRripSl");
}


static BurgnerIsofsItem *
_burgner_isofs_rrip_sl_create (GString *buf)
{
	BurgnerIsofsItem *i = burgner_isofs_rrip_sl_new ();
	gpointer b = g_malloc (buf->len);

	memcpy (b, buf->str, buf->len);

	g_object_set (i,
			"ca", b,
			"ca_len", buf->len,
			"continue", TRUE,
			NULL);
	g_object_set_data_full (G_OBJECT (i), "ca_buf", b, g_free);

	g_string_truncate (buf, 0);

	return i;
}

GSList *
burgner_isofs_rrip_sl_new_slist_from_symlink (const gchar *symlink)
{
	GList *path, *it; GSList *r = NULL;
	GString *buf;
	gchar *p;
	gboolean absolute = FALSE;

	g_return_val_if_fail (symlink != NULL, NULL);

	p = g_strdup (symlink);
	path = _g_uri_path_parser_static_parse_path (p, TRUE, &absolute);

	buf = g_string_sized_new (250);

	if (absolute) {
		g_string_append_c (buf, __BURGNER_ISOFS_IMPL_RRIP_SL_COMPONENT_FLAG_ROOT);
		g_string_append_c (buf, 0);
	}

	for (it = path; it; it = it->next) {
		if (buf->len > 248) {
			r = g_slist_prepend (r, _burgner_isofs_rrip_sl_create (buf));
		}

		if (strcmp (it->data, ".") == 0) {
			g_string_append_c (buf, __BURGNER_ISOFS_IMPL_RRIP_SL_COMPONENT_FLAG_CURRENT);
			g_string_append_c (buf, 0);
		} else if (strcmp (it->data, "..") == 0) {
			g_string_append_c (buf, __BURGNER_ISOFS_IMPL_RRIP_SL_COMPONENT_FLAG_PARENT);
			g_string_append_c (buf, 0);
		} else {
			gchar *c = it->data;
			guint l = strlen (it->data);

			while (l) {
				guint rl;

				if (buf->len >= 248) {
					r = g_slist_prepend (r, _burgner_isofs_rrip_sl_create (buf));
				}

				rl = MIN (l, 248 - buf->len);
				l -= rl;

				g_string_append_c (buf, l ? __BURGNER_ISOFS_IMPL_RRIP_SL_COMPONENT_FLAG_CONTINUE : 0);
				g_string_append_c (buf, rl);
				g_string_append_len (buf, c, rl);

				c += rl;
			}
		}
	}

	if (buf->len)
		r = g_slist_prepend (r, _burgner_isofs_rrip_sl_create (buf));

	if (r)
		g_object_set (r->data, "continue", FALSE, NULL);


	if (path)
		g_list_free (path);

	if (p)
		g_free (p);

	g_string_free (buf, TRUE);

	return g_slist_reverse (r);
}

