/*
 * BurgnerIsofsSuspCe - BurgnerIsofsSuspCe object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-susp-continuation-area.h"
#include "burgner-isofs-susp-ce.h"


static BurgnerIsofsSuspEntryClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_CA,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsSuspCe *self = BURGNER_ISOFS_SUSP_CE (object);

	if (self->ca) {
		g_object_unref (self->ca);
		self->ca = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsSuspCe *self = BURGNER_ISOFS_SUSP_CE (object);

	switch (prop_id) {
		case PROP_CA:
			if (self->ca)
				g_object_unref (self->ca);
			self->ca = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->ca)
				self->ca = BURGNER_ISOFS_ITEM (g_object_ref (self->ca));
			g_object_notify (object, "ca");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsSuspCe *self = BURGNER_ISOFS_SUSP_CE (object);

	switch (prop_id) {
		case PROP_CA:
			g_value_set_object (value, self->ca);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bisce_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = 28;
}

static void
_bisce_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	__BurgnerIsofsImplSuspCE *impl;
	BurgnerIsofsSuspCe *self = BURGNER_ISOFS_SUSP_CE (item);
	guint32 s_n, b_o, n_s, n_b;
	gboolean ret;

	g_return_if_fail (self->ca != NULL);
	g_return_if_fail (size == 28);

	impl = (__BurgnerIsofsImplSuspCE *) buf;
	memset (impl, 0, sizeof (*impl));

	s_n = b_o = n_s = n_b = 0;
	ret = burgner_isofs_map_table_get_addr (table, self->ca, &s_n, &b_o);
	g_return_if_fail (ret != FALSE);

	burgner_isofs_item_get_size (self->ca, &n_s, &n_b);
	n_b += n_s*2048;

	impl->header.sig[0] = 'C';
	impl->header.sig[1] = 'E';
	burgner_isofs_set_711 (&impl->header.len, 28);
	burgner_isofs_set_711 (&impl->header.version, 1);

	burgner_isofs_set_733 (impl->sector, s_n);
	burgner_isofs_set_733 (impl->offset, b_o);
	burgner_isofs_set_733 (impl->length, n_b);

#if 0
	burgner_isofs_writer_proxy_write (proxy, (guchar*) &impl, 28);
#endif
}


static void
class_init (BurgnerIsofsSuspCeClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bisce_get_size;
	item_class->write_sync = _bisce_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_CA,
					 g_param_spec_object ("ca",
					            	      "ca",
							      "Continuation Area",
							      BURGNER_TYPE_ISOFS_SUSP_CONTINUATION_AREA,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsSuspCe *self)
{
	self->ca = NULL;
}

GType
burgner_isofs_susp_ce_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsSuspCeClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsSuspCe),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_SUSP_ENTRY, "BurgnerIsofsSuspCe", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_susp_ce_new (void)
{
	BurgnerIsofsSuspCe *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_SUSP_CE, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_susp_ce_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsSuspCe");
}

