/*
 * BurgnerIsofsSuspContinuationArea - BurgnerIsofsSuspContinuationArea object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-susp-entry.h"
#include "burgner-isofs-susp-continuation-area.h"


static BurgnerIsofsItemClass *parent_class = NULL;

static void
finalize (GObject *object)
{
	BurgnerIsofsSuspContinuationArea *self = BURGNER_ISOFS_SUSP_CONTINUATION_AREA (object);

	if (self->entries) {
		g_slist_foreach (self->entries, (GFunc) g_object_unref, NULL);
		g_slist_free (self->entries);
		self->entries = NULL;
	}

	if (self->n_bytes) {
		g_slist_free (self->n_bytes);
		self->n_bytes = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_bisca_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	BurgnerIsofsSuspContinuationArea *ca = BURGNER_ISOFS_SUSP_CONTINUATION_AREA (item);
	GSList *it;
	guint32 b = 0, n_b;

	for (it = ca->entries; it != NULL; it = it->next) {
		n_b = 0;
		burgner_isofs_item_get_size (BURGNER_ISOFS_ITEM (it->data),
					     NULL, &n_b);
		b += n_b;
	}

	if (n_sectors)
		*n_sectors = b >> 11;
	if (n_bytes)
		*n_bytes = b & 0x7ff;
}

static void
_bisca_prepare (BurgnerIsofsItem *item,
		BurgnerIsofsMapTable *table,
		guint32 sector_n,
		guint32 byte_o,
		guint32 last_sector_n,
		guint32 last_byte_o)
{
	BurgnerIsofsSuspContinuationArea *ca = BURGNER_ISOFS_SUSP_CONTINUATION_AREA (item);
	BurgnerIsofsItem *ent;
	GSList *it;
	guint32 s = sector_n, b = byte_o, n_b, l_s, l_b;

	if (ca->n_bytes) {
		g_slist_free (ca->n_bytes);
		ca->n_bytes = NULL;
	}

	for (it = ca->entries; it != NULL; it = it->next) {
		if ((s > last_sector_n) ||
		    (s == last_sector_n && b >= last_byte_o)) {
			g_warning ("Not enough space!");
			break;
		}

		ent = BURGNER_ISOFS_ITEM (it->data);
		burgner_isofs_map_table_set (table, ent, s, b);

		n_b = 0;
		burgner_isofs_item_get_size (ent, NULL, &n_b);
		ca->n_bytes = g_slist_prepend (ca->n_bytes, GUINT_TO_POINTER (n_b));
	
		l_s = s;
		l_b = b;
		b += n_b;
		if (b >= 2048) {
			s += b / 2048;
			b %= 2048;
		}
		if ((s > last_sector_n) ||
		    (s == last_sector_n && b > last_byte_o)) {
			g_warning ("Too big item!");
			break;
		}
		burgner_isofs_item_prepare (ent, table, l_s, l_b, s, b);
	}

	ca->n_bytes = g_slist_reverse (ca->n_bytes);
}

static void
_bisca_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	BurgnerIsofsSuspContinuationArea *ca = BURGNER_ISOFS_SUSP_CONTINUATION_AREA (item);
	GSList *it, *it_b;
	gpointer buf_end = buf + size;

	{
		guint l, l_b;
		l = g_slist_length (ca->entries);
		l_b = g_slist_length (ca->n_bytes);
		g_return_if_fail (l == l_b);
	}

	it_b = ca->n_bytes;
	for (it = ca->entries; it != NULL; it = it->next, it_b = it_b->next) {
#if 0
		burgner_isofs_writer_proxy_begin_item (proxy, BURGNER_ISOFS_ITEM (it->data));
		burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it->data),
					       proxy,
					       table,
					       0,
					       GPOINTER_TO_UINT (it_b->data));
		burgner_isofs_writer_proxy_end_item (proxy, BURGNER_ISOFS_ITEM (it->data));
#else
		burgner_isofs_item_write_sync (BURGNER_ISOFS_ITEM (it->data),
					       table, buf, GPOINTER_TO_UINT (it_b->data));
		buf += GPOINTER_TO_UINT (it_b->data);
#endif
	}

	g_assert (buf == buf_end);
}

static void
class_init (BurgnerIsofsSuspContinuationAreaClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bisca_get_size;
	item_class->prepare = _bisca_prepare;
	item_class->write_sync = _bisca_write_sync;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_none;

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsSuspContinuationArea *self)
{
	self->entries = NULL;
	self->n_bytes = NULL;
}

GType
burgner_isofs_susp_continuation_area_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsSuspContinuationAreaClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsSuspContinuationArea),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsSuspContinuationArea", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_susp_continuation_area_new (void)
{
	BurgnerIsofsSuspContinuationArea *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_SUSP_CONTINUATION_AREA, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_susp_continuation_area_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsSuspContinuationArea");
}

void
burgner_isofs_susp_continuation_area_append (BurgnerIsofsSuspContinuationArea *self,
					     BurgnerIsofsItem *entry)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (self));
	g_return_if_fail (entry != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (entry));

	self->entries = g_slist_append (self->entries, g_object_ref (entry));
}

void
burgner_isofs_susp_continuation_area_prepend (BurgnerIsofsSuspContinuationArea *self,
					      BurgnerIsofsItem *entry)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (self));
	g_return_if_fail (entry != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (entry));

	self->entries = g_slist_prepend (self->entries, g_object_ref (entry));
}

void
burgner_isofs_susp_continuation_area_prepend_slist_take_ownership (BurgnerIsofsSuspContinuationArea *self,
								   GSList *l)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (self));
	g_return_if_fail (l != NULL);

	self->entries = g_slist_concat (l, self->entries);
}

void
burgner_isofs_susp_continuation_area_insert (BurgnerIsofsSuspContinuationArea *self,
					     BurgnerIsofsItem *entry,
					     gint pos)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (self));
	g_return_if_fail (entry != NULL);
	g_return_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (entry));

	if (pos < 0)
		self->entries = g_slist_append (self->entries, g_object_ref (entry));
	else
		self->entries = g_slist_insert (self->entries, g_object_ref (entry), pos);
}

GSList *
burgner_isofs_susp_continuation_area_get_entries (BurgnerIsofsSuspContinuationArea *ca)
{
	g_return_val_if_fail (ca != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_SUSP_CONTINUATION_AREA (ca), NULL);

	return ca->entries;
}

