/*
 * BurgnerIsofsSuspEntry - BurgnerIsofsSuspEntry object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-isofs-i18n.h"

#include <glib-object.h>
#include "burgner-isofs-impl.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-susp-entry.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SIG,
	PROP_VERSION,
	LAST_ARG
};


static void
finalize (GObject *object)
{
	BurgnerIsofsSuspEntry *self = BURGNER_ISOFS_SUSP_ENTRY (object);

	self->sig[0] = 'S';
	self->sig[1] = 'T';
	self->version = 1;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsSuspEntry *self = BURGNER_ISOFS_SUSP_ENTRY (object);

	switch (prop_id) {
		case PROP_SIG: {
			const gchar *s = g_value_get_string (value);
			g_return_if_fail (s != NULL);
			g_return_if_fail (*s != 0);
			g_return_if_fail (*(s+1) != 0);
			self->sig[0] = *s;
			self->sig[0] = *(s+1);
			g_object_notify (object, "sig");
			} break;
		case PROP_VERSION:
			self->version = g_value_get_uint (value);
			g_object_notify (object, "version");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsSuspEntry *self = BURGNER_ISOFS_SUSP_ENTRY (object);

	switch (prop_id) {
		case PROP_SIG: {
			gchar *s = g_strdup_printf ("%c%c", self->sig[0], self->sig[1]);
			g_value_set_string_take_ownership (value, s);
			} break;
		case PROP_VERSION:
			g_value_set_uint (value, self->version);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bise_get_size (BurgnerIsofsItem *item,
		guint32 *n_sectors,
		guint32 *n_bytes)
{
	if (n_sectors)
		*n_sectors = 0;
	if (n_bytes)
		*n_bytes = 4;
}

static void
_bise_write_sync (BurgnerIsofsItem *item,
		  BurgnerIsofsMapTable *table,
		  gpointer buf, guint32 size)
{
	__BurgnerIsofsImplSuspEntryHeader *impl;
	BurgnerIsofsSuspEntry *ent = BURGNER_ISOFS_SUSP_ENTRY (item);

	g_assert (size == 4);

	impl = (__BurgnerIsofsImplSuspEntryHeader *) buf;

	impl->sig[0] = ent->sig[0];
	impl->sig[1] = ent->sig[1];
	burgner_isofs_set_711 (&impl->len, 4);
	burgner_isofs_set_711 (&impl->version, ent->version);

#if 0
	burgner_isofs_writer_proxy_write (proxy, (guchar*) &impl, 4);
#endif
}

static void
class_init (BurgnerIsofsSuspEntryClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bise_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_none;
	item_class->write_sync = _bise_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_SIG,
					 g_param_spec_string ("sig",
					            	      _("sig"),
							      _("Signature"),
							      "ST",
							      G_PARAM_READABLE|G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_VERSION,
					 g_param_spec_uint ("version",
					            	    _("version"),
							    _("Version"),
							    1,
							    0xff,
							    1,
							    G_PARAM_READABLE|G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsSuspEntry *self)
{
	self->sig[0] = 'S';
	self->sig[1] = 'T';
	self->version = 1;
}

GType
burgner_isofs_susp_entry_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsSuspEntryClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsSuspEntry),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsSuspEntry", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_susp_entry_new (void)
{
	BurgnerIsofsSuspEntry *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_SUSP_ENTRY, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_susp_entry_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsSuspEntry");
}

BurgnerIsofsItem *
burgner_isofs_susp_st_new (void)
{
	BurgnerIsofsItem *ent = burgner_isofs_susp_entry_new ();
	g_object_set (ent,
		      "sig", "ST",
		      "version", 1,
		      NULL);
	return ent;
}

BurgnerIsofsItem *
burgner_isofs_rrip_re_new (void)
{
	BurgnerIsofsItem *ent = burgner_isofs_susp_entry_new ();
	g_object_set (ent,
		      "sig", "RE",
		      "version", 1,
		      NULL);
	return ent;
}

BurgnerIsofsItem *
burgner_isofs_susp_entry_dup (BurgnerIsofsSuspEntry *ent)
{
	g_return_val_if_fail (ent != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_ISOFS_SUSP_ENTRY (ent), NULL);

	if (BURGNER_ISOFS_SUSP_ENTRY_GET_CLASS (ent)->dup)
		return BURGNER_ISOFS_SUSP_ENTRY_GET_CLASS (ent)->dup (ent);

	return NULL;
}

