/*
 * BurgnerIsofsTreeVolumeDescriptor - BurgnerIsofsTreeVolumeDescriptor object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-utils.h"
#include "burgner-isofs-volume-descriptor-utils.h"
#include "burgner-isofs-typed-path-table.h"
#include "burgner-isofs-directory-record.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-tree-volume-descriptor.h"


static BurgnerIsofsVolumeDescriptorClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SYSTEM_ID,
	PROP_VOLUME_ID,
	PROP_LAST_ITEM,
	PROP_L_PT,
	PROP_M_PT,
	PROP_ROOT_DIR_REC,
	PROP_VOLUME_SET_ID,
	PROP_PUBLISHER_ID,
	PROP_DATA_PREPARER_ID,
	PROP_APPLICATION_ID,
	PROP_COPYRIGHT_FILE_ID,
	PROP_ABSTRACT_FILE_ID,
	PROP_BIBLIOGRAPHIC_FILE_ID,
	PROP_C_DATE,
	PROP_M_DATE,
	PROP_EX_DATE,
	PROP_EFF_DATE,
	PROP_VOLUME_SET_SIZE,
	PROP_VOLUME_SEQUENCE_NUMBER,
	PROP_LB_SIZE,
	PROP_ESCAPE_SEQ,
	PROP_FILE_STRUCTURE_VERSION,
	PROP_VOLUME_FLAGS,
	PROP_IS_SVD,
	LAST_ARG
};

static inline guint
_bitvd_strlen (gchar *str)
{
	return str ? strlen (str) : 0;
}


static void
finalize (GObject *object)
{
	BurgnerIsofsTreeVolumeDescriptor *self = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR (object);

#define FREE_STR(i)	G_STMT_START { \
	if (self->##i) { \
		g_free (self->##i); \
		self->##i = NULL; \
	}		} G_STMT_END

#define FREE_ID(i)	G_STMT_START { \
	if (self->##i##.type != BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR_IDENTIFIER_TYPE_NONE) \
		burgner_isofs_tree_volume_descriptor_identifier_static_free (&self->##i); \
			} G_STMT_END

#define FREE_DATE(i)	G_STMT_START { \
	self->##i = burgner_isofs_tree_volume_descriptor_date_void; \
			} G_STMT_END

#define FREE_ITEM(i)	G_STMT_START { \
	if (self->##i) { \
		g_object_unref (self->##i); \
		self->##i = NULL; \
	}		} G_STMT_END

	FREE_STR (system_id);
	FREE_STR (volume_id);
	FREE_STR (volume_set_id);
	FREE_STR (copyright_file_id);
	FREE_STR (abstract_file_id);
	FREE_STR (bibliographic_file_id);

	FREE_ITEM (last_item);
	FREE_ITEM (l_pt);
	FREE_ITEM (m_pt);

	FREE_ID (publisher_id);
	FREE_ID (data_preparer_id);
	FREE_ID (application_id);

	FREE_DATE (c_date);
	FREE_DATE (m_date);
	FREE_DATE (ex_date);
	FREE_DATE (eff_date);

	if (self->escape_seq) {
		(void) g_string_free (self->escape_seq, TRUE);
		self->escape_seq = NULL;
	}

	self->volume_set_size = 1;
	self->volume_sequence_number = 1;
	self->lb_size = 2048;
	self->file_structure_version = 1;
	self->volume_flags = 0;
	self->is_svd = 0;


	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsTreeVolumeDescriptor *self = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR (object);

	switch (prop_id) {
#define SET_PROP_STR(I,i)	\
		case PROP_##I: \
			if (self->##i) { \
				g_free (self->##i); \
				self->##i = NULL; \
			} \
			self->##i = (gchar*) g_value_get_string (value); \
			if (self->##i) \
				self->##i = g_strdup (self->##i); \
			g_object_notify (object, #i); \
			break;

#define SET_PROP_ID(I,i)	\
		case PROP_##I: { \
			BurgnerIsofsTreeVolumeDescriptorIdentifier *id = (BurgnerIsofsTreeVolumeDescriptorIdentifier *) g_value_get_pointer (value); \
			burgner_isofs_tree_volume_descriptor_identifier_static_free (&self->##i); \
			if (id) \
				burgner_isofs_tree_volume_descriptor_identifier_static_copy (id, &self->##i); \
			else \
				self->##i = burgner_isofs_tree_volume_descriptor_identifier_void; \
			} \
			g_object_notify (object, #i); \
			break;

#define SET_PROP_DATE(I,i)	\
		case PROP_##I: { \
			BurgnerIsofsTreeVolumeDescriptorDate *d = (BurgnerIsofsTreeVolumeDescriptorDate *) g_value_get_pointer (value); \
			if (d) \
				self->##i = *d; \
			else \
				self->##i = burgner_isofs_tree_volume_descriptor_date_void; \
			} \
			g_object_notify (object, #i); \
			break;

#define SET_PROP_ITEM(I,i)	\
		case PROP_##I: \
			if (self->##i) \
				g_object_unref (self->##i); \
			self->##i = (BurgnerIsofsItem *) g_value_get_object (value); \
			if (self->##i) \
				self->##i = BURGNER_ISOFS_ITEM (g_object_ref (self->##i)); \
			g_object_notify (object, #i); \
			break;

		SET_PROP_STR (SYSTEM_ID, system_id);
		SET_PROP_STR (VOLUME_ID, volume_id);
		SET_PROP_ITEM (LAST_ITEM, last_item);
		SET_PROP_ITEM (L_PT, l_pt);
		SET_PROP_ITEM (M_PT, m_pt);
		SET_PROP_ITEM (ROOT_DIR_REC, root_dir_rec);
		SET_PROP_STR (VOLUME_SET_ID, volume_set_id);
		SET_PROP_ID (PUBLISHER_ID, publisher_id);
		SET_PROP_ID (DATA_PREPARER_ID, data_preparer_id);
		SET_PROP_ID (APPLICATION_ID, application_id);
		SET_PROP_STR (COPYRIGHT_FILE_ID, copyright_file_id);
		SET_PROP_STR (ABSTRACT_FILE_ID, abstract_file_id);
		SET_PROP_STR (BIBLIOGRAPHIC_FILE_ID, bibliographic_file_id);
		SET_PROP_DATE (C_DATE, c_date);
		SET_PROP_DATE (M_DATE, m_date);
		SET_PROP_DATE (EX_DATE, ex_date);
		SET_PROP_DATE (EFF_DATE, eff_date);

		case PROP_VOLUME_SET_SIZE:
			self->volume_set_size = g_value_get_uint (value);
			g_object_notify (object, "volume_set_size");
			break;
		case PROP_VOLUME_SEQUENCE_NUMBER:
			self->volume_sequence_number = g_value_get_uint (value);
			g_object_notify (object, "volume_sequence_number");
			break;
		case PROP_LB_SIZE:
			self->lb_size = g_value_get_uint (value);
			g_object_notify (object, "lb_size");
			break;
		case PROP_ESCAPE_SEQ: {
			GString *s = (GString *) g_value_get_boxed (value);
			if (self->escape_seq) {
				(void) g_string_free (self->escape_seq, TRUE);
				self->escape_seq = NULL;
			}
			if (s)
				self->escape_seq = g_string_new_len (s->str, s->len);
			g_object_notify (object, "escape_seq");
			} break;
		case PROP_FILE_STRUCTURE_VERSION:
			self->file_structure_version = g_value_get_uint (value);
			g_object_notify (object, "file_structure_version");
			break;
		case PROP_VOLUME_FLAGS:
			self->volume_flags = g_value_get_uint (value);
			g_object_notify (object, "volume_flags");
			break;
		case PROP_IS_SVD:
			self->is_svd = g_value_get_boolean (value)?1:0;
			g_object_notify (object, "is_svd");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsTreeVolumeDescriptor *self = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR (object);

	switch (prop_id) {
#define GET_PROP_STR(I,i)	\
		case PROP_##I: \
			g_value_set_string (value, self->##i); \
			break;

#define GET_PROP_ID(I,i)	\
		case PROP_##I: \
			g_value_set_pointer (value, &self->##i); \
			break;

#define GET_PROP_DATE(I,i)	\
		case PROP_##I: \
			g_value_set_pointer (value, &self->##i); \
			break;

#define GET_PROP_ITEM(I,i)	\
		case PROP_##I: \
			g_value_set_object (value, self->##i); \
			break;

#define GET_PROP_UINT(I,i)	\
		case PROP_##I: \
			g_value_set_uint (value, self->##i); \
			break;
		
		GET_PROP_STR (SYSTEM_ID, system_id);
		GET_PROP_STR (VOLUME_ID, volume_id);
		GET_PROP_ITEM (LAST_ITEM, last_item);
		GET_PROP_ITEM (L_PT, l_pt);
		GET_PROP_ITEM (M_PT, m_pt);
		GET_PROP_ITEM (ROOT_DIR_REC, root_dir_rec);
		GET_PROP_STR (VOLUME_SET_ID, volume_set_id);
		GET_PROP_ID (PUBLISHER_ID, publisher_id);
		GET_PROP_ID (DATA_PREPARER_ID, data_preparer_id);
		GET_PROP_ID (APPLICATION_ID, application_id);
		GET_PROP_STR (COPYRIGHT_FILE_ID, copyright_file_id);
		GET_PROP_STR (ABSTRACT_FILE_ID, abstract_file_id);
		GET_PROP_STR (BIBLIOGRAPHIC_FILE_ID, bibliographic_file_id);
		GET_PROP_DATE (C_DATE, c_date);
		GET_PROP_DATE (M_DATE, m_date);
		GET_PROP_DATE (EX_DATE, ex_date);
		GET_PROP_DATE (EFF_DATE, eff_date);
		GET_PROP_UINT (VOLUME_SET_SIZE, volume_set_size);
		GET_PROP_UINT (VOLUME_SEQUENCE_NUMBER, volume_sequence_number);
		GET_PROP_UINT (LB_SIZE, lb_size);
		GET_PROP_UINT (FILE_STRUCTURE_VERSION, file_structure_version);
		GET_PROP_UINT (VOLUME_FLAGS, volume_flags);

		case PROP_ESCAPE_SEQ:
			g_value_set_boxed (value, self->escape_seq);
			break;
		case PROP_IS_SVD:
			g_value_set_boolean (value, self->is_svd?TRUE:FALSE);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bitvd_prepare (BurgnerIsofsItem *item,
		BurgnerIsofsMapTable *table,
		guint32 sector_n,
		guint32 byte_o,
		guint32 last_sector_n,
		guint32 last_byte_o)
{
	BurgnerIsofsTreeVolumeDescriptor *tvd = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR (item);

	g_return_if_fail (tvd->root_dir_rec != NULL);

	burgner_isofs_item_prepare (tvd->root_dir_rec,
				    table,
				    sector_n, byte_o+156,
				    sector_n, byte_o+190);
}

static void
_bitvd_str_write_n_pad (guchar *buf,
			guint len,
			guchar *str,
			guint str_len)
{
	if (str_len > len)
		str_len = len;

	if (str_len)
		memcpy (buf, str, str_len);

	if (str_len != len)
		memset (buf+str_len, 0x20, len-str_len);
}

static void
_bitvd_id_write_n_pad (guchar *buf,
		       guint len,
		       BurgnerIsofsTreeVolumeDescriptorIdentifier *id)
{
	gchar *str;
	guint str_len;

	if (id) {
		str = id->str;
		str_len = _bitvd_strlen (str);
	} else {
		str = NULL;
		str_len = 0;
	}


	if (str_len > len)
		str_len = len;

	if (str_len)
		memcpy (buf, str, str_len);

	if (str_len != len)
		memset (buf+str_len, 0x20, len-str_len);
}

static void
_bitvd_buf_write_n_pad (guchar *buf,
			guint len,
			guchar *str,
			guint str_len,
			guint8 pad_b)
{
	if (str_len > len)
		str_len = len;

	if (str_len)
		memcpy (buf, str, str_len);

	if (str_len != len)
		memset (buf+str_len, pad_b, len-str_len);
}

static void
_bitvd_date_write (guchar *buf,
		   guint len,
		   BurgnerIsofsTreeVolumeDescriptorDate *date)
{
	gchar s[17];

	g_return_if_fail (len >= 17);

	if (!date) {
		memset (buf, 0, 17);
		return;
	}

	(void) g_snprintf (s, 16L,
			   "%4.4d%2.2d%2.2d%2.2d%2.2d%2.2d%2.2d",
			   date->year, date->month, date->day,
			   date->hour, date->minute, date->second,
			   date->hundredth);
	burgner_isofs_set_712 (&(s[16]), date->offset_from_gmt);
	memcpy (buf, s, 17);
}

static void
_bitvd_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	__BurgnerIsofsImplTreeVolumeDescriptorData *impl;
	BurgnerIsofsTreeVolumeDescriptor *tvd = BURGNER_ISOFS_TREE_VOLUME_DESCRIPTOR (item);
	gboolean ret;
	guint32 addr;
	gpointer buf_end = buf + size;

	g_return_if_fail (size == 2048);
	g_return_if_fail (tvd->last_item != NULL);
	g_return_if_fail (tvd->root_dir_rec != NULL);

	impl = (__BurgnerIsofsImplTreeVolumeDescriptorData *) buf;
	memset (impl, 0, sizeof (*impl));

	burgner_isofs_set_711 (&impl->header.type,
				tvd->is_svd ?
				    BURGNER_ISOFS_VOLUME_DESCRIPTOR_TYPE_SUPPLEMENTARY :
				    BURGNER_ISOFS_VOLUME_DESCRIPTOR_TYPE_PRIMARY);

	impl->header.std_id[0] = 'C';
	impl->header.std_id[1] = 'D';
	impl->header.std_id[2] = '0';
	impl->header.std_id[3] = '0';
	impl->header.std_id[4] = '1';
	burgner_isofs_set_711 (&impl->header.version, 1);

	if (tvd->is_svd)
		impl->volume_flags = tvd->volume_flags;

	_bitvd_str_write_n_pad (impl->system_identifier, 32,
				tvd->system_id, _bitvd_strlen (tvd->system_id));
	_bitvd_str_write_n_pad (impl->volume_identifier, 32,
				tvd->volume_id, _bitvd_strlen (tvd->volume_id));

	addr = 0;
	ret = burgner_isofs_map_table_get_addr (table, tvd->last_item, &addr, NULL);
	g_return_if_fail (ret != FALSE);
	burgner_isofs_set_733 (impl->volume_space_size, addr);

	_bitvd_buf_write_n_pad (impl->escape_sequences, 32,
				tvd->escape_seq ? tvd->escape_seq->str : NULL,
				tvd->escape_seq ? tvd->escape_seq->len : 0, 0);

	burgner_isofs_set_723 (impl->volume_set_size, tvd->volume_set_size);
	burgner_isofs_set_723 (impl->volume_sequence_number, tvd->volume_sequence_number);
	burgner_isofs_set_723 (impl->logical_block_size, tvd->lb_size);

	addr = 0;
	if (tvd->l_pt) {
		burgner_isofs_item_get_size (tvd->l_pt, NULL, &addr);
		g_return_if_fail (addr != 0);
	} else if (tvd->m_pt) {
		burgner_isofs_item_get_size (tvd->m_pt, NULL, &addr);
		g_return_if_fail (addr != 0);
	}
	burgner_isofs_set_733 (impl->path_table_size, addr);

	if (tvd->l_pt) {
		addr = 0;
		ret = burgner_isofs_map_table_get_addr (table, tvd->l_pt, &addr, NULL);
		g_return_if_fail (ret != FALSE);
		burgner_isofs_set_733 (impl->l_path_table_location, addr);
	}

	if (tvd->m_pt) {
		addr = 0;
		ret = burgner_isofs_map_table_get_addr (table, tvd->m_pt, &addr, NULL);
		g_return_if_fail (ret != FALSE);
		burgner_isofs_set_733 (impl->m_path_table_location, addr);
	}

	buf += 156;

	burgner_isofs_item_write_sync (tvd->root_dir_rec, table, buf, 34);
	buf += 34;

	_bitvd_str_write_n_pad (impl->volume_set_identifier, 128,
				tvd->volume_set_id, _bitvd_strlen (tvd->volume_set_id));
	_bitvd_id_write_n_pad (impl->publisher_identifier, 128,
			       &tvd->publisher_id);
	_bitvd_id_write_n_pad (impl->data_preparer_identifier, 128,
			       &tvd->data_preparer_id);
	_bitvd_id_write_n_pad (impl->application_identifier, 128,
			       &tvd->application_id);

	_bitvd_str_write_n_pad (impl->copyright_file_identifier, 37,
				tvd->copyright_file_id, _bitvd_strlen (tvd->copyright_file_id));
	_bitvd_str_write_n_pad (impl->abstract_file_identifier, 37,
				tvd->abstract_file_id, _bitvd_strlen (tvd->abstract_file_id));
	_bitvd_str_write_n_pad (impl->bibliographic_file_identifier, 37,
				tvd->bibliographic_file_id, _bitvd_strlen (tvd->bibliographic_file_id));

	_bitvd_date_write (impl->volume_creation_date, 17,
			   &tvd->c_date);
	_bitvd_date_write (impl->volume_modification_date, 17,
			   &tvd->m_date);
	_bitvd_date_write (impl->volume_expiration_date, 17,
			   &tvd->ex_date);
	_bitvd_date_write (impl->volume_effective_date, 17,
			   &tvd->eff_date);

	burgner_isofs_set_711 (&impl->file_structure_version,
			       tvd->file_structure_version);

	buf += 1858;

	g_assert (buf == buf_end);
}


static void
class_init (BurgnerIsofsTreeVolumeDescriptorClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->prepare = _bitvd_prepare;
	item_class->write_sync = _bitvd_write_sync;

	/* Properties */
#define INSTALL_PROP_STR(I,i,d)		\
	g_object_class_install_property (object_class, \
					 PROP_##I, \
					 g_param_spec_string (#i, \
					            	      #i, \
							      d, \
							      NULL, \
							      G_PARAM_READABLE | G_PARAM_WRITABLE));

#define INSTALL_PROP_ID(I,i,d)		\
	g_object_class_install_property (object_class, \
					 PROP_##I, \
					 g_param_spec_pointer (#i, \
					            	       #i, \
							       d, \
							       G_PARAM_READABLE | G_PARAM_WRITABLE));

#define INSTALL_PROP_DATE(I,i,d)	\
	g_object_class_install_property (object_class, \
					 PROP_##I, \
					 g_param_spec_pointer (#i, \
					            	       #i, \
							       d, \
							       G_PARAM_READABLE | G_PARAM_WRITABLE));

#define INSTALL_PROP_ITEM(I,i,d,t)	\
	g_object_class_install_property (object_class, \
					 PROP_##I, \
					 g_param_spec_object (#i, \
					            	      #i, \
							      d, \
							      t, \
							      G_PARAM_READABLE | G_PARAM_WRITABLE));

	INSTALL_PROP_STR (SYSTEM_ID, system_id, "Sytem Identifier");
	INSTALL_PROP_STR (VOLUME_ID, volume_id, "Volume Identifier");
	INSTALL_PROP_ITEM (LAST_ITEM, last_item, "Last Item", BURGNER_TYPE_ISOFS_ITEM);
	INSTALL_PROP_ITEM (L_PT, l_pt, "Type L Path Table", BURGNER_TYPE_ISOFS_TYPED_PATH_TABLE);
	INSTALL_PROP_ITEM (M_PT, m_pt, "Type M Path Table", BURGNER_TYPE_ISOFS_TYPED_PATH_TABLE);
	INSTALL_PROP_ITEM (ROOT_DIR_REC, root_dir_rec, "Root Directory Record", BURGNER_TYPE_ISOFS_DIRECTORY_RECORD);
	INSTALL_PROP_STR (VOLUME_SET_ID, volume_set_id, "Volume Set Identifier");
	INSTALL_PROP_ID (PUBLISHER_ID, publisher_id, "Publisher Identifier");
	INSTALL_PROP_ID (DATA_PREPARER_ID, data_preparer_id, "Data Preparer Identifier");
	INSTALL_PROP_ID (APPLICATION_ID, application_id, "Application Identifier");
	INSTALL_PROP_STR (COPYRIGHT_FILE_ID, copyright_file_id, "Copyright File Identifier");
	INSTALL_PROP_STR (ABSTRACT_FILE_ID, abstract_file_id, "Abstract File Identifier");
	INSTALL_PROP_STR (BIBLIOGRAPHIC_FILE_ID, bibliographic_file_id, "Bibliographic File Identifier");
	INSTALL_PROP_DATE (C_DATE, c_date, "Creation Date");
	INSTALL_PROP_DATE (M_DATE, m_date, "Modification Date");
	INSTALL_PROP_DATE (EX_DATE, ex_date, "Expiration Date");
	INSTALL_PROP_DATE (EFF_DATE, eff_date, "Effective Date");

	g_object_class_install_property (object_class,
					 PROP_VOLUME_SET_SIZE,
					 g_param_spec_uint ("volume_set_size",
					            	    "volume_set_size",
							    "Volume Set Size",
							    1,
							    0xffff,
							    1,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_VOLUME_SEQUENCE_NUMBER,
					 g_param_spec_uint ("volume_sequence_number",
					                    "volume_sequence_number",
							    "Volume Sequence Number",
							    1,
							    0xffff,
							    1,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_LB_SIZE,
					 g_param_spec_uint ("lb_size",
					                    "lb_size",
							    "Logical Block Size",
							    1,
							    0xffff,
							    2048,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_ESCAPE_SEQ,
					 g_param_spec_pointer ("escape_seq",
					            	       "escape_seq",
							       "Escape Sequence (SVD only)",
							       G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_FILE_STRUCTURE_VERSION,
					 g_param_spec_uint ("file_structure_version",
					                    "file_structure_version",
							    "File Structure Version",
							    1,
							    0xff,
							    1,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_VOLUME_FLAGS,
					 /* TODO: use g_param_spec_flags or enum */
					 g_param_spec_uint ("volume_flags",
					                    "volume_flags",
							    "Volume Flags (SVD only)",
							    0,
							    0xff,
							    0,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_IS_SVD,
					 g_param_spec_boolean ("is_svd",
					            	       "is_svd",
							       "Is SVD",
							       FALSE,
							       G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsTreeVolumeDescriptor *self)
{
	self->system_id = self->volume_id = self->volume_set_id =
		self->copyright_file_id = self->abstract_file_id =
		self->bibliographic_file_id = NULL;

	self->last_item = self->l_pt = self->m_pt = self->root_dir_rec = NULL;

	self->publisher_id = self->data_preparer_id = self->application_id =
		burgner_isofs_tree_volume_descriptor_identifier_void;

	self->c_date = self->m_date = self->eff_date = self->ex_date =
		burgner_isofs_tree_volume_descriptor_date_void;

	self->volume_set_size = self->volume_sequence_number = 1;
	self->lb_size = 2048;

	self->escape_seq = NULL;

	self->file_structure_version = 1;

	self->volume_flags = 0;

	self->is_svd = 0;
	
}

GType
burgner_isofs_tree_volume_descriptor_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsTreeVolumeDescriptorClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsTreeVolumeDescriptor),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_VOLUME_DESCRIPTOR, "BurgnerIsofsTreeVolumeDescriptor", &info, 0 /*G_TYPE_FLAG_ABSTRACT*/);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_tree_volume_descriptor_new (void)
{
	BurgnerIsofsTreeVolumeDescriptor *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_TREE_VOLUME_DESCRIPTOR, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_tree_volume_descriptor_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsTreeVolumeDescriptor");
}

