/*
 * 
 *
 * This file is part of the libburgnerisofs library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include "burgner-isofs-identifier.h"
#include "burgner-isofs-tree.h"

static void
_burgner_isofs_tree_value_free (gpointer v)
{
	g_list_free (*((GList **) v));
	g_free (v);
}

BurgnerIsofsTree *
burgner_isofs_tree_new (void)
{
	BurgnerIsofsTree *tree = g_new0 (BurgnerIsofsTree, 1);
	BurgnerIsofsFileInfo *root_info = burgner_isofs_file_info_new ();

	root_info->type = BURGNER_ISOFS_FILE_TYPE_DIRECTORY;
	root_info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
				 BURGNER_ISOFS_FILE_PERMISSION_OWNER_EXEC |
				 BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
				 BURGNER_ISOFS_FILE_PERMISSION_GROUP_EXEC |
				 BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ |
				 BURGNER_ISOFS_FILE_PERMISSION_OTHERS_EXEC;
	root_info->is_directory = 1;

	tree->root = g_node_new (root_info);
	root_info->node = tree->root;

	tree->name_to_node = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, _burgner_isofs_tree_value_free);

	tree->refcount = 1;

	return tree;
}

BurgnerIsofsTree *
burgner_isofs_tree_ref (BurgnerIsofsTree *tree)
{
	g_return_val_if_fail (tree != NULL, NULL);

	tree->refcount++;
	return tree;
}

void
burgner_isofs_tree_unref (BurgnerIsofsTree *tree)
{
	g_return_if_fail (tree != NULL);
	g_return_if_fail (tree->refcount > 0);

	tree->refcount--;
	if (tree->refcount == 0)
		burgner_isofs_tree_free (tree);
}


static gboolean
_burgner_isofs_tree_root_foreach_unref (GNode *node,
					gpointer user_data)
{
	BurgnerIsofsFileInfo *info = (BurgnerIsofsFileInfo *) user_data;
	info->node = NULL;
	burgner_isofs_file_info_unref (info);
	return FALSE;
}

void
burgner_isofs_tree_free (BurgnerIsofsTree *tree)
{
	g_return_if_fail (tree != NULL);

	if (tree->root) {
		g_node_traverse (tree->root,
				 G_PRE_ORDER,
				 G_TRAVERSE_ALL,
				 -1,
				 _burgner_isofs_tree_root_foreach_unref,
				 NULL);
		g_node_destroy (tree->root);
	}

	if (tree->name_to_node)
		g_hash_table_destroy (tree->name_to_node);

	g_free (tree);
}



BurgnerIsofsFileInfo *
burgner_isofs_tree_get_root_info (BurgnerIsofsTree *tree)
{
	g_return_val_if_fail (tree != NULL, NULL);
	g_return_val_if_fail (tree->root != NULL, NULL);

	return (BurgnerIsofsFileInfo *) tree->root->data;
}


static GNode *
_burgner_isofs_tree_get_children_from_list (GSList *l,
					    GNode *parent)
{
	while (l) {
		if (((GNode *) l->data)->parent == parent)
			return (GNode *) l->data;
	}

	return NULL;
}

BurgnerIsofsFileInfo *
burgner_isofs_tree_get_info_from_path (BurgnerIsofsTree *tree,
					const gchar *path,
					gboolean create)
{
	BurgnerIsofsFileInfo *ret;
	GNode *parent;
	gchar **p;
	gint i;

	g_return_val_if_fail (tree != NULL, NULL);
	g_return_val_if_fail (path != NULL, NULL);

	while (*path == '/')
		path++;

	ret = (BurgnerIsofsFileInfo *) tree->root->data;

	parent = tree->root;
	p = g_strsplit (path, "/", -1);

	if (!p[i])
		return ret;

	for (i = 0; p[i+1] != NULL; i++) {
		GNode *c;
		GSList **l;

		if (!*p[i])
			continue;

		l = g_hash_table_lookup (tree->name_to_node, p[i]);

		if (l && *l && (c = _burgner_isofs_tree_get_children_from_list (*l, parent))) {
			parent = c;
			ret = (BurgnerIsofsFileInfo *) c->data;
			continue;
		}

		if (!create) {
			g_strfreev (p);
			return NULL;
		}

		while (p[i+1]) {
			BurgnerIsofsFileInfo *info = burgner_isofs_file_info_new ();

			info->is_directory = 1;
			info->type = BURGNER_ISOFS_FILE_TYPE_DIRECTORY;
			info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
						BURGNER_ISOFS_FILE_PERMISSION_OWNER_EXEC |
						BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
						BURGNER_ISOFS_FILE_PERMISSION_GROUP_EXEC |
						BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ |
						BURGNER_ISOFS_FILE_PERMISSION_OTHERS_EXEC;
			info->name = p[i];
			p[i] = NULL;

			c = g_node_new (info);
			info->node = c;
			g_node_prepend (parent, c);

			if (!(l = g_hash_table_lookup (tree->name_to_node, info->name))) {
				l = g_new0 (GSList*, 1);
				g_hash_table_insert (tree->name_to_node, g_strdup (info->name), l);
			}
			*l = g_slist_prepend (*l, c);

			ret = info;
			parent = c;
			i++;
		}

		break;
	}

	if (*p[i]) {
		BurgnerIsofsFileInfo *info = burgner_isofs_file_info_new ();
		GNode *c;

		info->type = BURGNER_ISOFS_FILE_TYPE_REGULAR;
		info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
					BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
					BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ;
		info->name = p[i];
		p[i] = NULL;

		c = g_node_new (info);
		info->node = c;
		g_node_prepend (parent, c);

		ret = info;
	}

	g_strfreev (p);
	return ret;
}

BurgnerIsofsFileInfo *
burgner_isofs_tree_add_file (BurgnerIsofsTree *tree,
			     BurgnerIsofsFileInfo *parent_info,
			     const gchar *name,
			     gboolean in_prev_session)
{
	BurgnerIsofsFileInfo *info;
	GNode *node;

	g_return_val_if_fail (tree != NULL, NULL);
	g_return_val_if_fail (parent_info != NULL, NULL);
	g_return_val_if_fail (parent_info->node != NULL, NULL);
	g_return_val_if_fail (tree->root == parent_info->node || g_node_is_ancestor (tree->root, parent_info->node), NULL);
	g_return_val_if_fail (name != NULL, NULL);

	info = burgner_isofs_file_info_new ();
	info->name = g_strdup (name);
	info->type = BURGNER_ISOFS_FILE_TYPE_REGULAR;
	info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
				BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
				BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ;
	info->is_in_prev_session = in_prev_session;

	node = g_node_new (info);
	info->node = node;
	g_node_prepend (parent_info->node, node);

	return info;
}

BurgnerIsofsFileInfo *
burgner_isofs_tree_add_directory (BurgnerIsofsTree *tree,
				  BurgnerIsofsFileInfo *parent_info,
				  const gchar *name,
				  gboolean in_prev_session)
{
	BurgnerIsofsFileInfo *info;
	GNode *node;
	GSList **l;

	g_return_val_if_fail (tree != NULL, NULL);
	g_return_val_if_fail (parent_info != NULL, NULL);
	g_return_val_if_fail (parent_info->node != NULL, NULL);
	g_return_val_if_fail (tree->root == parent_info->node || g_node_is_ancestor (tree->root, parent_info->node), NULL);
	g_return_val_if_fail (name != NULL, NULL);

	info = burgner_isofs_file_info_new ();
	info->name = g_strdup (name);
	info->type = BURGNER_ISOFS_FILE_TYPE_DIRECTORY;
	info->permissions = BURGNER_ISOFS_FILE_PERMISSION_OWNER_READ |
				BURGNER_ISOFS_FILE_PERMISSION_OWNER_EXEC |
				BURGNER_ISOFS_FILE_PERMISSION_GROUP_READ |
				BURGNER_ISOFS_FILE_PERMISSION_GROUP_EXEC |
				BURGNER_ISOFS_FILE_PERMISSION_OTHERS_READ |
				BURGNER_ISOFS_FILE_PERMISSION_OTHERS_EXEC;
	info->is_in_prev_session = in_prev_session;
	info->is_directory = 1;

	node = g_node_new (info);
	info->node = node;
	g_node_prepend (parent_info->node, node);

	l = (GSList **) g_hash_table_lookup (tree->name_to_node, info->name);
	if (!l) {
		l = g_new0 (GSList*, 1);
		g_hash_table_insert (tree->name_to_node, g_strdup (name), l);
	}

	*l = g_slist_prepend (*l, node);

	return info;
}


static guint
_get_num_length (guint num, guint base)
{
	guint b = base, n = 1;

	while (num >= b) {
		n++;
		b *= base;
	}

	return n;
}

static gboolean
_burgner_isofs_tree_create_id_traverse_func (GNode *node,
					     gpointer user_data)
{
	BurgnerIsofsIdentifierOptions options = GPOINTER_TO_INT (user_data);
	GNode *node_it;
	GList *l = NULL;

	for (node_it = node->children; node_it != NULL; node_it = node_it->next) {
		BurgnerIsofsFileInfo *info = (BurgnerIsofsFileInfo *) node_it->data;

		g_return_val_if_fail (info->name != NULL, TRUE);

		if (info->identifier)
			g_free (info->identifier);

		if (info->is_directory) {
			info->identifier = burgner_isofs_directory_identifier_from_name (info->name, options);
		} else {
			info->identifier= burgner_isofs_file_identifier_from_name (info->name, options);
		}

		l = g_list_prepend (l, info->identifier);
	}

	if (options & BURGNER_ISOFS_IDENTIFIER_UNIQUE) {
		GList *it;

		l = g_list_sort (l, (GCompareFunc) strcmp);

		if (it) {
			for (it = l; it->next != NULL; it = it->next) {
				if (strcmp (it->data, it->next->data) == 0) {
					GList *sl = NULL, *sl_it;
					GString *str;
					gint i;
					GList *n_it;
					guint n, nn, name_l, name_part_l;
					gchar *dot;

					sl = g_list_prepend (sl, it->data);
					sl = g_list_prepend (sl, it->next->data);

					for (n_it = it->next->next; n_it && strcmp (it->data, n_it->data) == 0;
					     n_it = n_it->next) {
						sl = g_list_prepend (sl, n_it->data);
					}

					n = g_list_length (sl);
					n = _get_num_length (n, 10) + 1;

					name_l = strlen (it->data);

					dot = strrchr ((gchar *) it->data, '.');
					name_part_l = dot ? dot - ((gchar *) it->data) : name_l;
					nn = n;

					do {
						GList *s_it;
						gint pref_l = name_part_l - nn;

						if (pref_l <= 0) {
							/* Aaarghh */
							g_list_free (sl);
							return FALSE;
						}

						n = nn;

						for (s_it = n_it;
						     s_it && (strncmp ((gchar *) s_it->data, (gchar *) it->data, pref_l) == 0) && (strcmp (((gchar *) s_it->data) + pref_l + n, ((gchar *) it->data) + pref_l + n) == 0);
						     s_it = s_it->next) {
							sl = g_list_prepend (sl, s_it->data);
						}

						for (s_it = it->prev;
						     s_it && strncmp ((gchar *) s_it->data, (gchar *) it->data, pref_l) == 0 && strncmp (((gchar *) s_it->data) + pref_l + n, ((gchar *) it->data) + pref_l + n, (name_l - pref_l) - n) == 0;
						     s_it = s_it->next) {
							sl = g_list_prepend (sl, s_it->data);
						}

						nn = g_list_length (sl);
						nn = _get_num_length (nn, 10) + 1;
					} while (nn != n);

					str = g_string_sized_new (n+1);
					for (sl_it = sl, i = 0; sl_it; sl_it = sl_it->next, i++) {
						g_string_printf (str, "%d", i);
						*(((gchar *) sl_it->data) + name_part_l - n) = '~';
						memset (((gchar *) sl_it->data) + name_part_l + 1 - n, '0', n - 1);
						memcpy (((gchar *) sl_it->data) + name_part_l - str->len, str->str, str->len);
					}
					g_string_free (str, TRUE);

					g_list_free (sl);
				}
			}
		}
	}

	g_list_free (l);

	return FALSE;
}

void
burgner_isofs_tree_create_identifiers (BurgnerIsofsTree *tree,
				       BurgnerIsofsIdentifierOptions options)
{
	g_return_if_fail (tree != NULL);

	g_node_traverse (tree->root,
			 G_PRE_ORDER, G_TRAVERSE_NON_LEAFS, -1,
			 _burgner_isofs_tree_create_id_traverse_func, GINT_TO_POINTER (options));
}

GNode *
burgner_isofs_tree_get_root_node (BurgnerIsofsTree *tree)
{
	g_return_val_if_fail (tree != NULL, NULL);
	return tree->root;
}

