/*
 * BurgnerIsofsTypedPathTable - BurgnerIsofsTypedPathTable object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-path-table.h"
#include "burgner-isofs-typed-path-table.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_PATH_TABLE,
	PROP_TYPE,
	LAST_ARG
};


static void
finalize (GObject *object)
{
	BurgnerIsofsTypedPathTable *self = BURGNER_ISOFS_TYPED_PATH_TABLE (object);

	if (self->pt) {
		g_object_unref (self->pt);
		self->pt = NULL;
	}

	self->type = BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsTypedPathTable *self = BURGNER_ISOFS_TYPED_PATH_TABLE (object);

	switch (prop_id) {
		case PROP_PATH_TABLE:
			if (self->pt == g_value_get_object (value))
				break;

			if (self->pt)
				g_object_unref (self->pt);
			self->pt = (BurgnerIsofsItem *) g_value_get_object (value);
			if (self->pt)
				self->pt = BURGNER_ISOFS_ITEM (g_object_ref (self->pt));

			g_object_notify (object, "path_table");
			break;
		case PROP_TYPE:
			self->type = g_value_get_enum (value);
			g_object_notify (object, "type");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsTypedPathTable *self = BURGNER_ISOFS_TYPED_PATH_TABLE (object);

	switch (prop_id) {
		case PROP_PATH_TABLE:
			g_value_set_object (value, self->pt);
			break;
		case PROP_TYPE:
			g_value_set_enum (value, self->type);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bitpt_get_size (BurgnerIsofsItem *item,
		 guint32 *n_sectors,
		 guint32 *n_bytes)
{
	BurgnerIsofsTypedPathTable *tpt = BURGNER_ISOFS_TYPED_PATH_TABLE (item);

	g_return_if_fail (tpt->pt != NULL);

	burgner_isofs_item_get_size (tpt->pt, n_sectors, n_bytes);
}

/* XXX Implement prepare  ...? :/ */
		
static void
_bitpt_write_sync (BurgnerIsofsItem *item,
		   BurgnerIsofsMapTable *table,
		   gpointer buf, guint32 size)
{
	BurgnerIsofsTypedPathTable *tpt = BURGNER_ISOFS_TYPED_PATH_TABLE (item);

	g_return_if_fail (tpt->pt != NULL);

	g_object_set (tpt->pt,
		      "type", tpt->type,
		      NULL);

#if 0
	burgner_isofs_writer_proxy_begin_item (proxy, tpt->pt);
	burgner_isofs_item_write_sync (tpt->pt,
				       proxy,
				       table,
				       n_sectors,
				       n_bytes);
	burgner_isofs_writer_proxy_end_item (proxy, tpt->pt);
#else
	burgner_isofs_item_write_sync (tpt->pt, table, buf, size);
#endif
}

static void
_bitpt_prepare (BurgnerIsofsItem *item,
		BurgnerIsofsMapTable *table,
		guint32 sector_n,
		guint32 byte_o,
		guint32 last_sector_n,
		guint32 last_byte_o)
{
	BurgnerIsofsTypedPathTable *tpt = BURGNER_ISOFS_TYPED_PATH_TABLE (item);

	g_return_if_fail (tpt->pt != NULL);

	burgner_isofs_item_prepare (tpt->pt, table, sector_n, byte_o, last_sector_n, last_byte_o);
}

static void
class_init (BurgnerIsofsTypedPathTableClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bitpt_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_sector_start_aligned;
	item_class->prepare=  _bitpt_prepare;
	item_class->write_sync = _bitpt_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_PATH_TABLE,
					 g_param_spec_object ("path_table",
					                      "path_table",
							      "Path Table",
							      BURGNER_TYPE_ISOFS_PATH_TABLE,
							      G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_TYPE,
					 g_param_spec_enum ("type",
					                    "type",
							    "Type",
							    BURGNER_TYPE_ISOFS_PATH_TABLE_RECORD_TYPE,
							    BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsTypedPathTable *self)
{
	self->pt = NULL;
	self->type = BURGNER_ISOFS_PATH_TABLE_RECORD_TYPE_L;
}

GType
burgner_isofs_typed_path_table_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsTypedPathTableClass),
			NULL, NULL, (GClassInitFunc) class_init, NULL, NULL,
			sizeof (BurgnerIsofsTypedPathTable), 0, (GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsTypedPathTable", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_typed_path_table_new (void)
{
	BurgnerIsofsTypedPathTable *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_TYPED_PATH_TABLE, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_typed_path_table_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsTypedPathTable");
}

