/*
 * BurgnerIsofsVoidItem - BurgnerIsofsVoidItem object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-void-item.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SIZE,
	LAST_ARG
};

static void
finalize (GObject *object)
{
	BurgnerIsofsVoidItem *self = BURGNER_ISOFS_VOID_ITEM (object);

	self->n_b = 0;
	self->flags = BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED | BURGNER_ISOFS_ITEM_BLANK;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsVoidItem *self = BURGNER_ISOFS_VOID_ITEM (object);

	switch (prop_id) {
		case PROP_SIZE:
			self->n_b = g_value_get_uint (value);
			g_object_notify (object, "size");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsVoidItem *self = BURGNER_ISOFS_VOID_ITEM (object);

	switch (prop_id) {
		case PROP_SIZE:
			g_value_set_uint (value, self->n_b);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bivi_get_size (BurgnerIsofsItem *item,
		guint32 *n_sectors,
		guint32 *n_bytes)
{
	BurgnerIsofsVoidItem *vi = BURGNER_ISOFS_VOID_ITEM (item);

	if (n_sectors)
		*n_sectors = vi->n_b >> 11;
	if (n_bytes)
		*n_bytes = vi->n_b & 0x7ff;
}

static void
_bivi_write_sync (BurgnerIsofsItem *item,
		  BurgnerIsofsMapTable *table,
		  gpointer buf, guint32 size)
{
	BurgnerIsofsVoidItem *vi = BURGNER_ISOFS_VOID_ITEM (item);

	g_return_if_fail (size == vi->n_b);

#if 0
	burgner_isofs_writer_proxy_write_n_byte (proxy, 0, vi->n_b);
#else
	memset (buf, 0, vi->n_b);
#endif
}

static BurgnerIsofsItemFlags
_burgner_isofs_void_item_get_flags (BurgnerIsofsItem *item)
{
	return BURGNER_ISOFS_VOID_ITEM (item)->flags;
}

static void
class_init (BurgnerIsofsVoidItemClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = set_property;
	object_class->get_property = get_property;
	object_class->finalize = finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = _bivi_get_size;
	item_class->write_sync = _bivi_write_sync;
	item_class->get_flags = _burgner_isofs_void_item_get_flags;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_SIZE,
					 g_param_spec_uint ("size",
					            	    "size",
							    "Size",
							    0,
							    G_MAXUINT,
							    0,
							    G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsVoidItem *self)
{
	self->flags = BURGNER_ISOFS_ITEM_SECTOR_START_ALIGNED | BURGNER_ISOFS_ITEM_BLANK;
	self->n_b = 0;
}

GType
burgner_isofs_void_item_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsVoidItemClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsVoidItem),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsVoidItem", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_void_item_new (void)
{
	BurgnerIsofsVoidItem *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_VOID_ITEM, NULL);


	return BURGNER_ISOFS_ITEM (self);
}


GQuark
burgner_isofs_void_item_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsVoidItem");
}

BurgnerIsofsItemFlags
burgner_isofs_void_item_add_flags (BurgnerIsofsVoidItem *self, BurgnerIsofsItemFlags flags)
{
	g_return_val_if_fail (self != NULL, BURGNER_ISOFS_ITEM_FLAGS_NONE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_VOID_ITEM (self), BURGNER_ISOFS_ITEM_FLAGS_NONE);

	self->flags |= flags;
	return self->flags;
}

BurgnerIsofsItemFlags
burgner_isofs_void_item_remove_flags (BurgnerIsofsVoidItem *self, BurgnerIsofsItemFlags flags)
{
	g_return_val_if_fail (self != NULL, BURGNER_ISOFS_ITEM_FLAGS_NONE);
	g_return_val_if_fail (BURGNER_IS_ISOFS_VOID_ITEM (self), BURGNER_ISOFS_ITEM_FLAGS_NONE);

	self->flags &= ~flags;
	return self->flags;
}

