/*
 * BurgnerIsofsVolumeDescriptor - BurgnerIsofsVolumeDescriptor object
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib-object.h>
#include "burgner-isofs-volume-descriptor-utils.h"
#include "burgner-isofs-item-utils.h"
#include "burgner-isofs-utils.h"
#include "burgner-isofs-impl.h"
#include "burgner-isofs-volume-descriptor.h"


static BurgnerIsofsItemClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SECTOR,
	LAST_ARG
};

static void
_bivd_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsVolumeDescriptor *self = BURGNER_ISOFS_VOLUME_DESCRIPTOR (object);

	switch (prop_id) {
		case PROP_SECTOR:
			self->sector = g_value_get_int (value);
			g_object_notify (object, "sector");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_bivd_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerIsofsVolumeDescriptor *self = BURGNER_ISOFS_VOLUME_DESCRIPTOR (object);

	switch (prop_id) {
		case PROP_SECTOR:
			g_value_set_int (value, self->sector);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}


static void
vd_get_size (BurgnerIsofsItem *item,
	     guint32 *n_sectors,
	     guint32 *n_bytes)
{
	*n_sectors = 1;
	*n_bytes = 0;
}

static void
vd_write_sync (BurgnerIsofsItem *item,
	       BurgnerIsofsMapTable *table,
	       gpointer buf, guint32 size)
{
	__BurgnerIsofsImplVolumeDescriptorHeader *header;

	g_return_if_fail (size == 2048);

	header = (__BurgnerIsofsImplVolumeDescriptorHeader *) buf;
	memset (header, 0, sizeof (*header));

	burgner_isofs_set_711 (&header->type, BURGNER_ISOFS_VOLUME_DESCRIPTOR_TYPE_SET_TERMINATOR);
	header->std_id[0] = 'C';
	header->std_id[1] = 'D';
	header->std_id[2] = '0';
	header->std_id[3] = '0';
	header->std_id[4] = '1';
	burgner_isofs_set_711 (&header->version, 1);

#if 0
	burgner_isofs_writer_proxy_write (proxy, (guchar*) &header, 7);
	burgner_isofs_writer_proxy_write_n_byte (proxy, 0, 2041);
#else
	memset (buf+7, 0, 2041);
#endif
}

static void
_bivd_finalize (GObject *object)
{
	BurgnerIsofsVolumeDescriptor *vd = BURGNER_ISOFS_VOLUME_DESCRIPTOR (object);

	vd->sector = -1;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
class_init (BurgnerIsofsVolumeDescriptorClass *klass)
{
	GObjectClass *object_class;
	BurgnerIsofsItemClass *item_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _bivd_set_property;
	object_class->get_property = _bivd_get_property;
	object_class->finalize = _bivd_finalize;

	item_class = (BurgnerIsofsItemClass *) klass;
	item_class->get_size = vd_get_size;
	item_class->get_flags = burgner_isofs_item_utils_get_flags_sector_start_aligned;
	item_class->write_sync = vd_write_sync;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_SECTOR,
					 g_param_spec_int ("sector",
					                   "sector",
							   "Sector",
							   G_MININT,
							   G_MAXINT,
							   -1,
							   G_PARAM_READABLE | G_PARAM_WRITABLE));

	parent_class = g_type_class_peek_parent (klass);
}

static void
init (BurgnerIsofsVolumeDescriptor *vd)
{
	vd->sector = -1;
}

GType
burgner_isofs_volume_descriptor_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerIsofsVolumeDescriptorClass),
			NULL,
			NULL,
			(GClassInitFunc) class_init,
			NULL,
			NULL,
			sizeof (BurgnerIsofsVolumeDescriptor),
			0,
			(GInstanceInitFunc) init,
		};

		type = g_type_register_static (BURGNER_TYPE_ISOFS_ITEM, "BurgnerIsofsVolumeDescriptor", &info, 0);
	}

	return type;
}

BurgnerIsofsItem *
burgner_isofs_volume_descriptor_new (void)
{
	BurgnerIsofsVolumeDescriptor *self;

	self = g_object_new (BURGNER_TYPE_ISOFS_VOLUME_DESCRIPTOR, NULL);


	return BURGNER_ISOFS_ITEM (self);

}

GQuark
burgner_isofs_volume_descriptor_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsVolumeDescriptor");
}


gint
burgner_isofs_volume_descriptor_cmp (gconstpointer a,
				     gconstpointer b)
{
	BurgnerIsofsVolumeDescriptor *va, *vb;

	g_return_val_if_fail (a != NULL, 1);
	g_return_val_if_fail (BURGNER_IS_ISOFS_VOLUME_DESCRIPTOR (a), 1);
	g_return_val_if_fail (b != NULL, -1);
	g_return_val_if_fail (BURGNER_IS_ISOFS_VOLUME_DESCRIPTOR (b), -1);

	va = BURGNER_ISOFS_VOLUME_DESCRIPTOR (a);
	vb = BURGNER_ISOFS_VOLUME_DESCRIPTOR (b);

	if (va->sector >= 0) {
		if (vb->sector < 0)
			return -1;
		return va->sector - vb->sector;
	}

	if (vb->sector >= 0)
		return 1;

	return 0;
}
