/*
 * BurgnerIsofsWriter - BurgnerIsofsWriter object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib.h>
#include "burgner-isofs-list-iteration.h"
#include "burgner-isofs-writer.h"

#if 0
#define _BURGNER_ISOFS_WRITER_USE_MEM_CHUNK
#define _BURGNER_ISOFS_WRITER_N_PREALLOC 2
#endif

#ifdef _BURGNER_ISOFS_WRITER_USE_MEM_CHUNK
static GMemChunk *_burgner_isofs_writer_mem_chunk = NULL;
G_LOCK_DEFINE_STATIC (_burgner_isofs_writer_mem_chunk);
#endif

BurgnerIsofsWriter *
burgner_isofs_writer_new (GList *list)
{
	BurgnerIsofsWriter *self;

#ifdef _BURGNER_ISOFS_WRITER_USE_MEM_CHUNK
	G_LOCK (_burgner_isofs_writer_mem_chunk);
	if (!_burgner_isofs_writer_mem_chunk) {
		_burgner_isofs_writer_mem_chunk = g_mem_chunk_new ("BurgnerIsofsWriter mem chunk",
		 			sizeof (BurgnerIsofsWriter),
					_BURGNER_ISOFS_WRITER_N_PREALLOC * sizeof (BurgnerIsofsWriter),
					G_ALLOC_AND_FREE);
	}
	self = (BurgnerIsofsWriter *) g_mem_chunk_alloc0 (_burgner_isofs_writer_mem_chunk);
	G_UNLOCK (_burgner_isofs_writer_mem_chunk);
#else
	self = g_new0 (BurgnerIsofsWriter, 1);
#endif

	self->refcount = 1;

	self->items = list;

	return self;
}


BurgnerIsofsWriter *
burgner_isofs_writer_ref (BurgnerIsofsWriter *self)
{
	g_return_val_if_fail (self != NULL, NULL);
	self->refcount++;
	return self;
}

void
burgner_isofs_writer_unref (BurgnerIsofsWriter *self)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (self->refcount != 0);

	self->refcount--;
	if (self->refcount == 0) {
		burgner_isofs_writer_free (self);
	}
}

void
burgner_isofs_writer_free (BurgnerIsofsWriter *self)
{
	g_return_if_fail (self != NULL);

	if (self->map_table) {
		g_object_unref (self->map_table);
		self->map_table = NULL;
	}

	if (self->tmp_buf) {
		g_string_free (self->tmp_buf, TRUE);
		self->tmp_buf = NULL;
	}

#ifdef _BURGNER_ISOFS_WRITER_USE_MEM_CHUNK
	G_LOCK (_burgner_isofs_writer_mem_chunk);
	g_mem_chunk_free (_burgner_isofs_writer_mem_chunk, self);
	G_UNLOCK (_burgner_isofs_writer_mem_chunk);
#else
	g_free (self);
#endif
}

GQuark
burgner_isofs_writer_error_quark (void)
{
	return g_quark_from_string ("BurgnerIsofsWriter");
}

static gboolean
_burgner_isofs_writer_prepare_foreach_item_func (GList *list,
						 GList *actual,
						 BurgnerIsofsItem *item,
						 BurgnerIsofsItemFlags flags,
						 guint32 padding,
						 guint32 s_c, guint32 b_o,
						 guint32 n_s, guint32 n_b,
						 guint32 a_s, guint32 a_b,
						 gpointer user_data)
{
	BurgnerIsofsMapTable *map_table = BURGNER_ISOFS_MAP_TABLE (user_data);

	burgner_isofs_map_table_set (map_table, item, s_c, b_o);
	burgner_isofs_item_prepare (item, map_table, s_c, b_o, a_s, a_b);

	return FALSE;
}

void
burgner_isofs_writer_prepare (BurgnerIsofsWriter *writer,
			      guint32 first_sector_lba)
{
	guint32 s_c = first_sector_lba, b_o = 0;

	g_return_if_fail (writer != NULL);
	g_return_if_fail (writer->map_table == NULL);
	g_return_if_fail (writer->next == NULL);
	g_return_if_fail (writer->items != NULL);

	writer->map_table = burgner_isofs_map_table_new ();

	burgner_isofs_list_iteration_foreach (writer->items, &s_c, &b_o,
						_burgner_isofs_writer_prepare_foreach_item_func,
						writer->map_table);

	writer->next = writer->items;
}

typedef struct {
	BurgnerIsofsWriter *self;

	gpointer buf;
	guint32 buf_size;

	guint next_is_async : 1;
	guint next_is_async_set : 1;
} _BurgnerIsofsWriter_WriteSync_Data;

static gboolean
_burgner_isofs_writer_write_sync_foreach_func (GList *list,
					       GList *actual,
					       BurgnerIsofsItem *item,
					       BurgnerIsofsItemFlags flags,
					       guint32 padding,
					       guint32 s_c, guint32 b_o,
					       guint32 n_s, guint32 n_b,
					       guint32 a_s, guint32 a_b,
					       gpointer user_data)
{
	_BurgnerIsofsWriter_WriteSync_Data *d = (_BurgnerIsofsWriter_WriteSync_Data *) user_data;
	guint32 size;

	if (padding) {
		if (padding <= d->buf_size) {
			memset (d->buf, 0, padding);
			d->buf += padding;
			d->buf_size -= padding;
		} else {
			memset (d->buf, 0, d->buf_size);

			d->next_is_async = 0;
			d->next_is_async_set = 1;

			d->self->blank = padding - d->buf_size;

			d->buf_size = 0;

			d->self->next = actual;
			d->self->s_c = s_c;
			d->self->b_o = b_o;

			return TRUE;
		}
	}

	size = (n_s << 11) + n_b;

	if (flags & BURGNER_ISOFS_ITEM_WRITE_ASYNC) {
		d->next_is_async = 1;
		d->next_is_async_set = 1;

		d->self->next = actual;
		d->self->s_c = s_c;
		d->self->b_o = b_o;

		return TRUE;
	} else if (flags & BURGNER_ISOFS_ITEM_BLANK) {
		if (size <= d->buf_size) {
			memset (d->buf, 0, size);
			d->buf += size;
			d->buf_size -= size;
		} else {
			memset (d->buf, 0, d->buf_size);

			d->next_is_async = 0;
			d->next_is_async_set = 1;

			d->self->blank = size - d->buf_size;

			d->buf_size = 0;

			d->self->next = actual->next;
			d->self->s_c = a_s;
			d->self->b_o = a_b;

			return TRUE;
		}
	} else {
		if (size <= d->buf_size) {
			burgner_isofs_item_write_sync (item, d->self->map_table,
							d->buf, size);
			d->buf += size;
			d->buf_size -= size;
		} else {
			if (!d->self->tmp_buf) {
				d->self->tmp_buf = g_string_sized_new (size);
			} else {
				g_string_set_size (d->self->tmp_buf, size);
			}

			burgner_isofs_item_write_sync (item, d->self->map_table,
							d->self->tmp_buf->str, size);

			memcpy (d->buf, d->self->tmp_buf->str, d->buf_size);
			g_string_erase (d->self->tmp_buf, 0, d->buf_size);

			d->next_is_async = 0;
			d->next_is_async_set = 1;

			d->buf_size = 0;

			d->self->next = actual->next;
			d->self->s_c = a_s;
			d->self->b_o = a_b;

			return TRUE;
		}
	}

	if (!actual->next) {
		d->self->next = NULL;
		d->self->s_c = a_s;
		d->self->b_o = a_b;
	}

	return FALSE;
}

guint32
burgner_isofs_writer_write_sync (BurgnerIsofsWriter *writer,
				 gpointer b, guint32 b_size,
				 gboolean *next_is_async)
{
	_BurgnerIsofsWriter_WriteSync_Data d;
	gpointer buf = b;
	guint32 n_s, n_b, buf_size = b_size;

	g_return_val_if_fail (writer != NULL, 0);
	g_return_val_if_fail (writer->async_item == NULL, 0);
	g_return_val_if_fail (writer->map_table != NULL, 0);

	if (writer->blank) {
		if (writer->blank <= buf_size) {
			memset (buf, 0, writer->blank);
			buf += writer->blank;
			buf_size -= writer->blank;
			writer->blank = 0;
		} else {
			memset (buf, 0, buf_size);
			writer->blank -= buf_size;

			if (next_is_async)
				*next_is_async = FALSE;

			return b_size;
		}
	} else if (writer->tmp_buf && (writer->tmp_buf_offset < writer->tmp_buf->len)) {
		guint s = writer->tmp_buf->len - writer->tmp_buf_offset;
		
		if (s <= buf_size) {
			memcpy (buf, &writer->tmp_buf->str[writer->tmp_buf_offset], s);
			g_string_truncate (writer->tmp_buf, 0);
			writer->tmp_buf_offset = 0;

			buf += s;
			buf_size -= s;
		} else {
			memcpy (buf, &writer->tmp_buf->str[writer->tmp_buf_offset], buf_size);
			writer->tmp_buf_offset += buf_size;

			if (next_is_async)
				*next_is_async = FALSE;

			return b_size;
		}
	}

	if (!writer->next) {
		if (next_is_async)
			*next_is_async = FALSE;
		return b_size - buf_size;
	}

	d.self = writer;
	d.buf = buf;
	d.buf_size = buf_size;
	d.next_is_async = d.next_is_async_set = 0;

	n_s = writer->s_c;
	n_b = writer->b_o;

	burgner_isofs_list_iteration_foreach (writer->next, &n_s, &n_b,
					      _burgner_isofs_writer_write_sync_foreach_func, &d);

	if (!d.next_is_async_set) {
		g_assert (!writer->next);
		if (next_is_async)
			*next_is_async = FALSE;
	} else if (next_is_async) {
		*next_is_async = d.next_is_async;
	}

	return b_size - d.buf_size;
}

static void
_burgner_isofs_writer_write_async_cb (BurgnerIsofsItem *item,
				      gpointer write_async_data,
				      gpointer buf,
				      guint32 bytes_requested,
				      guint32 bytes_written,
				      GError *error,
				      gboolean end,
				      gpointer user_data)
{
	BurgnerIsofsWriter *self = (BurgnerIsofsWriter *) user_data;
	BurgnerIsofsWriterWriteAsyncCallback cb = self->cb;
	gpointer cb_data = self->cb_data;
	guint32 buf_size = self->buf_size;

	g_assert (buf == self->buf);
	g_assert (bytes_requested == self->buf_size);
	g_assert (item == self->async_item);
	g_assert (bytes_written <= bytes_requested);

	self->cb = NULL; self->cb_data = NULL;
	self->buf = NULL; self->buf_size = 0;

	if (error) {
		if (cb)
			cb (self, buf, buf_size, 0, error, cb_data);
	} else {
		self->write_async_data = write_async_data;

		if (!(self->async_remaining -= bytes_written)) {
			self->async_item = NULL;
			self->write_async_data = NULL;
			self->async_size = 0;
		}

		if (cb)
			cb (self, buf, buf_size, bytes_written, NULL, cb_data);
	}
}

static void
_burgner_isofs_writer_write_async_init_cb (BurgnerIsofsItem *item,
					   gpointer write_async_data,
					   GError *error,
					   gpointer user_data)
{
	BurgnerIsofsWriter *self = (BurgnerIsofsWriter *) user_data;

	if (error) {
		BurgnerIsofsWriterWriteAsyncCallback cb = self->cb;
		gpointer cb_data = self->cb_data, buf = self->buf;
		guint32 buf_size = self->buf_size;

		self->cb = NULL; self->cb_data = NULL;
		self->buf = NULL; self->buf_size = 0;

		g_assert (self->next->prev);
		g_assert (self->next->prev->data == self->async_item);
		self->next = self->next->prev;

		self->async_item = NULL;
		self->write_async_data = NULL;
		self->async_size = self->async_remaining = 0;

		if (cb)
			cb (self, buf, buf_size, 0, error, cb_data);
	} else {
		self->write_async_data = write_async_data;

		burgner_isofs_item_write_async (self->async_item, self->map_table,
						self->write_async_data,
						self->buf, MIN (self->buf_size, self->async_remaining),
						_burgner_isofs_writer_write_async_cb, self);
	}
}

void
burgner_isofs_writer_write_async (BurgnerIsofsWriter *writer,
				  gpointer buf, guint32 buf_size,
				  BurgnerIsofsWriterWriteAsyncCallback cb,
				  gpointer cb_data)
{
	g_return_if_fail (writer != NULL);
	g_return_if_fail (writer->blank == 0);
	g_return_if_fail ((writer->tmp_buf == NULL) || !writer->tmp_buf->len);
	g_return_if_fail (writer->buf == NULL);
	g_return_if_fail (writer->buf_size == 0);
	g_return_if_fail (writer->cb == NULL);
	g_return_if_fail (writer->map_table != NULL);

	if (writer->async_item) {
		writer->buf = buf;
		writer->buf_size = buf_size;
		writer->cb = cb;
		writer->cb_data = cb_data;

		burgner_isofs_item_write_async (writer->async_item, writer->map_table,
						writer->write_async_data,
						buf, MIN (buf_size, writer->async_remaining),
						_burgner_isofs_writer_write_async_cb, writer);
	} else {
		BurgnerIsofsItemFlags f;
		guint32 n_s = 0, n_b = 0;

		g_return_if_fail (writer->next != NULL);

		writer->async_item = BURGNER_ISOFS_ITEM (writer->next->data);
		f = burgner_isofs_item_get_flags (writer->async_item);

		g_return_if_fail (f & BURGNER_ISOFS_ITEM_WRITE_ASYNC);

		burgner_isofs_item_get_size (writer->async_item, &n_s, &n_b);
		writer->async_size = writer->async_remaining = (n_s << 11) + n_b;

		writer->next = writer->next->next;
		writer->s_c += n_s;
		writer->b_o += n_b;
		writer->s_c += writer->b_o >> 11;
		writer->b_o &= 0x7ff;

		writer->buf = buf;
		writer->buf_size = buf_size;
		writer->cb = cb;
		writer->cb_data = cb_data;

		burgner_isofs_item_write_async_init (writer->async_item, writer->map_table,
						     _burgner_isofs_writer_write_async_init_cb, writer);
	}
}

static gboolean
_burgner_isofs_writer_to_write_sync_foreach_func (GList *list,
						  GList *actual,
						  BurgnerIsofsItem *item,
						  BurgnerIsofsItemFlags flags,
						  guint32 padding,
						  guint32 s_c, guint32 b_o,
						  guint32 n_s, guint32 n_b,
						  guint32 a_s, guint32 a_b,
						  gpointer user_data)
{
	guint32 *r = (guint32 *) user_data;

	*r += padding;

	if (flags & BURGNER_ISOFS_ITEM_WRITE_ASYNC)
		return TRUE;

	*r += (n_s << 11) + n_b;

	return FALSE;
}

guint32
burgner_isofs_writer_to_write_sync (BurgnerIsofsWriter *writer)
{
	guint32 r, n_s, n_b;

	g_return_val_if_fail (writer != NULL, 0);
	g_return_val_if_fail (writer->map_table != NULL, 0);

	if (writer->async_item) {
		return 0;
	}

	if (writer->blank) {
		r = writer->blank;
	} else if (writer->tmp_buf && writer->tmp_buf->len) {
		r = writer->tmp_buf->len;
	} else {
		r = 0;
	}

	n_s = writer->s_c;
	n_b = writer->b_o;

	if (writer->next) {
		burgner_isofs_list_iteration_foreach (writer->next, &n_s, &n_b,
						      _burgner_isofs_writer_to_write_sync_foreach_func, &r);
	}

	return r;
}

guint32
burgner_isofs_writer_to_write_async (BurgnerIsofsWriter *writer)
{
	g_return_val_if_fail (writer != NULL, 0);
	g_return_val_if_fail (writer->map_table != NULL, 0);

	if (writer->async_item) {
		return writer->async_remaining;
	}

	if (writer->blank ||
	    (writer->tmp_buf && writer->tmp_buf->len)) {
		return 0;
	}

	if (writer->next) {
		BurgnerIsofsItem *i = BURGNER_ISOFS_ITEM (writer->next->data);

		if (burgner_isofs_item_get_flags (i) & BURGNER_ISOFS_ITEM_WRITE_ASYNC) {
			guint32 n_s = 0, n_b = 0;

			burgner_isofs_item_get_size (i, &n_s, &n_b);

			return (n_s << 11) + n_b;
		}
	}

	return 0;
}

gboolean
burgner_isofs_writer_end_reached (BurgnerIsofsWriter *writer)
{
	g_return_val_if_fail (writer != NULL, 0);
	g_return_val_if_fail (writer->map_table != NULL, 0);

	return (!writer->next && !writer->async_item && !writer->blank && (!writer->tmp_buf || !writer->tmp_buf->len));
}

