/*
 * BurgnerMmcWriteStage - BurgnerMmcWriteStage object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-mmc-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "../libburgnerutils/burgner-fifo-reader.h"
#include "../libburgnerscsi/burgner-scsi-status-code.h"
#include "../libburgnerscsi/burgner-scsi-sense.h"
#include "burgner-mmc-write.h"
#include "burgner-mmc-read-buffer-capacity.h"
#include "burgner-mmc-write-stage.h"


#define _BURGNER_MMC_WRITE_STAGE_WRITE_CMD_TIMEOUT		(100000)
#define _BURGNER_MMC_WRITE_STAGE_READ_BUF_CAP_CMD_TIMEOUT	(1000)


enum {
	_BURGNER_MMC_WRITE_STAGE_STATE_NONE,
	_BURGNER_MMC_WRITE_STAGE_STATE_WRITING,
	_BURGNER_MMC_WRITE_STAGE_STATE_READING_BUFFER_CAPACITY,
	_BURGNER_MMC_WRITE_STAGE_STATE_IDLE
};

static BurgnerScsiStageClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_DEVICE,
	PROP_FIFO,
	PROP_DATA_BLOCK_SIZE,
	PROP_START_LBA,
	PROP_READ_BUF_CAP,
	PROP_MAX_WRITE_BETWEEN_READ_BUF_CAP,
	PROP_MIN_TIME_BETWEEN_READ_BUF_CAP,
	PROP_CDB_BUF,
	PROP_CDB_BUF_SIZE,
	PROP_SENSE_BUF,
	PROP_SENSE_BUF_SIZE,
	PROP_READ_BUF_CAP_DATA_BUF,
	PROP_READ_BUF_CAP_DATA_BUF_SIZE,
	PROP_PROGRESS_CB,
	PROP_BUFFER_CAPACITY_CB,
	PROP_USER_DATA,
	BURGNER_MMC_WRITE_STAGE_LAST_ARG
};

static void
burgner_mmc_write_stage_finalize (GObject *object)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (object);

	if (self->device) {
		g_object_unref (self->device);
		self->device = NULL;
	}

	if (self->fifo)
		self->fifo = NULL;

	self->write_count = 0;
	self->next_lba = 0;
	self->data_block_size = 2048;

	self->cdb_buf = self->sense_buf = self->read_buf_cap_data_buf = NULL;
	self->cdb_buf_size = self->sense_buf_size = 0;
	self->read_buf_cap_data_buf_size = 0;

	self->max_write_between_read_buf_cap = 20;
	self->min_time_between_read_buf_cap = 1000;

	self->progress_cb = NULL;
	self->buffer_cb = NULL;
	self->user_data = NULL;

	if (self->timeout_id) {
		g_source_remove (self->timeout_id);
		self->timeout_id = 0;
	}

	if (self->idle_id) {
		g_source_remove (self->idle_id);
		self->idle_id = 0;
	}

	self->pad = 0;

	self->read_buf_cap = 1;
	self->end = 0;
	self->state = _BURGNER_MMC_WRITE_STAGE_STATE_NONE;
	self->cancel = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void _burgner_mmc_write_stage_setup_write_command (BurgnerMmcWriteStage *self);
static void _burgner_mmc_write_stage_setup_read_buf_cap_command (BurgnerMmcWriteStage *self);

static void
burgner_mmc_write_stage_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (object);

	switch (prop_id) {
		case PROP_DEVICE:
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			if (self->device)
				g_object_unref (self->device);
			self->device = (BurgnerScsiDevice *) g_value_get_object (value);
			if (self->device)
				self->device = BURGNER_SCSI_DEVICE (g_object_ref (self->device));
			g_object_notify (object, "device");
			break;
		case PROP_FIFO:
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->fifo = (BurgnerFifo *) g_value_get_object (value);
			g_object_notify (object, "fifo");
			break;
		case PROP_DATA_BLOCK_SIZE: {
			guint d = self->data_block_size;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->data_block_size = g_value_get_uint (value);
			if (d != self->data_block_size)
				g_object_notify (object, "data_block_size");
			} break;
		case PROP_START_LBA: {
			guint d = self->next_lba;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->next_lba = g_value_get_uint (value);
			if (d != self->next_lba)
				g_object_notify (object, "start_lba");
			} break;
		case PROP_READ_BUF_CAP: {
			gboolean d = self->read_buf_cap;
			self->read_buf_cap = g_value_get_boolean (value)? 1 : 0;
			if (d != self->read_buf_cap) {
				g_object_notify (object, "read_buf_cap");
				if (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_IDLE) {
					if (!self->read_buf_cap) {
						if (self->timeout_id) {
							g_source_remove (self->timeout_id);
							self->timeout_id = 0;
						}

						if (burgner_fifo_to_read (self->fifo, NULL)) {
							_burgner_mmc_write_stage_setup_write_command (self);
						}
					} else if (self->read_buf_cap_data_buf && self->read_buf_cap_data_buf_size) {
						_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
					}
				}
			}
			} break;
		case PROP_MAX_WRITE_BETWEEN_READ_BUF_CAP: {
			gint d = self->max_write_between_read_buf_cap;
			self->max_write_between_read_buf_cap = g_value_get_int (value);
			if (d != self->max_write_between_read_buf_cap)
				g_object_notify (object, "max_write_between_read_buf_cap");
			} break;
		case PROP_MIN_TIME_BETWEEN_READ_BUF_CAP: {
			gint d = self->min_time_between_read_buf_cap;
			self->min_time_between_read_buf_cap = g_value_get_int (value);
			if (d != self->min_time_between_read_buf_cap)
				g_object_notify (object, "min_time_between_read_buf_cap");
			} break;
		case PROP_CDB_BUF: {
			gpointer d = self->cdb_buf;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->cdb_buf = g_value_get_pointer (value);
			if (d != self->cdb_buf)
				g_object_notify (object, "cdb_buf");
			} break;
		case PROP_CDB_BUF_SIZE: {
			guint d = self->cdb_buf_size;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->cdb_buf_size = g_value_get_uint (value);
			if (d != self->cdb_buf_size)
				g_object_notify (object, "cdb_buf_size");
			} break;
		case PROP_SENSE_BUF: {
			gpointer d = self->sense_buf;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->sense_buf = g_value_get_pointer (value);
			if (d != self->sense_buf)
				g_object_notify (object, "sense_buf");
			} break;
		case PROP_SENSE_BUF_SIZE: {
			guint d = self->sense_buf_size;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->sense_buf_size = g_value_get_uint (value);
			if (d != self->sense_buf_size)
				g_object_notify (object, "sense_buf_size");
			} break;
		case PROP_READ_BUF_CAP_DATA_BUF: {
			gpointer d = self->read_buf_cap_data_buf;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->read_buf_cap_data_buf = g_value_get_pointer (value);
			if (d != self->read_buf_cap_data_buf)
				g_object_notify (object, "read_buf_cap_data_buf");
			} break;
		case PROP_READ_BUF_CAP_DATA_BUF_SIZE: {
			guint d = self->read_buf_cap_data_buf_size;
			g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
			self->read_buf_cap_data_buf_size = g_value_get_uint (value);
			if (d != self->read_buf_cap_data_buf_size)
				g_object_notify (object, "read_buf_cap_data_buf_size");
			} break;
		case PROP_PROGRESS_CB: {
			gpointer d = self->progress_cb;
			self->progress_cb = (BurgnerMmcWriteStageProgressCallback) g_value_get_pointer (value);
			if (d != (gpointer) self->progress_cb)
				g_object_notify (object, "progress_cb");
			} break;
		case PROP_BUFFER_CAPACITY_CB: {
			gpointer d = self->buffer_cb;
			self->buffer_cb = (BurgnerMmcWriteStageBufferCapacityCallback) g_value_get_pointer (value);
			if (d != (gpointer) self->buffer_cb)
				g_object_notify (object, "buffer_capacity_cb");
			} break;
		case PROP_USER_DATA: {
			gpointer d = self->user_data;
			self->user_data = g_value_get_pointer (value);
			if (d != self->user_data)
				g_object_notify (object, "user_data");
			} break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}


static void
burgner_mmc_write_stage_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (object);

	switch (prop_id) {
		case PROP_DEVICE:
			g_value_set_object (value, self->device);
			break;
		case PROP_FIFO:
			g_value_set_object (value, self->fifo);
			break;
		case PROP_DATA_BLOCK_SIZE:
			g_value_set_uint (value, self->data_block_size);
			break;
		case PROP_START_LBA:
			g_value_set_uint (value, self->next_lba);
			break;
		case PROP_READ_BUF_CAP:
			g_value_set_boolean (value, self->read_buf_cap);
			break;
		case PROP_MAX_WRITE_BETWEEN_READ_BUF_CAP:
			g_value_set_int (value, self->max_write_between_read_buf_cap);
			break;
		case PROP_MIN_TIME_BETWEEN_READ_BUF_CAP:
			g_value_set_int (value, self->min_time_between_read_buf_cap);
			break;
		case PROP_CDB_BUF:
			g_value_set_pointer (value, self->cdb_buf);
			break;
		case PROP_CDB_BUF_SIZE:
			g_value_set_uint (value, self->cdb_buf_size);
			break;
		case PROP_SENSE_BUF:
			g_value_set_pointer (value, self->sense_buf);
			break;
		case PROP_SENSE_BUF_SIZE:
			g_value_set_uint (value, self->sense_buf_size);
			break;
		case PROP_READ_BUF_CAP_DATA_BUF:
			g_value_set_pointer (value, self->read_buf_cap_data_buf);
			break;
		case PROP_READ_BUF_CAP_DATA_BUF_SIZE:
			g_value_set_uint (value, self->read_buf_cap_data_buf_size);
			break;
		case PROP_PROGRESS_CB:
			g_value_set_pointer (value, self->progress_cb);
			break;
		case PROP_BUFFER_CAPACITY_CB:
			g_value_set_pointer (value, self->buffer_cb);
			break;
		case PROP_USER_DATA:
			g_value_set_pointer (value, self->user_data);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_mmc_write_stage_send_write_command_cb
			(BurgnerScsiDevice *device,
			 gpointer cdb, guint cdb_size,
			 gpointer sense, guint sense_size, guint sense_used,
			 gpointer data, guint data_size, guint data_used,
			 BurgnerScsiDirection direction,
			 guint timeout,
			 gint status,
			 GError *error,
			 gpointer user_data)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (user_data);

	g_assert (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_WRITING);
	g_assert (self->device == device);
	g_assert (self->cdb_buf == cdb);
	g_assert (self->sense_buf == sense);
	g_assert (self->sense_buf_size == sense_size);
	g_assert (self->timeout_id == 0);

	if (error || status != BURGNER_SCSI_STATUS_CODE_GOOD) {
		BurgnerScsiSense *s = NULL;

		if (sense) {
			memset (sense + sense_used, 0, sense_size - sense_used);
			(void) burgner_scsi_sense_new_from_buf
							(&s,
							 (guchar *) sense,
							 sense_size,
							 NULL);
		}

		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self),
				       FALSE, error, s);
		if (s)
			burgner_scsi_sense_unref (s);
		return;
	}

	if (self->progress_cb)
		self->progress_cb (self, device,
				   data_used / self->data_block_size,
				   self->user_data);

	if (self->cancel) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self),
				       TRUE, NULL, NULL);
		return;
	}

	burgner_fifo_read (self->fifo, data_used - self->pad);

	if (self->end) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self),
				       FALSE, NULL, NULL);
		return;
	}

	if (self->read_buf_cap &&
	    self->write_count >= self->max_write_between_read_buf_cap) {
		_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
	} else if (burgner_fifo_to_read (self->fifo, NULL)) {
		_burgner_mmc_write_stage_setup_write_command (self);
	} else if (self->read_buf_cap) {
		_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
	} else {
		self->state = _BURGNER_MMC_WRITE_STAGE_STATE_IDLE;
	}
}

static gboolean
_burgner_mmc_write_stage_timeout_func (BurgnerMmcWriteStage *self)
{
	self->timeout_id = 0;
	_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
	return FALSE;
}

static void
_burgner_mmc_write_stage_read_buf_cap_command_cb
			(BurgnerScsiDevice *device,
			 gpointer cdb, guint cdb_size,
			 gpointer sense, guint sense_size, guint sense_used,
			 gpointer data, guint data_size, guint data_used,
			 BurgnerScsiDirection direction,
			 guint timeout,
			 gint status,
			 GError *error,
			 gpointer user_data)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (user_data);

	g_assert (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_READING_BUFFER_CAPACITY);
	g_assert (self->device == device);
	g_assert (self->cdb_buf == cdb);
	g_assert (self->sense_buf == sense);
	g_assert (self->sense_buf_size == sense_size);
	g_assert (self->timeout_id == 0);
	g_assert (self->read_buf_cap_data_buf == data);
	g_assert (self->read_buf_cap_data_buf_size == data_size);

	if (error || status != BURGNER_SCSI_STATUS_CODE_GOOD) {
		if (self->buffer_cb)
			self->buffer_cb (self,
					 device,
					 0, 0,
					 error,
					 sense_used,
					 self->user_data);
	}

	if (self->buffer_cb) {
		guint32 b = 0, t = 0;

		memset ((guchar *) data + data_used, 0, data_size - data_used);

		burgner_mmc_read_buffer_capacity_parse_data (data,
							     data_size,
							     &b, &t);

		self->buffer_cb (self, device, b, t, NULL, sense_used,
				 self->user_data);
	}

	if (self->cancel || self->end) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), self->cancel, NULL, NULL);
	} else if (burgner_fifo_to_read (self->fifo, NULL)) {
		_burgner_mmc_write_stage_setup_write_command (self);
	} else if (self->read_buf_cap) {
		if (self->min_time_between_read_buf_cap > 0) {
			self->state = _BURGNER_MMC_WRITE_STAGE_STATE_IDLE;

			self->timeout_id = g_timeout_add
				(self->min_time_between_read_buf_cap,
				 (GSourceFunc) _burgner_mmc_write_stage_timeout_func,
				 self);
		} else {
			_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
		}
	} else {
		self->state = _BURGNER_MMC_WRITE_STAGE_STATE_IDLE;
	}
}

static void
_burgner_mmc_write_stage_setup_write_command (BurgnerMmcWriteStage *self)
{
	guchar *data = NULL;
	guint to_write = burgner_fifo_to_read (self->fifo, &data), pad;

	g_return_if_fail ((to_write % self->data_block_size) == 0 || self->fifo->end_reached);

	/*
	 * This is a bad and dirty (but needed and working) hack. :|
	 * At least, it works fifos which gives a buffer of which size
	 * is a multiple of data_block_size.
	 */
	if (self->fifo->end_reached &&
	    (pad = to_write % self->data_block_size)) {
		pad = self->data_block_size - pad;
		memset (data + to_write, 0, pad);
		to_write += pad;
		self->pad = pad;
	} else {
		self->pad = 0;
	}

	self->state = _BURGNER_MMC_WRITE_STAGE_STATE_WRITING;
	burgner_mmc_write_write_cdb (self->cdb_buf,
				     BURGNER_MMC_WRITE_CDB_SIZE,
				     to_write,
				     self->data_block_size,
				     self->next_lba,
				     FALSE);
	self->next_lba += to_write / self->data_block_size;
	self->write_count += 1;

	burgner_scsi_device_send_cmd (self->device,
				      self->cdb_buf, BURGNER_MMC_WRITE_CDB_SIZE,
				      self->sense_buf, self->sense_buf_size,
				      data, to_write,
				      BURGNER_SCSI_DIRECTION_TO_DEVICE,
				      _BURGNER_MMC_WRITE_STAGE_WRITE_CMD_TIMEOUT,
				      _burgner_mmc_write_stage_send_write_command_cb,
				      self);
}

static void
_burgner_mmc_write_stage_setup_read_buf_cap_command (BurgnerMmcWriteStage *self)
{
	self->state = _BURGNER_MMC_WRITE_STAGE_STATE_READING_BUFFER_CAPACITY;
	burgner_mmc_read_buffer_capacity_write_cdb
			(self->cdb_buf,
			 BURGNER_MMC_READ_BUFFER_CAPACITY_CDB_SIZE,
			 self->read_buf_cap_data_buf_size);
	
	burgner_scsi_device_send_cmd
			(self->device,
			 self->cdb_buf,
			 BURGNER_MMC_READ_BUFFER_CAPACITY_CDB_SIZE,
			 self->sense_buf, self->sense_buf_size,
			 self->read_buf_cap_data_buf,
			 self->read_buf_cap_data_buf_size,
			 BURGNER_SCSI_DIRECTION_FROM_DEVICE,
			 _BURGNER_MMC_WRITE_STAGE_READ_BUF_CAP_CMD_TIMEOUT,
			 _burgner_mmc_write_stage_read_buf_cap_command_cb,
			 self);
}

static void
_burgner_mmc_write_stage_real_start (BurgnerScsiStage *stage)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (stage);

	g_return_if_fail (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_NONE);
	g_return_if_fail (self->device != NULL);
	g_return_if_fail (self->fifo != NULL);
	g_return_if_fail (self->write_count == 0);
	g_return_if_fail (self->data_block_size != 0);
	g_return_if_fail (self->cdb_buf != NULL);
	g_return_if_fail (self->cdb_buf_size >= 10);
	g_return_if_fail (!self->read_buf_cap || self->read_buf_cap_data_buf);
	g_return_if_fail (!self->read_buf_cap_data_buf_size || self->read_buf_cap_data_buf);
	g_return_if_fail (!self->sense_buf_size || self->sense_buf);
	g_return_if_fail (self->timeout_id == 0);

	g_object_set (self->fifo, "reader", self, NULL);

	if (burgner_fifo_to_read (self->fifo, NULL)) {
		_burgner_mmc_write_stage_setup_write_command (self);
	} else if (self->read_buf_cap) {
		_burgner_mmc_write_stage_setup_read_buf_cap_command (self);
	} else {
		self->state = _BURGNER_MMC_WRITE_STAGE_STATE_IDLE;
	}
}

static gboolean
_burgner_mmc_write_stage_cancel_idle_func (BurgnerMmcWriteStage *self)
{
	self->idle_id = 0;

	burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self),
			       TRUE, NULL, NULL);

	return FALSE;
}

static void
_burgner_mmc_write_stage_cancel (BurgnerScsiStage *stage)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (stage);

	g_return_if_fail (self->state != _BURGNER_MMC_WRITE_STAGE_STATE_NONE);

	if (self->state != _BURGNER_MMC_WRITE_STAGE_STATE_IDLE) {
		self->cancel = 1;
		return;
	}

	if (self->timeout_id) {
		g_source_remove (self->timeout_id);
		self->timeout_id = 0;
	}

	self->idle_id = g_idle_add ((GSourceFunc) _burgner_mmc_write_stage_cancel_idle_func, self);
}

static gboolean
_burgner_mmc_write_stage_cancellable (BurgnerScsiStage *stage)
{
	return TRUE;
}

static void
_burgner_mmc_write_stage_end (BurgnerScsiStage *stage,
			      gboolean cancelled,
			      GError *error,
			      BurgnerScsiSense *sense)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (stage);

	if (self->write_count)
		self->write_count = 0;

	if (self->timeout_id) {
		g_source_remove (self->timeout_id);
		self->timeout_id = 0;
	}

	if (self->idle_id) {
		g_source_remove (self->idle_id);
		self->idle_id = 0;
	}

	if (self->state != _BURGNER_MMC_WRITE_STAGE_STATE_NONE)
		self->state = _BURGNER_MMC_WRITE_STAGE_STATE_NONE;

	if (parent_class->end)
		parent_class->end (stage, cancelled, error, sense);
}


static void
burgner_mmc_write_stage_class_init (BurgnerMmcWriteStageClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiStageClass *stage_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_mmc_write_stage_set_property;
	object_class->get_property = burgner_mmc_write_stage_get_property;
	object_class->finalize = burgner_mmc_write_stage_finalize;

	stage_class = (BurgnerScsiStageClass *) klass;
	stage_class->real_start = _burgner_mmc_write_stage_real_start;
	stage_class->cancel = _burgner_mmc_write_stage_cancel;
	stage_class->cancellable = _burgner_mmc_write_stage_cancellable;
	stage_class->end = _burgner_mmc_write_stage_end;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_DEVICE,
					 g_param_spec_object ("device",
					                _("device"),
							_("A BurgnerScsiDevice"),
							BURGNER_TYPE_SCSI_DEVICE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_FIFO,
					 g_param_spec_object ("fifo",
					                _("fifo"),
							_("A BurgnerFifo"),
							BURGNER_TYPE_FIFO,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_DATA_BLOCK_SIZE,
					 g_param_spec_uint ("data_block_size",
					                _("data_block_size"),
							_("Data Block Size"),
							0,
							2448,
							2048,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_START_LBA,
					 /* FIXME: start lba may be negative */
					 g_param_spec_uint ("start_lba",
					                _("start_lba"),
							_("Start LBA"),
							0,
							0xffffffff,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_READ_BUF_CAP,
					 g_param_spec_boolean ("read_buf_cap",
					                _("read_buf_cap"),
							_("Read Buffer Capacity"),
							TRUE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_MAX_WRITE_BETWEEN_READ_BUF_CAP,
					 g_param_spec_int ("max_write_between_read_buf_cap",
					                _("max_write_between_read_buf_cap"),
							_("Maximum number of write between two reading of buffer capacity"),
							-1,
							G_MAXINT,
							20,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_MIN_TIME_BETWEEN_READ_BUF_CAP,
					 g_param_spec_int ("min_time_between_read_buf_cap",
					                _("min_time_between_read_buf_cap"),
							_("Minimum time (in milliseconds) between two consecutive reading of buffer capacity"),
							-1,
							G_MAXINT,
							1000,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CDB_BUF,
					 g_param_spec_pointer ("cdb_buf",
					                _("cdb_buf"),
							_("A buffer which will contain the CDB"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CDB_BUF_SIZE,
					 g_param_spec_uint ("cdb_buf_size",
					                _("cdb_buf_size"),
							_("Size of the buffer which will contain the CDB (should be 10, or more)"),
							0,
							G_MAXUINT,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SENSE_BUF,
					 g_param_spec_pointer ("sense_buf",
					                _("sense_buf"),
							_("A buffer which will contain the sense"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SENSE_BUF_SIZE,
					 g_param_spec_uint ("sense_buf_size",
					                _("sense_buf_size"),
							_("Size of the buffer which will contain the sense (should be 18, or more)"),
							0,
							G_MAXUINT,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_READ_BUF_CAP_DATA_BUF,
					 g_param_spec_pointer ("read_buf_cap_data_buf",
					                _("read_buf_cap_data_buf"),
							_("A buffer which will contain the data returned by the READ BUFFER CAPACITY command"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_READ_BUF_CAP_DATA_BUF_SIZE,
					 g_param_spec_uint ("read_buf_cap_data_buf_size",
					                _("read_buf_cap_data_buf_size"),
							_("Size of the read_buf_cap_data_buf"),
							0,
							G_MAXUINT,
							12,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_PROGRESS_CB,
					 g_param_spec_pointer ("progress_cb",
					                _("progress_cb"),
							_("A BurgnerMmcWriteStageProgressCallback"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_BUFFER_CAPACITY_CB,
					 g_param_spec_pointer ("buffer_capacity_cb",
					                _("buffer_capacity_cb"),
							_("A BurgnerMmcWriteStageBufferCapacityCallback"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_USER_DATA,
					 g_param_spec_pointer ("user_data",
					                _("user_data"),
							_("User data to be passed to progress_cb and buffer_capacity_cb"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));

	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_mmc_write_stage_init (BurgnerMmcWriteStage *self)
{
	self->device = NULL;
	self->fifo = NULL;
	self->write_count = 0;
	self->next_lba = 0;
	self->data_block_size = 2048;

	self->pad = 0;

	self->cdb_buf = NULL;
	self->cdb_buf_size = 0;

	self->sense_buf = NULL;
	self->sense_buf_size = 0;

	self->read_buf_cap_data_buf = NULL;
	self->read_buf_cap_data_buf_size = 0;

	self->max_write_between_read_buf_cap = 20;
	self->min_time_between_read_buf_cap = 1000;

	self->progress_cb = NULL;
	self->buffer_cb = NULL;
	self->user_data = NULL;

	self->timeout_id = self->idle_id = 0;

	self->read_buf_cap = 1;
	self->end = 0;
	self->state = _BURGNER_MMC_WRITE_STAGE_STATE_NONE;
	self->cancel = 0;
}

static void
_burgner_mmc_write_stage_to_read (BurgnerFifoReader *reader,
				  BurgnerFifo *fifo,
				  const guchar *to_read,
				  guint size)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (reader);

	if (self->cancel || self->state != _BURGNER_MMC_WRITE_STAGE_STATE_IDLE)
		return;

	if (self->timeout_id) {
		g_source_remove (self->timeout_id);
		self->timeout_id = 0;
	}

	_burgner_mmc_write_stage_setup_write_command (self);
}

static void
_burgner_mmc_write_stage_end_reached (BurgnerFifoReader *reader,
				      BurgnerFifo *fifo)
{
	BurgnerMmcWriteStage *self = BURGNER_MMC_WRITE_STAGE (reader);

	g_return_if_fail (self->state != _BURGNER_MMC_WRITE_STAGE_STATE_NONE);

	self->end = 1;

	if (self->state == _BURGNER_MMC_WRITE_STAGE_STATE_IDLE) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self),
				       FALSE, NULL, NULL);
	}
}

static void
_burgner_mmc_write_stage_fifo_reader_iface_init (BurgnerFifoReaderIface *iface)
{
	iface->to_read = _burgner_mmc_write_stage_to_read;
	iface->end_reached = _burgner_mmc_write_stage_end_reached;
}

GType
burgner_mmc_write_stage_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerMmcWriteStageClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_mmc_write_stage_class_init,
			NULL,
			NULL,
			sizeof (BurgnerMmcWriteStage),
			0,
			(GInstanceInitFunc) burgner_mmc_write_stage_init,
		};

		static const GInterfaceInfo fifo_reader_info = {
			(GInterfaceInitFunc) _burgner_mmc_write_stage_fifo_reader_iface_init,
			NULL,
			NULL
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_STAGE, "BurgnerMmcWriteStage", &info, 0);
		g_type_add_interface_static (type, BURGNER_TYPE_FIFO_READER, &fifo_reader_info);
	}

	return type;
}

BurgnerScsiStage *
burgner_mmc_write_stage_new (void)
{
	BurgnerMmcWriteStage *self;

	self = g_object_new (BURGNER_TYPE_MMC_WRITE_STAGE, NULL);


	return BURGNER_SCSI_STAGE (self);
}


GQuark
burgner_mmc_write_stage_error_quark (void)
{
	return g_quark_from_string ("BurgnerMmcWriteStage");
}

