/*
 * 
 *
 * This file is part of the libburgnerscsi library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <string.h>

#include "burgner-scsi-driver-description-parser.h"


typedef enum {
	BSDD_PARSER_STATE_OUT = 0,
	BSDD_PARSER_STATE_IN_MODULE_DESC,
	BSDD_PARSER_STATE_IN_INFO,
	BSDD_PARSER_STATE_IN_TYPES,
	BSDD_PARSER_STATE_IN_TYPE,
	BSDD_PARSER_STATE_IN_INTERFACE
} BsddParser_State;

struct _BurgnerScsiDriverDescriptionParser {
	GMarkupParseContext *parse_ctx;
	gchar *filename;

	BurgnerModule *module;
	BurgnerScsiDriverInfo *info;
	gchar *driver_type;

	guint unknown_count;
	BsddParser_State state;
	gboolean error;
};



static void
burgner_scsi_driver_description_parser_start_element
				(GMarkupParseContext *context,
				 const gchar *element_name,
				 const gchar **attribute_names,
				 const gchar **attribute_values,
				 gpointer user_data,
				 GError **error)
{
	BurgnerScsiDriverDescriptionParser *data = (BurgnerScsiDriverDescriptionParser *) user_data;

	if (data->unknown_count) {
		data->unknown_count++;
		return;
	}

	switch (data->state) {
		case BSDD_PARSER_STATE_OUT:
			if (strcmp (element_name, "module_desc") != 0) {
				/*
				 * FIXME: should we use
				 * G_MARKUP_ERROR_UNKNOWN_ELEMENT ?
				 */
				g_set_error (error,
					     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
					     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_ILLEGAL_ROOT_ELEMENT,
					     _("%s: invalid driver description (illegal root element)"),
					     data->filename ? data->filename :
					     	_("[unknown]"));
				return;
			}
			data->state = BSDD_PARSER_STATE_IN_MODULE_DESC;
			break;
		case BSDD_PARSER_STATE_IN_MODULE_DESC:
			if (strcmp (element_name, "info") == 0) {
				guint i;

				data->state = BSDD_PARSER_STATE_IN_INFO;

				if (data->info) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_MULTIPLE_INFO_ELEMENT,
						     _("%s: invalid driver description (multiple info element)"),
						     data->filename ? data->filename : _("[unknown]"));
					return;
				} else if (!attribute_names) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_INFO_ELEMENT,
						     _("%s: invalid driver description (incomplete info element)"),
						     data->filename ? data->filename : _("[unknown]"));
					return;
				}

				data->info = burgner_scsi_driver_info_new ();

				for (i = 0; attribute_names[i]; i++) {
					if (!data->info->id &&
					    strcmp (attribute_names[i],
					    	    "id") == 0) {
						data->info->id = g_strdup (attribute_values[i]);
					} else if (!data->info->name &&
						   strcmp (attribute_names[i],
						   	   "name") == 0) {
						data->info->name = g_strdup (attribute_values[i]);
					} else if (!data->info->version &&
						   strcmp (attribute_names[i],
						   	   "version") == 0) {
						data->info->version = g_strdup (attribute_values[i]);
					} else if (!data->info->description &&
						   strcmp (attribute_names[i],
						   	 "description") == 0) {
						data->info->description = g_strdup (attribute_values[i]);
					} else if (!data->info->authors &&
						   strcmp (attribute_names[i],
						   	   "authors") == 0) {
						data->info->authors = g_strsplit (attribute_values[i], ",", -1);
					} else if (!data->info->need_privilege &&
						   strcmp (attribute_names[i],
						       "need_privilege") == 0) {
						if (g_ascii_strcasecmp (attribute_values[i], "true") == 0)
							data->info->need_privilege = TRUE;
					}
				}

				if (!data->info->id ||
				    !data->info->name ||
				    !data->info->version ||
				    !data->info->authors ||
				    !data->info->description) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_INFO_ELEMENT,
						     _("%s: invalid driver description (incomplete info element)"),
						     data->filename ? data->filename : _("[unknown]"));
					return;
				}
			} else if (strcmp (element_name, "types") == 0) {
				guint i;

				data->state = BSDD_PARSER_STATE_IN_TYPES;

				g_assert ((!data->driver_type) || (data->module));
				if (data->driver_type) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_MULTIPLE_TYPES_ELEMENT,
						     _("%s: invalid driver description (multiple types element)"),
							data->filename ? data->filename :
							_("[unknown]"));
					return;
				}

				for (i = 0; attribute_names[i]; i++) {
					if (strcmp (attribute_names[i], "driver_type") != 0) {
						continue;
					}

					data->driver_type = g_strdup (attribute_values[i]);
					break;
				}

				if (!data->driver_type) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_TYPES_ELEMENT,
						     _("%s: invalid driver description (incomplete 'types' element)"),
							data->filename ? data->filename :
							_("[unknown]"));
					return;
				}
			} else {
				data->unknown_count++;
			}
			break;
		case BSDD_PARSER_STATE_IN_INFO:
			data->unknown_count++;
			break;
		case BSDD_PARSER_STATE_IN_TYPES:
			if (strcmp (element_name, "type") == 0) {
				gchar *name = NULL, *info_symbol = NULL;
				GType parent_type = G_TYPE_INVALID;
				guint i;

				data->state = BSDD_PARSER_STATE_IN_TYPE;

				for (i = 0; attribute_names[i]; i++) {
					if (!name &&
					    strcmp (attribute_names[i], "name") == 0) {
						name = (gchar *) attribute_values[i];
					} else if (parent_type == G_TYPE_INVALID &&
						   strcmp (attribute_names[i], "parent") == 0) {
						parent_type = g_type_from_name (attribute_values[i]);
						if (parent_type == 0) {
							g_set_error (error,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INVALID_PARENT_TYPE,
								     _("%s: invalid driver description (%s: invalid parent type)"),
								     data->filename ? data->filename : _("[unknown]"),
								     attribute_values[i]);
							return;
						}
					} else if (!info_symbol &&
						   strcmp (attribute_names[i], "info_symbol") == 0) {
						info_symbol = (gchar *) attribute_values[i];
					}
				}

				if (!name || !info_symbol || parent_type == G_TYPE_INVALID) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_TYPE_ELEMENT,
					     	     _("%s: invalid driver description (incomplete type element)"),
						     data->filename ? data->filename :
						     _("[unknown]"));
					return;
				}

				if (!data->module) {
					data->module = burgner_module_new ();
				}

				burgner_module_add_type (data->module,
							 name,
							 parent_type,
							 info_symbol);
			} else 	if (strcmp (element_name, "interface") == 0) {
				gchar *info_symbol = NULL;
				GType instance_type = G_TYPE_INVALID, interface_type = G_TYPE_INVALID;
				guint i;

				data->state = BSDD_PARSER_STATE_IN_INTERFACE;

				for (i = 0; attribute_names[i]; i++) {
					if (interface_type == G_TYPE_INVALID &&
					    strcmp (attribute_names[i], "interface") == 0) {
						interface_type = g_type_from_name (attribute_values[i]);
						if (interface_type == 0) {
							g_set_error (error,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INVALID_INTERFACE_TYPE,
								     _("%s: invalid driver description (%s: invalid interface type)"),
								     data->filename ? data->filename : _("[unknown]"),
								     attribute_values[i]);
							return;
						}
					} else if (instance_type == G_TYPE_INVALID &&
						   strcmp (attribute_names[i], "parent") == 0) {
						instance_type = g_type_from_name (attribute_values[i]);
						if (instance_type == 0) {
							g_set_error (error,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
								     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INVALID_INSTANCE_TYPE,
								     _("%s: invalid driver description (%s: invalid instance type)"),
								     data->filename ? data->filename : _("[unknown]"),
								     attribute_values[i]);
							return;
						}
					} else if (!info_symbol &&
						   strcmp (attribute_names[i], "info_symbol") == 0) {
						info_symbol = (gchar *) attribute_values[i];
					}
				}

				if (interface_type == G_TYPE_INVALID ||
				    instance_type == G_TYPE_INVALID ||
				    !info_symbol) {
					g_set_error (error,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
						     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_INTERFACE_ELEMENT,
					     	     _("%s: invalid driver description (incomplete 'interface' element)"),
						     data->filename ? data->filename :
						     _("[unknown]"));
					return;
				}

				if (!data->module) {
					data->module = burgner_module_new ();
				}

				burgner_module_add_interface (data->module,
							      instance_type,
							      interface_type,
							      info_symbol);
			} else {
				data->unknown_count++;
			}
			break;
		case BSDD_PARSER_STATE_IN_TYPE:
		case BSDD_PARSER_STATE_IN_INTERFACE:
			data->unknown_count++;
			break;
		default:
			g_assert_not_reached ();
			break;
	}
}

static void
burgner_scsi_driver_description_parser_end_element
				(GMarkupParseContext *context,
				 const gchar *element_name,
				 gpointer user_data,
				 GError **error)
{
	BurgnerScsiDriverDescriptionParser *data = (BurgnerScsiDriverDescriptionParser *) user_data;

	if (data->unknown_count) {
		data->unknown_count--;
		return;
	}

	switch (data->state) {
		case BSDD_PARSER_STATE_OUT:
			g_return_if_reached ();
			break;
		case BSDD_PARSER_STATE_IN_MODULE_DESC:
			data->state = BSDD_PARSER_STATE_OUT;

			if (!data->driver_type ||
			    !data->module ||
			    !data->info) {
				g_set_error (error,
					     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR,
					     BURGNER_SCSI_DRIVER_DESCRIPTION_PARSER_ERROR_INCOMPLETE_DRIVER_DESCRIPTION,
					     _("%s: invalid driver description (incomplete driver description)"),
					     data->filename ? data->filename : _("[unknown]"));
				return;
			}

			data->info->driver_type = data->driver_type;
			data->driver_type = NULL;
			break;
		case BSDD_PARSER_STATE_IN_INFO:
		case BSDD_PARSER_STATE_IN_TYPES:
			data->state = BSDD_PARSER_STATE_IN_MODULE_DESC;
			break;
		case BSDD_PARSER_STATE_IN_TYPE:
		case BSDD_PARSER_STATE_IN_INTERFACE:
			data->state = BSDD_PARSER_STATE_IN_TYPES;
			break;
		default:
			g_return_if_reached ();
			break;
	}
}

static const GMarkupParser burgner_scsi_driver_description_parser_markup_parser = {
	burgner_scsi_driver_description_parser_start_element,
	burgner_scsi_driver_description_parser_end_element,
	NULL,
	NULL,
	NULL
};


BurgnerScsiDriverDescriptionParser *
burgner_scsi_driver_description_parser_new (gchar *filename)
{
	BurgnerScsiDriverDescriptionParser *dp = g_new0 (BurgnerScsiDriverDescriptionParser, 1);

	dp->parse_ctx = g_markup_parse_context_new (&burgner_scsi_driver_description_parser_markup_parser, 0, dp, NULL);
	if (filename)
		dp->filename = g_strdup (filename);

	return dp;
}

void
burgner_scsi_driver_description_parser_free (BurgnerScsiDriverDescriptionParser *dp)
{
	g_return_if_fail (dp != NULL);

	if (dp->parse_ctx)
		g_markup_parse_context_free (dp->parse_ctx);
	if (dp->module)
		g_object_unref (dp->module);
	if (dp->info)
		burgner_scsi_driver_info_unref (dp->info);
	if (dp->driver_type)
		g_free (dp->driver_type);
	if (dp->filename)
		g_free (dp->filename);

	g_free (dp);
}

gboolean
burgner_scsi_driver_description_parser_parse (BurgnerScsiDriverDescriptionParser *dp,
					      gchar *text,
					      guint size,
					      GError **error)
{
	GError *sub_err = NULL;

	g_return_val_if_fail (dp != NULL, FALSE);
	g_return_val_if_fail ((size == 0) || (text != NULL), FALSE);
	g_return_val_if_fail (dp->parse_ctx != NULL, FALSE);

	if (!g_markup_parse_context_parse (dp->parse_ctx,
					   text, size,
					   &sub_err)) {
		g_propagate_error (error, sub_err);
		g_markup_parse_context_free (dp->parse_ctx);
		dp->parse_ctx = NULL;

		if (dp->module) {
			g_object_unref (dp->module);
			dp->module = NULL;
		}
		if (dp->info) {
			burgner_scsi_driver_info_unref (dp->info);
			dp->info = NULL;
		}
		if (dp->driver_type) {
			g_free (dp->driver_type);
			dp->driver_type = NULL;
		}
	
		return FALSE;
	}

	return TRUE;
}

gboolean
burgner_scsi_driver_description_parser_end (BurgnerScsiDriverDescriptionParser *dp,
					    BurgnerModule **module,
					    BurgnerScsiDriverInfo **info,
					    GError **error)
{
	GError *sub_err = NULL;

	g_return_val_if_fail (dp != NULL, FALSE);
	g_return_val_if_fail (dp->parse_ctx != NULL, FALSE);


	if (!g_markup_parse_context_end_parse (dp->parse_ctx,
					       &sub_err)) {
		g_propagate_error (error, sub_err);
		g_markup_parse_context_free (dp->parse_ctx);
		dp->parse_ctx = NULL;

		if (dp->module) {
			g_object_unref (dp->module);
			dp->module = NULL;
		}
		if (dp->info) {
			burgner_scsi_driver_info_unref (dp->info);
			dp->info = NULL;
		}
		if (dp->driver_type) {
			g_free (dp->driver_type);
			dp->driver_type = NULL;
		}

		return FALSE;
	}

	g_markup_parse_context_free (dp->parse_ctx);
	dp->parse_ctx = NULL;

	g_return_val_if_fail (dp->module != NULL, FALSE);
	g_return_val_if_fail (dp->info != NULL, FALSE);
	g_return_val_if_fail (dp->driver_type == NULL, FALSE);

	if (module)
		*module = BURGNER_MODULE (g_object_ref (dp->module));
	g_object_unref (dp->module);
	dp->module = NULL;

	if (info)
		*info = burgner_scsi_driver_info_ref (dp->info);
	burgner_scsi_driver_info_unref (dp->info);
	dp->info = NULL;

	return TRUE;
}


GQuark
burgner_scsi_driver_description_parser_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiDriverDescriptionParser");
}

