/*
 * BurgnerScsiDriverManager - BurgnerScsiDriverManager object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <libintl.h>
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) x

#include <string.h>
#include <glib-object.h>
#include <libgnomevfs/gnome-vfs.h>
#include "burgner-scsi-privilege.h"
#include "burgner-scsi-marshal.h"
#include "burgner-scsi-driver-description-parser.h"
#include "burgner-scsi-driver-manager.h"


#define check(o)		G_STMT_START {	\
	g_return_if_fail (o);	\
	g_return_if_fail (BURGNER_IS_SCSI_DRIVER_MANAGER (o));	\
				} G_STMT_END
#define check_with_val(o,v)	G_STMT_START {	\
	g_return_val_if_fail ((o), (v));	\
	g_return_val_if_fail (BURGNER_IS_SCSI_DRIVER_MANAGER (o), (v));	\
				} G_STMT_END

static GObjectClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_PRIVILEGE,
	LAST_ARG
};

enum {
	LOAD_DRIVER_INFO_LIST_ASYNC_DONE,
	LAST_SIGNAL
};

static guint burgner_scsi_driver_manager_signals[LAST_SIGNAL] = { 0, };

static void
burgner_scsi_driver_manager_finalize (GObject *object)
{
	BurgnerScsiDriverManager *self = BURGNER_SCSI_DRIVER_MANAGER (object);

	if (self->privilege) {
		g_object_unref (self->privilege);
		self->privilege = NULL;
	}

	/* ... */

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
burgner_scsi_driver_manager_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiDriverManager *self = BURGNER_SCSI_DRIVER_MANAGER (object);

	switch (prop_id) {
		case PROP_PRIVILEGE:
			if (self->privilege)
				g_object_unref (self->privilege);
			self->privilege = (BurgnerScsiPrivilege *) g_value_get_object (value);
			if (self->privilege)
				self->privilege = BURGNER_SCSI_PRIVILEGE (g_object_ref (self->privilege));
			g_object_notify (object, "privilege");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_scsi_driver_manager_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiDriverManager *self = BURGNER_SCSI_DRIVER_MANAGER (object);

	switch (prop_id) {
		case PROP_PRIVILEGE:
			g_value_set_object (value, self->privilege);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_scsi_driver_manager_class_init (BurgnerScsiDriverManagerClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_scsi_driver_manager_set_property;
	object_class->get_property = burgner_scsi_driver_manager_get_property;
	object_class->finalize = burgner_scsi_driver_manager_finalize;

	/* Signals */
	burgner_scsi_driver_manager_signals[LOAD_DRIVER_INFO_LIST_ASYNC_DONE] = 
		g_signal_new ("load_driver_info_list_async_done",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (BurgnerScsiDriverManagerClass, load_driver_info_list_async_done),
			      NULL, NULL,
			      burgner_scsi_marshal_VOID__POINTER_POINTER_POINTER_POINTER,
			      G_TYPE_NONE, 4,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER);

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_PRIVILEGE,
					 g_param_spec_object ("privilege",
					                _("privilege"),
							_("A BurgnerScsiPrivilege"),
							BURGNER_TYPE_SCSI_PRIVILEGE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_scsi_driver_manager_init (BurgnerScsiDriverManager *self)
{
	self->privilege = NULL;
	self->info_list = NULL;
	self->module_id_to_info =
			g_hash_table_new_full (g_str_hash, g_str_equal,
					       NULL, NULL);
}

GType
burgner_scsi_driver_manager_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiDriverManagerClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_scsi_driver_manager_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiDriverManager),
			0,
			(GInstanceInitFunc) burgner_scsi_driver_manager_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerScsiDriverManager", &info, 0);
	}

	return type;
}

BurgnerScsiDriverManager *
burgner_scsi_driver_manager_new (void)
{
	BurgnerScsiDriverManager *self;

	self = g_object_new (BURGNER_TYPE_SCSI_DRIVER_MANAGER, NULL);


	return self;
}


GQuark
burgner_scsi_driver_manager_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiDriverManager");
}


gboolean
burgner_scsi_driver_manager_load_driver_info_list_sync (BurgnerScsiDriverManager *dm,
					gchar *directory,
				        GList **driver_info_list,
				        GError **err)
{
	GError *dir_err = NULL;
	GDir *dir;
	GString *s;
	gchar *dir_entry;
	guint directory_len;

	check_with_val (dm, FALSE);
	g_return_val_if_fail (directory != NULL, FALSE);

	dir = g_dir_open ((const gchar *) directory, 0, &dir_err);
	if (!dir) {
		g_propagate_error (err, dir_err);
		return FALSE;
	}

	s = g_string_new (directory);
	g_string_append (s, G_DIR_SEPARATOR_S);
	directory_len = s->len;

	while ((dir_entry = (gchar *) g_dir_read_name (dir))) {
		BurgnerScsiDriverDescriptionParser *ddp;
		BurgnerModule *module = NULL;
		BurgnerScsiDriverInfo *info = NULL;
		GError *sub_err = NULL;
		gchar *content = NULL, *modpath;
		guint length = 0;

		g_string_truncate (s, directory_len);
		g_string_append (s, dir_entry);

		if ((s->len - directory_len) < 5 ||
		    strcasecmp (s->str + (s->len - 5), ".desc") != 0)
			continue;

		if (!g_file_test (s->str, G_FILE_TEST_IS_REGULAR))
			continue;

		if (!g_file_get_contents (s->str,
					  &content,
					  &length,
					  &sub_err)) {
			/* 
			 * FIXME: this only reports the first error that
			 * happened.
			 */
			if (err && !*err)
				g_propagate_error (err, sub_err);
			g_error_free (sub_err);
			continue;
		}

		sub_err = NULL;
		ddp = burgner_scsi_driver_description_parser_new (s->str);
		if (!burgner_scsi_driver_description_parser_parse (ddp,
						   content, length,
						   &sub_err) ||
		    !burgner_scsi_driver_description_parser_end (ddp,
		    						 &module,
								 &info,
								 &sub_err)) {
			burgner_scsi_driver_description_parser_free (ddp);
			g_free (content);
			if (err && !*err)
				g_propagate_error (err, sub_err);
			g_error_free (sub_err);
			continue;
		}

		modpath = g_module_build_path (directory,
					       info->id);
		g_object_set (module,
			      "module_path", modpath,
			      NULL);
		g_free (modpath);

		burgner_module_register_types_ifaces (module);

		g_hash_table_insert (dm->module_id_to_info,
				     info->id,
				     info);
		dm->info_list = g_list_prepend (dm->info_list, info);
		burgner_scsi_driver_description_parser_free (ddp);
		g_free (content);
	}

	g_string_free (s, TRUE);
	g_dir_close (dir);

	return TRUE;
}


#define HAVE_GNOMEVFS
#ifdef HAVE_GNOMEVFS

static GList *_bsdm_closing_handles = NULL;

static void
_bsdm_close_handle_close_cb (GnomeVFSAsyncHandle *handle,
			     GnomeVFSResult result,
			     gpointer user_data)
{
	_bsdm_closing_handles = g_list_remove (_bsdm_closing_handles, handle);
}

static void
_bsdm_close_handle (GnomeVFSAsyncHandle *handle)
{
	g_return_if_fail (handle != NULL);
	_bsdm_closing_handles = g_list_prepend (_bsdm_closing_handles, handle);
	gnome_vfs_async_close (handle,
			       _bsdm_close_handle_close_cb,
			       NULL);
}


typedef struct {
	BurgnerScsiDriverManager *dm;

	GnomeVFSAsyncHandle *dir_handle;
	GnomeVFSURI *dir_uri;

	gchar *directory;

	GList *file_uri_list;
	GList *file_data_list;

	GMemChunk *file_data_mem_chunk;
	GMemChunk *file_buffer_mem_chunk;

	GList *modules;
	GList *driver_infos;
} _BsdmLoadAsync_Data;

#define _BSDM_LOAD_ASYNC_FILE_DATA_BUFFER_SIZE (512)
typedef struct {
	_BsdmLoadAsync_Data *data;

	GnomeVFSAsyncHandle *handle;
	GnomeVFSURI *uri;
	gpointer buffer;

	BurgnerScsiDriverDescriptionParser *parser;
} _BsdmLoadAsync_FileData;

static void
_bsdm_load_async_error (_BsdmLoadAsync_Data *data,
			GError *error)
{
	g_return_if_fail (data != NULL);

	burgner_scsi_driver_manager_load_driver_info_list_async_done (data->dm,
								      data,
								      data->directory,
								      NULL,
								      error);

	if (data->dir_handle)
		gnome_vfs_async_cancel (data->dir_handle);

	if (data->dir_uri)
		gnome_vfs_uri_unref (data->dir_uri);

	if (data->directory)
		g_free (data->directory);

	if (data->file_uri_list) {
		g_list_foreach (data->file_uri_list,
				(GFunc) gnome_vfs_uri_unref,
				NULL);
		g_list_free (data->file_uri_list);
	}

	if (data->file_data_list) {
		GList *it;
		for (it = data->file_data_list; it != NULL; it = it->next) {
			_BsdmLoadAsync_FileData *fdata = (_BsdmLoadAsync_FileData *) it->data;
			if (fdata->handle) {
				gnome_vfs_async_cancel (fdata->handle);
				_bsdm_close_handle (fdata->handle);
				gnome_vfs_uri_unref (fdata->uri);
				g_mem_chunk_free (data->file_buffer_mem_chunk, fdata->buffer);
				burgner_scsi_driver_description_parser_free (fdata->parser);
				g_mem_chunk_free (data->file_data_mem_chunk, fdata);
			}
		}
		g_list_free (data->file_data_list);
	}

	if (data->file_data_mem_chunk)
		g_mem_chunk_destroy (data->file_data_mem_chunk);

	if (data->file_buffer_mem_chunk)
		g_mem_chunk_destroy (data->file_buffer_mem_chunk);

	if (data->modules) {
		g_list_foreach (data->modules, (GFunc) g_object_unref, NULL);
		g_list_free (data->modules);
	}

	if (data->driver_infos) {
		g_list_foreach (data->driver_infos, (GFunc) burgner_scsi_driver_info_unref, NULL);
		g_list_free (data->driver_infos);
	}

	g_free (data);
}

static void
_bsdm_load_async_done (_BsdmLoadAsync_Data *data)
{
	GList *it;

	g_return_if_fail (data != NULL);
	g_return_if_fail (data->dir_handle == NULL);
	g_return_if_fail (data->file_uri_list == NULL);
	g_return_if_fail (data->file_data_list == NULL);

	burgner_scsi_driver_manager_load_driver_info_list_async_done (data->dm,
								      data,
								      data->directory,
								      data->driver_infos,
								      NULL);

	for (it = data->modules; it != NULL; it = it->next)
		burgner_module_register_types_ifaces (BURGNER_MODULE (it->data));
	/*
	 * As the gtype system doesn't seem to ref its GTypePlugin-s, we mustn't
	 * unref our modules once they've been registered. :\
	 */
	g_list_free (data->modules);

	for (it = data->driver_infos; it != NULL; it = it->next) {
		BurgnerScsiDriverInfo *info = (BurgnerScsiDriverInfo *) it->data;
		g_hash_table_insert (data->dm->module_id_to_info,
				     info->id, info);
	}

	data->dm->info_list = g_list_concat (data->dm->info_list, data->driver_infos);

	if (data->dir_uri)
		gnome_vfs_uri_unref (data->dir_uri);

	if (data->directory)
		g_free (data->directory);

	if (data->file_data_mem_chunk)
		g_mem_chunk_destroy (data->file_data_mem_chunk);

	if (data->file_buffer_mem_chunk)
		g_mem_chunk_destroy (data->file_buffer_mem_chunk);

	g_free (data);
}

static void
_bsdm_load_async_start_file (_BsdmLoadAsync_Data *data,
			     GnomeVFSURI *uri);

static void
_bsdm_load_async_file_read_cb (GnomeVFSAsyncHandle *handle,
			       GnomeVFSResult result,
			       gpointer buffer,
			       GnomeVFSFileSize bytes_requested,
			       GnomeVFSFileSize bytes_read,
			       _BsdmLoadAsync_FileData *fdata)
{
	if (result != GNOME_VFS_OK &&
	    result != GNOME_VFS_ERROR_EOF) {
		_bsdm_load_async_error (fdata->data,
					g_error_new (BURGNER_SCSI_DRIVER_MANAGER_ERROR,
						     BURGNER_SCSI_DRIVER_MANAGER_ERROR_READING_FILE,
						     _("%s: error while reading file (%s)"),
						     gnome_vfs_uri_get_path (fdata->uri),
						     gnome_vfs_result_to_string (result)));
		return;
	}

	if (bytes_read) {
		GError *err = NULL;
		if (!burgner_scsi_driver_description_parser_parse (fdata->parser,
								   (gchar *) buffer,
							           (guint) bytes_read,
							           &err)) {
			_bsdm_load_async_error (fdata->data, err);
			return;
		}
	}

	if (result == GNOME_VFS_ERROR_EOF) {
		BurgnerModule *module = NULL;
		BurgnerScsiDriverInfo *info = NULL;
		GError *err = NULL;
		_BsdmLoadAsync_Data *data;
		gchar *modpath;
		guint l;

		if (!burgner_scsi_driver_description_parser_end (fdata->parser,
								 &module,
								 &info,
								 &err)) {
			_bsdm_load_async_error (fdata->data, err);
			return;
		}

		g_return_if_fail (module != NULL);
		g_return_if_fail (info != NULL);

		modpath = g_module_build_path (fdata->data->directory,
					       info->id);
		g_object_set (module,
			      "module_path", modpath,
			      NULL);
		g_free (modpath);

		fdata->data->modules = g_list_prepend (fdata->data->modules,
						       module);
		fdata->data->driver_infos = g_list_prepend (fdata->data->driver_infos, info);

		gnome_vfs_uri_unref (fdata->uri);
		burgner_scsi_driver_description_parser_free (fdata->parser);
		g_mem_chunk_free (fdata->data->file_buffer_mem_chunk, fdata->buffer);
		_bsdm_close_handle (fdata->handle);

		data = fdata->data;
		data->file_data_list = g_list_remove (data->file_data_list, fdata);
		g_mem_chunk_free (data->file_data_mem_chunk, fdata);

		if ((l = g_list_length (data->file_data_list)) < 4) {
			if (data->file_uri_list) {
				GnomeVFSURI *uri = (GnomeVFSURI *) data->file_uri_list->data;
				g_assert (uri);
				data->file_uri_list = g_list_remove (data->file_uri_list, uri);
				_bsdm_load_async_start_file (data, uri);
			} else {
				if (l == 0 && !data->dir_handle) {
					_bsdm_load_async_done (data);
				}
			}
		}
	} else {
		gnome_vfs_async_read (fdata->handle,
				      fdata->buffer,
				      _BSDM_LOAD_ASYNC_FILE_DATA_BUFFER_SIZE,
				      (GnomeVFSAsyncReadCallback) _bsdm_load_async_file_read_cb,
				      fdata);
	}
}

static void
_bsdm_load_async_file_open_cb (GnomeVFSAsyncHandle *handle,
			       GnomeVFSResult result,
			       _BsdmLoadAsync_FileData *fdata)
{
	if (result != GNOME_VFS_OK) {
		_bsdm_load_async_error (fdata->data,
					g_error_new (BURGNER_SCSI_DRIVER_MANAGER_ERROR,
						     BURGNER_SCSI_DRIVER_MANAGER_ERROR_OPENING_FILE,
						     _("%s: error while opening file (%s)"),
						     gnome_vfs_uri_get_path (fdata->uri),
						     gnome_vfs_result_to_string (result)));
		return;
	}

	fdata->parser = burgner_scsi_driver_description_parser_new ((gchar *) gnome_vfs_uri_get_path (fdata->uri));
	fdata->buffer = g_mem_chunk_alloc (fdata->data->file_buffer_mem_chunk);

	gnome_vfs_async_read (fdata->handle,
			      fdata->buffer,
			      _BSDM_LOAD_ASYNC_FILE_DATA_BUFFER_SIZE,
			      (GnomeVFSAsyncReadCallback) _bsdm_load_async_file_read_cb,
			      fdata);
}

static void
_bsdm_load_async_start_file (_BsdmLoadAsync_Data *data,
			     GnomeVFSURI *uri)
{
	_BsdmLoadAsync_FileData *fdata;

	g_return_if_fail (data != NULL);
	g_return_if_fail (uri != NULL);

	fdata = (_BsdmLoadAsync_FileData *) g_mem_chunk_alloc0 (data->file_data_mem_chunk);
	fdata->data = data;
	fdata->uri = gnome_vfs_uri_ref (uri);

	data->file_data_list = g_list_prepend (data->file_data_list, fdata);

	gnome_vfs_async_open_uri (&fdata->handle,
				  fdata->uri,
				  GNOME_VFS_OPEN_READ,
				  8,
				  (GnomeVFSAsyncOpenCallback) _bsdm_load_async_file_open_cb,
				  fdata);
}

static void
_bsdm_load_async_load_dir_cb (GnomeVFSAsyncHandle *handle,
			      GnomeVFSResult result,
			      GList *list,
			      guint entries_read,
			      _BsdmLoadAsync_Data *data)
{
	GList *it;

	if (result != GNOME_VFS_OK &&
	    result != GNOME_VFS_ERROR_EOF) {
		GError *err = g_error_new (BURGNER_SCSI_DRIVER_MANAGER_ERROR,
					   BURGNER_SCSI_DRIVER_MANAGER_ERROR_LOADING_DIRECTORY,
					   _("%s: error while loading directory (%s)"),
					   data->directory ? data->directory : _("[unknown]"),
					   gnome_vfs_result_to_string (result));
		_bsdm_load_async_error (data, err);
		return;
	}

	for (it = list; it != NULL; it = it->next) {
		GnomeVFSFileInfo *info = (GnomeVFSFileInfo *) it->data;
		GnomeVFSURI *uri;
		guint l;

		if (!(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_TYPE) ||
		    info->type != GNOME_VFS_FILE_TYPE_REGULAR ||
		    (l = strlen (info->name)) < 5 ||
		    strcmp (info->name + l - 5, ".desc") != 0)
			continue;

		uri = gnome_vfs_uri_append_file_name (data->dir_uri, info->name);

		if (g_list_length (data->file_data_list) < 4)
			_bsdm_load_async_start_file (data, uri);
		else
			data->file_uri_list = g_list_prepend (data->file_uri_list, gnome_vfs_uri_ref (uri));

		gnome_vfs_uri_unref (uri);
	}

	if (result == GNOME_VFS_ERROR_EOF) {
		data->dir_handle = NULL;

		g_assert ((!data->file_uri_list) || data->file_data_list);
		if (!data->file_data_list) {
			_bsdm_load_async_done (data);
		}
	}
}


BurgnerScsiDriverManagerLoadDriverInfoListAsyncId
burgner_scsi_driver_manager_load_driver_info_list_async (BurgnerScsiDriverManager *dm,
							 gchar *directory)
{
	_BsdmLoadAsync_Data *data;
	GString *s;

	check_with_val (dm, NULL);
	g_return_val_if_fail (directory != NULL, NULL);

	data = g_new0 (_BsdmLoadAsync_Data, 1);
	dm->async_data_list = g_list_prepend (dm->async_data_list, data);

	data->dm = dm;
	data->directory = g_strdup (directory);

	data->file_data_mem_chunk = g_mem_chunk_new ("_BsdmLoadAsync_FileData mem chunk",
						     sizeof (_BsdmLoadAsync_FileData),
						     sizeof (_BsdmLoadAsync_FileData) * 4,
						     G_ALLOC_AND_FREE);
	data->file_buffer_mem_chunk = g_mem_chunk_new ("_BsdmLoadAsync_FileData buffer mem chunk",
						       _BSDM_LOAD_ASYNC_FILE_DATA_BUFFER_SIZE,
						       _BSDM_LOAD_ASYNC_FILE_DATA_BUFFER_SIZE * 4,
						       G_ALLOC_AND_FREE);


	s = g_string_new ("file://");

	if (!g_path_is_absolute (directory)) {
		gchar *cwd = g_get_current_dir ();
		g_string_append (s, cwd);
		g_free (cwd);
	}
	g_string_append (s, G_DIR_SEPARATOR_S);
	g_string_append (s, directory);
	data->dir_uri = gnome_vfs_uri_new (s->str);
	g_string_free (s, TRUE);

	gnome_vfs_async_load_directory_uri (&data->dir_handle,
					    data->dir_uri,
					    GNOME_VFS_FILE_INFO_DEFAULT,
					    1,
					    8,
					    (GnomeVFSAsyncDirectoryLoadCallback) _bsdm_load_async_load_dir_cb,
					    data);

	return (BurgnerScsiDriverManagerLoadDriverInfoListAsyncId) data;
}

#else

BurgnerScsiDriverManagerLoadDriverInfoListAsyncId
burgner_scsi_driver_manager_load_driver_info_list_async (BurgnerScsiDriverManager *dm,
							 gchar *directory)
{
	return NULL;
}

#endif


GList *
burgner_scsi_driver_manager_get_driver_info_list (BurgnerScsiDriverManager *dm)
{
	check_with_val (dm, NULL);
	return dm->info_list;
}

BurgnerScsiDriver *
burgner_scsi_driver_manager_get_driver (BurgnerScsiDriverManager *dm,
					gchar *driver_id)
{
	BurgnerScsiDriver *driver;
	BurgnerScsiDriverInfo *info;
	GType driver_type;

	check_with_val (dm, NULL);
	g_return_val_if_fail (driver_id != NULL, NULL);
	g_return_val_if_fail (*driver_id != 0, NULL);

	info = (BurgnerScsiDriverInfo *) g_hash_table_lookup (dm->module_id_to_info, driver_id);
	if (!info)
		return NULL;

	g_return_val_if_fail (info->driver_type != NULL, NULL);

	driver_type = g_type_from_name (info->driver_type);
	g_return_val_if_fail (driver_type != 0, NULL);

	driver = BURGNER_SCSI_DRIVER (g_object_new (driver_type, NULL));
	g_return_val_if_fail (driver != NULL, NULL);

	if (info->need_privilege && dm->privilege)
		burgner_scsi_driver_set_privilege (driver, dm->privilege);

	return driver;
}

GList *
burgner_scsi_driver_manager_get_all_drivers (BurgnerScsiDriverManager *dm)
{
	GList *driver_list = NULL, *it;

	check_with_val (dm, NULL);

	for (it = dm->info_list; it != NULL; it = it->next) {
		BurgnerScsiDriverInfo *info = (BurgnerScsiDriverInfo *) it->data;
		BurgnerScsiDriver *driver;
		GType driver_type;

		g_return_val_if_fail (info->driver_type != NULL, NULL);

		driver_type = g_type_from_name (info->driver_type);
		g_return_val_if_fail (driver_type != 0, NULL);

		driver = BURGNER_SCSI_DRIVER (g_object_new (driver_type, NULL));
		g_return_val_if_fail (driver != NULL, NULL);

		if (info->need_privilege && dm->privilege)
			burgner_scsi_driver_set_privilege (driver, dm->privilege);

		driver_list = g_list_prepend (driver_list, driver);
	}

	return driver_list;
}

BurgnerScsiDriver *
burgner_scsi_driver_manager_get_driver_from_uri (BurgnerScsiDriverManager *dm,
						 GUri *uri)
{
	const gchar *driver_id;

	check_with_val (dm, NULL);
	g_return_val_if_fail (uri != NULL, NULL);

	driver_id = uri->authority;
	g_return_val_if_fail (driver_id != NULL, NULL);

	return burgner_scsi_driver_manager_get_driver (dm, (gchar *) driver_id);
}

void
burgner_scsi_driver_manager_load_driver_info_list_async_done (BurgnerScsiDriverManager *dm,
						  BurgnerScsiDriverManagerLoadDriverInfoListAsyncId async_id,
						  gchar *directory,
						  GList *driver_info_list,
						  GError *err)
{
	check (dm);

	g_signal_emit (dm,
		       burgner_scsi_driver_manager_signals[LOAD_DRIVER_INFO_LIST_ASYNC_DONE],
		       0,
		       async_id,
		       directory,
		       driver_info_list,
		       err);
}

