/*
 * BurgnerScsiInquiryStage - BurgnerScsiInquiryStage object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <string.h>
#include <glib-object.h>
#include "burgner-scsi-inquiry.h"
#include "burgner-scsi-status-code.h"
#include "burgner-scsi-inquiry-stage.h"


static BurgnerScsiSendCommandStageClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_DATA_BUF,
	PROP_DATA_BUF_SIZE,
	BURGNER_SCSI_INQUIRY_STAGE_LAST_ARG
};

static void
burgner_scsi_inquiry_stage_finalize (GObject *object)
{
	BurgnerScsiInquiryStage *self = BURGNER_SCSI_INQUIRY_STAGE (object);

	self->data_buf = NULL;
	self->data_buf_size = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
burgner_scsi_inquiry_stage_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiInquiryStage *self = BURGNER_SCSI_INQUIRY_STAGE (object);

	switch (prop_id) {
		case PROP_DATA_BUF:
			self->data_buf = g_value_get_pointer (value);
			g_object_notify (object, "data_buf");
			break;
		case PROP_DATA_BUF_SIZE:
			self->data_buf_size = g_value_get_uint (value);
			g_object_notify (object, "data_buf_size");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_scsi_inquiry_stage_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiInquiryStage *self = BURGNER_SCSI_INQUIRY_STAGE (object);

	switch (prop_id) {
		case PROP_DATA_BUF:
			g_value_set_pointer (value, self->data_buf);
			break;
		case PROP_DATA_BUF_SIZE:
			g_value_set_uint (value, self->data_buf_size);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_scsi_inquiry_stage_get_command_parameters
				(BurgnerScsiSendCommandStage *sc,
				guchar *cdb_buf, guint cdb_buf_size,
				BurgnerScsiDirection *direction,
				guchar **data, guint *data_size,
				guint *timeout)
{
	BurgnerScsiInquiryStage *self = BURGNER_SCSI_INQUIRY_STAGE (sc);

	g_return_if_fail (cdb_buf_size >= BURGNER_SCSI_INQUIRY_CDB_SIZE);

	burgner_scsi_inquiry_simple_write_cdb (cdb_buf,
					       cdb_buf_size,
					       self->data_buf_size);

	*direction = BURGNER_SCSI_DIRECTION_FROM_DEVICE;
	*data = self->data_buf;
	*data_size = self->data_buf_size;
	*timeout = 10;
}

static gboolean
_burgner_scsi_inquiry_stage_command_result (BurgnerScsiSendCommandStage *sc,
					   guint status,
					   BurgnerScsiSense *sense,
					   guint data_used,
					   GError *command_error,
					   GError **error)
{
	BurgnerScsiInquiryStage *self = BURGNER_SCSI_INQUIRY_STAGE (sc);

	if (BURGNER_SCSI_SEND_COMMAND_STAGE_CLASS (parent_class)->command_result
			(sc, status, sense, data_used, command_error, error)) {
		if (self->data_buf && (self->data_buf_size > data_used))
			memset (((guchar *) self->data_buf) + data_used,
				0,
				self->data_buf_size - data_used);
		return TRUE;
	}

	return FALSE;
}

static void
burgner_scsi_inquiry_stage_class_init (BurgnerScsiInquiryStageClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiSendCommandStageClass *sc_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_scsi_inquiry_stage_set_property;
	object_class->get_property = burgner_scsi_inquiry_stage_get_property;
	object_class->finalize = burgner_scsi_inquiry_stage_finalize;

	sc_class = (BurgnerScsiSendCommandStageClass *) klass;
	sc_class->get_command_parameters = _burgner_scsi_inquiry_stage_get_command_parameters;
	sc_class->command_result = _burgner_scsi_inquiry_stage_command_result;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_DATA_BUF,
					 g_param_spec_pointer ("data_buf",
					                _("Data Buffer"),
							_("A buffer which will contains the inquiry data"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_DATA_BUF_SIZE,
					 g_param_spec_uint ("data_buf_size",
					                _("Data Buffer Size"),
							_("Size of the buffer which will contains the inquiry data"),
							0,
							0xff,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_scsi_inquiry_stage_init (BurgnerScsiInquiryStage *self)
{
	self->data_buf = NULL;
	self->data_buf_size = 0;
}

GType
burgner_scsi_inquiry_stage_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiInquiryStageClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_scsi_inquiry_stage_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiInquiryStage),
			0,
			(GInstanceInitFunc) burgner_scsi_inquiry_stage_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_SEND_COMMAND_STAGE, "BurgnerScsiInquiryStage", &info, 0);
	}

	return type;
}

BurgnerScsiStage *
burgner_scsi_inquiry_stage_new (void)
{
	BurgnerScsiInquiryStage *self;

	self = g_object_new (BURGNER_TYPE_SCSI_INQUIRY_STAGE, NULL);


	return BURGNER_SCSI_STAGE (self);
}


GQuark
burgner_scsi_inquiry_stage_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiInquiryStage");
}

