/*
 * BurgnerScsiLinuxDevice - BurgnerScsiLinuxDevice object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <glib-object.h>
#include "../libburgnerutils/burgner-ring-fifo.h"
#include "burgner-scsi-linux-device.h"


static BurgnerScsiDeviceClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_FD,
	PROP_DMA_SIZE,
	PROP_THREAD,
	BURGNER_SCSI_LINUX_DEVICE_LAST_ARG
};

static void
_burgner_scsi_linux_device_finalize (GObject *object)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (object);

	self->fd = -1;
	self->dma_size = 0;

	if (self->thread) {
		g_object_unref (self->thread);
		self->thread = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_scsi_linux_device_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (object);

	switch (prop_id) {
		case PROP_FD:
			self->fd = g_value_get_int (value);
			g_object_notify (object, "fd");
			break;
		case PROP_DMA_SIZE:
			self->dma_size = g_value_get_uint (value);
			g_object_notify (object, "dma_size");
			break;
		case PROP_THREAD:
			if (self->thread)
				g_object_unref (self->thread);
			self->thread = (BurgnerThread *) g_value_get_object (value);
			if (self->thread)
				self->thread = BURGNER_THREAD (g_object_ref (self->thread));
			g_object_notify (object, "thread");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_scsi_linux_device_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (object);

	switch (prop_id) {
		case PROP_FD:
			g_value_set_int (value, self->fd);
			break;
		case PROP_DMA_SIZE:
			g_value_set_uint (value, self->dma_size);
			break;
		case PROP_THREAD:
			g_value_set_object (value, self->thread);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static gboolean
_burgner_scsi_linux_device_send_cmd_sync (BurgnerScsiDevice *device,
					  gpointer cdb, guint cdb_size,
					  gpointer sense, guint sense_size, guint *sense_used,
					  gpointer data, guint data_size, guint *data_used,
					  BurgnerScsiDirection direction,
					  guint timeout,
					  gint *status,
					  GError **error)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (device);

	g_return_val_if_fail (self->fd >= 0, FALSE);

	return burgner_scsi_linux_device_send_cmd_real
				(self, self->fd,
				 cdb, cdb_size,
				 sense, sense_size, sense_used,
				 data, data_size, data_used,
				 direction, timeout, status,
				 error);
}


typedef struct {
	gpointer cdb;
	guint cdb_size;
	gpointer sense;
	gint sense_size;
	guint sense_used;
	gpointer data;
	guint data_size;
	guint data_used;
	BurgnerScsiDirection direction;
	guint timeout;
	gint status;

	GError *error;

	BurgnerScsiDeviceSendCmdCallback cb;
	gpointer cb_data;

	BurgnerScsiLinuxDevice *self;
} _BurgnerScsiLinuxDevice_SendCmd_Data;

static gboolean
_burgner_scsi_linux_device_send_cmd_thread_func (gpointer user_data)
{
	_BurgnerScsiLinuxDevice_SendCmd_Data *data = (_BurgnerScsiLinuxDevice_SendCmd_Data *) user_data;

	burgner_scsi_linux_device_send_cmd_real
				(data->self, data->self->fd,
				 data->cdb, data->cdb_size,
				 data->sense, data->sense_size, &data->sense_used,
				 data->data, data->data_size, &data->data_used,
				 data->direction,
				 data->timeout,
				 &data->status,
				 &data->error);

	return FALSE;
}

static gboolean
_burgner_scsi_linux_device_send_cmd_thread_cb_func (BurgnerThread *thread,
						    BurgnerThreadCallId id,
						    BurgnerThreadFunction function,
						    gpointer function_data,
						    gpointer user_data)
{
	_BurgnerScsiLinuxDevice_SendCmd_Data *data = (_BurgnerScsiLinuxDevice_SendCmd_Data *) user_data;

	data->self->send_cmd_data_list = g_list_remove (data->self->send_cmd_data_list, data);

	if (data->cb)
		data->cb (BURGNER_SCSI_DEVICE (data->self),
			  data->cdb, data->cdb_size,
			  data->sense, data->sense_size, data->sense_used,
			  data->data, data->data_size, data->data_used,
			  data->direction,
			  data->timeout,
			  data->status,
			  data->error,
			  data->cb_data);

	if (data->error)
		g_error_free (data->error);

	g_free (data);
	return FALSE;
}


static void
_burgner_scsi_linux_device_send_cmd (BurgnerScsiDevice *device,
				     gpointer cdb, guint cdb_size,
				     gpointer sense, guint sense_size,
				     gpointer data_b, guint data_size,
				     BurgnerScsiDirection direction,
				     guint timeout,
				     BurgnerScsiDeviceSendCmdCallback cb,
				     gpointer cb_data)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (device);
	_BurgnerScsiLinuxDevice_SendCmd_Data *data;

	g_return_if_fail (self->fd >= 0);
	g_return_if_fail (self->thread != NULL);

	data = (_BurgnerScsiLinuxDevice_SendCmd_Data *) g_new0 (_BurgnerScsiLinuxDevice_SendCmd_Data, 1);

	data->cdb = cdb;
	data->cdb_size = cdb_size;
	data->sense = sense;
	data->sense_size = sense_size;
	data->data = data_b;
	data->data_size = data_size;
	data->direction = direction;
	data->timeout = timeout;
	data->cb = cb;
	data->cb_data = cb_data;
	data->self = self;

	burgner_thread_call (self->thread,
			     _burgner_scsi_linux_device_send_cmd_thread_func,
			     data,
			     _burgner_scsi_linux_device_send_cmd_thread_cb_func,
			     data);
}


static BurgnerFifo *
_burgner_scsi_linux_device_get_fifo (BurgnerScsiDevice *device)
{
	BurgnerScsiLinuxDevice *self = BURGNER_SCSI_LINUX_DEVICE (device);
	BurgnerFifo *fifo;
	guint cs;

	cs = self->dma_size > 0 ? self->dma_size : 8192;

	fifo = burgner_ring_fifo_new ();
	g_object_set (fifo,
		      "read_chunk_size", cs,
		      "write_chunk_size", cs,
		      NULL);

	return fifo;
}


static void
_burgner_scsi_linux_device_class_init (BurgnerScsiLinuxDeviceClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiDeviceClass *device_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_scsi_linux_device_set_property;
	object_class->get_property = _burgner_scsi_linux_device_get_property;
	object_class->finalize = _burgner_scsi_linux_device_finalize;

	device_class = (BurgnerScsiDeviceClass *) klass;
	device_class->send_cmd_sync = _burgner_scsi_linux_device_send_cmd_sync;
	device_class->send_cmd = _burgner_scsi_linux_device_send_cmd;
	device_class->get_fifo = _burgner_scsi_linux_device_get_fifo;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_FD,
					 g_param_spec_int ("fd",
					                _("fd"),
							_("A file descriptor"),
							-1,
							G_MAXINT,
							-1,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_DMA_SIZE,
					 g_param_spec_uint ("dma_size",
					                _("dma_size"),
							_("DMA size"),
							0,
							G_MAXUINT,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_THREAD,
					 g_param_spec_object ("thread",
					                _("thread"),
							_("A BurgnerThread"),
							BURGNER_TYPE_THREAD,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_scsi_linux_device_init (BurgnerScsiLinuxDevice *self)
{
	self->fd = -1;
	self->dma_size = 0;
	self->thread = NULL;
	self->send_cmd_data_list = NULL;
}

GType
burgner_scsi_linux_device_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiLinuxDeviceClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_scsi_linux_device_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiLinuxDevice),
			0,
			(GInstanceInitFunc) _burgner_scsi_linux_device_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_DEVICE, "BurgnerScsiLinuxDevice", &info, 0);
	}

	return type;
}


GQuark
burgner_scsi_linux_device_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiLinuxDevice");
}

gboolean
burgner_scsi_linux_device_send_cmd_real (BurgnerScsiLinuxDevice *device,
				   gint fd,
				   gpointer cdb, guint cdb_size,
				   gpointer sense, guint sense_size, guint *sense_used,
				   gpointer data, guint data_size, guint *data_used,
				   BurgnerScsiDirection direction,
				   guint timeout, gint *status,
				   GError **error)
{
	g_return_val_if_fail (device != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DEVICE (device), FALSE);

	if (BURGNER_SCSI_LINUX_DEVICE_GET_CLASS (device)->send_cmd_real)
		return BURGNER_SCSI_LINUX_DEVICE_GET_CLASS (device)->send_cmd_real (device, fd, cdb, cdb_size, sense, sense_size, sense_used, data, data_size, data_used, direction, timeout, status, error);

	return FALSE;
}

