/*
 * BurgnerScsiLinuxDriver - BurgnerScsiLinuxDriver object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <glib-object.h>
#include "burgner-scsi-linux-driver.h"


static BurgnerScsiDriverClass *parent_class = NULL;

static void
_burgner_scsi_linux_driver_finalize (GObject *object)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (object);

	g_return_if_fail (self->devices == NULL);
	g_return_if_fail (self->open_device_data_list == NULL);
	g_return_if_fail (self->close_device_data_list == NULL);
	g_return_if_fail (self->get_uri_data_list == NULL);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static BurgnerScsiDevice *
_burgner_scsi_linux_driver_open_device_sync (BurgnerScsiDriver *driver,
					     const GUri *uri,
					     gint requested_dma_size,
					     gint *real_dma_size,
					     GError **error)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);
	BurgnerScsiDevice *device;
	BurgnerThread *thread;
	gint fd, dma = 0;
	gchar *file;

	g_return_val_if_fail (uri != NULL, NULL);

	file = burgner_scsi_linux_driver_get_file_name_from_uri (self, uri, error);
	g_return_val_if_fail (file != NULL, NULL);

	thread = burgner_thread_new (error);
	if (!thread)
		return NULL;

	fd = burgner_scsi_linux_driver_open_device_real
					(self,
					 file,
					 requested_dma_size, &dma,
					 error);
	if (fd < 0) {
		g_free (file);
		g_object_unref (thread);
		return NULL;
	}

	if (real_dma_size)
		*real_dma_size = dma;

	device = burgner_scsi_linux_driver_create_device (self);
	g_object_set (device,
		      "fd", fd,
		      "dma_size", dma,
		      "thread", thread,
		      NULL);
	g_object_unref (thread);
	g_free (file);

	self->devices = g_list_prepend (self->devices, device);

	return BURGNER_SCSI_DEVICE (g_object_ref (device));
}

static gboolean
_burgner_scsi_linux_driver_close_device_sync (BurgnerScsiDriver *driver,
					      BurgnerScsiDevice *device,
					      GError **error)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);
	GList *l;
	gint fd = -1;

	l = g_list_find (self->devices, device);
	g_return_val_if_fail (l != NULL, FALSE);

	g_object_get (device,
		      "fd", &fd,
		      NULL);
	g_return_val_if_fail (fd >= 0, FALSE);

	if (!burgner_scsi_linux_driver_close_device_real (self, fd, error))
		return FALSE;

	g_object_set (device,
		      "fd", -1,
		      NULL);
	self->devices = g_list_delete_link (self->devices, l);
	g_object_unref (device);

	return TRUE;
}


typedef struct {
	const gchar *file;
	gint fd;

	GError *error;

	BurgnerScsiDriverOpenDeviceCallback cb;
	gpointer cb_data;
	GUri *uri;
	gint requested_dma_size;
	gint real_dma_size;

	guint idle_id;

	BurgnerThread *thread;
	BurgnerScsiLinuxDriver *self;
} _BurgnerScsiLinuxDriver_OpenDevice_Data;

static gboolean
_burgner_scsi_linux_driver_open_device_thread_func (gpointer thread_data)
{
	_BurgnerScsiLinuxDriver_OpenDevice_Data *data = (_BurgnerScsiLinuxDriver_OpenDevice_Data *) thread_data;

	data->error = NULL;
	data->fd = burgner_scsi_linux_driver_open_device_real
					(data->self,
					 data->file,
					 data->requested_dma_size,
					 &data->real_dma_size,
					 &data->error);

	return FALSE;
}

static gboolean
_burgner_scsi_linux_driver_open_device_thread_cb_func (BurgnerThread *thread,
						       BurgnerThreadCallId id,
						       BurgnerThreadFunction function,
						       gpointer function_data,
						       gpointer user_data)
{
	_BurgnerScsiLinuxDriver_OpenDevice_Data *data = (_BurgnerScsiLinuxDriver_OpenDevice_Data *) user_data;
	BurgnerScsiDevice *device;
	
	data->self->open_device_data_list = g_list_remove (data->self->open_device_data_list, data);

	if (data->fd < 0) {
		if (data->cb)
			data->cb (BURGNER_SCSI_DRIVER (data->self),
				  data,
				  data->uri,
				  NULL,
				  data->requested_dma_size,
				  data->real_dma_size,
				  data->error,
				  data->cb_data);

		if (data->error)
			g_error_free (data->error);
		if (data->uri)
			g_uri_unref (data->uri);
		if (data->thread)
			g_object_unref (data->thread);

		g_free (data);
		return FALSE;
	}

	device = burgner_scsi_linux_driver_create_device (data->self);
	g_object_set (device,
		      "fd", data->fd,
		      "dma_size", data->real_dma_size,
		      "thread", data->thread,
		      NULL);
	data->self->devices = g_list_prepend (data->self->devices, device);

	if (data->cb)
		data->cb (BURGNER_SCSI_DRIVER (data->self),
			  data,
			  data->uri,
			  device,
			  data->requested_dma_size,
			  data->real_dma_size,
			  data->error,
			  data->cb_data);

	if (data->uri)
		g_uri_unref (data->uri);
	if (data->thread)
		g_object_unref (data->thread);
	g_free (data);

	return FALSE;
}

static gboolean
_burgner_scsi_linux_driver_open_device_idle_func (gpointer user_data)
{
	_BurgnerScsiLinuxDriver_OpenDevice_Data *data = (_BurgnerScsiLinuxDriver_OpenDevice_Data *) user_data;

	data->idle_id = 0;

	data->file = burgner_scsi_linux_driver_get_file_name_from_uri (data->self,
								       data->uri,
								       &data->error);
	if (!data->file) {
		data->self->open_device_data_list = g_list_remove (data->self->open_device_data_list, data);

		if (data->cb)
			data->cb (BURGNER_SCSI_DRIVER (data->self),
				  (BurgnerScsiDriverOpId) data,
				  data->uri,
				  NULL,
				  data->requested_dma_size,
				  data->real_dma_size,
				  data->error,
				  data->cb_data);

		if (data->error)
			g_error_free (data->error);
		if (data->uri)
			g_uri_unref (data->uri);

		g_free (data);
		return FALSE;
	}

	data->error = NULL;
	data->thread = burgner_thread_new (&data->error);
	if (!data->thread) {
		data->self->open_device_data_list = g_list_remove (data->self->open_device_data_list, data);

		if (data->cb)
			data->cb (BURGNER_SCSI_DRIVER (data->self),
				  (BurgnerScsiDriverOpId) data,
				  data->uri,
				  NULL,
				  data->requested_dma_size,
				  data->real_dma_size,
				  data->error,
				  data->cb_data);

		if (data->error)
			g_error_free (data->error);
		if (data->uri)
			g_uri_unref (data->uri);

		g_free (data);
		return FALSE;
	}

	burgner_thread_call (data->thread,
			     _burgner_scsi_linux_driver_open_device_thread_func,
			     data,
			     _burgner_scsi_linux_driver_open_device_thread_cb_func,
			     data);

	return FALSE;
}


static BurgnerScsiDriverOpId
_burgner_scsi_linux_driver_open_device (BurgnerScsiDriver *driver,
					const GUri *uri,
					gint requested_dma_size,
					BurgnerScsiDriverOpenDeviceCallback cb,
					gpointer cb_data)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);
	_BurgnerScsiLinuxDriver_OpenDevice_Data *data;

	data = g_new0 (_BurgnerScsiLinuxDriver_OpenDevice_Data, 1);

	data->self = self;
	data->cb = cb;
	data->cb_data = cb_data;
	data->uri = g_uri_clone (uri);
	data->requested_dma_size = requested_dma_size;

	self->open_device_data_list = g_list_prepend (self->open_device_data_list, data);

	data->idle_id = g_idle_add (_burgner_scsi_linux_driver_open_device_idle_func, data);

	return (BurgnerScsiDriverOpId) data;
}


typedef struct {
	BurgnerScsiDevice *device;
	gint fd;

	GError *error;

	BurgnerScsiDriverCloseDeviceCallback cb;
	gpointer cb_data;

	BurgnerScsiLinuxDriver *self;
	BurgnerThread *thread;
} _BurgnerScsiLinuxDriver_CloseDevice_Data;

static gboolean
_burgner_scsi_linux_driver_close_device_thread_func (gpointer thread_data)
{
	_BurgnerScsiLinuxDriver_CloseDevice_Data *data = (_BurgnerScsiLinuxDriver_CloseDevice_Data *) thread_data;

	if (burgner_scsi_linux_driver_close_device_real (data->self, data->fd, &data->error)) {
		data->fd = -1;
	}

	return FALSE;
}

static gboolean
_burgner_scsi_linux_driver_close_device_thread_cb_func (BurgnerThread *thread,
							BurgnerThreadCallId id,
							BurgnerThreadFunction function,
							gpointer function_data,
							gpointer user_data)
{
	_BurgnerScsiLinuxDriver_CloseDevice_Data *data = (_BurgnerScsiLinuxDriver_CloseDevice_Data *) user_data;

	data->self->close_device_data_list = g_list_remove (data->self->close_device_data_list, data);

	if (data->fd >= 0) {
		if (data->cb)
			data->cb (BURGNER_SCSI_DRIVER (data->self),
				  data,
				  data->device,
				  data->error,
				  data->cb_data);

		if (data->thread)
			g_object_unref (data->thread);
		if (data->error)
			g_error_free (data->error);
		if (data->device)
			g_object_unref (data->device);

		g_free (data);
		return FALSE;
	}

	if (data->cb)
		data->cb (BURGNER_SCSI_DRIVER (data->self),
			  data,
			  data->device,
			  NULL,
			  data->cb_data);

	if (data->thread)
		g_object_unref (data->thread);
	if (data->error)
		g_error_free (data->error);
	if (data->device)
		g_object_unref (data->device);

	g_free (data);
	return FALSE;
}


static BurgnerScsiDriverOpId
_burgner_scsi_linux_driver_close_device (BurgnerScsiDriver *driver,
					 BurgnerScsiDevice *device,
					 BurgnerScsiDriverCloseDeviceCallback cb,
					 gpointer cb_data)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);
	_BurgnerScsiLinuxDriver_CloseDevice_Data *data;
	BurgnerThread *thread = NULL;
	GList *l;
	gint fd = -1;

	l = g_list_find (self->devices, device);
	g_return_val_if_fail (l != NULL, NULL);

	g_object_get (device,
		      "fd", &fd,
		      "thread", &thread,
		      NULL);
	g_return_val_if_fail (fd >= 0, NULL);
	g_return_val_if_fail (thread != NULL, NULL);

	self->devices = g_list_delete_link (self->devices, l);

	data = g_new0 (_BurgnerScsiLinuxDriver_CloseDevice_Data, 1);

	data->device = device;
	data->fd = fd;
	data->cb = cb;
	data->cb_data = cb_data;
	data->self = self;
	data->thread = BURGNER_THREAD (g_object_ref (thread));

	self->close_device_data_list = g_list_prepend (self->close_device_data_list, data);

	burgner_thread_call (thread,
			     _burgner_scsi_linux_driver_close_device_thread_func,
			     data,
			     _burgner_scsi_linux_driver_close_device_thread_cb_func,
			     data);

	return data;
}

static GList *
_burgner_scsi_linux_driver_get_uri_list_sync (BurgnerScsiDriver *driver,
					      GError **error)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);

	return burgner_scsi_linux_driver_get_uri_list_real (self, error);
}


typedef struct {
	GList *l;

	BurgnerScsiDriverGetUriListCallback cb;
	gpointer cb_data;

	GError *error;

	guint idle_id;

	BurgnerScsiLinuxDriver *self;
	BurgnerThread *thread;
} _BurgnerScsiLinuxDriver_GetUriList_Data;

static gboolean
_burgner_scsi_linux_driver_get_uri_list_thread_func (gpointer thread_data)
{
	_BurgnerScsiLinuxDriver_GetUriList_Data *data = (_BurgnerScsiLinuxDriver_GetUriList_Data *) thread_data;

	data->l = burgner_scsi_linux_driver_get_uri_list_real (data->self, &data->error);

	return FALSE;
}

static gboolean
_burgner_scsi_linux_driver_get_uri_list_thread_cb_func (BurgnerThread *thread,
							BurgnerThreadCallId id,
							BurgnerThreadFunction function,
							gpointer function_data,
							gpointer user_data)
{
	_BurgnerScsiLinuxDriver_GetUriList_Data *data = (_BurgnerScsiLinuxDriver_GetUriList_Data *) user_data;

	g_assert (!data->error || !data->l);

	data->self->get_uri_data_list = g_list_remove (data->self->get_uri_data_list, data);

	if (data->cb)
		data->cb (BURGNER_SCSI_DRIVER (data->self),
			  data->l,
			  data->error,
			  data->cb_data);

	if (data->error)
		g_error_free (data->error);

	if (data->thread)
		g_object_unref (data->thread);

	g_free (data);
	return FALSE;
}


static gboolean
_burgner_scsi_linux_driver_get_uri_list_idle_func (gpointer user_data)
{
	_BurgnerScsiLinuxDriver_GetUriList_Data *data = (_BurgnerScsiLinuxDriver_GetUriList_Data *) user_data;
	GError *error = NULL;

	data->idle_id = 0;

	if (!(data->thread = burgner_thread_new (&error))) {
		data->self->get_uri_data_list = g_list_remove (data->self->get_uri_data_list, data);

		if (data->cb)
			data->cb (BURGNER_SCSI_DRIVER (data->self),
				  NULL,
				  error,
				  data->cb_data);

		g_free (data);
		return FALSE;
	}

	burgner_thread_call (data->thread,
			     _burgner_scsi_linux_driver_get_uri_list_thread_func,
			     data,
			     _burgner_scsi_linux_driver_get_uri_list_thread_cb_func,
			     data);

	return FALSE;
}

static BurgnerScsiDriverOpId
_burgner_scsi_linux_driver_get_uri_list (BurgnerScsiDriver *driver,
					 BurgnerScsiDriverGetUriListCallback cb,
					 gpointer cb_data)
{
	BurgnerScsiLinuxDriver *self = BURGNER_SCSI_LINUX_DRIVER (driver);
	_BurgnerScsiLinuxDriver_GetUriList_Data *data;

	data = (_BurgnerScsiLinuxDriver_GetUriList_Data *) g_new0 (_BurgnerScsiLinuxDriver_GetUriList_Data, 1);

	data->cb = cb;
	data->cb_data = cb_data;
	data->self = self;

	self->get_uri_data_list = g_list_prepend (self->get_uri_data_list, data);

	data->idle_id = g_idle_add (_burgner_scsi_linux_driver_get_uri_list_idle_func, data);

	return data;
}

static gboolean
_burgner_scsi_linux_driver_need_privilege (BurgnerScsiDriver *driver)
{
	return FALSE;
}

static gchar *
_burgner_scsi_linux_driver_get_file_name_from_uri_default (BurgnerScsiLinuxDriver *driver,
							   const GUri *uri,
							   GError **error)
{
	g_return_val_if_fail (uri != NULL, NULL);

	if (!g_uri_get_scheme (uri) ||
	    g_ascii_strcasecmp (g_uri_get_scheme (uri), "file") != 0) {
		g_set_error (error,
			     BURGNER_SCSI_DRIVER_ERROR,
			     BURGNER_SCSI_DRIVER_ERROR_UNSUPPORTED_URI_SCHEME,
			     _("Unsupported URI scheme: '%s'"),
			             uri->scheme);
		return NULL;
	}

	return g_strdup (g_uri_get_path (uri));
}

static gboolean
_burgner_scsi_linux_driver_close_device_real_default (BurgnerScsiLinuxDriver *driver,
						      gint fd,
						      GError **error)
{
	if (close (fd) != 0) {
		g_set_error (error,
			     BURGNER_SCSI_DRIVER_ERROR,
			     BURGNER_SCSI_DRIVER_ERROR_CLOSING_DEVICE,
			     _("Error while closing device: %s"),
			     		strerror (errno));
		return FALSE;
	}

	return TRUE;
}



static void
_burgner_scsi_linux_driver_class_init (BurgnerScsiLinuxDriverClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiDriverClass *driver_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = _burgner_scsi_linux_driver_finalize;

	driver_class = (BurgnerScsiDriverClass *) klass;
	driver_class->open_device_sync = _burgner_scsi_linux_driver_open_device_sync;
	driver_class->close_device_sync = _burgner_scsi_linux_driver_close_device_sync;
	driver_class->open_device = _burgner_scsi_linux_driver_open_device;
	driver_class->close_device = _burgner_scsi_linux_driver_close_device;
	driver_class->get_uri_list_sync = _burgner_scsi_linux_driver_get_uri_list_sync;
	driver_class->get_uri_list = _burgner_scsi_linux_driver_get_uri_list;
	driver_class->need_privilege = _burgner_scsi_linux_driver_need_privilege;

	/* Virtual functions */
	klass->get_file_name_from_uri = _burgner_scsi_linux_driver_get_file_name_from_uri_default;
	klass->close_device_real = _burgner_scsi_linux_driver_close_device_real_default;


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_scsi_linux_driver_init (BurgnerScsiLinuxDriver *self)
{
	self->devices = self->open_device_data_list =
		self->close_device_data_list =
		self->get_uri_data_list = NULL;
}

GType
burgner_scsi_linux_driver_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiLinuxDriverClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_scsi_linux_driver_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiLinuxDriver),
			0,
			(GInstanceInitFunc) _burgner_scsi_linux_driver_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_DRIVER, "BurgnerScsiLinuxDriver", &info, 0);
	}

	return type;
}


GQuark
burgner_scsi_linux_driver_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiLinuxDriver");
}

gchar *
burgner_scsi_linux_driver_get_file_name_from_uri (BurgnerScsiLinuxDriver *driver,
					  const GUri *uri,
					  GError **error)
{
	g_return_val_if_fail (driver != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DRIVER (driver), NULL);

	if (BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->get_file_name_from_uri)
		return BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->get_file_name_from_uri (driver, uri, error);

	return NULL;
}

gint
burgner_scsi_linux_driver_open_device_real (BurgnerScsiLinuxDriver *driver,
				  const gchar *file,
				  gint requested_dma_size,
				  gint *real_dma_size,
				  GError **error)
{
	g_return_val_if_fail (driver != NULL, -1);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DRIVER (driver), -1);

	if (BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->open_device_real)
		return BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->open_device_real (driver, file, requested_dma_size, real_dma_size, error);

	return -1;
}

gboolean
burgner_scsi_linux_driver_close_device_real (BurgnerScsiLinuxDriver *driver,
				       gint fd,
				       GError **error)
{
	g_return_val_if_fail (driver != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DRIVER (driver), FALSE);

	if (BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->close_device_real)
		return BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->close_device_real (driver, fd, error);

	return FALSE;
}


GList *
burgner_scsi_linux_driver_get_uri_list_real (BurgnerScsiLinuxDriver *driver,
				     GError **error)
{
	g_return_val_if_fail (driver != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DRIVER (driver), NULL);

	if (BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->get_uri_list_real)
		return BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->get_uri_list_real (driver, error);

	return NULL;
}


BurgnerScsiDevice *
burgner_scsi_linux_driver_create_device (BurgnerScsiLinuxDriver *driver)
{
	g_return_val_if_fail (driver != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_SCSI_LINUX_DRIVER (driver), NULL);

	if (BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->create_device)
		return BURGNER_SCSI_LINUX_DRIVER_GET_CLASS (driver)->create_device (driver);

	return NULL;
}

