/*
 * 
 *
 * This file is part of the libburgnerscsi library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <string.h>

#include "burgner-scsi-spc-impl.h"
#include "burgner-scsi-utils.h"
#include "burgner-scsi-mode-parameter-list.h"

gboolean
burgner_scsi_mode_parameter_list_write_list (gboolean is_10,
					     gpointer buf, guint buf_size,
					     guint *buf_used,
					     GList *pages)
{
	guchar *b;
	GList *it;
	guint l, s;

	g_return_val_if_fail (buf != NULL, FALSE);
	g_return_val_if_fail (buf_size > (is_10 ? 8 : 4), FALSE);
	g_return_val_if_fail (pages != NULL, FALSE);

	b = (guchar *) buf;
	s = buf_size;

	/*
	 * When using MODE SELECT command, the Mode Parameter List should
	 * be NULL.
	 */
	l = is_10 ? 8 : 4;
	memset (b, 0, l);
	b += l;
	s -= l;

	for (it = pages; it != NULL; it = it->next) {
		_BurgnerScsiSpcImplModePageHeader *header = (_BurgnerScsiSpcImplModePageHeader *) it->data;

		l = header->page_length + 2;

		if (l > s) {
			if (buf_used)
				*buf_used = buf_size - s;
			return FALSE;
		}

		memcpy (b, it->data, l);

		s -= l;
		b += l;
	}

	if (buf_used)
		*buf_used = buf_size - s;

	return TRUE;
}

gboolean
burgner_scsi_mode_parameter_list_write_args (gboolean is_10,
					     gpointer buf, guint buf_size,
					     guint *buf_used,
					     gconstpointer *first_page,
					     ...)
{
	GList *l = NULL;
	gconstpointer p;
	va_list args;
	gboolean ret;

	g_return_val_if_fail (buf != NULL, FALSE);
	g_return_val_if_fail (buf_size > (is_10 ? 8 : 4), FALSE);
	g_return_val_if_fail (first_page != NULL, FALSE);

	va_start (args, first_page);

	p = first_page;

	do {
		g_return_val_if_fail (p != NULL, FALSE);
		l = g_list_prepend (l, (gpointer) p);
	} while ((p = va_arg (args, gconstpointer)));

	l = g_list_reverse (l);

	ret = burgner_scsi_mode_parameter_list_write_list (is_10,
							   buf,
							   buf_size,
							   buf_used,
							   l);

	g_list_free (l);
	return ret;
}


/**
 * burgner_scsi_mode_parameter_list_parse:
 * @is_10: must be %TRUE if buf was filled by a MODE SENSE 10 command; %FALSE else (filled by a MODE SENSE 6 command)
 * @buf: data from the MODE SENSE {6,10} command
 * @buf_size: size of @buf
 * @buf_missing: if non-NULL, will contain the size of the missing part of buf (if not equal to 0, you should issue a new MODE SENSE command, with a data buffer of which size is @buf_size + *@buf_missing)
 * @pages: if non-NULL, will contain the list of pages (pointers into buf) that you must free with #g_list_free when no more needed
 * @error: return location for a #GError
 *
 * Parses the mode parameter list contains in @buf. If an error occurred, returns %FALSE and set *@error; else returns %TRUE.
 *
 * Returns: %TRUE, if no error occurred
 */
gboolean
burgner_scsi_mode_parameter_list_parse (gboolean is_10,
					gpointer buf, guint buf_size,
					guint *buf_missing,
					GList **pages,
					GError **err)
{
	GList *p;
	guchar *b;
	guint br, dr;

	g_return_val_if_fail (buf != NULL, FALSE);
	g_return_val_if_fail (buf_size >= (is_10 ? 8 : 4), FALSE);

	p = NULL;
	b = (guchar *) buf;
	br = buf_size;

	if (is_10) {
		_BurgnerScsiSpcImplModeParameterHeader10 *impl =
			(_BurgnerScsiSpcImplModeParameterHeader10 *) buf;
		guint16 ds = burgner_scsi_get_uint16 (impl->mode_data_length);
		b += 8;
		br -= 8;
		dr = ds - 6;
	} else {
		_BurgnerScsiSpcImplModeParameterHeader6 *impl =
			(_BurgnerScsiSpcImplModeParameterHeader6 *) buf;
		guint ds = impl->mode_data_length;
		b += 4;
		br -= 4;
		dr = ds - 3;
	}

	while (br >= 2 && dr >= 2) {
		_BurgnerScsiSpcImplModePageHeader *impl =
			(_BurgnerScsiSpcImplModePageHeader *) b;

		if (dr >= impl->page_length) {
			if (br >= impl->page_length) {
				p = g_list_prepend (p, b);
				b += impl->page_length + 2;
				br -= impl->page_length + 2;
				dr -= impl->page_length + 2;
			} else {
				if (buf_missing)
					*buf_missing = dr - br;

				if (pages) {
					*pages = p;
				} else {
					g_list_free (p);
				}

				return TRUE;
			}
		} else {
			g_set_error (err,
				     BURGNER_SCSI_MODE_PARAMETER_LIST_ERROR,
				     BURGNER_SCSI_MODE_PARAMETER_LIST_ERROR_MALFORMED_PARAMETER_LIST,
				     _("Malformed Parameter List"));
			if (p)
				g_list_free (p);
			if (buf_missing)
				*buf_missing = 0;

			return FALSE;
		}
	}

	if (dr == 0) {
		if (pages) {
			*pages = p;
		} else {
			g_list_free (p);
		}
		if (buf_missing)
			*buf_missing = 0;
		return TRUE;
	} else if (dr >= 2) {
		g_assert (br < 2);
		if (buf_missing)
			*buf_missing = dr - br;
		if (pages) {
			*pages = p;
		} else {
			g_list_free (p);
		}
		return TRUE;
	} else {
		g_set_error (err,
			     BURGNER_SCSI_MODE_PARAMETER_LIST_ERROR,
			     BURGNER_SCSI_MODE_PARAMETER_LIST_ERROR_MALFORMED_PARAMETER_LIST,
			     _("Malformed Parameter List"));
		return FALSE;
	}

	g_return_val_if_reached (FALSE);
	return FALSE;
}

GQuark
burgner_scsi_mode_parameter_list_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiModeParameterList");
}

