/*
 * BurgnerScsiSendCommandStage - BurgnerScsiSendCommandStage object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <stdio.h>
#include <string.h>
#include <glib-object.h>
#include "burgner-scsi-status-code.h"
#include "burgner-scsi-stage-error-utils.h"
#include "burgner-scsi-send-command-stage.h"


static BurgnerScsiStageClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_DEVICE,
	PROP_CDB_BUF,
	PROP_CDB_BUF_SIZE,
	PROP_SENSE_BUF,
	PROP_SENSE_BUF_SIZE,
	BURGNER_SCSI_SEND_COMMAND_STAGE_LAST_ARG
};

static void
burgner_scsi_send_command_stage_finalize (GObject *object)
{
	BurgnerScsiSendCommandStage *self = BURGNER_SCSI_SEND_COMMAND_STAGE (object);

	if (self->device) {
		g_object_unref (self->device);
		self->device = NULL;
	}

	if (self->cdb_buf)
		self->cdb_buf = NULL;
	if (self->cdb_buf_size)
		self->cdb_buf_size = 0;
	if (self->sense_buf)
		self->sense_buf = NULL;
	if (self->sense_buf_size)
		self->sense_buf_size = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
burgner_scsi_send_command_stage_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiSendCommandStage *self = BURGNER_SCSI_SEND_COMMAND_STAGE (object);

	switch (prop_id) {
		case PROP_DEVICE:
			if (self->device)
				g_object_unref (self->device);
			self->device = (BurgnerScsiDevice *) g_value_get_object (value);
			if (self->device)
				self->device = BURGNER_SCSI_DEVICE (g_object_ref (self->device));
			g_object_notify (object, "device");
			break;
		case PROP_CDB_BUF:
			self->cdb_buf = (guchar *) g_value_get_pointer (value);
			g_object_notify (object, "cdb_buf");
			break;
		case PROP_CDB_BUF_SIZE:
			self->cdb_buf_size = g_value_get_uint (value);
			g_object_notify (object, "cdb_buf_size");
			break;
		case PROP_SENSE_BUF:
			self->sense_buf = g_value_get_pointer (value);
			g_object_notify (object, "sense_buf");
			break;
		case PROP_SENSE_BUF_SIZE:
			self->sense_buf_size = g_value_get_uint (value);
			g_object_notify (object, "sense_buf_size");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_scsi_send_command_stage_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiSendCommandStage *self = BURGNER_SCSI_SEND_COMMAND_STAGE (object);

	switch (prop_id) {
		case PROP_DEVICE:
			g_value_set_object (value, self->device);
			break;
		case PROP_CDB_BUF:
			g_value_set_pointer (value, self->cdb_buf);
			break;
		case PROP_CDB_BUF_SIZE:
			g_value_set_uint (value, self->cdb_buf_size);
			break;
		case PROP_SENSE_BUF:
			g_value_set_pointer (value, self->sense_buf);
			break;
		case PROP_SENSE_BUF_SIZE:
			g_value_set_uint (value, self->sense_buf_size);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_scsi_send_command_stage_dev_send_cmd_cb (BurgnerScsiDevice *device,
						 gpointer cdb, guint cdb_size,
						 gpointer sense_buf, guint sense_size, guint sense_used,
						 gpointer data, guint data_size, guint data_used,
						 BurgnerScsiDirection direction,
						 guint timeout,
						 guint status,
						 GError *command_error,
						 BurgnerScsiSendCommandStage *self)
{
	BurgnerScsiSense *sense = NULL;
	GError *error = NULL;

	if (sense_used) {
		memset (((guchar *) sense_buf) + sense_used, 0, sense_size - sense_used);
		(void) burgner_scsi_sense_new_from_buf (&sense, sense_buf, sense_size, NULL);
	}

	/*
	 * Horrible hack!
	 *
	 * The BurgnerScsiSendCommandStage should have a "unit_attention_behavior"
	 * parameter, that should specify its beavior when a unit attention condition
	 * occurs. It should:
	 * 	- ignore it (and send the command again), or
	 * 	- emit a signal (called "unit_attention") and let the user choose whether
	 * 	   the command should be sent again, or
	 * 	- make as if this condition didn't append (and emit the "end" signal).
	 */
	if (sense && sense->sense_key == BURGNER_SCSI_SENSE_KEY_UNIT_ATTENTION) {
		burgner_scsi_sense_unref (sense);
		fprintf (stderr, "Got unit attention\n");
		burgner_scsi_device_send_cmd (device,
					      cdb, cdb_size,
					      sense_buf, sense_size,
					      data, data_size,
					      direction,
					      timeout,
					      (BurgnerScsiDeviceSendCmdCallback) _burgner_scsi_send_command_stage_dev_send_cmd_cb,
					      self);
		return;
	}

	if (!burgner_scsi_send_command_stage_command_result (self, status, sense, data_used, command_error, &error)) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), FALSE, error, sense);
		if (error)
			g_error_free (error);
	} else {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), FALSE, NULL, sense);
	}

	if (sense)
		burgner_scsi_sense_unref (sense);
}

static void
_burgner_scsi_send_command_stage_real_start (BurgnerScsiStage *stage)
{
	BurgnerScsiSendCommandStage *self = BURGNER_SCSI_SEND_COMMAND_STAGE (stage);
	BurgnerScsiDirection direction = BURGNER_SCSI_DIRECTION_NONE;
	guchar *data = NULL;
	guint data_size = 0, timeout = 10;

	g_return_if_fail (self->device != NULL);
	g_return_if_fail (self->cdb_buf != NULL);
	g_return_if_fail (self->cdb_buf_size != 0);
	g_return_if_fail ((self->sense_buf_size == 0) || (self->sense_buf != NULL));

	burgner_scsi_send_command_stage_get_command_parameters (self,
								self->cdb_buf, self->cdb_buf_size,
								&direction,
								&data, &data_size,
								&timeout);

	burgner_scsi_device_send_cmd (self->device,
				      self->cdb_buf, self->cdb_buf_size,
				      self->sense_buf, self->sense_buf_size,
				      data, data_size,
				      direction,
				      timeout,
				      (BurgnerScsiDeviceSendCmdCallback) _burgner_scsi_send_command_stage_dev_send_cmd_cb,
				      self);
}

static gboolean
_burgner_scsi_send_command_stage_cancellable (BurgnerScsiStage *stage)
{
	return FALSE;
}

static gboolean
_burgner_scsi_send_command_stage_default_command_result
				(BurgnerScsiSendCommandStage *self,
				 guint status,
				 BurgnerScsiSense *sense,
				 guint data_used,
				 GError *command_error,
				 GError **error)
{
	if (command_error) {
		if (error)
			*error = g_error_copy (command_error);
		return FALSE;
	}

	if (status != BURGNER_SCSI_STATUS_CODE_GOOD) {
		burgner_scsi_stage_set_error_from_status_code (error, status);
		return FALSE;
	}

	return TRUE;
}

static void
burgner_scsi_send_command_stage_class_init (BurgnerScsiSendCommandStageClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiStageClass *stage_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_scsi_send_command_stage_set_property;
	object_class->get_property = burgner_scsi_send_command_stage_get_property;
	object_class->finalize = burgner_scsi_send_command_stage_finalize;

	stage_class = (BurgnerScsiStageClass *) klass;
	stage_class->real_start = _burgner_scsi_send_command_stage_real_start;
	stage_class->cancellable = _burgner_scsi_send_command_stage_cancellable;

	/* Virtual functions */
	klass->command_result = _burgner_scsi_send_command_stage_default_command_result;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_DEVICE,
					 g_param_spec_object ("device",
					                _("device"),
							_("A BurgnerScsiDevice"),
							BURGNER_TYPE_SCSI_DEVICE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CDB_BUF,
					 g_param_spec_pointer ("cdb_buf",
					                _("cdb_buf"),
							_("CDB buffer"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CDB_BUF_SIZE,
					 g_param_spec_uint ("cdb_buf_size",
					                _("cdb_buf_size"),
							_("CDB buffer size"),
							6,
							16,
							6,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SENSE_BUF,
					 g_param_spec_pointer ("sense_buf",
					                _("sense_buf"),
							_("Sense Buffer"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SENSE_BUF_SIZE,
					 g_param_spec_uint ("sense_buf_size",
					                _("sense_buf_size"),
							_("Sense Buffer Size"),
							0,
							G_MAXUINT,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_scsi_send_command_stage_init (BurgnerScsiSendCommandStage *self)
{
	self->device = NULL;
	self->cdb_buf = self->sense_buf = NULL;
	self->cdb_buf_size = 0;
	self->sense_buf_size = 0;
}

GType
burgner_scsi_send_command_stage_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiSendCommandStageClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_scsi_send_command_stage_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiSendCommandStage),
			0,
			(GInstanceInitFunc) burgner_scsi_send_command_stage_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_STAGE, "BurgnerScsiSendCommandStage", &info, G_TYPE_FLAG_ABSTRACT);
	}

	return type;
}

GQuark
burgner_scsi_send_command_stage_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiSendCommandStage");
}

void
burgner_scsi_send_command_stage_get_command_parameters (BurgnerScsiSendCommandStage *self,
						       guchar *cdb_buf, guint cdb_buf_size,
						       BurgnerScsiDirection *direction,
						       guchar **data, guint *data_size,
						       guint *timeout)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_SEND_COMMAND_STAGE (self));

	if (BURGNER_SCSI_SEND_COMMAND_STAGE_GET_CLASS (self)->get_command_parameters)
		BURGNER_SCSI_SEND_COMMAND_STAGE_GET_CLASS (self)->get_command_parameters
					(self, cdb_buf, cdb_buf_size, direction, data, data_size, timeout);
}

gboolean
burgner_scsi_send_command_stage_command_result (BurgnerScsiSendCommandStage *self,
					       guint status,
					       BurgnerScsiSense *sense,
					       guint data_used,
					       GError *command_error,
					       GError **error)
{
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_SEND_COMMAND_STAGE (self), FALSE);

	if (BURGNER_SCSI_SEND_COMMAND_STAGE_GET_CLASS (self)->command_result)
		return BURGNER_SCSI_SEND_COMMAND_STAGE_GET_CLASS (self)->command_result
					(self, status, sense, data_used, command_error, error);

	return TRUE;
}

