/*
 * 
 *
 * This file is part of the libburgnerscsi library.
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <string.h>
#include "burgner-scsi-utils.h"
#include "burgner-scsi-spc-impl.h"
#include "burgner-scsi-sense.h"

BurgnerScsiSense *
burgner_scsi_sense_new (void)
{
	BurgnerScsiSense *sense = g_new0 (BurgnerScsiSense, 1);
	sense->ref_count = 1;
	return sense;
}

void
burgner_scsi_sense_free (BurgnerScsiSense *sense)
{
	g_return_if_fail (sense != NULL);

	g_free (sense);
}

BurgnerScsiSense *
burgner_scsi_sense_ref (BurgnerScsiSense *sense)
{
	g_return_val_if_fail (sense != NULL, NULL);
	sense->ref_count += 1;
	return sense;
}

void
burgner_scsi_sense_unref (BurgnerScsiSense *sense)
{
	g_return_if_fail (sense != NULL);

	if ((sense->ref_count--) == 0) {
		burgner_scsi_sense_free (sense);
	}
}


gboolean
burgner_scsi_sense_parse (BurgnerScsiSense *sense,
			  guchar *sense_buf,
			  guint sense_buf_len,
			  GError **error)
{
	_BurgnerScsiSpcImplSenseData *sense_data;

	g_return_val_if_fail (sense != NULL, FALSE);
	g_return_val_if_fail (sense_buf != NULL, FALSE);

	if (sense_buf_len < 18) {
		g_set_error (error,
			     BURGNER_SCSI_SENSE_ERROR,
			     BURGNER_SCSI_SENSE_ERROR_BUFFER_TOO_SMALL,
			     _("Sense buffer is too small (should be at least 18 bytes long)"));
		return FALSE;
	}

	sense_data = (_BurgnerScsiSpcImplSenseData *) sense_buf;

	switch (sense_data->response_code) {
		case 0x70:
			sense->response_code = BURGNER_SCSI_SENSE_RESPONSE_CODE_CURRENT_ERROR;
			break;
		case 0x71:
			sense->response_code = BURGNER_SCSI_SENSE_RESPONSE_CODE_DEFERRED_ERROR;
			break;
		case 0x7f:
			sense->response_code = BURGNER_SCSI_SENSE_RESPONSE_CODE_VENDOR_SPECIFIC;
			break;
		default:
			g_set_error (error,
				     BURGNER_SCSI_SENSE_ERROR,
				     BURGNER_SCSI_SENSE_ERROR_INVALID_RESPONSE_CODE,
				     _("Invalid response code"));
			return FALSE;
	}

	sense->sense_key = sense_data->sense_key;

	
	sense->information = sense_data->information_valid ? burgner_scsi_get_uint32 (sense_data->information) : 0;

	sense->asc = sense->ascq = 0;
	sense->sksv = 0;
	memset (&sense->sks, 0, sizeof (sense->sks));

	if (sense_data->additional_sense_length >= 5) {
		sense->asc = sense_data->asc;

		if (sense_data->additional_sense_length >= 6) {
			sense->ascq = sense_data->ascq;

			if (sense_data->additional_sense_length >= 10 &&
			    sense_data->sks.unknown.sksv) {
				sense->sksv = 1;

				switch (sense->sense_key) {
					case BURGNER_SCSI_SENSE_KEY_ILLEGAL_REQUEST:
						sense->sks.illegal_request.bit_pointer_valid = sense_data->sks.field_pointer.bit_pointer_valid;
						sense->sks.illegal_request.bit_pointer = sense_data->sks.field_pointer.bit_pointer;
						sense->sks.illegal_request.in_cdb = sense_data->sks.field_pointer.c_d;
						sense->sks.illegal_request.field_pointer = burgner_scsi_get_uint16 (sense_data->sks.field_pointer.field_pointer);
						break;
					case BURGNER_SCSI_SENSE_KEY_RECOVERED_ERROR:
					case BURGNER_SCSI_SENSE_KEY_MEDIUM_ERROR:
					case BURGNER_SCSI_SENSE_KEY_HARDWARE_ERROR:
						sense->sks.actual_retry_count = burgner_scsi_get_uint16 (sense_data->sks.actual_retry_count.actual_retry_count);
						break;
					case BURGNER_SCSI_SENSE_KEY_NO_SENSE:
					case BURGNER_SCSI_SENSE_KEY_NOT_READY:
						sense->sks.progress_indication = burgner_scsi_get_uint16 (sense_data->sks.progress_indication.progress_indication);
						break;
					default:
						sense->sksv = 0;
						break;
				}
			}
		}
	}

	return TRUE;
}

gboolean
burgner_scsi_sense_new_from_buf (BurgnerScsiSense **sense,
				 guchar *sense_buf,
				 guint sense_buf_len,
				 GError **error)
{
	BurgnerScsiSense *s;

	g_return_val_if_fail (sense_buf != NULL, FALSE);
	g_return_val_if_fail (sense_buf_len != 0, FALSE);

	s = burgner_scsi_sense_new ();
	g_return_val_if_fail (s != NULL, FALSE);

	if (!burgner_scsi_sense_parse (s, sense_buf, sense_buf_len, error)) {
		burgner_scsi_sense_unref (s);
		return FALSE;
	}

	if (sense)
		*sense = s;
	else
		burgner_scsi_sense_unref (s);

	return TRUE;
}

GQuark
burgner_scsi_sense_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiSense");
}

static const gchar *
_burgner_scsi_sense_response_code_to_string (BurgnerScsiSenseResponseCode rc)
{
	switch (rc) {
		case BURGNER_SCSI_SENSE_RESPONSE_CODE_CURRENT_ERROR:
			return _("Current error");
		case BURGNER_SCSI_SENSE_RESPONSE_CODE_DEFERRED_ERROR:
			return _("Deferred error");
		case BURGNER_SCSI_SENSE_RESPONSE_CODE_VENDOR_SPECIFIC:
			return _("Vendor specific");
		default:
			break;
	}

	g_return_val_if_reached ("");
	return "";
}

static const gchar *
_burgner_scsi_sense_key_to_string (BurgnerScsiSenseKey key)
{
	switch (key) {
		case BURGNER_SCSI_SENSE_KEY_NO_SENSE:
			return _("No sense");
		case BURGNER_SCSI_SENSE_KEY_RECOVERED_ERROR:
			return _("Recovered error");
		case BURGNER_SCSI_SENSE_KEY_NOT_READY:
			return _("Not ready");
		case BURGNER_SCSI_SENSE_KEY_MEDIUM_ERROR:
			return _("Medium error");
		case BURGNER_SCSI_SENSE_KEY_HARDWARE_ERROR:
			return _("Hardware error");
		case BURGNER_SCSI_SENSE_KEY_ILLEGAL_REQUEST:
			return _("Illegal request");
		case BURGNER_SCSI_SENSE_KEY_UNIT_ATTENTION:
			return _("Unit attention");
		case BURGNER_SCSI_SENSE_KEY_DATA_PROTECT:
			return _("Data protect");
		case BURGNER_SCSI_SENSE_KEY_BLANK_CHECK:
			return _("Blank check");
		case BURGNER_SCSI_SENSE_KEY_VENDOR_SPECIFIC:
			return _("Vendor specific");
		case BURGNER_SCSI_SENSE_KEY_COPY_ABORTED:
			return _("Copy aborted");
		case BURGNER_SCSI_SENSE_KEY_ABORTED_COMMAND:
			return _("Aborted command");
		case BURGNER_SCSI_SENSE_KEY_OBSOLETE:
			return _("Obsolete");
		case BURGNER_SCSI_SENSE_KEY_VOLUME_OVERFLOW:
			return _("Volume overflow");
		case BURGNER_SCSI_SENSE_KEY_MISCOMPARE:
			return _("Miscompare");
		case BURGNER_SCSI_SENSE_KEY_RESERVED:
			return _("Reserved");
		default:
			break;
	}

	g_return_val_if_reached ("");
	return "";
}

gchar *
burgner_scsi_sense_get_detailed_description_string (BurgnerScsiSense *sense)
{
	g_return_val_if_fail (sense != NULL, NULL);

	return g_strdup_printf (_("Sense key: %s, Response code: %s, asc: 0x%x, ascq: 0x%x"),
					_burgner_scsi_sense_key_to_string (sense->sense_key),
					_burgner_scsi_sense_response_code_to_string (sense->response_code),
					sense->asc,
					sense->ascq);
}

