/*
 * BurgnerScsiStageManagerCallback - BurgnerScsiStageManagerCallback object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <glib-object.h>
#include "burgner-scsi-stage-manager-callback.h"


static BurgnerScsiStageManagerClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_GET_NEXT_STAGE_FUNC,
	PROP_CONTINUE_FUNC,
	PROP_USER_DATA,
	BURGNER_SCSI_STAGE_MANAGER_CALLBACK_LAST_ARG
};

static void
_burgner_scsi_stage_manager_callback_finalize (GObject *object)
{
	BurgnerScsiStageManagerCallback *self = BURGNER_SCSI_STAGE_MANAGER_CALLBACK (object);

	self->get_next_stage_func = NULL;
	self->continue_func = NULL;
	self->user_data = NULL;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_scsi_stage_manager_callback_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiStageManagerCallback *self = BURGNER_SCSI_STAGE_MANAGER_CALLBACK (object);

	switch (prop_id) {
		case PROP_GET_NEXT_STAGE_FUNC:
			self->get_next_stage_func = g_value_get_pointer (value);
			g_object_notify (object, "get_next_stage_func");
			break;
		case PROP_CONTINUE_FUNC:
			self->continue_func = g_value_get_pointer (value);
			g_object_notify (object, "continue_func");
			break;
		case PROP_USER_DATA:
			self->user_data = g_value_get_pointer (value);
			g_object_notify (object, "user_data");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_scsi_stage_manager_callback_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiStageManagerCallback *self = BURGNER_SCSI_STAGE_MANAGER_CALLBACK (object);

	switch (prop_id) {
		case PROP_GET_NEXT_STAGE_FUNC:
			g_value_set_pointer (value, self->get_next_stage_func);
			break;
		case PROP_CONTINUE_FUNC:
			g_value_set_pointer (value, self->continue_func);
			break;
		case PROP_USER_DATA:
			g_value_set_pointer (value, self->user_data);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static BurgnerScsiStage *
_burgner_scsi_stage_manager_callback_get_next_stage
				(BurgnerScsiStageManager *sm,
				 gint *stage_id,
				 BurgnerScsiStage *actual_stage,
				 gint actual_stage_id)
{
	BurgnerScsiStageManagerCallback *self = BURGNER_SCSI_STAGE_MANAGER_CALLBACK (sm);

	if (self->get_next_stage_func) {
		return self->get_next_stage_func (self,
						  stage_id,
						  actual_stage,
						  actual_stage_id,
						  self->user_data);
	}

	return NULL;
}

static gboolean
_burgner_scsi_stage_manager_callback_continue
				(BurgnerScsiStageManager *sm,
				 BurgnerScsiStage *stage,
				 gint stage_id,
				 GError *stage_error,
				 BurgnerScsiSense *sense,
				 GError **error)
{
	BurgnerScsiStageManagerCallback *self = BURGNER_SCSI_STAGE_MANAGER_CALLBACK (sm);

	if (self->continue_func) {
		return self->continue_func (self,
					    stage, stage_id,
					    stage_error, sense, error,
					    self->user_data);
	}

	return TRUE;
}


static void
_burgner_scsi_stage_manager_callback_class_init (BurgnerScsiStageManagerCallbackClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiStageManagerClass *sm_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_scsi_stage_manager_callback_set_property;
	object_class->get_property = _burgner_scsi_stage_manager_callback_get_property;
	object_class->finalize = _burgner_scsi_stage_manager_callback_finalize;

	sm_class = (BurgnerScsiStageManagerClass *) klass;
	sm_class->get_next_stage = _burgner_scsi_stage_manager_callback_get_next_stage;
	sm_class->_continue = _burgner_scsi_stage_manager_callback_continue;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_GET_NEXT_STAGE_FUNC,
					 g_param_spec_pointer ("get_next_stage_func",
					                _("get_next_stage_func"),
							_("A BurgnerScsiStageManagerCallback function, returning the next stage"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_CONTINUE_FUNC,
					 g_param_spec_pointer ("continue_func",
					                _("continue_func"),
							_("A BurgnerScsiStageManagerCallback function, called after each stage, and which must detect errors"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_USER_DATA,
					 g_param_spec_pointer ("user_data",
					                _("user_data"),
							_("User data passed to get_next_stage_func and continue_func"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_scsi_stage_manager_callback_init (BurgnerScsiStageManagerCallback *self)
{
	self->get_next_stage_func = NULL;
	self->continue_func = NULL;
	self->user_data = NULL;
}

GType
burgner_scsi_stage_manager_callback_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiStageManagerCallbackClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_scsi_stage_manager_callback_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiStageManagerCallback),
			0,
			(GInstanceInitFunc) _burgner_scsi_stage_manager_callback_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_STAGE_MANAGER, "BurgnerScsiStageManagerCallback", &info, 0);
	}

	return type;
}

BurgnerScsiStage *
burgner_scsi_stage_manager_callback_new (void)
{
	BurgnerScsiStageManagerCallback *self;

	self = g_object_new (BURGNER_TYPE_SCSI_STAGE_MANAGER_CALLBACK, NULL);


	return BURGNER_SCSI_STAGE (self);
}


GQuark
burgner_scsi_stage_manager_callback_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiStageManagerCallback");
}

