/*
 * BurgnerScsiStageManager - BurgnerScsiStageManager object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-scsi-marshal.h"
#include "burgner-scsi-stage-manager.h"


#define check(o)		G_STMT_START {	\
	g_return_if_fail (o);	\
	g_return_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (o));	\
				} G_STMT_END
#define check_with_val(o,v)	G_STMT_START {	\
	g_return_val_if_fail ((o), (v));	\
	g_return_val_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (o), (v));	\
				} G_STMT_END

static BurgnerScsiStageClass *parent_class = NULL;

enum {
	START_STAGE,
	END_STAGE,
	BURGNER_SCSI_STAGE_MANAGER_LAST_SIGNAL
};

static guint burgner_scsi_stage_manager_signals[BURGNER_SCSI_STAGE_MANAGER_LAST_SIGNAL] = { 0, };

static void
burgner_scsi_stage_manager_finalize (GObject *object)
{
	BurgnerScsiStageManager *self = BURGNER_SCSI_STAGE_MANAGER (object);

	g_assert (!self->end_signal_id || self->stage);
	g_assert (self->stage_id == -1 || self->stage);

	if (self->stage) {
		if (self->end_signal_id) {
			g_signal_handler_disconnect (self->stage, self->end_signal_id);
			self->end_signal_id = 0;
		}

		g_object_unref (self->stage);
		self->stage = NULL;
		self->stage_id = -1;
	}

	if (self->cancel)
		self->cancel = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void _burgner_scsi_stage_manager_real_start (BurgnerScsiStage *p_stage);

static void _burgner_scsi_stage_manager_cancel (BurgnerScsiStage *stage);

static gboolean
_burgner_scsi_stage_manager_cancellable (BurgnerScsiStage *stage)
{
	return TRUE;
}

static void
burgner_scsi_stage_manager_class_init (BurgnerScsiStageManagerClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiStageClass *stage_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = burgner_scsi_stage_manager_finalize;

	stage_class = (BurgnerScsiStageClass *) klass;
	stage_class->real_start = _burgner_scsi_stage_manager_real_start;
	stage_class->cancel = _burgner_scsi_stage_manager_cancel;
	stage_class->cancellable = _burgner_scsi_stage_manager_cancellable;

	/* Signals */
	burgner_scsi_stage_manager_signals[START_STAGE] = g_signal_new ("start_stage",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerScsiStageManagerClass, start_stage),
						NULL, NULL,
						burgner_scsi_marshal_VOID__OBJECT_INT,
						G_TYPE_NONE, 2,
						BURGNER_TYPE_SCSI_STAGE,
						G_TYPE_INT);
	burgner_scsi_stage_manager_signals[END_STAGE] = g_signal_new ("end_stage",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerScsiStageManagerClass, end_stage),
						NULL, NULL,
						burgner_scsi_marshal_VOID__OBJECT_INT_POINTER_POINTER,
						G_TYPE_NONE, 4,
						BURGNER_TYPE_SCSI_STAGE,
						G_TYPE_INT,
						G_TYPE_POINTER,
						G_TYPE_POINTER);

	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_scsi_stage_manager_init (BurgnerScsiStageManager *self)
{
	self->end_signal_id = 0;
	self->stage = NULL;
	self->stage_id = -1;
	self->cancel = 0;
}

GType
burgner_scsi_stage_manager_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiStageManagerClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_scsi_stage_manager_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiStageManager),
			0,
			(GInstanceInitFunc) burgner_scsi_stage_manager_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_STAGE, "BurgnerScsiStageManager", &info, G_TYPE_FLAG_ABSTRACT);
	}

	return type;
}

GQuark
burgner_scsi_stage_manager_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiStageManager");
}

static void
_burgner_scsi_stage_manager_stage_end_signal_cb (BurgnerScsiStage *stage,
						gboolean cancelled,
						GError *stage_error,
						BurgnerScsiSense *sense,
						BurgnerScsiStageManager *self)
{
	GError *error = NULL;
	BurgnerScsiStage *next_stage = NULL;
	gint stage_id = -1, s_id = self->stage_id;
	gboolean r;

	g_assert (self->stage == stage);
	g_assert (!cancelled || self->cancel);

	if (self->cancel) {
		self->cancel = 0;

		if (!cancelled)
			burgner_scsi_stage_manager_end_stage (self, self->stage, self->stage_id, stage_error, sense);

		g_signal_handler_disconnect (stage, self->end_signal_id);
		self->end_signal_id = 0;

		g_object_unref (self->stage);
		self->stage = NULL;
		self->stage_id = -1;
		
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), TRUE, NULL, NULL);

		return;
	}

	r = burgner_scsi_stage_manager_continue (self, stage, s_id, stage_error, sense, &error);

	burgner_scsi_stage_manager_end_stage (self, self->stage, self->stage_id, stage_error, sense);

	if (r) {
		next_stage = burgner_scsi_stage_manager_get_next_stage
				(self, &stage_id, self->stage, self->stage_id);
	}

	g_signal_handler_disconnect (stage, self->end_signal_id);
	self->end_signal_id = 0;

	g_object_unref (self->stage);
	self->stage = NULL;
	self->stage_id = -1;

	if (!r) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), FALSE, error, sense);
		return;
	} else if (!next_stage) {
		burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), FALSE, NULL, NULL);
		return;
	}

	self->stage = next_stage;
	self->stage_id = stage_id;

	self->end_signal_id = g_signal_connect_data (self->stage, "end",
						       (GCallback) _burgner_scsi_stage_manager_stage_end_signal_cb,
						       self, NULL, 0);

	burgner_scsi_stage_manager_start_stage (self, self->stage, self->stage_id);
	burgner_scsi_stage_start (self->stage, FALSE);
}

static void
_burgner_scsi_stage_manager_real_start (BurgnerScsiStage *p_stage)
{
	BurgnerScsiStageManager *self = BURGNER_SCSI_STAGE_MANAGER (p_stage);
	BurgnerScsiStage *stage;
	gint stage_id = -1;

	g_return_if_fail (self->stage == NULL);
	g_return_if_fail (self->stage_id == -1);

	if (self->cancel == 1) {
		self->cancel = 0;
		burgner_scsi_stage_end (p_stage, TRUE, NULL, NULL);
	}

	stage = burgner_scsi_stage_manager_get_next_stage (self, &stage_id, NULL, -1);
	if (!stage) {
		burgner_scsi_stage_end (p_stage, FALSE, NULL, NULL);
	}

	self->stage = stage;
	self->stage_id = stage_id;

	self->end_signal_id = g_signal_connect_data (stage, "end",
						       (GCallback) _burgner_scsi_stage_manager_stage_end_signal_cb,
						       self, NULL, 0);

	burgner_scsi_stage_manager_start_stage (self, stage, stage_id);
	burgner_scsi_stage_start (stage, FALSE);
}

static void
_burgner_scsi_stage_manager_cancel (BurgnerScsiStage *stage)
{
	BurgnerScsiStageManager *self = BURGNER_SCSI_STAGE_MANAGER (stage);

	if (!self->cancel) {
		self->cancel = 1;
		if (self->stage && burgner_scsi_stage_cancellable (self->stage)) {
			burgner_scsi_stage_cancel (self->stage);
		}
	}
}

void
burgner_scsi_stage_manager_start_stage (BurgnerScsiStageManager *self,
				       BurgnerScsiStage *stage,
				       gint stage_id)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (self));
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	g_signal_emit (self, burgner_scsi_stage_manager_signals[START_STAGE], 0,
		       stage, stage_id);
}

void
burgner_scsi_stage_manager_end_stage (BurgnerScsiStageManager *self,
				     BurgnerScsiStage *stage,
				     gint stage_id,
				     GError *error,
				     BurgnerScsiSense *sense)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (self));
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	g_signal_emit (self, burgner_scsi_stage_manager_signals[END_STAGE], 0,
		       stage, stage_id, error, sense);
}

BurgnerScsiStage *
burgner_scsi_stage_manager_get_next_stage (BurgnerScsiStageManager *self,
					  gint *stage_id,
					  BurgnerScsiStage *actual_stage,
					  gint actual_stage_id)
{
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (self), NULL);
	g_return_val_if_fail (!actual_stage || BURGNER_IS_SCSI_STAGE (actual_stage), NULL);

	if (BURGNER_SCSI_STAGE_MANAGER_GET_CLASS (self)->get_next_stage)
		return BURGNER_SCSI_STAGE_MANAGER_GET_CLASS (self)->get_next_stage
					(self, stage_id, actual_stage, actual_stage_id);

	return NULL;
}

gboolean
burgner_scsi_stage_manager_continue (BurgnerScsiStageManager *self,
				    BurgnerScsiStage *stage,
				    gint stage_id,
				    GError *stage_error,
				    BurgnerScsiSense *sense,
				    GError **error)
{
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_STAGE_MANAGER (self), FALSE);
	g_return_val_if_fail (stage != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_STAGE (stage), FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

	if (BURGNER_SCSI_STAGE_MANAGER_GET_CLASS (self)->_continue)
		return BURGNER_SCSI_STAGE_MANAGER_GET_CLASS (self)->_continue
					(self, stage, stage_id, stage_error, sense, error);

	return FALSE;
}

#undef check
#undef check_with_val

