/*
 * BurgnerScsiStage - BurgnerScsiStage object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib-object.h>
#include "burgner-scsi-marshal.h"
#include "burgner-scsi-stage.h"


static GObjectClass *parent_class = NULL;

enum {
	BURGNER_SCSI_STAGE_END,
	BURGNER_SCSI_STAGE_LAST_SIGNAL
};

static guint burgner_scsi_stage_signals[BURGNER_SCSI_STAGE_LAST_SIGNAL] = { 0, };

static void
burgner_scsi_stage_finalize (GObject *object)
{
	BurgnerScsiStage *self = BURGNER_SCSI_STAGE (object);

	if (self->idle_id) {
		g_source_remove (self->idle_id);
		self->idle_id = 0;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
_bms_idle_func (BurgnerScsiStage *stage)
{
	stage->idle_id = 0;

	g_assert (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start);

	BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start (stage);

	return FALSE;
}

static void
_burgner_scsi_stage_start_default_impl (BurgnerScsiStage *stage,
				       gboolean from_idle)
{
	g_return_if_fail (stage->idle_id == 0);

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start) {
		if (from_idle) {
			stage->idle_id = g_idle_add ((GSourceFunc) _bms_idle_func,
						     stage);
		} else {
			BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start (stage);
		}
	}
}

static void
_burgner_scsi_stage_cancel_default_impl (BurgnerScsiStage *stage)
{
	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start) {
		if (stage->idle_id) {
			g_source_remove (stage->idle_id);
			stage->idle_id = 0;
		} else {
			if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_cancel)
				BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_cancel
									(stage);
		}
	}
}

static void
burgner_scsi_stage_class_init (BurgnerScsiStageClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = burgner_scsi_stage_finalize;

	/* Virtual Functions */
	klass->start = _burgner_scsi_stage_start_default_impl;
	klass->cancel = _burgner_scsi_stage_cancel_default_impl;

	/* Signals */
	burgner_scsi_stage_signals[BURGNER_SCSI_STAGE_END] =
		g_signal_new ("end",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (BurgnerScsiStageClass, end),
			      NULL, NULL,
			      burgner_scsi_marshal_VOID__BOOLEAN_POINTER_POINTER,
			      G_TYPE_NONE, 3,
			      G_TYPE_BOOLEAN,
			      G_TYPE_POINTER,
			      G_TYPE_POINTER);

	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_scsi_stage_init (BurgnerScsiStage *self)
{
	self->idle_id = 0;
}

GType
burgner_scsi_stage_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiStageClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_scsi_stage_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiStage),
			0,
			(GInstanceInitFunc) burgner_scsi_stage_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerScsiStage", &info, G_TYPE_FLAG_ABSTRACT);
	}

	return type;
}

GQuark
burgner_scsi_stage_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiStage");
}

void
burgner_scsi_stage_start (BurgnerScsiStage *stage,
			 gboolean from_idle)
{
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->start)
		BURGNER_SCSI_STAGE_GET_CLASS (stage)->start (stage, from_idle);
}


void
burgner_scsi_stage_real_start (BurgnerScsiStage *stage)
{
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start)
		BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_start (stage);
}

void
burgner_scsi_stage_cancel (BurgnerScsiStage *stage)
{
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->cancel)
		BURGNER_SCSI_STAGE_GET_CLASS (stage)->cancel (stage);
}

void
burgner_scsi_stage_real_cancel (BurgnerScsiStage *stage)
{
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_cancel)
		BURGNER_SCSI_STAGE_GET_CLASS (stage)->real_cancel (stage);
}

void
burgner_scsi_stage_end (BurgnerScsiStage *stage,
		       gboolean cancelled,
		       GError *error,
		       BurgnerScsiSense *sense)
{
	g_return_if_fail (stage != NULL);
	g_return_if_fail (BURGNER_IS_SCSI_STAGE (stage));

	g_object_ref (stage);
	g_signal_emit (stage,
		       burgner_scsi_stage_signals[BURGNER_SCSI_STAGE_END],
		       0, cancelled, error, sense);
	g_object_unref (stage);
}

gboolean
burgner_scsi_stage_cancellable (BurgnerScsiStage *stage)
{
	g_return_val_if_fail (stage != NULL, FALSE);
	g_return_val_if_fail (BURGNER_IS_SCSI_STAGE (stage), FALSE);

	if (BURGNER_SCSI_STAGE_GET_CLASS (stage)->cancellable)
		BURGNER_SCSI_STAGE_GET_CLASS (stage)->cancellable (stage);

	return FALSE;
}

