/*
 * BurgnerScsiWaitStage - BurgnerScsiWaitStage object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-scsi-i18n.h"

#include <glib-object.h>
#include "burgner-scsi-wait-stage.h"


static BurgnerScsiStageClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_TIME,
	PROP_PRIORITY,
	BURGNER_SCSI_WAIT_STAGE_LAST_ARG
};

static void
_burgner_scsi_wait_stage_finalize (GObject *object)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (object);

	if (self->timeout_id) {
		g_source_remove (self->timeout_id);
		self->timeout_id = 0;
	}

	self->time = 1000;
	self->priority = G_PRIORITY_DEFAULT_IDLE;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_scsi_wait_stage_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (object);

	switch (prop_id) {
		case PROP_TIME:
			self->time = g_value_get_uint (value);
			g_object_notify (object, "time");
			break;
		case PROP_PRIORITY:
			self->priority = g_value_get_int (value);
			g_object_notify (object, "priority");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_scsi_wait_stage_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (object);

	switch (prop_id) {
		case PROP_TIME:
			g_value_set_uint (value, self->time);
			break;
		case PROP_PRIORITY:
			g_value_set_int (value, self->priority);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static gboolean
_burgner_scsi_wait_stage_timeout_func (gpointer user_data)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (user_data);

	g_return_val_if_fail (self->timeout_id != 0, FALSE);
	self->timeout_id = 0;

	burgner_scsi_stage_end (BURGNER_SCSI_STAGE (self), FALSE, NULL, NULL);
	return FALSE;
}

static void
_burgner_scsi_wait_stage_real_start (BurgnerScsiStage *stage)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (stage);

	g_return_if_fail (self->timeout_id == 0);

	self->timeout_id = g_timeout_add_full (self->priority,
					       self->time,
					       _burgner_scsi_wait_stage_timeout_func,
					       self, NULL);
}

static void
_burgner_scsi_wait_stage_real_cancel (BurgnerScsiStage *stage)
{
	BurgnerScsiWaitStage *self = BURGNER_SCSI_WAIT_STAGE (stage);

	g_return_if_fail (self->timeout_id != 0);

	g_source_remove (self->timeout_id);
	self->timeout_id = 0;
}

static gboolean
_burgner_scsi_wait_stage_cancellable (BurgnerScsiStage *stage)
{
	return TRUE;
}

static void
_burgner_scsi_wait_stage_class_init (BurgnerScsiWaitStageClass *klass)
{
	GObjectClass *object_class;
	BurgnerScsiStageClass *stage_class = (BurgnerScsiStageClass *) klass;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_scsi_wait_stage_set_property;
	object_class->get_property = _burgner_scsi_wait_stage_get_property;
	object_class->finalize = _burgner_scsi_wait_stage_finalize;

	stage_class = (BurgnerScsiStageClass *) klass;
	stage_class->real_start = _burgner_scsi_wait_stage_real_start;
	stage_class->real_cancel = _burgner_scsi_wait_stage_real_cancel;
	stage_class->cancellable = _burgner_scsi_wait_stage_cancellable;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_TIME,
					 g_param_spec_uint ("time",
					                _("time"),
							_("Time to wait"),
							0,
							G_MAXUINT,
							1000,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_PRIORITY,
					 g_param_spec_int ("priority",
					                _("priority"),
							_("The priority of the idle source of the timeout"),
							0,
							G_MAXINT,
							G_PRIORITY_DEFAULT_IDLE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_scsi_wait_stage_init (BurgnerScsiWaitStage *self)
{
	self->time = 1000;
	self->priority = G_PRIORITY_DEFAULT_IDLE;
	self->timeout_id = 0;
}

GType
burgner_scsi_wait_stage_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerScsiWaitStageClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_scsi_wait_stage_class_init,
			NULL,
			NULL,
			sizeof (BurgnerScsiWaitStage),
			0,
			(GInstanceInitFunc) _burgner_scsi_wait_stage_init,
		};

		type = g_type_register_static (BURGNER_TYPE_SCSI_STAGE, "BurgnerScsiWaitStage", &info, 0);
	}

	return type;
}

BurgnerScsiStage *
burgner_scsi_wait_stage_new (void)
{
	BurgnerScsiWaitStage *self;

	self = g_object_new (BURGNER_TYPE_SCSI_WAIT_STAGE, NULL);


	return BURGNER_SCSI_STAGE (self);
}


GQuark
burgner_scsi_wait_stage_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiWaitStage");
}

