/*
 * BurgnerScsiLinuxCdromDevice - BurgnerScsiLinuxCdromDevice object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../../burgner-scsi-i18n.h"

#include <string.h>
#include <errno.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>

#include <glib-object.h>
#include "../../burgner-scsi-status-code.h"
#include "burgner-scsi-linux-cdrom-device.h"


static BurgnerScsiLinuxDeviceClass *parent_class = NULL;


static guchar
_burgner_scsi_linux_cdrom_device_scsi_direction_to_data_direction (BurgnerScsiDirection dir)
{
	guchar r = CGC_DATA_UNKNOWN;

	switch (dir) {
		case BURGNER_SCSI_DIRECTION_NONE:
			r = CGC_DATA_NONE;
			break;
		case BURGNER_SCSI_DIRECTION_FROM_DEVICE:
			r = CGC_DATA_READ;
			break;
		case BURGNER_SCSI_DIRECTION_TO_DEVICE:
			r = CGC_DATA_WRITE;
			break;
		default:
			g_return_val_if_reached (CGC_DATA_UNKNOWN);
			break;
	}

	return r;
}

static gboolean
_burgner_scsi_linux_cdrom_device_send_cmd_real (BurgnerScsiLinuxDevice *device,
						gint fd,
						gpointer cdb, guint cdb_size,
						gpointer sense, guint sense_size, guint *sense_used,
						gpointer data, guint data_size, guint *data_used,
						BurgnerScsiDirection direction,
						guint timeout,
						gint *status,
						GError **error)
{
	struct cdrom_generic_command cmd;
	struct request_sense sense_struct;

	g_return_val_if_fail (cdb_size <= CDROM_PACKET_SIZE, FALSE);

	memset (&cmd, 0, sizeof (cmd));
	memset (&sense_struct, 0, sizeof (sense_struct));

	memcpy (cmd.cmd, cdb, MIN (CDROM_PACKET_SIZE, cdb_size));
	cmd.buffer = data;
	cmd.buflen = data_size;
	cmd.sense = &sense_struct;
	cmd.data_direction = _burgner_scsi_linux_cdrom_device_scsi_direction_to_data_direction (direction);
	cmd.quiet = 1;
	cmd.timeout = G_MAXINT;

	if (ioctl (fd, CDROM_SEND_PACKET, &cmd) != 0 &&
	    ((guchar *) &sense_struct)[0] == 0 &&
	    ((guchar *) &sense_struct)[2] == 0) {
		g_set_error (error,
			     BURGNER_SCSI_DEVICE_ERROR,
			     BURGNER_SCSI_DEVICE_ERROR_SENDING_CMD,
			     _("Error while sending command: %s"),
			     		strerror (errno));
		return FALSE;
	}

	if (sense_used) {
		*sense_used = MIN (sense_struct.add_sense_len + 8, sense_size);
		memcpy (sense, &sense_struct, *sense_used);
	}

	/*
	 * Seen in cdrtools-2.0/libscg/scsi-linux-ata.c:
	 * cmd.buffer may change (?) ...
	 */
	if (direction == BURGNER_SCSI_DIRECTION_FROM_DEVICE &&
	    cmd.buffer != data) {
		memcpy (data, cmd.buffer, MIN (data_size, cmd.buflen));
	}

	/*
	 * ... and cmd.buflen may be higher than the value we set before the ioctl :|
	 */
	if (data_used) {
		*data_used = cmd.buflen <= data_size ? cmd.buflen : data_size;
	}

	if (status) {
		/* Don't know how to get the status... :( */
		*status = sense_struct.error_code ? BURGNER_SCSI_STATUS_CODE_CHECK_CONDITION : BURGNER_SCSI_STATUS_CODE_GOOD;
	}

	return TRUE;
}


static void
_burgner_scsi_linux_cdrom_device_class_init (BurgnerScsiLinuxCdromDeviceClass *klass)
{
	BurgnerScsiLinuxDeviceClass *device_class;

	device_class = (BurgnerScsiLinuxDeviceClass *) klass;
	device_class->send_cmd_real = _burgner_scsi_linux_cdrom_device_send_cmd_real;
	parent_class = g_type_class_peek_parent (klass);
}


GTypeInfo burgner_scsi_linux_cdrom_device_info = {
	sizeof (BurgnerScsiLinuxCdromDeviceClass),
	NULL,
	NULL,
	(GClassInitFunc) _burgner_scsi_linux_cdrom_device_class_init,
	NULL,
	NULL,
	sizeof (BurgnerScsiLinuxCdromDevice),
	0,
	NULL,
};

GType
burgner_scsi_linux_cdrom_device_get_type (void)
{
	return g_type_from_name ("BurgnerScsiLinuxCdromDevice");
}


BurgnerScsiDevice *
burgner_scsi_linux_cdrom_device_new (void)
{
	BurgnerScsiLinuxCdromDevice *self;

	self = g_object_new (BURGNER_TYPE_SCSI_LINUX_CDROM_DEVICE, NULL);


	return BURGNER_SCSI_DEVICE (self);
}


GQuark
burgner_scsi_linux_cdrom_device_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiLinuxCdromDevice");
}

