/*
 * BurgnerScsiLinuxCdromDriver - BurgnerScsiLinuxCdromDriver object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../../burgner-scsi-i18n.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <glib-object.h>
#include "burgner-scsi-linux-cdrom-device.h"
#include "burgner-scsi-linux-cdrom-driver.h"


#define DRIVER_ID	"linux-cdrom"

#define DMA_SIZE	65536


static BurgnerScsiLinuxDriverClass *parent_class = NULL;


static gint
_burgner_scsi_linux_cdrom_driver_open_device_real (BurgnerScsiLinuxDriver *driver,
						   const gchar *file,
						   gint requested_dma_size,
						   gint *real_dma_size,
						   GError **error)
{
	struct stat stat_buf;
	int fd;

	memset (&stat_buf, 0, sizeof (stat_buf));
	if (stat (file, &stat_buf) != 0) {
		g_set_error (error,
			     BURGNER_SCSI_DRIVER_ERROR,
			     BURGNER_SCSI_DRIVER_ERROR_BAD_FILE,
			     _("Error while getting informations about %s: %s"),
			     		file,
			     		strerror (errno));
		return -1;
	}

	if (!S_ISBLK (stat_buf.st_mode)) {
		g_set_error (error,
			     BURGNER_SCSI_DRIVER_ERROR,
			     BURGNER_SCSI_DRIVER_ERROR_BAD_FILE,
			     _("%s is not a block device"),
			     file);
		return -1;
	}

	if ((fd = open (file, O_RDONLY | O_NONBLOCK)) < 0) {
		g_set_error (error,
			     BURGNER_SCSI_DRIVER_ERROR,
			     BURGNER_SCSI_DRIVER_ERROR_OPENING_DEVICE,
			     _("Error while opening %s: %s"),
			     		file,
			     		strerror (errno));
		return -1;
	}

	if (real_dma_size)
		*real_dma_size = DMA_SIZE;

	return fd;
}

static GList *
_burgner_scsi_linux_cdrom_driver_get_uri_list_real (BurgnerScsiLinuxDriver *driver,
						    GError **error)
{
	GList *l = NULL;
	GString *s;
	GDir *dir;
	const gchar *ent_name;

	dir = g_dir_open ("/dev/cdroms", 0, error);
	if (!dir)
		return NULL;

	s = g_string_sized_new (50);

	while ((ent_name = g_dir_read_name (dir))) {
		struct stat stat_buf;
		GUri *uri;

		g_string_printf (s, "/dev/cdroms/%s", ent_name);

		if (stat (s->str, &stat_buf) != 0)
			continue;

		g_string_prepend (s, "file://" DRIVER_ID);
		uri = g_uri_new ();
		g_uri_set (uri, s->str, FALSE, FALSE, NULL);
		l = g_list_prepend (l, uri);
	}

	g_string_free (s, TRUE);
	return g_list_reverse (l);
}

static BurgnerScsiDevice *
_burgner_scsi_linux_cdrom_driver_create_device (BurgnerScsiLinuxDriver *driver)
{
	return burgner_scsi_linux_cdrom_device_new ();
}


static void
_burgner_scsi_linux_cdrom_driver_class_init (BurgnerScsiLinuxCdromDriverClass *klass)
{
	BurgnerScsiLinuxDriverClass *driver_class;

	driver_class = (BurgnerScsiLinuxDriverClass *) klass;
	driver_class->open_device_real = _burgner_scsi_linux_cdrom_driver_open_device_real;
	driver_class->get_uri_list_real = _burgner_scsi_linux_cdrom_driver_get_uri_list_real;
	driver_class->create_device = _burgner_scsi_linux_cdrom_driver_create_device;

	parent_class = g_type_class_peek_parent (klass);
}


GTypeInfo burgner_scsi_linux_cdrom_driver_info = {
	sizeof (BurgnerScsiLinuxCdromDriverClass),
	NULL,
	NULL,
	(GClassInitFunc) _burgner_scsi_linux_cdrom_driver_class_init,
	NULL,
	NULL,
	sizeof (BurgnerScsiLinuxCdromDriver),
	0,
	NULL,
};

GType
burgner_scsi_linux_cdrom_driver_get_type (void)
{
	return g_type_from_name ("BurgnerScsiLinuxCdromDriver");
}


BurgnerScsiDriver *
burgner_scsi_linux_cdrom_driver_new (void)
{
	BurgnerScsiLinuxCdromDriver *self;

	self = g_object_new (BURGNER_TYPE_SCSI_LINUX_CDROM_DRIVER, NULL);


	return BURGNER_SCSI_DRIVER (self);
}


GQuark
burgner_scsi_linux_cdrom_driver_error_quark (void)
{
	return g_quark_from_string ("BurgnerScsiLinuxCdromDriver");
}

