/*
 * BurgnerFifoBigEndianFilter - BurgnerFifoBigEndianFilter object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-utils-i18n.h"

#include <glib-object.h>
#include "burgner-fifo-big-endian-filter.h"


static BurgnerFifoFilterClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_SAMPLE_SIZE,
	PROP_SAMPLE_SIGNED,
	BURGNER_FIFO_BIG_ENDIAN_FILTER_LAST_ARG
};

static void
_burgner_fifo_big_endian_filter_finalize (GObject *object)
{
	BurgnerFifoBigEndianFilter *self = BURGNER_FIFO_BIG_ENDIAN_FILTER (object);

	self->sample_size = 2;
	self->sample_signed = 1;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_fifo_big_endian_filter_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoBigEndianFilter *self = BURGNER_FIFO_BIG_ENDIAN_FILTER (object);
	guint i;

	switch (prop_id) {
		case PROP_SAMPLE_SIZE:
			i = g_value_get_uint (value);
			g_return_if_fail ((i == 1) || (i == 2) || (i == 4) || (i == 8));

			self->sample_size = i;
			
			g_object_notify (object, "sample_size");
			break;
		case PROP_SAMPLE_SIGNED:
			self->sample_signed = g_value_get_boolean (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_big_endian_filter_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoBigEndianFilter *self = BURGNER_FIFO_BIG_ENDIAN_FILTER (object);

	switch (prop_id) {
		case PROP_SAMPLE_SIZE:
			g_value_set_uint (value, self->sample_size);
			break;
		case PROP_SAMPLE_SIGNED:
			g_value_set_boolean (value, self->sample_signed);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_big_endian_filter_filter (BurgnerFifoFilter *filter,
				    gpointer buf, guint32 size)
{
	BurgnerFifoBigEndianFilter *self = BURGNER_FIFO_BIG_ENDIAN_FILTER (filter);
	gpointer buf_end = buf + size;

	g_return_if_fail ((size % self->sample_size) == 0);

	if (self->sample_size == 1)
		return;

	while (buf != buf_end) {
		if (self->sample_size == 2) {
				*((guint16 *) buf) = GUINT16_SWAP_LE_BE (*((guint16 *) buf));
		} else if (self->sample_size == 4) {
				*((guint32 *) buf) = GUINT32_SWAP_LE_BE (*((guint32 *) buf));
		} else if (self->sample_size == 8) {
				*((guint64 *) buf) = GUINT64_SWAP_LE_BE (*((guint64 *) buf));
		} else {
			g_return_if_reached ();
		}

		buf += self->sample_size;
	}
}

static void
_burgner_fifo_big_endian_filter_class_init (BurgnerFifoBigEndianFilterClass *klass)
{
	GObjectClass *object_class;
	BurgnerFifoFilterClass *filter_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_fifo_big_endian_filter_set_property;
	object_class->get_property = _burgner_fifo_big_endian_filter_get_property;
	object_class->finalize = _burgner_fifo_big_endian_filter_finalize;

	filter_class = (BurgnerFifoFilterClass *) klass;
	filter_class->filter = _burgner_fifo_big_endian_filter_filter;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_SAMPLE_SIZE,
					 g_param_spec_uint ("sample_size",
					                _("sample_size"),
							_("Size of one sample"),
							1,
							8,
							2,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SAMPLE_SIGNED,
					 g_param_spec_boolean ("sample_signed",
					                _("sample_signed"),
							_("Whether samples are signed or not"),
							TRUE,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_fifo_big_endian_filter_init (BurgnerFifoBigEndianFilter *self)
{
	self->sample_size = 2;
	self->sample_signed = 1;
}

GType
burgner_fifo_big_endian_filter_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerFifoBigEndianFilterClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_fifo_big_endian_filter_class_init,
			NULL,
			NULL,
			sizeof (BurgnerFifoBigEndianFilter),
			0,
			(GInstanceInitFunc) _burgner_fifo_big_endian_filter_init,
		};

		type = g_type_register_static (BURGNER_TYPE_FIFO_FILTER, "BurgnerFifoBigEndianFilter", &info, 0);
	}

	return type;
}

BurgnerFifoFilter *
burgner_fifo_big_endian_filter_new (void)
{
	BurgnerFifoBigEndianFilter *self;

	self = g_object_new (BURGNER_TYPE_FIFO_BIG_ENDIAN_FILTER, NULL);


	return BURGNER_FIFO_FILTER (self);
}


GQuark
burgner_fifo_big_endian_filter_error_quark (void)
{
	return g_quark_from_string ("BurgnerFifoBigEndianFilter");
}

