/*
 * BurgnerFifoFdWriter - BurgnerFifoFdWriter object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-utils-i18n.h"

#include <unistd.h>
#include <errno.h>

#include <glib-object.h>
#include "burgner-utils-marshal.h"
#include "burgner-fifo-fd-writer.h"


static GObjectClass *parent_class = NULL;

enum {
	BURGNER_FIFO_FD_WRITER_PROP_0,
	PROP_FIFO,
	PROP_FD,
	PROP_SIZE,
	BURGNER_FIFO_FD_WRITER_LAST_ARG
};

enum {
	ERROR,
	LAST_SIGNAL
};

static gulong _burgner_fifo_fd_writer_signals[LAST_SIGNAL] = { 0, };


static void
_burgner_fifo_fd_writer_finalize (GObject *object)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (object);

	self->fifo = NULL;
	self->fd = -1;

	self->size = self->remaining = 0;

	if (self->channel) {
		if (self->watch_id) {
			g_source_remove (self->watch_id);
			self->watch_id = 0;
		}
		g_io_channel_unref (self->channel);
		self->channel = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

#if 0
static gboolean
_burgner_fifo_fd_writer_do_read_write (BurgnerFifoFdWriter *self)
{
	guchar *buf = NULL;
	guint to_write;
	gint b_read;

	while ((to_write = ))
}
#endif


static gboolean _burgner_fifo_fd_writer_watch_func (GIOChannel *channel,
				    GIOCondition condition,
				    gpointer user_data);

static gboolean
_burgner_fifo_fd_writer_idle_func (gpointer user_data)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (user_data);
	guchar *b = NULL;
	guint s;

	self->idle_id = 0;

	g_return_val_if_fail (self->fifo != NULL, FALSE);
	g_return_val_if_fail (self->fd >= 0, FALSE);

	if (self->error)	
		return FALSE;

	while ((self->remaining || !self->size) &&
	       (s = (burgner_fifo_to_write (self->fifo, &b)))) {
		gint r = read (self->fd, b, self->size ? MIN (s, self->remaining) : s);

		if (r > 0) {
			if (self->size) {
				g_assert (r <= self->remaining);
				self->remaining -= r;
			}

			burgner_fifo_wrote (self->fifo, r);
			continue;
		}

		if (r == 0) {
			g_return_val_if_fail (!self->size || (self->remaining == 0), FALSE);

			self->error = 1;
			burgner_fifo_end (self->fifo);
		} else if (errno != EAGAIN) {
			self->error = 1;
			g_signal_emit (self,
				       _burgner_fifo_fd_writer_signals[ERROR],
				       0,
				       errno);
		} else if (!self->watch_id) {
			self->watch_id = g_io_add_watch (self->channel,
							 G_IO_IN,
							 _burgner_fifo_fd_writer_watch_func,
							 self);
		}

		return FALSE;
	}

	if (self->size && (self->remaining == 0)) {
		self->error = 1;
		burgner_fifo_end (self->fifo);
	}

	return FALSE;
}

#if 0
static void
_burgner_fifo_fd_writer_do_write (BurgnerFifoFdWriter *self,
				  gpointer b, guint s)
{
	gpointer buf = b;
	guint size = s;

	do {
		gint r = read (self->fd, buf, size);

		if (r > 0) {
			burgner_fifo_wrote (self->fifo, r);
			continue;
		}

		if (r == 0) {
			self->error = 1;
			burgner_fifo_end (self->fifo);
		} else if (errno != EAGAIN) {
			self->error = 1;
			g_signal_emit (self,
				       _burgner_fifo_fd_writer_signals[ERROR],
				       0,
				       errno);
		} else {
			self->watch_id = g_io_add_watch (self->channel,
							 G_IO_IN,
							 _burgner_fifo_fd_writer_watch_func,
							 self);
		}
	} while ((s = burgner_fifo_to_write (self->fifo, (guchar **) &b)));
}
#endif

static void
_burgner_fifo_fd_writer_add_idle (BurgnerFifoFdWriter *self)
{
	if ((!self->error) &&
	    (self->fifo) &&
	    (self->fd >= 0) &&
	    (self->idle_id == 0))
		self->idle_id = g_idle_add (_burgner_fifo_fd_writer_idle_func,
					    self);
}

static gboolean
_burgner_fifo_fd_writer_watch_func (GIOChannel *channel,
				    GIOCondition condition,
				    gpointer user_data)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (user_data);

	self->watch_id = 0;

	if (self->fifo) {
		_burgner_fifo_fd_writer_add_idle (self);
	}

	return FALSE;
}

static void
_burgner_fifo_fd_writer_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (object);

	switch (prop_id) {
		case PROP_FIFO:
			g_return_if_fail (self->fifo == NULL);
			self->fifo = BURGNER_FIFO (g_value_get_object (value));
			g_return_if_fail (self->fifo != NULL);

			g_object_set (self->fifo,
				      "writer", self,
				      NULL);

			_burgner_fifo_fd_writer_add_idle (self);

			g_object_notify (object, "fifo");
			break;
		case PROP_FD:
			if (self->watch_id) {
				g_assert (self->channel);
				g_source_remove (self->watch_id);
				self->watch_id = 0;
			}

			if (self->channel) {
				g_io_channel_unref (self->channel);
				self->channel = NULL;
			}

			if (self->error)
				self->error = 0;

			self->remaining = self->size;

			self->fd = g_value_get_int (value);

			if (self->fd >= 0) {
				self->channel = g_io_channel_unix_new (self->fd);
				_burgner_fifo_fd_writer_add_idle (self);
			} else if (self->idle_id) {
				g_source_remove (self->idle_id);
				self->idle_id = 0;
			}

			g_object_notify (object, "fd");
			break;
		case PROP_SIZE:
			self->size = g_value_get_uint (value);
			self->remaining = self->size;
			g_object_notify (object, "size");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_fd_writer_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (object);

	switch (prop_id) {
		case PROP_FIFO:
			g_value_set_object (value, self->fifo);
			break;
		case PROP_FD:
			g_value_set_int (value, self->fd);
			break;
		case PROP_SIZE:
			g_value_set_uint (value, self->size);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_fd_writer_class_init (BurgnerFifoFdWriterClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_fifo_fd_writer_set_property;
	object_class->get_property = _burgner_fifo_fd_writer_get_property;
	object_class->finalize = _burgner_fifo_fd_writer_finalize;

	/* Signals */
	_burgner_fifo_fd_writer_signals[ERROR] = g_signal_new ("error",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerFifoFdWriterClass, error),
						NULL, NULL,
						burgner_utils_marshal_VOID__INT,
						G_TYPE_NONE, 1,
						G_TYPE_INT);

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_FIFO,
					 g_param_spec_object ("fifo",
					                _("fifo"),
							_("A BurgnerFifo"),
							BURGNER_TYPE_FIFO,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_FD,
					 g_param_spec_int ("fd",
					                _("fd"),
							_("A file descriptor, opened in non-blocking mode"),
							-1,
							G_MAXINT,
							-1,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_SIZE,
					 g_param_spec_uint ("size",
					                _("size"),
							_("The size of the file to write, or 0 if unknown"),
							0,
							G_MAXUINT,
							0,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_fifo_fd_writer_init (BurgnerFifoFdWriter *self)
{
	self->fifo = NULL;

	self->fd = -1;
	self->channel = NULL;
	self->watch_id = 0;

	self->idle_id = 0;

	self->size = self->remaining = 0;

	self->error = 0;
}

static void
_burgner_fifo_fd_writer_to_write (BurgnerFifoWriter *writer,
				  BurgnerFifo *fifo,
				  guchar *to_write,
				  guint size)
{
	BurgnerFifoFdWriter *self = BURGNER_FIFO_FD_WRITER (writer);
	gint r;

	g_assert (fifo == self->fifo);
	g_return_if_fail (self->fd >= 0);

	r = read (self->fd, to_write, self->size ? MIN (size, self->remaining) : size);

	g_return_if_fail (r != 0);

	if (r > 0) {
		if (self->watch_id) {
			g_source_remove (self->watch_id);
			self->watch_id = 0;
		}

		if (self->size) {
			g_return_if_fail (r <= self->remaining);
			self->remaining -= r;
		}

		burgner_fifo_wrote (self->fifo, r);
	} else if (r == 0) {
		g_assert (!self->size);

		if (self->watch_id) {
			g_source_remove (self->watch_id);
			self->watch_id = 0;
		}

		self->error = 1;
		burgner_fifo_end (self->fifo);
		return;
	} else if (errno != EAGAIN) {
		if (self->watch_id) {
			g_source_remove (self->watch_id);
			self->watch_id = 0;
		}

		self->error = 1;
		g_signal_emit (self,
			       _burgner_fifo_fd_writer_signals[ERROR],
			       0,
			       errno);
		return;
	}
	
	if (self->size && !self->remaining) {
		if (self->watch_id) {
			g_source_remove (self->watch_id);
			self->watch_id = 0;
		}

		self->error = 1;
		burgner_fifo_end (self->fifo);
	} else if (!self->watch_id) {
		self->watch_id = g_io_add_watch (self->channel,
						 G_IO_IN,
						 _burgner_fifo_fd_writer_watch_func,
						 self);
	}
}

static void
_burgner_fifo_fd_writer_writer_iface_init (BurgnerFifoWriterIface *iface)
{
	iface->to_write = _burgner_fifo_fd_writer_to_write;
}

GType
burgner_fifo_fd_writer_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerFifoFdWriterClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_fifo_fd_writer_class_init,
			NULL,
			NULL,
			sizeof (BurgnerFifoFdWriter),
			0,
			(GInstanceInitFunc) _burgner_fifo_fd_writer_init,
		};
		static GInterfaceInfo iface_info = {
			(GInterfaceInitFunc) _burgner_fifo_fd_writer_writer_iface_init,
			NULL,
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerFifoFdWriter", &info, 0);
		g_type_add_interface_static (type, BURGNER_TYPE_FIFO_WRITER, &iface_info);
	}

	return type;
}

BurgnerFifoWriter *
burgner_fifo_fd_writer_new (void)
{
	BurgnerFifoFdWriter *self;

	self = g_object_new (BURGNER_TYPE_FIFO_FD_WRITER, NULL);


	return BURGNER_FIFO_WRITER (self);
}


GQuark
burgner_fifo_fd_writer_error_quark (void)
{
	return g_quark_from_string ("BurgnerFifoFdWriter");
}

