/*
 * BurgnerFifoGnomeVfsWriter - BurgnerFifoGnomeVfsWriter object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-utils-i18n.h"

#include <glib-object.h>
#include "burgner-utils-marshal.h"
#include "burgner-fifo.h"
#include "burgner-fifo-gnome-vfs-writer.h"


static GObjectClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_FIFO,
	PROP_HANDLE,
	BURGNER_FIFO_GNOME_VFS_WRITER_LAST_ARG
};

enum {
	ERROR,
	BURGNER_FIFO_GNOME_VFS_WRITER_LAST_SIGNAL
};

static guint burgner_fifo_gnome_vfs_writer_signals[BURGNER_FIFO_GNOME_VFS_WRITER_LAST_SIGNAL] = { 0, };

static void
_burgner_fifo_gnome_vfs_writer_to_write (BurgnerFifoWriter *writer,
					 BurgnerFifo *fifo,
					 guchar *to_write,
					 guint size);

static void
_burgner_fifo_gnome_vfs_writer_finalize (GObject *object)
{
	BurgnerFifoGnomeVfsWriter *self = BURGNER_FIFO_GNOME_VFS_WRITER (object);

	if (self->fifo)
		self->fifo = NULL;

	if (self->handle) {
		if (self->reading)
			gnome_vfs_async_cancel (self->handle);
		self->handle = NULL;
	}

	self->reading = self->error = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_fifo_gnome_vfs_writer_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoGnomeVfsWriter *self = BURGNER_FIFO_GNOME_VFS_WRITER (object);

	switch (prop_id) {
		case PROP_FIFO:
			g_return_if_fail (self->fifo == NULL);
			self->fifo = BURGNER_FIFO (g_value_get_object (value));
			g_return_if_fail (self->fifo != NULL);
			g_object_set (self->fifo,
				      "writer", self,
				      NULL);
			g_object_notify (object, "fifo");
			break;
		case PROP_HANDLE: {
			gboolean err = self->error;
			g_return_if_fail (self->reading == 0);
			self->handle = (GnomeVFSAsyncHandle *) g_value_get_pointer (value);
			self->error = 0;
			g_object_notify (object, "handle");
			if (err) {
				guchar *b = NULL;
				guint s;

				g_return_if_fail (self->fifo != NULL);

				s = burgner_fifo_to_write (self->fifo, &b);
				if (s) {
					_burgner_fifo_gnome_vfs_writer_to_write
						(BURGNER_FIFO_WRITER (self),
						 self->fifo,
						 b, s);
				}
			}
			} break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_gnome_vfs_writer_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifoGnomeVfsWriter *self = BURGNER_FIFO_GNOME_VFS_WRITER (object);

	switch (prop_id) {
		case PROP_FIFO:
			g_value_set_object (value, self->fifo);
			break;
		case PROP_HANDLE:
			g_value_set_pointer (value, self->handle);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
_burgner_fifo_gnome_vfs_writer_class_init (BurgnerFifoGnomeVfsWriterClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = _burgner_fifo_gnome_vfs_writer_set_property;
	object_class->get_property = _burgner_fifo_gnome_vfs_writer_get_property;
	object_class->finalize = _burgner_fifo_gnome_vfs_writer_finalize;

	/* Signals */
	burgner_fifo_gnome_vfs_writer_signals[ERROR] = g_signal_new ("error",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerFifoGnomeVfsWriterClass, error),
						NULL, NULL,
						burgner_utils_marshal_VOID__INT,
						G_TYPE_NONE, 1,
						G_TYPE_INT);

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_FIFO,
					 g_param_spec_object ("fifo",
					                _("fifo"),
							_("A BurgnerFifo"),
							BURGNER_TYPE_FIFO,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_HANDLE,
					 g_param_spec_pointer ("handle",
					                _("handle"),
							_("A GnomeVFSAsyncHandle from which data will be read"),
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_fifo_gnome_vfs_writer_init (BurgnerFifoGnomeVfsWriter *self)
{
	self->fifo = NULL;
	self->handle = NULL;
	self->reading = self->error = 0;
}

static void
_burgner_fifo_gnome_vfs_writer_read_cb (GnomeVFSAsyncHandle *handle,
					GnomeVFSResult result,
					gpointer buffer,
					GnomeVFSFileSize bytes_requested,
					GnomeVFSFileSize bytes_read,
					gpointer user_data)
{
	BurgnerFifoGnomeVfsWriter *self = BURGNER_FIFO_GNOME_VFS_WRITER (user_data);

	g_return_if_fail (self->reading == 1);
	g_return_if_fail (self->handle == handle);
	g_return_if_fail (self->fifo != NULL);

	self->reading = 0;

	if (result == GNOME_VFS_OK) {
		burgner_fifo_wrote (self->fifo, bytes_read);
	} else if (result == GNOME_VFS_ERROR_EOF) {
		burgner_fifo_end (self->fifo);
	} else {
		self->error = 1;
		g_signal_emit (self,
			       burgner_fifo_gnome_vfs_writer_signals[ERROR],
			       0,
			       result);
	}
}


static void
_burgner_fifo_gnome_vfs_writer_to_write (BurgnerFifoWriter *writer,
					 BurgnerFifo *fifo,
					 guchar *to_write,
					 guint size)
{
	BurgnerFifoGnomeVfsWriter *self = BURGNER_FIFO_GNOME_VFS_WRITER (writer);

	g_return_if_fail (self->fifo == fifo);
	g_return_if_fail (self->reading == 0);

	if (self->error || !self->handle)
		return;

	self->reading = 1;
	gnome_vfs_async_read (self->handle,
			      to_write,
			      size,
			      _burgner_fifo_gnome_vfs_writer_read_cb,
			      self);
}

static void
_burgner_fifo_gnome_vfs_writer_fifo_writer_iface_init (BurgnerFifoWriterIface *iface)
{
	iface->to_write = _burgner_fifo_gnome_vfs_writer_to_write;
}

GType
burgner_fifo_gnome_vfs_writer_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerFifoGnomeVfsWriterClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_fifo_gnome_vfs_writer_class_init,
			NULL,
			NULL,
			sizeof (BurgnerFifoGnomeVfsWriter),
			0,
			(GInstanceInitFunc) _burgner_fifo_gnome_vfs_writer_init,
		};
		static GInterfaceInfo iface_info = {
			(GInterfaceInitFunc) _burgner_fifo_gnome_vfs_writer_fifo_writer_iface_init,
			NULL,
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerFifoGnomeVfsWriter", &info, 0);
		g_type_add_interface_static (type, BURGNER_TYPE_FIFO_WRITER, &iface_info);
	}

	return type;
}

BurgnerFifoWriter *
burgner_fifo_gnome_vfs_writer_new (void)
{
	BurgnerFifoGnomeVfsWriter *self;

	self = g_object_new (BURGNER_TYPE_FIFO_GNOME_VFS_WRITER, NULL);


	return BURGNER_FIFO_WRITER (self);
}


GQuark
burgner_fifo_gnome_vfs_writer_error_quark (void)
{
	return g_quark_from_string ("BurgnerFifoGnomeVfsWriter");
}

