/*
 * BurgnerFifo - BurgnerFifo object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <libintl.h>
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) x

#include <glib-object.h>
#include "burgner-fifo.h"
#include "burgner-fifo-reader.h"
#include "burgner-fifo-writer.h"


#define check(o)		G_STMT_START {	\
	g_return_if_fail (o);	\
	g_return_if_fail (BURGNER_IS_FIFO (o));	\
				} G_STMT_END
#define check_with_val(o,v)	G_STMT_START {	\
	g_return_val_if_fail ((o), (v));	\
	g_return_val_if_fail (BURGNER_IS_FIFO (o), (v));	\
				} G_STMT_END

static GObjectClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_READER,
	PROP_WRITER,
	PROP_FILTER,
	LAST_ARG
};

static void
burgner_fifo_finalize (GObject *object)
{
	BurgnerFifo *self = BURGNER_FIFO (object);

	if (self->reader) {
		g_object_unref (self->reader);
		self->reader = NULL;
	}

	if (self->writer) {
		g_object_unref (self->writer);
		self->writer = NULL;
	}

	if (self->filter) {
		g_object_unref (self->filter);
		self->filter = NULL;
	}

	if (self->end_reached)
		self->end_reached = FALSE;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
burgner_fifo_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifo *self = BURGNER_FIFO (object);

	switch (prop_id) {
		case PROP_READER:
			if (self->reader)
				g_object_unref (self->reader);
			self->reader = (BurgnerFifoReader *) g_value_get_object (value);
			if (self->reader)
				self->reader = BURGNER_FIFO_READER (g_object_ref (self->reader));
			g_object_notify (object, "reader");
			break;
		case PROP_WRITER:
			if (self->writer)
				g_object_unref (self->writer);
			self->writer = (BurgnerFifoWriter *) g_value_get_object (value);
			if (self->writer)
				self->writer = BURGNER_FIFO_WRITER (g_object_ref (self->writer));
			g_object_notify (object, "writer");
			break;
		case PROP_FILTER:
			if (self->filter)
				g_object_unref (self->filter);
			self->filter = (BurgnerFifoFilter *) g_value_get_object (value);
			if (self->filter)
				self->filter = BURGNER_FIFO_FILTER (g_object_ref (self->filter));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_fifo_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerFifo *self = BURGNER_FIFO (object);

	switch (prop_id) {
		case PROP_READER:
			g_value_set_object (value, self->reader);
			break;
		case PROP_WRITER:
			g_value_set_object (value, self->writer);
			break;
		case PROP_FILTER:
			g_value_set_object (value, self->filter);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static gboolean
burgner_fifo_impl_end_reached (BurgnerFifo *fifo)
{
	return fifo->end_reached;
}

static void
burgner_fifo_impl_end (BurgnerFifo *fifo)
{
	g_return_if_fail (fifo->end_reached != TRUE);

	fifo->end_reached = TRUE;
}

static void
burgner_fifo_impl_reset (BurgnerFifo *fifo)
{
	fifo->end_reached = 0;
}


static void
burgner_fifo_class_init (BurgnerFifoClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_fifo_set_property;
	object_class->get_property = burgner_fifo_get_property;
	object_class->finalize = burgner_fifo_finalize;

	/* Virtual functions */
	klass->end_reached = burgner_fifo_impl_end_reached;
	klass->end = burgner_fifo_impl_end;
	klass->reset = burgner_fifo_impl_reset;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_READER,
					 g_param_spec_object ("reader",
					                _("reader"),
							_("A BurgnerFifoReader"),
							G_TYPE_OBJECT,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WRITER,
					 g_param_spec_object ("writer",
					                _("writer"),
							_("A BurgnerFifoWriter"),
							G_TYPE_OBJECT,
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_FILTER,
					 g_param_spec_object ("filter",
					                _("filter"),
							_("A BurgnerFifoFilter"),
							BURGNER_TYPE_FIFO_FILTER,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_fifo_init (BurgnerFifo *self)
{
	self->reader = NULL;
	self->writer = NULL;
	self->filter = NULL;
	self->end_reached = 0;
}

GType
burgner_fifo_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerFifoClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_fifo_class_init,
			NULL,
			NULL,
			sizeof (BurgnerFifo),
			0,
			(GInstanceInitFunc) burgner_fifo_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerFifo", &info, G_TYPE_FLAG_ABSTRACT);
	}

	return type;
}


GQuark
burgner_fifo_error_quark (void)
{
	return g_quark_from_string ("BurgnerFifo");
}


guint
burgner_fifo_to_read (BurgnerFifo *fifo, guchar **buf)
{
	check_with_val (fifo, 0);

	if (BURGNER_FIFO_GET_CLASS (fifo)->to_read)
		return BURGNER_FIFO_GET_CLASS (fifo)->to_read (fifo, buf);

	return 0;
}

void
burgner_fifo_read (BurgnerFifo *fifo, guint size)
{
	check (fifo);

	if (BURGNER_FIFO_GET_CLASS (fifo)->read)
		BURGNER_FIFO_GET_CLASS (fifo)->read (fifo, size);
}

gboolean
burgner_fifo_end_reached (BurgnerFifo *fifo)
{
	check_with_val (fifo, FALSE);

	if (BURGNER_FIFO_GET_CLASS (fifo)->end_reached)
		return BURGNER_FIFO_GET_CLASS (fifo)->end_reached (fifo);

	return FALSE;
}

guint
burgner_fifo_to_write (BurgnerFifo *fifo, guchar **buf)
{
	check_with_val (fifo, 0);

	if (BURGNER_FIFO_GET_CLASS (fifo)->to_write)
		return BURGNER_FIFO_GET_CLASS (fifo)->to_write (fifo, buf);

	return 0;
}

void
burgner_fifo_wrote (BurgnerFifo *fifo, guint size)
{
	check (fifo);

	if (BURGNER_FIFO_GET_CLASS (fifo)->wrote)
		BURGNER_FIFO_GET_CLASS (fifo)->wrote (fifo, size);
}

void
burgner_fifo_end (BurgnerFifo *fifo)
{
	check (fifo);

	if (BURGNER_FIFO_GET_CLASS (fifo)->end)
		BURGNER_FIFO_GET_CLASS (fifo)->end (fifo);
}

void
burgner_fifo_reset (BurgnerFifo *fifo)
{
	g_return_if_fail (fifo != NULL);
	g_return_if_fail (BURGNER_IS_FIFO (fifo));

	if (BURGNER_FIFO_GET_CLASS (fifo)->reset)
		BURGNER_FIFO_GET_CLASS (fifo)->reset (fifo);
}

void
burgner_fifo_set_data_block_size (BurgnerFifo *fifo, guint size)
{
	g_return_if_fail (fifo != NULL);
	g_return_if_fail (BURGNER_IS_FIFO (fifo));

	if (BURGNER_FIFO_GET_CLASS (fifo)->set_data_block_size)
		BURGNER_FIFO_GET_CLASS (fifo)->set_data_block_size (fifo, size);
}

guint
burgner_fifo_get_percent (BurgnerFifo *fifo)
{
	g_return_val_if_fail (fifo != NULL, 0);
	g_return_val_if_fail (BURGNER_IS_FIFO (fifo), 0);

	if (BURGNER_FIFO_GET_CLASS (fifo)->get_percent)
		return BURGNER_FIFO_GET_CLASS (fifo)->get_percent (fifo);

	return 0;
}

void
burgner_fifo_do_filter (BurgnerFifo *fifo, gpointer buf, guint32 size)
{
	g_return_if_fail (fifo != NULL);
	g_return_if_fail (BURGNER_IS_FIFO (fifo));

	if (fifo->filter)
		burgner_fifo_filter_filter (fifo->filter, buf, size);
}

void
burgner_fifo_wake_up (BurgnerFifo *fifo)
{
	g_return_if_fail (fifo != NULL);
	g_return_if_fail (BURGNER_IS_FIFO (fifo));

	if (BURGNER_FIFO_GET_CLASS (fifo)->wake_up)
		return BURGNER_FIFO_GET_CLASS (fifo)->wake_up (fifo);
}

