/*
 * BurgnerModule - BurgnerModule object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <libintl.h>
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) x

#include <glib-object.h>
#include "burgner-utils-marshal.h"
#include "burgner-module.h"

typedef struct {
	GType a;
	GType b;
} BurgnerModuleTwoTypes;

static BurgnerModuleTwoTypes *
burgner_module_two_types_new (void)
{
	return g_new0 (BurgnerModuleTwoTypes, 1);
}

static gboolean
burgner_module_two_types_equal_func (gconstpointer p,
				     gconstpointer q)
{
	BurgnerModuleTwoTypes *a = (BurgnerModuleTwoTypes *) p,
			      *b = (BurgnerModuleTwoTypes *) q;

#if 0
	return ((*a) == (*b) ? TRUE : FALSE);
#else
	return (a->a == b->a) && (a->b == b->b);
#endif
}

typedef struct {
	gchar *name;
	GType parent_type;
	gchar *info_symbol;
} BurgnerModuleTypeInfo;

typedef struct {
	GType instance_type;
	GType interface_type;
	gchar *info_symbol;
} BurgnerModuleIfaceInfo;

static BurgnerModuleTypeInfo *
burgner_module_type_info_new (gchar *name,
			      GType parent_type,
			      gchar *info_symbol)
{
	BurgnerModuleTypeInfo *i = g_new0 (BurgnerModuleTypeInfo, 1);

	i->name = g_strdup (name);
	i->parent_type = parent_type;
	i->info_symbol = g_strdup (info_symbol);

	return i;
}

static void
burgner_module_type_info_free (BurgnerModuleTypeInfo *i)
{
	if (i->name)
		g_free (i->name);
	if (i->info_symbol)
		g_free (i->info_symbol);
	g_free (i);
}

static BurgnerModuleIfaceInfo *
burgner_module_iface_info_new (GType instance_type,
			       GType interface_type,
			       gchar *info_symbol)
{
	BurgnerModuleIfaceInfo *i = g_new0 (BurgnerModuleIfaceInfo, 1);

	i->instance_type = instance_type;
	i->interface_type = interface_type;
	i->info_symbol = g_strdup (info_symbol);

	return i;
}

static void
burgner_module_iface_info_free (BurgnerModuleIfaceInfo *i)
{
	if (i->info_symbol)
		g_free (i->info_symbol);
	g_free (i);
}


#define check(o)		G_STMT_START {	\
	g_return_if_fail (o);	\
	g_return_if_fail (BURGNER_IS_MODULE (o));	\
				} G_STMT_END
#define check_with_val(o,v)	G_STMT_START {	\
	g_return_val_if_fail ((o), (v));	\
	g_return_val_if_fail (BURGNER_IS_MODULE (o), (v));	\
				} G_STMT_END

static GObjectClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_MODULE_PATH,
	LAST_ARG
};

enum {
	LOAD,
	UNLOAD,
	LAST_SIGNAL
};

static guint burgner_module_signals[LAST_SIGNAL] = { 0, };

static void
burgner_module_finalize (GObject *object)
{
	BurgnerModule *self = BURGNER_MODULE (object);

	if (self->type_list) {
		g_list_foreach (self->type_list,
				(GFunc) burgner_module_type_info_free,
				NULL);
		g_list_free (self->type_list);
		self->type_list = NULL;
	}

	if (self->iface_list) {
		g_list_foreach (self->iface_list,
				(GFunc) burgner_module_iface_info_free,
				NULL);
		g_list_free (self->iface_list);
		self->iface_list = NULL;
	}

	if (self->type_to_info_sym) {
		g_hash_table_destroy (self->type_to_info_sym);
		self->type_to_info_sym = NULL;
	}

	if (self->iface_to_info_sym) {
		g_hash_table_destroy (self->iface_to_info_sym);
		self->iface_to_info_sym = NULL;
	}

	if (self->module_path) {
		g_free (self->module_path);
		self->module_path = NULL;
	}

	if (self->module) {
		g_module_close (self->module);
		self->module = NULL;
	}

	if (self->use_count)
		self->use_count = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
burgner_module_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerModule *self = BURGNER_MODULE (object);

	switch (prop_id) {
		case PROP_MODULE_PATH:
			if (self->module_path)
				g_free (self->module_path);
			self->module_path = g_value_dup_string (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_module_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerModule *self = BURGNER_MODULE (object);

	switch (prop_id) {
		case PROP_MODULE_PATH:
			g_value_set_string (value, self->module_path);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_module_add_type_impl (BurgnerModule *module,
			      gchar *name,
			      GType parent_type,
			      gchar *info_symbol)
{
	BurgnerModuleTypeInfo *info;
	/*GType type;*/

	info = burgner_module_type_info_new (name, parent_type, info_symbol);
	module->type_list = g_list_prepend (module->type_list, info);
}

static void
burgner_module_add_interface_impl (BurgnerModule *module,
				   GType instance_type,
				   GType interface_type,
				   gchar *info_symbol)
{
	BurgnerModuleIfaceInfo *i;
	/*BurgnerModuleTwoTypes *tt;*/

	i = burgner_module_iface_info_new (instance_type,
					   interface_type,
					   info_symbol);
	module->iface_list = g_list_prepend (module->iface_list, i);
}

static void
burgner_module_register_types_ifaces_impl (BurgnerModule *module)
{
	GList *it;

	check (module);

	for (it = module->type_list; it != NULL; it = it->next) {
		BurgnerModuleTypeInfo *i = (BurgnerModuleTypeInfo *) it->data;
		GType type;

		type = g_type_register_dynamic (i->parent_type,
						i->name,
						G_TYPE_PLUGIN (module),
						0);
		g_return_if_fail (type != G_TYPE_INVALID);

		/*
		 * We use g_type_name as the key because the size of the GType type
		 * may be too long and g_type_name is unique, always returns the
		 * same value for a given type, and has a return value that has the
		 * right size.
		 */
		g_hash_table_insert (module->type_to_info_sym,
				     (gpointer) g_type_name (type),
				     g_strdup (i->info_symbol));
	}
	g_list_free (module->type_list);
	module->type_list = NULL;

	for (it = module->iface_list; it != NULL; it = it->next) {
		BurgnerModuleIfaceInfo *i = (BurgnerModuleIfaceInfo *) it->data;
		BurgnerModuleTwoTypes *tt;

		g_type_add_interface_dynamic (i->instance_type,
					      i->interface_type,
					      G_TYPE_PLUGIN (module));

		tt = burgner_module_two_types_new ();
		tt->a = i->instance_type;
		tt->b = i->interface_type;

		g_hash_table_insert (module->iface_to_info_sym,
				     tt, g_strdup (i->info_symbol));
	}
	g_list_free (module->iface_list);
	module->iface_list = NULL;
}


static void
burgner_module_class_init (BurgnerModuleClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->set_property = burgner_module_set_property;
	object_class->get_property = burgner_module_get_property;
	object_class->finalize = burgner_module_finalize;

	/* Virtual functions */
	klass->add_type = burgner_module_add_type_impl;
	klass->add_interface = burgner_module_add_interface_impl;
	klass->register_types_ifaces = burgner_module_register_types_ifaces_impl;

	/* Signals */
	burgner_module_signals[LOAD] = g_signal_new ("load",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerModuleClass, load),
						NULL, NULL,
						burgner_utils_marshal_VOID__STRING_POINTER_POINTER,
						G_TYPE_NONE, 3,
						G_TYPE_STRING,
						G_TYPE_POINTER,
						G_TYPE_POINTER);
	burgner_module_signals[UNLOAD] = g_signal_new ("unload",
						G_TYPE_FROM_CLASS (klass),
						G_SIGNAL_RUN_FIRST,
						G_STRUCT_OFFSET (BurgnerModuleClass, unload),
						NULL, NULL,
						burgner_utils_marshal_VOID__STRING_POINTER,
						G_TYPE_NONE, 2,
						G_TYPE_STRING,
						G_TYPE_POINTER);

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_MODULE_PATH,
					 g_param_spec_string ("module_path",
					                _("module_path"),
							_("The path of the module, to be passed to g_module_open"),
							NULL,
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_module_init (BurgnerModule *self)
{
	self->type_to_info_sym = g_hash_table_new_full (NULL, NULL,
							NULL,
							g_free);
	self->iface_to_info_sym = g_hash_table_new_full (NULL,
					burgner_module_two_types_equal_func, 
							 g_free,
							 g_free);

	self->module_path = NULL;
	self->module = NULL;

	self->use_count = 0;
}

static void
burgner_module_plugin_iface_complete_type_info (GTypePlugin *plugin,
						GType g_type,
						GTypeInfo *info,
						GTypeValueTable *value_table)
{
	BurgnerModule *module = BURGNER_MODULE (plugin);
	gchar *sym;

	g_return_if_fail (module->type_to_info_sym != NULL);

	sym = g_hash_table_lookup (module->type_to_info_sym, g_type_name (g_type));
	if (sym && *sym) {
		GTypeInfo *module_ti;

		g_return_if_fail (module->module != NULL);

		if (!g_module_symbol (module->module,
				      sym,
				      (gpointer*) &module_ti)) {
			g_warning ("Can't get symbol '%s' from module '%s'",
				   sym,
				   module->module_path);
			return;
		} else {
			*info = *module_ti;
		}
	}
}

static void
burgner_module_plugin_iface_complete_interface_info (GTypePlugin *plugin,
						     GType instance_type,
						     GType interface_type,
						     GInterfaceInfo *info)
{
	BurgnerModuleTwoTypes tt;
	BurgnerModule *module = BURGNER_MODULE (plugin);
	gchar *sym;

	g_return_if_fail (module->iface_to_info_sym != NULL);

	tt.a = instance_type;
	tt.b = interface_type;

	sym = g_hash_table_lookup (module->iface_to_info_sym, &tt);
	if (sym && *sym) {
		GInterfaceInfo *module_ii;

		g_return_if_fail (module->module != NULL);

		if (!g_module_symbol (module->module,
				      sym,
				      (gpointer*) &module_ii)) {
			g_warning ("Can't get symbol '%s' from module '%s'",
				   sym,
				   module->module_path);
			return;
		} else {
			*info = *module_ii;
		}
	}

}

static void
burgner_module_plugin_iface_init (GTypePluginClass *iface)
{
	iface->use_plugin = (GTypePluginUse) burgner_module_use;
	iface->unuse_plugin = (GTypePluginUnuse) burgner_module_unuse;
	iface->complete_type_info = burgner_module_plugin_iface_complete_type_info;
	iface->complete_interface_info = burgner_module_plugin_iface_complete_interface_info;
}

GType
burgner_module_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerModuleClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_module_class_init,
			NULL,
			NULL,
			sizeof (BurgnerModule),
			0,
			(GInstanceInitFunc) burgner_module_init,
		};

		static GInterfaceInfo plugin_info = {
			(GInterfaceInitFunc) burgner_module_plugin_iface_init,
			(GInterfaceFinalizeFunc) NULL,
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerModule", &info, 0);
		g_type_add_interface_static (type, G_TYPE_TYPE_PLUGIN, &plugin_info);
	}

	return type;
}

BurgnerModule *
burgner_module_new (void)
{
	BurgnerModule *self;

	self = g_object_new (BURGNER_TYPE_MODULE, NULL);


	return self;
}


GQuark
burgner_module_error_quark (void)
{
	return g_quark_from_string ("BurgnerModule");
}

void
burgner_module_use (BurgnerModule *module)
{
	check (module);
	module->use_count++;

	if (module->use_count == 1) {
		g_return_if_fail (module->module == NULL);
		g_return_if_fail (module->module_path != NULL);

		module->module = g_module_open (module->module_path, 0);
		if (!module->module) {
			GError *err = g_error_new (BURGNER_MODULE_ERROR,
					BURGNER_MODULE_ERROR_G_MODULE_ERROR,
			      _("Error occurred while loading module '%s': %s"),
			      			module->module_path,
						g_module_error ());
			burgner_module_load (module, module->module_path,
					     NULL, err);
			g_error_free (err);
			return;
		}

		burgner_module_load (module, module->module_path,
				     module->module, NULL);
	}
}

void
burgner_module_unuse (BurgnerModule *module)
{
	check (module);
	g_return_if_fail (module->use_count > 0);

	if (((--module->use_count) == 0) &&
	    module->module != NULL) {
	    	gboolean ret = g_module_close (module->module);
		module->module = NULL;

		if (!ret) {
			GError *err = g_error_new (BURGNER_MODULE_ERROR,
					BURGNER_MODULE_ERROR_G_MODULE_ERROR,
					_("Error while unloading '%s': %s"),
						module->module_path,
						g_module_error ());
			burgner_module_unload (module, module->module_path,
					       err);
			g_error_free (err);
		} else {
			burgner_module_unload (module, module->module_path,
						NULL);
		}
	}
}


void
burgner_module_add_type (BurgnerModule *module,
			 gchar *name,
			 GType parent_type,
			 gchar *info_symbol)
{
	check (module);
	g_return_if_fail (name != NULL);
	g_return_if_fail (*name != 0);
	g_return_if_fail (parent_type != G_TYPE_INVALID);

	if (BURGNER_MODULE_GET_CLASS (module)->add_type)
		BURGNER_MODULE_GET_CLASS (module)->add_type (module, name, parent_type, info_symbol);
}

void
burgner_module_add_interface (BurgnerModule *module,
			      GType instance_type,
			      GType interface_type,
			      gchar *info_symbol)
{
	check (module);
	g_return_if_fail (instance_type != G_TYPE_INVALID);
	g_return_if_fail (G_TYPE_IS_INTERFACE (interface_type));

	if (BURGNER_MODULE_GET_CLASS (module)->add_interface)
		BURGNER_MODULE_GET_CLASS (module)->add_interface (module, instance_type, interface_type, info_symbol);
}

void
burgner_module_load (BurgnerModule *module,
		     gchar *module_path,
		     GModule *gmodule,
		     GError *err)
{
	check (module);

	g_signal_emit (module, burgner_module_signals[LOAD], 0,
			module_path, gmodule, err);
}

void
burgner_module_unload (BurgnerModule *module,
		       gchar *module_path,
		       GError *error)
{
	check (module);

	g_signal_emit (module, burgner_module_signals[UNLOAD], 0,
			module_path, error);
}

void
burgner_module_register_types_ifaces (BurgnerModule *module)
{
	check (module);

	if (BURGNER_MODULE_GET_CLASS (module)->register_types_ifaces)
		BURGNER_MODULE_GET_CLASS (module)->register_types_ifaces (module);
}

