/*
 * BurgnerRingFifo - BurgnerRingFifo object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-utils-i18n.h"

#include <stdio.h>
#include <string.h>
#include <glib-object.h>
#include "burgner-fifo-reader.h"
#include "burgner-fifo-writer.h"
#include "burgner-ring-fifo.h"

static BurgnerFifoClass *parent_class = NULL;

enum {
	PROP_0,
	PROP_BUFFER,
	PROP_SIZE,
	PROP_READ_CHUNK_SIZE,
	PROP_WRITE_CHUNK_SIZE,
	LAST_ARG
};

static void
burgner_ring_fifo_finalize (GObject *object)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (object);

	if (self->buf) {
		g_free (self->buf);
		self->buf = NULL;
	}

	self->buf_end = self->real_buf_end = self->to_read = self->to_write = NULL; 

	self->read_chunk_size = self->write_chunk_size = 0;
	self->real_read_chunk_size = self->real_write_chunk_size = 0;

	self->data_block_size = 0;

	self->reading = self->writing = 0;
	self->equal_case = TO_WRITE;

	if (self->idle_id) {
		g_source_remove (self->idle_id);
		self->idle_id = 0;
	}

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
_brf_idle_func (BurgnerRingFifo *self)
{
	BurgnerFifo *fifo = BURGNER_FIFO (self);
	guint size;
	guchar *end;
	gboolean read_sth = FALSE;

	if (!fifo->writer) {
		self->idle_id = 0;
		return FALSE;
	}

	while (!self->writing &&
	       !fifo->end_reached &&
	       (size = (end = self->to_write > self->to_read ||
				(self->to_write == self->to_read &&
				 self->equal_case == TO_WRITE)
					? self->buf_end : self->to_read)
							- self->to_write) &&
	       (!self->write_chunk_size || size >= self->write_chunk_size ||
					   end == self->buf_end)) {
		guint to_write = self->write_chunk_size
					? MIN (self->write_chunk_size, size)
					: size;

		self->writing = to_write;
		burgner_fifo_writer_to_write (fifo->writer, fifo,
					      self->to_write, to_write);
	}


	if (!fifo->reader) {
		self->idle_id = 0;
		return FALSE;
	}

	while (!self->reading &&
	       (size = (end = self->to_read > self->to_write ||
				(self->to_read == self->to_write &&
				 self->equal_case == TO_READ)
					? self->buf_end : self->to_write)
							- self->to_read) &&
	       (!self->read_chunk_size || size >= self->read_chunk_size ||
		end == self->buf_end || 
		(fifo->end_reached && end == self->to_write))) {
		guint to_read = self->read_chunk_size
					? MIN (self->read_chunk_size, size)
					: size;


		read_sth = TRUE;
		self->reading = to_read;
		burgner_fifo_reader_to_read (fifo->reader, fifo,
					     self->to_read, to_read);
	}

	if (fifo->end_reached &&
	    self->to_read == self->to_write &&
	    self->equal_case == TO_WRITE) {
		burgner_fifo_reader_end_reached (fifo->reader, fifo);
	} else if (read_sth) {
		return TRUE;
	}

	self->idle_id = 0;
	return FALSE;
}

static void
_brf_add_idle (BurgnerRingFifo *self)
{
	if (!self->idle_id)
		self->idle_id = g_idle_add ((GSourceFunc) _brf_idle_func, self);
}

static void
burgner_ring_fifo_set_property (GObject *object,
	      guint prop_id,
	      const GValue *value,
	      GParamSpec *pspec)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (object);

	switch (prop_id) {
		case PROP_SIZE: {
			guint size;

			g_return_if_fail (self->buf == NULL);

			size = g_value_get_uint (value);
			self->buf = g_malloc (size);
			self->real_buf_end = self->buf + size;
			if (self->data_block_size) {
				self->buf_end = self->buf + size -
						(size % self->data_block_size);
			} else {
				self->buf_end = self->real_buf_end;
			}

			self->to_read = self->to_write = self->buf;

			g_object_notify (object, "size");
			g_object_notify (object, "buffer");
			_brf_add_idle (self);
			} break;
		case PROP_READ_CHUNK_SIZE: {
			guint old_v = self->real_read_chunk_size;
			self->real_read_chunk_size = g_value_get_uint (value);

			if (self->real_read_chunk_size != old_v) {
				guint d = self->read_chunk_size;
				if (self->data_block_size) {
					self->read_chunk_size = self->real_read_chunk_size - (self->real_read_chunk_size % self->data_block_size);
				} else {
					self->read_chunk_size = self->real_read_chunk_size;
				}

				if (d != self->read_chunk_size)
					g_object_notify (object, "read_chunk_size");

				if (self->read_chunk_size > d)
					_brf_add_idle (self);
			}

			} break;
		case PROP_WRITE_CHUNK_SIZE: {
			guint old_v = self->real_write_chunk_size;
			self->real_write_chunk_size = g_value_get_uint (value);

			if (self->real_write_chunk_size != old_v) {
				guint d = self->write_chunk_size;
				if (self->data_block_size) {
					self->write_chunk_size = self->real_write_chunk_size - (self->real_write_chunk_size % self->data_block_size);
				} else {
					self->write_chunk_size = self->real_write_chunk_size;
				}

				if (d != self->write_chunk_size)
					g_object_notify (object, "write_chunk_size");

				if (self->write_chunk_size > d)
					_brf_add_idle (self);
			}

			} break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
burgner_ring_fifo_get_property (GObject *object,
	      guint prop_id,
	      GValue *value,
	      GParamSpec *pspec)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (object);

	switch (prop_id) {
		case PROP_BUFFER:
			g_value_set_pointer (value, self->buf);
			break;
		case PROP_SIZE:
			g_value_set_uint (value, self->buf_end - self->buf);
			break;
		case PROP_READ_CHUNK_SIZE:
			g_value_set_uint (value, self->read_chunk_size);
			break;
		case PROP_WRITE_CHUNK_SIZE:
			g_value_set_uint (value, self->write_chunk_size);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}



static void
burgner_ring_fifo_notify (GObject *object,
			  GParamSpec *pspec)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (object);

	if (pspec->owner_type != BURGNER_TYPE_FIFO)
		return;

	if (strcmp (pspec->name, "reader") == 0) {
		if (self->reading) {
			g_warning ("You can set the BurgnerFifoReader of "
				   "a BurgnerFifo only one time");
			return;
		}

		_brf_add_idle (self);
	} else if (strcmp (pspec->name, "writer") == 0) {
		if (self->writing) {
			g_warning ("You can set the BurgnerFifoWriter of "
				   "a BurgnerFifo only one time");
			return;
		}

		_brf_add_idle (self);
	}
}


static guint
_burgner_ring_fifo_to_read (BurgnerFifo *fifo, guchar **buf)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);
	guint r = 0;

	g_return_val_if_fail (self->buf != NULL, 0);

	if (buf)
		*buf = self->to_read;

	_brf_add_idle (self);

	if (self->reading) {
		r = self->reading;
	}

	return r;
}

static void
_burgner_ring_fifo_read (BurgnerFifo *fifo, guint size)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);

	g_return_if_fail (self->buf != NULL);
	g_return_if_fail (size <= self->reading);

	self->to_read += size;
	if (self->to_read == self->buf_end)
		self->to_read = self->buf;

	self->reading = 0;

	if (self->to_read == self->to_write && size != 0)
		self->equal_case = TO_WRITE;

	_brf_add_idle (self);
}

static guint
_burgner_ring_fifo_to_write (BurgnerFifo *fifo, guchar **buf)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);
	guint r = 0;

	g_return_val_if_fail (self->buf != NULL, 0);

	if (buf)
		*buf = self->to_write;

	_brf_add_idle (self);

	if (self->writing) {
		r = self->writing;
	}

	return r;
}

static void
_burgner_ring_fifo_wrote (BurgnerFifo *fifo, guint size)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);

	g_return_if_fail (self->buf != NULL);
	g_return_if_fail (size <= self->writing);

	burgner_fifo_do_filter (fifo, self->to_write, size);

	self->to_write += size;
	if (self->to_write == self->buf_end)
		self->to_write = self->buf;

	self->writing = 0;

	if (self->to_write == self->to_read && size != 0)
		self->equal_case = TO_READ;

	_brf_add_idle (self);
}

static void
_burgner_ring_fifo_end (BurgnerFifo *fifo)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);

	self->writing = 0;

	_brf_add_idle (self);

	if (parent_class->end)
		parent_class->end (fifo);
}

static void
_burgner_ring_fifo_reset (BurgnerFifo *fifo)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);

	self->to_read = self->to_write = self->buf;
	self->reading = self->writing = 0;
	self->equal_case = TO_WRITE;
	if (self->idle_id) {
		gboolean ret = g_source_remove (self->idle_id);
		g_return_if_fail (ret != FALSE);
		self->idle_id = 0;
	}

	if (parent_class->reset)
		parent_class->reset (fifo);
}

static void
_burgner_ring_fifo_set_data_block_size (BurgnerFifo *fifo,
					guint data_block_size)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);

	if (self->data_block_size == data_block_size)
		return;

	g_return_if_fail (self->to_read == self->buf);
	g_return_if_fail (self->to_write == self->buf);
	g_return_if_fail (self->equal_case == TO_WRITE);
	g_return_if_fail (self->reading == 0);
	g_return_if_fail (self->writing == 0);

	self->data_block_size = data_block_size;

	if (self->real_buf_end)
		self->buf_end = self->real_buf_end - ((self->real_buf_end - self->buf) % self->data_block_size);

	if (self->real_read_chunk_size)
		self->read_chunk_size = self->real_read_chunk_size - (self->real_read_chunk_size % self->data_block_size);

	if (self->real_write_chunk_size)
		self->write_chunk_size = self->real_write_chunk_size - (self->real_write_chunk_size % self->data_block_size);

	_brf_add_idle (self);
}

static guint
_burgner_ring_fifo_get_percent (BurgnerFifo *fifo)
{
	BurgnerRingFifo *self = BURGNER_RING_FIFO (fifo);
	guint l = self->buf_end - self->buf;
	guint f;

	if (self->to_write > self->to_read) {
		f = self->to_write - self->to_read;
	} else if (self->to_write < self->to_read) {
		f = l - (self->to_read - self->to_write);
	} else if (self->equal_case == TO_WRITE) {
		return 0;
	} else {
		return 100;
	}

	return f * 100 / l;
}

static void
_burgner_ring_fifo_wake_up (BurgnerFifo *fifo)
{
	_brf_add_idle (BURGNER_RING_FIFO (fifo));
}

static void
burgner_ring_fifo_class_init (BurgnerRingFifoClass *klass)
{
	GObjectClass *object_class;
	BurgnerFifoClass *fifo_class;

	object_class = (GObjectClass *) klass;
	object_class->notify = burgner_ring_fifo_notify;
	object_class->set_property = burgner_ring_fifo_set_property;
	object_class->get_property = burgner_ring_fifo_get_property;
	object_class->finalize = burgner_ring_fifo_finalize;

	fifo_class = (BurgnerFifoClass *) klass;
	fifo_class->to_read = _burgner_ring_fifo_to_read;
	fifo_class->read = _burgner_ring_fifo_read;
	fifo_class->to_write = _burgner_ring_fifo_to_write;
	fifo_class->wrote = _burgner_ring_fifo_wrote;
	fifo_class->end = _burgner_ring_fifo_end;
	fifo_class->reset = _burgner_ring_fifo_reset;
	fifo_class->set_data_block_size = _burgner_ring_fifo_set_data_block_size;
	fifo_class->get_percent = _burgner_ring_fifo_get_percent;
	fifo_class->wake_up = _burgner_ring_fifo_wake_up;

	/* Properties */
	g_object_class_install_property (object_class,
					 PROP_BUFFER,
					 g_param_spec_pointer ("buffer",
							       _("buffer"),
							       _("Buffer"),
							       G_PARAM_READABLE));
	g_object_class_install_property (object_class,
					 PROP_SIZE,
					 g_param_spec_uint ("size",
					                _("size"),
							_("Size of internal buffer"),
							0,
							0xffffffff,
							0x2000, /* 8 kB */
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_READ_CHUNK_SIZE,
					 g_param_spec_uint ("read_chunk_size",
					                _("read_chunk_size"),
							_("Size of a chunk to be read"),
							0,
							0xffffffff,
							0x2000, /* 8 kB */
							G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (object_class,
					 PROP_WRITE_CHUNK_SIZE,
					 g_param_spec_uint ("write_chunk_size",
					                _("write_chunk_size"),
							_("Size of a chunk to be written"),
							0,
							0xffffffff,
							0x2000, /* 8 kB */
							G_PARAM_READABLE | G_PARAM_WRITABLE));


	parent_class = g_type_class_peek_parent (klass);
}

static void
burgner_ring_fifo_init (BurgnerRingFifo *self)
{
	self->buf = self->buf_end = self->to_read = self->to_write = NULL;
	self->real_buf_end = NULL;
	self->read_chunk_size = self->write_chunk_size = 0x2000;
	self->real_read_chunk_size = self->real_write_chunk_size = 0x2000;
	self->data_block_size = 0;
	self->equal_case = TO_WRITE;
	self->reading = self->writing = 0;
	self->idle_id = 0;
}

GType
burgner_ring_fifo_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerRingFifoClass),
			NULL,
			NULL,
			(GClassInitFunc) burgner_ring_fifo_class_init,
			NULL,
			NULL,
			sizeof (BurgnerRingFifo),
			0,
			(GInstanceInitFunc) burgner_ring_fifo_init,
		};

		type = g_type_register_static (BURGNER_TYPE_FIFO, "BurgnerRingFifo", &info, 0);
	}

	return type;
}

BurgnerFifo *
burgner_ring_fifo_new (void)
{
	BurgnerRingFifo *self;

	self = g_object_new (BURGNER_TYPE_RING_FIFO, NULL);

	return BURGNER_FIFO (self);
}


GQuark
burgner_ring_fifo_error_quark (void)
{
	return g_quark_from_string ("BurgnerRingFifo");
}

