/*
 * BurgnerThread - BurgnerThread object
 *
 * Copyright (c) 2003 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "burgner-utils-i18n.h"

#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <glib-object.h>
#include "burgner-thread.h"


struct _BurgnerThreadCall {
	guint o;

	BurgnerThreadFunction function;
	gpointer function_data;

	BurgnerThreadCallback callback;
	gpointer callback_data;
};

static GMemChunk *_burgner_thread_call_mem_chunk = NULL;
G_LOCK_DEFINE (_burgner_thread_call_mem_chunk);

static _BurgnerThreadCall *
_burgner_thread_call_new (BurgnerThreadFunction function,
			  gpointer function_data,
			  BurgnerThreadCallback callback,
			  gpointer callback_data)
{
	_BurgnerThreadCall *c;

	G_LOCK (_burgner_thread_call_mem_chunk);
	if (!_burgner_thread_call_mem_chunk) {
		_burgner_thread_call_mem_chunk = g_mem_chunk_new
					("_BurgnerThreadCall",
					 sizeof (_BurgnerThreadCall),
					 3 * sizeof (_BurgnerThreadCall),
					 G_ALLOC_AND_FREE);
	}

	c = g_mem_chunk_alloc (_burgner_thread_call_mem_chunk);
	G_UNLOCK (_burgner_thread_call_mem_chunk);

	c->o = 0x12345678;
	c->function = function;
	c->function_data = function_data;
	c->callback = callback;
	c->callback_data = callback_data;

	return c;
}

static void
_burgner_thread_call_free (_BurgnerThreadCall *c)
{
	g_return_if_fail (c != NULL);

	G_LOCK (_burgner_thread_call_mem_chunk);
	g_mem_chunk_free (_burgner_thread_call_mem_chunk, c);
	G_UNLOCK (_burgner_thread_call_mem_chunk);
}


static GObjectClass *parent_class = NULL;

static void
_burgner_thread_finalize (GObject *object)
{
	BurgnerThread *self = BURGNER_THREAD (object);

	if (self->thread)
		self->thread = NULL;

	if (self->pipe_watch_id) {
		g_source_remove (self->pipe_watch_id);
		self->pipe_watch_id = 0;
	}

	if (self->pipe_channel) {
		g_io_channel_unref (self->pipe_channel);
		self->pipe_channel = NULL;
	}

	if (self->rpipe_fd) {
		close (self->rpipe_fd);
		self->rpipe_fd = -1;
	}

	if (self->wpipe_fd) {
		close (self->wpipe_fd);
		self->wpipe_fd = -1;
	}

	if (self->call_list) {
		g_list_foreach (self->call_list,
				(GFunc) _burgner_thread_call_free,
				NULL);
		g_list_free (self->call_list);
		self->call_list = NULL;
	}

	if (self->thread_busy)
		self->thread_busy = 0;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
_burgner_thread_class_init (BurgnerThreadClass *klass)
{
	GObjectClass *object_class;

	object_class = (GObjectClass *) klass;
	object_class->finalize = _burgner_thread_finalize;

	parent_class = g_type_class_peek_parent (klass);
}

static void
_burgner_thread_init (BurgnerThread *self)
{
	self->thread = NULL;

	self->rpipe_fd = -1;
	self->wpipe_fd = -1;
	self->pipe_channel = NULL;
	self->pipe_watch_id = 0;

	self->actual_call = NULL;
	self->call_list = NULL;

	self->thread_busy = 0;
}

GType
burgner_thread_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static GTypeInfo info = {
			sizeof (BurgnerThreadClass),
			NULL,
			NULL,
			(GClassInitFunc) _burgner_thread_class_init,
			NULL,
			NULL,
			sizeof (BurgnerThread),
			0,
			(GInstanceInitFunc) _burgner_thread_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, "BurgnerThread", &info, 0);
	}

	return type;
}

typedef struct {
	gint r;
	gint w;
} _BurgnerThreadThreadPipes;

static gpointer
_burgner_thread_thread_func (gpointer data)
{
	_BurgnerThreadCall *c = NULL;
	_BurgnerThreadThreadPipes *pipes = (_BurgnerThreadThreadPipes *) data;
	gint rfd = pipes->r;
	gint wfd = pipes->w;

	g_assert (rfd >= 0);
	g_assert (wfd >= 0);

	g_free (pipes);

	while (read (rfd, &c, sizeof (gpointer)) == sizeof (gpointer)) {
		g_assert (c != NULL);
		g_assert (c->function != NULL);

		while (c->function (c->function_data));
		if (write (wfd, &c, sizeof (gpointer)) == sizeof (gpointer)) {
			continue;
		}

		g_warning ("BurgnerThread thread: write failed: %s", strerror (errno));
		break;
	}

	close (rfd);
	close (wfd);

	return NULL;
}

static gboolean
_burgner_thread_pipe_watch_func (GIOChannel *channel,
				 GIOCondition condition,
				 gpointer data)
{
	BurgnerThread *self = BURGNER_THREAD (data);

	g_assert (condition == G_IO_IN);

	if (self->actual_call) {
		_BurgnerThreadCall *c = NULL;

		if (read (self->rpipe_fd, &c, sizeof (gpointer)) != sizeof (gpointer)) {
			g_warning ("BurgnerThread: read failed: %s", strerror (errno));
			return FALSE;
		}
		g_assert (c == self->actual_call);

		g_object_ref (self);
		if (!c->callback ||
		    !(c->callback (self, c, c->function, c->function_data, c->callback_data))) {
			self->call_list = g_list_remove (self->call_list, c);
			_burgner_thread_call_free (c);
			if (self->call_list) {
				self->actual_call = self->call_list->data;
				self->call_list = g_list_delete_link (self->call_list, self->call_list);
			} else {
				self->actual_call = NULL;
			}
		}

		if (self->actual_call) {
			if (write (self->wpipe_fd, &self->actual_call, sizeof (gpointer)) != sizeof (gpointer)) {
				g_warning ("BurgnerThread: write failed: %s", strerror (errno));
			}
			g_object_unref (self);
			return TRUE;
		} else {
			self->pipe_watch_id = 0;
			g_object_unref (self);
			return FALSE;
		}
	}

	self->pipe_watch_id = 0;
	return FALSE;
}

BurgnerThread *
burgner_thread_new (GError **error)
{
	BurgnerThread *self;
	GThread *thread;
	_BurgnerThreadThreadPipes *pipes;
	gint r[2] = { 0, 0 }, w[2] = { 0, 0 };

	if (pipe (r) != 0 ||
	    pipe (w) != 0) {
		g_set_error (error,
			     BURGNER_THREAD_ERROR,
			     BURGNER_THREAD_ERROR_CREATING_PIPE,
			     _("Error while creating pipe: %s"),
			     	strerror (errno));
		return NULL;
	}

	pipes = g_new (_BurgnerThreadThreadPipes, 1);
	pipes->r = w[0];
	pipes->w = r[1];

	if (!(thread = g_thread_create (_burgner_thread_thread_func,
					pipes,
					FALSE, error))) {
		g_free (pipes);
		close (r[0]);
		close (r[1]);
		close (w[0]);
		close (w[1]);
		return NULL;
	}

	self = g_object_new (BURGNER_TYPE_THREAD, NULL);

	self->rpipe_fd = r[0];
	self->wpipe_fd = w[1];

	self->thread = thread;
	self->pipe_channel = g_io_channel_unix_new (self->rpipe_fd);

	return self;
}


GQuark
burgner_thread_error_quark (void)
{
	return g_quark_from_string ("BurgnerThread");
}

BurgnerThreadCallId
burgner_thread_call (BurgnerThread *thread,
		     BurgnerThreadFunction function,
		     gpointer function_data,
		     BurgnerThreadCallback callback,
		     gpointer callback_data)
{
	_BurgnerThreadCall *c;

	g_return_val_if_fail (thread != NULL, NULL);
	g_return_val_if_fail (BURGNER_IS_THREAD (thread), NULL);
	g_return_val_if_fail (thread->thread != NULL, NULL);
	g_return_val_if_fail (thread->rpipe_fd >= 0, NULL);
	g_return_val_if_fail (thread->wpipe_fd >= 0, NULL);
	g_return_val_if_fail (thread->pipe_channel != NULL, NULL);
	g_return_val_if_fail (function != NULL, NULL);

	c = _burgner_thread_call_new (function, function_data,
				      callback, callback_data);

	if (!thread->actual_call) {
		g_assert (thread->call_list == NULL);
		g_assert (thread->pipe_watch_id == 0);

		if (write (thread->wpipe_fd, &c, sizeof (gpointer))
						!= sizeof (gpointer)) {
			g_warning ("BurgnerThread: write failed: %s\n",
					strerror (errno));
			_burgner_thread_call_free (c);
			return NULL;
		} else {
			thread->actual_call = c;
			thread->pipe_watch_id = g_io_add_watch (thread->pipe_channel,
							   G_IO_IN,
							   _burgner_thread_pipe_watch_func,
							   thread);
		}
	} else {
		g_assert (thread->pipe_watch_id != 0);
		thread->call_list = g_list_append (thread->call_list, c);
	}

	return c;
}

