/*
 * 
 *
 * This file is part of the libburgnerscsi library.
 *
 * Copyright (c) 2002 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include "g-malloc-saver.h"

GMallocSaver *
g_malloc_saver_new (gulong chunk_size,
		    gint max_chunks,
		    gboolean do_0)
{
	GMallocSaver *saver;

	g_return_val_if_fail (chunk_size > 0, NULL);

	saver = g_new0 (GMallocSaver, 1);
	saver->refcount = 1;
	saver->chunks = NULL;
	saver->max_chunks = max_chunks;
	saver->chunk_size = chunk_size;
	saver->do_0 = (do_0 == TRUE);

	return saver;
}

void
g_malloc_saver_free (GMallocSaver *saver)
{
	g_return_if_fail (saver != NULL);

	if (saver->chunks) {
		g_slist_foreach (saver->chunks, (GFunc) g_free, NULL);
		g_slist_free (saver->chunks);
	}

	g_free (saver);
}


GMallocSaver *
g_malloc_saver_ref (GMallocSaver *saver)
{
	g_return_val_if_fail (saver != NULL, NULL);

	saver->refcount += 1;
	return saver;
}

void
g_malloc_saver_unref (GMallocSaver *saver)
{
	g_return_if_fail (saver != NULL);

	saver->refcount -= 1;

	if (saver->refcount == 0)
		g_malloc_saver_free (saver);
}


gpointer
g_malloc_saver_get_chunk (GMallocSaver *saver)
{
	gpointer c;

	g_return_val_if_fail (saver != NULL, NULL);

	if (saver->chunks) {
		c = saver->chunks->data;
		saver->chunks = g_slist_delete_link (saver->chunks,
						     saver->chunks);
	} else {
		c = g_malloc (saver->chunk_size);
	}

	if (saver->do_0)
		memset (c, 0, (size_t) saver->chunk_size);

	return c;
}

gpointer
g_malloc_saver_try_get_chunk (GMallocSaver *saver)
{
	gpointer c;

	g_return_val_if_fail (saver != NULL, NULL);

	if (saver->chunks) {
		c = saver->chunks->data;
		saver->chunks = g_slist_delete_link (saver->chunks,
						     saver->chunks);
	} else {
		c = g_try_malloc (saver->chunk_size);
		if (!c)
			return NULL;
	}

	if (saver->do_0)
		memset (c, 0, (size_t) saver->chunk_size);

	return c;
}

gpointer
g_malloc_saver_get_chunk_dont_malloc (GMallocSaver *saver)
{
	gpointer c;

	g_return_val_if_fail (saver != NULL, NULL);

	if (saver->chunks) {
		c = saver->chunks->data;
		saver->chunks = g_slist_delete_link (saver->chunks,
						     saver->chunks);
	} else {
		return NULL;
	}

	if (saver->do_0)
		memset (c, 0, (size_t) saver->chunk_size);

	return c;
}


void
g_malloc_saver_take_chunk (GMallocSaver *saver,
			   gpointer chunk)
{
	g_return_if_fail (saver != NULL);
	g_return_if_fail (chunk != NULL);

	if (g_slist_length (saver->chunks) >= saver->max_chunks) {
		g_free (chunk);
	} else {
		saver->chunks = g_slist_prepend (saver->chunks, chunk);
	}
}


gint
g_malloc_saver_max_chunks_add (GMallocSaver *saver,
			       gint n)
{
	g_return_val_if_fail (saver != NULL, 0);

	return (saver->max_chunks += n);
}

gint
g_malloc_saver_set_max_chunks (GMallocSaver *saver,
			       gint max_chunks)
{
	g_return_val_if_fail (saver != NULL, 0);

	return (saver->max_chunks = max_chunks);
}

