/*
 * GUri Dynamic Parser
 *
 * Copyright (c) 2001 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <libintl.h>
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) x

#include <string.h>
#include <glib.h>
#include "g-uri.h"
#include "g-uri-char-utils.h"

static inline gchar *
_g_uri_parser_dynamic_parse_fragment (gchar **uri,
					gboolean validate,
					GError **error)
{
	gchar *s = strchr (*uri, '#'), *f;
	if (!s)
		return NULL;

	*(s++) = '\0';

	f = g_strdup (s);

	if (validate) {
		/*
		 * RFC2396 (line 1510):
		 * 	fragment      = *uric
		 *
		 * RFC2396 (line 1519):
		 *	uric          = reserved | unreserved | escaped
		 *
		 * RFC2396 (line 1526):
		 *	escaped       = "%" hex hex
		 */
		gchar *c;

		for (c = f; *c != 0; c += 1) {
		        if (IS_URIC (*c)) {
			         continue;
			}

			if (*c != '%') {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INVALID_FRAGMENT,
					     _("Fragment contains an illegal character (%c)"),
					     *c);
				g_free (f);
				return NULL;
			}

			if (!*(c + 1) || !*(c + 2)) {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INCOMPLETE_ESCAPED_CHAR,
					     _("Fragment contains an incomplete escaped char"));
					     g_free (f);
					     return NULL;
			}

			if (!IS_HEX (*(c + 1)) || !IS_HEX (*(c + 2))) {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INVALID_ESCAPED_CHAR,
					     _("Fragment contains an invalid escaped char"));
				g_free (f);
				return NULL;
			}

			c += 2;
		}
	}

	return f;
}

static inline gchar *
_g_uri_parser_dynamic_parse_scheme (gchar **uri,
					gboolean validate,
					GError **error)
{
	gchar *it;
	gchar *scheme;

	for (it = *uri; *it != '\0'; it++) {
		if (it == *uri) {
			if (IS_ALPHA (*it))
				continue;

			if (*it == ':') {
				*it = 0;
				*uri += 1;
				scheme = g_strdup (it);
				return scheme;
			}

			return NULL;
		}

		if (IS_ALPHANUM (*it) ||
		    (*it == '+') || (*it == '-') ||
		    (*it == '.'))
			continue;

		if (*it == ':') {
			scheme = g_strndup (*uri, it - *uri);
			*uri = it + 1;
			return scheme;
		}
		break;
	}

	/* No scheme ... */

	return NULL;
}

static inline gchar *
_g_uri_parser_dynamic_parse_authority (gchar **uri,
					gboolean validate,
					GError **error)
{
	gchar *authority;
	if (**uri == '/' && *(*uri + 1) == '/') {
		gchar *sl = *uri + 2;
		if ((sl = strpbrk (sl, "/?"))) {
			authority = g_strndup (*uri + 2, sl - (*uri + 2));
			*uri = sl;
		} else {
			authority = g_strdup (*uri + 2);
			while (**uri)	*uri += 1;
		}

		return authority;
	}
	return NULL;
}

static inline gchar *
_g_uri_parser_dynamic_parse_query (gchar **uri,
					gboolean validate,
					GError **error)
{
	gchar *q = strchr (*uri, '?'), *query;
	if (!q)
		return NULL;

	*q = 0;
	query = g_strdup (++q);

	if (validate) {
		/*
		 * RFC2396 (line 1508):
		 * 	query         = *uric
		 *
		 * RFC2396 (line 1519):
		 *	uric          = reserved | unreserved | escaped
		 *
		 * RFC2396 (line 1526):
		 *	escaped       = "%" hex hex
		 */
		gchar *c = query;

		for (; *c != 0; c++) {
		        if (IS_URIC (*c)) {
			         continue;
			}

			if (*c != '%') {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INVALID_QUERY,
					     _("Query contains an illegal character (%c)"),
					     *c);
				g_free (query);
				return NULL;
			}

			if (!*(c + 1) || !*(c + 2)) {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INCOMPLETE_ESCAPED_CHAR,
					     _("Query contains an incomplete escaped char"));
				g_free (query);
				return NULL;
			}

			if (!IS_HEX (*(c + 1)) || !IS_HEX (*(c + 2))) {
				g_set_error (error,
					     G_URI_ERROR,
					     G_URI_ERROR_INVALID_ESCAPED_CHAR,
					     _("Query contains an invalid escaped char"));
				g_free (query);
				return NULL;
			}

			c += 2;
		}
	}

	return query;
}

static inline gchar *
_g_uri_parser_dynamic_parse_path (gchar **uri,
				      gboolean *is_absolute,
					gboolean validate,
					GError **error)
{
	/* path can't be undefined */
#if 0
	if (**uri == '\0') {
		return NULL;
	}
#endif

	if (**uri == '/' || **uri == '\0')
		*is_absolute = TRUE;
	else
		*is_absolute = FALSE;

	return g_strdup (*uri);
}

gboolean
_g_uri_parser_dynamic_parse_uri (gchar *c_uri,
				     gchar **scheme,
				     gchar **authority,
				     gchar **path,
				     gboolean *is_absolute,
				     gchar **query,
				     gchar **fragment,
				     gboolean validate,
				     GError **error)
{
	GError *err = NULL;
	gchar *foo = NULL, **uri = &c_uri;
	gboolean foo_b = FALSE;

	g_return_val_if_fail (c_uri != NULL, FALSE);


#define SET(a,b)	G_STMT_START {		\
	if (a)	*a = b; else g_free (b);	\
			} G_STMT_END

	foo = _g_uri_parser_dynamic_parse_fragment (uri, validate, &err);
	if (!foo && err) {
		g_free (foo);
		g_propagate_error (error, err);
		return FALSE;
	}
	SET (fragment, foo);

	foo = _g_uri_parser_dynamic_parse_scheme (uri, validate, &err);
	if (!foo && err) {
		g_free (foo);
		g_propagate_error (error, err);
		return FALSE;
	}
	SET (scheme, foo);

	foo = _g_uri_parser_dynamic_parse_authority (uri, validate, &err);
	if (!foo && err) {
		g_free (foo);
		g_propagate_error (error, err);
		return FALSE;
	}
	SET (authority, foo);

	foo = _g_uri_parser_dynamic_parse_query (uri, validate, &err);
	if (!foo && err) {
		g_free (foo);
		g_propagate_error (error, err);
		return FALSE;
	}
	SET (query, foo);

	foo = _g_uri_parser_dynamic_parse_path (uri, &foo_b, validate, &err);
	if (!foo && err) {
		g_free (foo);
		g_propagate_error (error, err);
		return FALSE;
	}
	SET (path, foo);
	if (is_absolute)
		*is_absolute = foo_b;

#undef SET

	return TRUE;
}

