/*
 * GUri - GUri struct
 *
 * Copyright (c) 2001 Alexandre Archambault
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *	Alexandre Archambault <aarchambault@free.fr>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib.h>
#include "g-uri.h"
#include "g-uri-parser-dynamic.h"
/*include "g-uri-parser-static.h"*/
#include "g-uri-path-parser.h"
#include "g-uri-resolver.h"


#define SELF(o)			(G_URI (o))
typedef GUri Self;


#if 1
#define G_URI_USE_MEMCHUNK
#define G_URI_CHUNK_SIZE		1024
#endif


#define check(o)		g_return_if_fail ((o) != NULL);
#define check_with_val(o,v)	g_return_val_if_fail ((o) != NULL, (v));



#ifdef G_URI_USE_MEMCHUNK

G_LOCK_DEFINE_STATIC (g_uri_mem_chunk);
GMemChunk *g_uri_mem_chunk = NULL;

#endif


GUri *
g_uri_new (void)
{
	GUri *self;

#ifndef G_URI_USE_MEMCHUNK
	self = g_new0 (GUri, 1);
#else
	G_LOCK (g_uri_mem_chunk);
	if (!g_uri_mem_chunk) {
		g_uri_mem_chunk = g_mem_chunk_new ("GUri mem chunk",
							  sizeof (GUri),
							  G_URI_CHUNK_SIZE,
							  G_ALLOC_AND_FREE);
	}
	self = g_chunk_new0 (GUri, g_uri_mem_chunk);
	G_UNLOCK (g_uri_mem_chunk);
#endif

	self->refcount = 1;

	return self;
}


void
g_uri_destroy (GUri *self)
{
	check (self)

#define FREE_IF_NEEDED(s)	G_STMT_START { \
	if (s) \
		g_free (s);	} G_STMT_END

	FREE_IF_NEEDED (self->scheme);
	FREE_IF_NEEDED (self->authority);
	FREE_IF_NEEDED (self->path);
	FREE_IF_NEEDED (self->query);
	FREE_IF_NEEDED (self->fragment);

	if (self->has_parsed_auth) {
		FREE_IF_NEEDED (self->parsed_authority.userinfo);
		FREE_IF_NEEDED (self->parsed_authority.host);
		FREE_IF_NEEDED (self->parsed_authority.port);
	}
#undef FREE_IF_NEEDED

#ifndef G_URI_USE_MEMCHUNK
	g_free (self);
#else
	G_LOCK (g_uri_mem_chunk);
	g_mem_chunk_free (g_uri_mem_chunk, self);
	G_UNLOCK (g_uri_mem_chunk);
#endif
}


GUri *
g_uri_ref (GUri *self)
{
	g_return_val_if_fail (self != NULL, NULL);

	self->refcount++;

	return self;
}

void
g_uri_unref (GUri *self)
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (self->refcount > 0);

	self->refcount--;

	if (self->refcount == 0) {
		g_uri_destroy (self);
	}
}


GUri *
g_uri_clone (const GUri *uri)
{
	GUri *n;
	/*gchar *s;*/

	g_return_val_if_fail (uri != NULL, NULL);

	n = g_uri_new ();

	n->scheme = g_strdup (uri->scheme);
	n->authority = g_strdup (uri->authority);
	n->path = g_strdup (uri->path);
	n->query = g_strdup (uri->query);
	n->fragment = g_strdup (uri->fragment);

	n->path_is_absolute = uri->path_is_absolute;

	if ((n->has_parsed_auth = uri->has_parsed_auth)) {
		n->parsed_authority.userinfo = g_strdup (uri->parsed_authority.userinfo);
		n->parsed_authority.host = g_strdup (uri->parsed_authority.host);
		n->parsed_authority.port = g_strdup (uri->parsed_authority.port);
	}

	return n;
}

gboolean
g_uri_is_void (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, TRUE);

	/* The fragment isn't considered as part of the uri */
	/* The uri isn't considered void even if *uri->{scheme,authority,query}
	 * are null, as if uri->{scheme,authority,query} are non-null the uri has
	 * contained  {":","//","?"}. */
	return ((!uri->path || *(uri->path) == 0) &&
		!uri->scheme &&
		!uri->authority &&
		!uri->query);
}

gchar *
g_uri_format (const GUri *uri,
		  GUriFormatOptions options)
{
	GString *s;
	gchar *ret;

	g_return_val_if_fail (uri != NULL, NULL);

	/* If authority is defined, path must be absolute */
	g_return_val_if_fail (uri->authority == NULL ||
			      uri->path_is_absolute, NULL);

	s = g_string_new (NULL);

	if (uri->scheme) {
		g_string_append (s, uri->scheme);
		g_string_append_c (s, ':');
	}

	if (uri->authority)
		g_string_append_printf (s, "//%s", uri->authority);

	if (!uri->path)
		g_string_append_c (s, '/');
		/*
		 * To be RFC1738 conform
		 */
	else
		/* If the path is absolute, it contains the beginning '/' */
		g_string_append (s, uri->path);

	if (uri->query) {
		g_string_append_c (s, '?');
		g_string_append (s, uri->query);
	}

	if (uri->fragment) {
		g_string_append_c (s, '#');
		g_string_append (s, uri->fragment);
	}

	ret = s->str;
	g_string_free (s, FALSE);

	return ret;
}

void
g_uri_reset (GUri *uri)
{
	g_return_if_fail (uri != NULL);

#define FREE_IF_NEEDED(s)	G_STMT_START { \
	if (s) { \
		g_free (s); \
		s = NULL; \
	}	} G_STMT_END

	FREE_IF_NEEDED (uri->scheme);
	FREE_IF_NEEDED (uri->authority);
	FREE_IF_NEEDED (uri->parsed_authority.userinfo);
	FREE_IF_NEEDED (uri->parsed_authority.host);
	FREE_IF_NEEDED (uri->parsed_authority.port);
	FREE_IF_NEEDED (uri->path);
	FREE_IF_NEEDED (uri->query);
	FREE_IF_NEEDED (uri->fragment);
#undef FREE_IF_NEEDED

	uri->has_parsed_auth = 0;
	uri->path_is_absolute = 0;
}

gboolean
g_uri_set (GUri *uri,
	   gchar *str,
	   gboolean validate,
	   gboolean parse_auth,
	   GError **error)
{
	gchar *c_str;
	gboolean ret, path_is_absolute;

	g_return_val_if_fail (str != NULL, FALSE);
	g_return_val_if_fail (uri != NULL, FALSE);

	path_is_absolute = uri->path_is_absolute;

	c_str = g_strdup (str);
	ret = _g_uri_parser_dynamic_parse_uri (c_str,
						&uri->scheme,
						&uri->authority,
						&uri->path,
						&path_is_absolute,
						&uri->query,
						&uri->fragment,
						validate,
						error);
	g_free (c_str);

	uri->path_is_absolute = path_is_absolute;

	if (!ret)
		return FALSE;

	if (parse_auth) {
		ret = g_uri_parse_authority (uri, error);
	}

	return ret;
}

void
g_uri_set_scheme (GUri *uri,
		      const gchar *scheme)
{
	g_return_if_fail (uri != NULL);

	if (uri->scheme)
		g_free (uri->scheme);

	uri->scheme = g_strdup (scheme);
}

void
g_uri_set_authority (GUri *uri,
		       const gchar *authority)
{
	g_return_if_fail (uri != NULL);

	if (uri->authority)
		g_free (uri->authority);

	uri->authority = g_strdup (authority);
}

void
g_uri_set_path (GUri *uri,
		    const gchar *path)
{
	g_return_if_fail (uri != NULL);

	if (uri->path)
		g_free (uri->path);

	uri->path = g_strdup (path);
}

void
g_uri_set_query (GUri *uri,
		     const gchar *query)
{
	g_return_if_fail (uri != NULL);

	if (uri->query)
		g_free (uri->query);

	uri->query = g_strdup (query);
}

void
g_uri_set_fragment (GUri *uri,
			const gchar *fragment)
{
	g_return_if_fail (uri != NULL);

	if (uri->fragment)
		g_free (uri->fragment);

	uri->fragment = g_strdup (fragment);
}

const gchar *
g_uri_get_scheme (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, NULL);
	return uri->scheme;
}

const gchar *
g_uri_get_authority (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, NULL);
	return uri->authority;
}

const gchar *
g_uri_get_path (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, NULL);
	return uri->path;
}

const gchar *
g_uri_get_query (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, NULL);
	return uri->query;
}

const gchar *
g_uri_get_fragment (const GUri *uri)
{
	g_return_val_if_fail (uri != NULL, NULL);
	return uri->fragment;
}

gboolean
g_uri_parse_authority (GUri *uri,
		       GError **error)
{
	gchar *chr, *auth;

	g_return_val_if_fail (uri != NULL, FALSE);


	if (uri->parsed_authority.userinfo)
		g_free (uri->parsed_authority.userinfo);

	if (uri->parsed_authority.host)
		g_free (uri->parsed_authority.host);

	if (uri->parsed_authority.port)
		g_free (uri->parsed_authority.port);


	if (!uri->authority || !*uri->authority)
		return TRUE;

	auth = uri->authority;

	if ((chr = strchr (auth, '@'))) {
		uri->parsed_authority.userinfo = g_strndup (auth, chr - auth);
		auth = chr + 1;
	}

	if ((chr = strchr (auth, ':'))) {
		uri->parsed_authority.host = g_strndup (auth, chr - auth);
		uri->parsed_authority.port = g_strdup (chr + 1);
	} else {
		uri->parsed_authority.host = g_strdup (auth);
	}

	/*
	 * FIXME: validate userinfo, host, and port as:
	 * 
 	 *      userinfo      = *( unreserved | escaped |
 	 *                         ";" | ":" | "&" | "=" | "+" | "$" | "," )
 	 *      hostport      = host [ ":" port ]
 	 *      host          = hostname | IPv4address
 	 *      hostname      = *( domainlabel "." ) toplabel [ "." ]
 	 *      domainlabel   = alphanum | alphanum *( alphanum | "-" ) alphanum
 	 *      toplabel      = alpha | alpha *( alphanum | "-" ) alphanum
 	 *      IPv4address   = 1*digit "." 1*digit "." 1*digit "." 1*digit
 	 *      port          = *digit
	 *
	 * and set *error if needed.
	 */

	return TRUE;
}

gboolean
g_uri_relative_resolve (GUri **uri,
			const GUri *r,
			const GUri *base,
			GError **err)
{
	GUri *n;

	g_return_val_if_fail (r != NULL, FALSE);
	g_return_val_if_fail (base != NULL, FALSE);

	n = g_uri_clone (r);

	if (!(_g_uri_relative_resolver_step_2 (n, base) ||
	      _g_uri_relative_resolver_step_3 (n, base) ||
	      _g_uri_relative_resolver_step_4 (n, base) ||
	      _g_uri_relative_resolver_step_5 (n, base)))
		_g_uri_relative_resolver_step_6 (n, base);

	if (uri)
		*uri = n;
	else
		g_uri_unref (n);

	return TRUE;
}


GQuark
g_uri_error_quark (void)
{
	return g_quark_from_string ("GUri");
}

