
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <libgnomevfs/gnome-vfs.h>

#include "../libburgnerscsi/burgner-scsi-init.h"

#include "../libburgnerscsi/burgner-scsi-driver-manager.h"
#include "../libburgnerscsi/burgner-scsi-driver-info.h"
#include "../libburgnerscsi/burgner-scsi-driver.h"
#include "../libburgnerscsi/burgner-scsi-device.h"

#include "../libburgnerscsi/burgner-scsi-inquiry.h"
#include "../libburgnerscsi/burgner-scsi-prevent-allow-medium-removal.h"

#include "../libburgnermmc/burgner-mmc-blank.h"
#include "../libburgnermmc/burgner-mmc-close-track-session.h"

#include "../libburgnermmc/burgner-mmc-read-track-information.h"
#include "../libburgnermmc/burgner-mmc-impl.h"
#include "../libburgnerscsi/burgner-scsi-utils.h"


static struct option long_options[] = {
	{ "list-drivers", 0, NULL, 0 },
	{ "list-devices", 1, NULL, 0 },
	{ "list-all-devices", 0, NULL, 0 },
	{ "device", 1, NULL, 'd' },
	{ "devices-do-inquiry", 0, NULL, 0 },
	{ "inquiry", 0, NULL, 0 },
	{ "lock-tray", 0, NULL, 0 },
	{ "unlock-tray", 0, NULL, 0 },
	{ "blank", 1, NULL, 0 },
	{ "help", 0, NULL, 'h' },
	{ "directory", 1, NULL, 0 },
	{ "close-track", 0, NULL, 0 },
	{ "get-nwa", 0, NULL, 0 },
	{ NULL, 0, NULL, 0 }
};

static gboolean list_drivers = FALSE;
static gboolean list_all_devices = FALSE;
static gboolean devices_do_inquiry = FALSE;
static gboolean inquiry = FALSE;
static gboolean lock_tray = FALSE;
static gboolean unlock_tray = FALSE;
static gboolean close_track = FALSE;
static gboolean get_nwa = FALSE;

static enum {
	NONE = 0,
	COMPLETE,
	MINIMAL
} blank = NONE;

static gboolean list_devices = FALSE;
static gchar *driver_id;

static gchar *device_uri;

static GList *directories = NULL;


static BurgnerScsiDriverManager *driver_manager = NULL;
static BurgnerScsiDriver *driver = NULL;
static BurgnerScsiDevice *device = NULL;

static BurgnerScsiPrivilege *privilege = NULL;


static void
print_help (void) {
	g_print ("burgner-scsi-test " VERSION "\n"
		 "\tOptions:\n"
		 "\t\t--help or -h    Print this help\n\n"
		 "\t\t--directory <directory>	Add <directory> to the driver directory list\n"
		 "\t\t--list-drivers    List all available drivers\n"
		 "\t\t--list-devices <driver>    List all devices from driver <driver>\n"
		 "\t\t--list-all-devices    List all devices from all drivers\n"
		 "\t\t--devices-do-inquiry    Do an inquiry on each listed device\n"
		 "\n"
		 "\t\t--device <device> or -d <device>    Specify the device to use (<device> must be an URI)\n"
		 "\t\t--inquiry    Do an inquiry on <device>\n"
		 "\t\t--close-track    Close track\n"
		 "\t\t--get-nwa    Get the Next Writable Address (NWA)\n"
		 "\t\t--lock-tray    Try to lock tray on <device>\n"
		 "\t\t--unlock-tray  Try to unlock tray on <device>\n"
		 "\t\t--blank <complete/minimal>    Try to blank completely or minimaly disc in <device>\n");
}

static void
do_list_drivers (void)
{
	GList *l;
	/*
	 * We musn't modify this list, this function returns the internal
	 * list of the driver manager
	 */

	g_print ("-> Drivers\n");

	for (l = burgner_scsi_driver_manager_get_driver_info_list (driver_manager);
	     l; l = l->next) {
		BurgnerScsiDriverInfo *info = (BurgnerScsiDriverInfo *) l->data;

		g_print ("*** %s:\n"
			 "\tid: %s\n"
			 "\tname: %s\n"
			 "\tversion: %s\n"
			 "\tdescription: %s\n"
			 "\tauthors:\n",
			 	info->name ? info->name : "[no name]",
			 	info->id ? info->id : "[no name]",
			 	info->name ? info->name : "[no name]",
			 	info->version ? info->version : "[no name]",
			 	info->description ? info->description : "[no name]");

		if (info->authors && *info->authors) {
			guint i;
			for (i = 0; info->authors[i]; i++) {
				g_print ("\t\t%s\n", info->authors[i]);
			}
		} else {
			g_print ("\t\t[no authors]\n");
		}

		g_print ("\tdriver type: %s\n"
			 "\tneed privilege: %s\n\n",
			 	info->driver_type ? info->driver_type : "[no driver_type]",
				info->need_privilege ? "yes" : "no");
	}
}

static void
do_list_devices (void)
{
	BurgnerScsiDriver *driver = burgner_scsi_driver_manager_get_driver
						(driver_manager, driver_id);
	GList *uri_list, *it;
	GError *err = NULL;

	if (!driver) {
		fprintf (stderr, "Can't get driver %s\n", driver_id);
		return;
	}

	g_print ("-> %s's devices\n", driver_id);

	uri_list = burgner_scsi_driver_get_uri_list_sync (driver, &err);
	if (!uri_list && err) {
		fprintf (stderr, "Error while getting device list from driver %s: %s\n",
					driver_id,
					err->message ? err->message : "[no message]");
		exit (1);
	}

	for (it = uri_list; it != NULL; it = it->next) {
		gchar *s = g_uri_format ((GUri *) it->data,
					 G_URI_FORMAT_OPTIONS_NONE);
		g_print ("*** %s\n", s);
		g_free (s);
		g_uri_unref ((GUri *) it->data);
	}
	g_list_free (uri_list);
	g_object_unref (driver);
}

static void
do_list_all_devices (void)
{
	GList *drivers = burgner_scsi_driver_manager_get_all_drivers (driver_manager), *it;

	g_print ("-> Devices\n");

	for (it = drivers; it != NULL; it = it->next) {
		GList *uri_list, *uri_it;
		GError *err = NULL;

		uri_list = burgner_scsi_driver_get_uri_list_sync
				(BURGNER_SCSI_DRIVER (it->data),
				 &err);
		if (!uri_list && err) {
			fprintf (stderr, "Error while getting device list: %s\n",
						err->message ? err->message : "[no message]");
			g_object_unref (driver);
			exit (1);
		}

		for (uri_it = uri_list; uri_it != NULL; uri_it = uri_it->next) {
			gchar *s = g_uri_format ((GUri *) uri_it->data,
						 G_URI_FORMAT_OPTIONS_NONE);
			g_print ("*** %s\n", s);
			g_free (s);
			g_uri_unref (uri_it->data);
		}

		g_list_free (uri_list);
		g_object_unref (it->data);
	}

	g_list_free (drivers);
}

static void
init_driver_manager (void)
{
	GError *err = NULL;
	GList *it;

	g_assert (driver_manager == NULL);

	driver_manager = burgner_scsi_driver_manager_new ();
	g_assert (driver_manager != NULL);

	g_object_set (driver_manager,
		      "privilege", privilege,
		      NULL);

	directories = g_list_prepend (directories, LIBDIR "/burgner-0.1/libburgnerscsi/drivers"); 
	for (it = directories; it != NULL; it = it->next) {
		if (!burgner_scsi_driver_manager_load_driver_info_list_sync
				(driver_manager,
				 (gchar *) it->data,
				 NULL,
				 &err)) {
			fprintf (stderr,
				 "Error while loading drivers from %s: %s\n",
				 (gchar *) it->data,
				 err && err->message
				 	? err->message
					: "[No error or message]");
			g_error_free (err);
			err = NULL;
		} else if (err) {
			fprintf (stderr,
				 "Ignored error while loading drivers from %s: %s\n",
				 (gchar *) it->data,
				 err && err->message
				 	? err->message
					: "[No error or message]");
			g_error_free (err);
			err = NULL;
		}
	}
}

#define SENSE_BUF_SIZE		(18)
static guchar sense_buf[SENSE_BUF_SIZE] = { 0, };

#define CDB_BUF_SIZE		(12)
static guchar cdb_buf[CDB_BUF_SIZE] = { 0, };


static void
do_inquiry (void)
{
	static guchar data_buf[BURGNER_SCSI_INQUIRY_DATA_SIZE] = { 0, };
	GError *err = NULL;
	gchar *vendor = NULL, *product = NULL;
	guint status = 0;
	BurgnerScsiDeviceType dev_type = BURGNER_SCSI_DEVICE_TYPE_UNKNOWN;

	g_assert (BURGNER_SCSI_INQUIRY_CDB_SIZE <= CDB_BUF_SIZE);

	memset (&sense_buf, 0, SENSE_BUF_SIZE);
	memset (&cdb_buf, 0, BURGNER_SCSI_INQUIRY_CDB_SIZE);
	memset (&data_buf, 0, BURGNER_SCSI_INQUIRY_DATA_SIZE);

	burgner_scsi_inquiry_simple_write_cdb ((gpointer) cdb_buf,
					       BURGNER_SCSI_INQUIRY_CDB_SIZE,
					       BURGNER_SCSI_INQUIRY_DATA_SIZE);

	if (!burgner_scsi_device_send_cmd_sync (device,
						cdb_buf, BURGNER_SCSI_INQUIRY_CDB_SIZE,
						sense_buf, SENSE_BUF_SIZE, NULL,
						data_buf, BURGNER_SCSI_INQUIRY_DATA_SIZE, NULL,
						BURGNER_SCSI_DIRECTION_FROM_DEVICE,
						1000,
						&status, &err)) {
		fprintf (stderr, "Error while asking inquiry to %s\n", device_uri);
		exit (1);
	}

	burgner_scsi_inquiry_parse_data (data_buf, BURGNER_SCSI_INQUIRY_DATA_SIZE,
					 &dev_type,
					 &vendor, &product);

	g_print ("-> Inquiry\n"
		 "device type: %s\n"
		 "vendor: '%s'\n"
		 "product: '%s'\n",
		 	burgner_scsi_device_type_to_string (dev_type),
			vendor ? vendor : "[unknown]",
			product ? product : "[unknown]");
}

static void
do_close_track (void)
{
	GError *err = NULL;
	guint status = 0;

	g_assert (BURGNER_MMC_CLOSE_TRACK_SESSION_CDB_SIZE <= CDB_BUF_SIZE);

	memset (&sense_buf, 0, SENSE_BUF_SIZE);
	memset (&cdb_buf, 0, BURGNER_MMC_CLOSE_TRACK_SESSION_CDB_SIZE);

	burgner_mmc_close_track_write_cdb (cdb_buf, BURGNER_MMC_CLOSE_TRACK_SESSION_CDB_SIZE,
					   BURGNER_MMC_CLOSE_TRACK_LAST_TRACK_NUM, FALSE);

	if (!burgner_scsi_device_send_cmd_sync (device,
						cdb_buf, BURGNER_MMC_CLOSE_TRACK_SESSION_CDB_SIZE,
						sense_buf, SENSE_BUF_SIZE, NULL,
						NULL, 0, NULL,
						BURGNER_SCSI_DIRECTION_NONE,
						400000,
						&status, &err)) {
		fprintf (stderr, "Error while closing track\n");
		exit (1);
	}
}

static void
do_lock_unlock_tray (gboolean lock)
{
	GError *err = NULL;
	guint status = 0;

	g_assert (BURGNER_SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL_CDB_SIZE <= CDB_BUF_SIZE);

	memset (&sense_buf, 0, SENSE_BUF_SIZE);
	memset (&cdb_buf, 0, BURGNER_SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL_CDB_SIZE);

	burgner_scsi_prevent_allow_medium_removal_write_cdb ((gpointer) cdb_buf,
							     BURGNER_SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL_CDB_SIZE,
							     lock);

	if (!burgner_scsi_device_send_cmd_sync (device,
						cdb_buf, BURGNER_SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL_CDB_SIZE,
						sense_buf, SENSE_BUF_SIZE, NULL,
						NULL, 0, NULL,
						BURGNER_SCSI_DIRECTION_NONE,
						1000,
						&status, &err)) {
		fprintf (stderr, "Error while (un)locking %s tray\n", device_uri);
		exit (1);
	}
}

static void
do_blank (void)
{
	GError *err = NULL;
	guint status = 0;

	g_assert (blank != NONE);
	g_assert (BURGNER_MMC_BLANK_CDB_SIZE <= CDB_BUF_SIZE);

	g_print ("-> %s\n", blank == COMPLETE
				? "Blanking disc"
				: "Blanking disc minimally");

	memset (&sense_buf, 0, SENSE_BUF_SIZE);
	memset (&cdb_buf, 0, BURGNER_MMC_BLANK_CDB_SIZE);

	burgner_mmc_blank_write_cdb ((gpointer) cdb_buf,
				     BURGNER_MMC_BLANK_CDB_SIZE,
				     blank == COMPLETE
				     	? BURGNER_MMC_BLANK_TYPE_DISC
					: BURGNER_MMC_BLANK_TYPE_MINIMALLY_DISC,
				     0);

	if (!burgner_scsi_device_send_cmd_sync (device,
						cdb_buf, BURGNER_MMC_BLANK_CDB_SIZE,
						sense_buf, SENSE_BUF_SIZE, NULL,
						NULL, 0, NULL,
						BURGNER_SCSI_DIRECTION_NONE,
						4800000, /* 80min * 60s/min * 1000ms/s = 4800000ms */
						&status, &err)) {
		fprintf (stderr, "Error while blanking %s\n", device_uri);
		exit (1);
	}
}

static void
do_get_nwa (void)
{
	_BurgnerMmcImplReadTrackInformationData data;
	GError *err = NULL;
	gint status = 0;

	g_assert (BURGNER_MMC_READ_TRACK_INFORMATION_CDB_SIZE <= CDB_BUF_SIZE);

	memset (&sense_buf, 0, SENSE_BUF_SIZE);
	memset (&cdb_buf, 0, BURGNER_MMC_READ_TRACK_INFORMATION_CDB_SIZE);
	memset (&data, 0, sizeof (data));

	burgner_mmc_read_track_information_write_cdb (cdb_buf, BURGNER_MMC_READ_TRACK_INFORMATION_CDB_SIZE,
							sizeof (data), 0xff);

	if (!burgner_scsi_device_send_cmd_sync (device,
						cdb_buf, BURGNER_MMC_READ_TRACK_INFORMATION_CDB_SIZE,
						sense_buf, SENSE_BUF_SIZE, NULL,
						&data, sizeof (data), NULL,
						BURGNER_SCSI_DIRECTION_FROM_DEVICE,
						BURGNER_MMC_READ_TRACK_INFORMATION_TIMEOUT,
						&status, &err)) {
		fprintf (stderr, "Error while asking track information to %s\n", device_uri);
		exit (1);
	}

	if (!data.nwa_v) {
		fprintf (stdout, "-1\n");
	} else {
		fprintf (stdout, "%" G_GUINT32_FORMAT "\n", burgner_scsi_get_uint32 (data.nwa));
	}
}

int
main (int argc, char **argv)
{
	int c, opt_index = 0;

	g_type_init ();
	gnome_vfs_init ();
	burgner_scsi_init ();

	if (burgner_scsi_privilege_supported) {
		GError *err = NULL;

		privilege = burgner_scsi_privilege_new ();

		if (!burgner_scsi_privilege_init (privilege, &err)) {
			fprintf (stderr, "Error while initializing privileges: %s\n", err && err->message ? err->message : "[No description]");
			g_object_unref (privilege);
			privilege = NULL;
		} else if (!burgner_scsi_privilege_remove (privilege, &err)) {
			fprintf (stderr, "Error while removing privileges: %s\n", err && err->message ? err->message : "[No description]");
			g_object_unref (privilege);
			privilege = NULL;
		}
	}

	while ((c = getopt_long (argc, argv, "hd:",
				 long_options, &opt_index)) != -1) {
		switch (c) {
			case 'd':
				if (device_uri) {
					print_help ();
					return 0;
				}
				device_uri = g_strdup (optarg);
				break;
			case 'h':
				print_help ();
				return 0;
			case 0:
				switch (opt_index) {
					case 0:
						list_drivers = TRUE;
						break;
					case 1:
						if (driver_id ||
						    !optarg ||
						    !*optarg) {
							print_help ();
							return 0;
						}

						driver_id = g_strdup (optarg);
						list_devices = TRUE;
						
						break;
					case 2:
						list_all_devices = TRUE;
						break;
#if 0
					case 3:
						if (device_uri) {
							print_help ();
							return 0;
						}
						device_uri = g_strdup (optarg);
						break;
#endif
					case 4:
						devices_do_inquiry = TRUE;
						break;
					case 5:
						inquiry = TRUE;
						break;
					case 6:
						lock_tray = TRUE;
						break;
					case 7:
						unlock_tray = TRUE;
						break;
					case 8:
						if (blank != NONE ||
						    !optarg ||
						    !*optarg) {
							print_help ();
							return 0;
						}

						if (strcmp (optarg, "minimal") == 0) {
							blank = MINIMAL;
						} else if (strcmp (optarg, "complete") == 0) {
							blank = COMPLETE;
						} else {
							print_help ();
							return 0;
						}
						break;
					case 9:
						print_help ();
						return 0;
					case 10:
						directories = g_list_prepend (directories, g_strdup (optarg));
						break;
					case 11:
						close_track = TRUE;
						break;
					case 12:
						get_nwa = TRUE;
						break;
					default:
						g_assert_not_reached ();
						break;
					
				}
				break;
			default:
				print_help ();
				return 1;
		}
	}

	directories = g_list_reverse (directories);

	init_driver_manager ();

	if (!device_uri) {
		if (inquiry || lock_tray || unlock_tray || blank) {
			print_help ();
			return 0;
		}
	} else {
		GUri *duri = g_uri_new ();
		GError *err = NULL;

		if (!g_uri_set (duri, device_uri, FALSE, FALSE, &err) ||
		    !duri->authority) {
			fprintf (stderr, "%s: invalid device uri: %s\n",
						device_uri,
						err && err->message
							? err->message
							: "[no error or message]");
			return 1;
		}

		driver = burgner_scsi_driver_manager_get_driver_from_uri
				(driver_manager, duri);
		if (!driver) {
			fprintf (stderr, "Can't get driver of %s\n", device_uri);
			return 1;
		}

		device = burgner_scsi_driver_open_device_sync (driver, duri, 65536, NULL, &err);
		if (!device) {
			fprintf (stderr, "%s: unable to open device: %s\n",
						device_uri,
						err && err->message
							? err->message
							: "[no error or message]");
			return 1;
		}
	}

	if (list_drivers)
		do_list_drivers ();

	if (list_devices)
		do_list_devices ();

	if (list_all_devices)
		do_list_all_devices ();

	if (inquiry)
		do_inquiry ();

	if (lock_tray)
		do_lock_unlock_tray (TRUE);

	if (close_track)
		do_close_track ();

	if (get_nwa)
		do_get_nwa ();

	if (unlock_tray)
		do_lock_unlock_tray (FALSE);

	if (blank)
		do_blank ();

	if (device) {
		GError *err = NULL;

		g_assert (driver);

		if (!burgner_scsi_driver_close_device_sync (driver,
							    device,
							    &err)) {
			fprintf (stderr, "Error while closing %s: %s\n",
						device_uri,
						err && err->message
							? err->message
							: "[no error or message]");
			return 1;
		}
		g_object_unref (device);
		g_object_unref (driver);
	}

	g_object_unref (driver_manager);

	return 0;
}

